
exception Incomplete
exception Impossible



structure NotRandom =
struct

val i = ref 0

fun nextInt() = (i := (!i) + 1;
		 (!i) - 1)

fun nextString() = Int.toString(nextInt())

end


(* 
 
 * A MutableSymbolSet is a mutable set of symbols.

 * The structure keeps track of all the symbols that have been
   allocated through:

     fresh : string -> symbol
     fromString : string -> symbol
   
 * Because the set of all symbols is infinite, attempts to iterate
   over it fail.  This set allows us to iterate over only the portion
   we care about--the portion of symbols actually being used.

 *)
functor MutableSymbolSet () =
struct


exception Top
exception Bottom


type super = string

val mems = ref ([] : super list)

val eq = (op =) : super * super -> bool
val wt = (op <) : super * super -> bool
val st = (op >) : super * super -> bool

fun bot () = raise Bottom
fun top () = raise Top

fun join (s1,s2) = if (eq(s1,s2)) then s1 else (top ())
fun meet (s1,s2) = if (eq(s1,s2)) then s1 else (bot ())

fun members() = LazyList.fromList (! mems)

fun is_member e = List.exists (fn e' => eq(e,e')) (!mems)


fun fresh s = 
    let
	val r = NotRandom.nextString()

	val s' = s ^ "/" ^ r
    in
	if List.exists (fn s' => eq(s',s)) (!mems)
	then fresh s
	else s'
    end


fun fromString s = 
    let 
	val _ = if List.exists (fn s' => 
				   eq(s',s)) (!mems)
		then ()
		else mems := s::(!mems)
    in
	s
    end


val from = fromString

fun toString s = s

end




(*
 * Symbol is the superset of all VARs.
 *)
structure Symbol =
struct

structure SymbolSet = MutableSymbolSet()

open SymbolSet

type symbol = super

end




(*
 
 * Label is used for freshifying allocated labels.

 * We have separate structures for LAM labels, and CALL labels which
   refer back to this structure.

 *)
structure Label =
struct

structure SymbolSet = MutableSymbolSet()

open SymbolSet

type symbol = super

end





(* 
 
 * S-Expressions are utilized by the front end.
 
 *)
signature SEXP =
sig
    datatype sexp = SYM of Symbol.symbol
		  | CONS of {car : sexp , cdr : sexp}
		  | STR of string
		  | INT of int
		  | FUN of sexp_fun
		  | NIL 	      

    withtype sexp_fun = sexp -> sexp

    val fromString : string -> sexp
end 





structure SExp : SEXP =
struct



exception ParserError of string


		    
datatype sexp = SYM of Symbol.symbol
	      | CONS of {car : sexp , cdr : sexp}
	      | STR of string
	      | INT of int
	      | FUN of sexp_fun
	      | NIL 

withtype sexp_fun = sexp -> sexp 



datatype token = LPAR
	       | RPAR
	       | QUOTE
	       | QQUOTE
	       | UNQUOTE
	       | UNSQUOTE
	       | DOT
	       | EOF
	       | SEXP of sexp



fun nextId(id,c::r) = 
    if c = #" "  orelse
       c = #"\n" orelse
       c = #"\t" orelse
       c = #"`"  orelse
       c = #"'"  orelse
       c = #"("  orelse
       c = #")"  orelse
       c = #";"  orelse
       c = #","  
    then (id,c::r)
    else nextId(id ^ (implode [c]), r)
  | nextId(id,[]) = (id,[])
		

fun isInt([#"0"]) = true
  | isInt([#"1"]) = true
  | isInt([#"2"]) = true
  | isInt([#"3"]) = true
  | isInt([#"4"]) = true
  | isInt([#"5"]) = true
  | isInt([#"6"]) = true
  | isInt([#"7"]) = true
  | isInt([#"8"]) = true
  | isInt([#"9"]) = true
  | isInt(#"0"::r) = isInt r
  | isInt(#"1"::r) = isInt r
  | isInt(#"2"::r) = isInt r
  | isInt(#"3"::r) = isInt r
  | isInt(#"4"::r) = isInt r
  | isInt(#"5"::r) = isInt r
  | isInt(#"6"::r) = isInt r
  | isInt(#"7"::r) = isInt r
  | isInt(#"8"::r) = isInt r
  | isInt(#"9"::r) = isInt r

  | isInt(_) = false


fun nextToken s =
let
    fun nt (#" "::r) = nt r
      | nt (#"\n"::r) = nt r
      | nt (#"\t"::r) = nt r
      | nt (#"."::r) = (DOT, r)
      | nt (#"`"::r) = (QQUOTE, r)
      | nt (#"'"::r) = (QUOTE, r)
      | nt (#"("::r) = (LPAR, r)
      | nt (#")"::r) = (RPAR, r)
      | nt (#","::(#"@")::r) = (UNSQUOTE, r)
      | nt (#","::r) = (UNQUOTE, r)
      | nt (c::r) = 
	let
	    val (str,rest) = nextId("",c::r)
	in
	    if isInt (explode str)
	    then (SEXP (INT (case (Int.fromString str) of
				 SOME i => i
			       | NONE => raise Impossible)),rest)
	    else (SEXP (SYM (Symbol.fromString str)),rest)
	end
      | nt ([]) = (EOF,[])
in
    nt s
end
    




fun tokenListFromString s =
    let
	fun tlfcl cl =
	    case (nextToken cl) of
		(EOF,cl') => []
	      | (nt,cl') => nt::(tlfcl cl')
    in
	tlfcl (explode s)
    end


fun parseTopLevelSExpList [] = ([],[])
  | parseTopLevelSExpList tl = 
    let
	val (nsexp,rest) = parseSExp tl

	val (nsexpl,rest') = parseTopLevelSExpList rest
    in
	(nsexp::nsexpl,rest')
    end

and parseSExpList [] = raise (ParserError("exected SExp list; got EOF"))
  | parseSExpList (RPAR::tl) = (NIL,tl)
  | parseSExpList tl = 
    let
	val (car,rest) = parseSExp tl

	val (cdr,rest') = parseSExpList rest
    in
	(CONS {car=car, cdr=cdr},rest')
    end

and parseSExp [] = raise (ParserError("expected SExp; got EOF"))
  | parseSExp (s as (LPAR::tl)) = parseSExpList tl
  | parseSExp ((SEXP s)::tl) = (s,tl)


fun fromString s = 
    let
	val (sexp,tl) = parseSExp (tokenListFromString s)
    in
	sexp
    end 



end






structure LambdaCalcExp = 
struct

open SExp


type var = Symbol.symbol



datatype lcexp = REF of var
	       | APP of {f : lcexp, 
			 args : lcexp list}
	       | LAM of {vars : var list, 
			 body : lcexp}
	       | LETREC of { varlams : (var * lcexp) list,
			     body : lcexp }
	       | SEXP of sexp
			


fun parseLExp (SYM s) = REF s
  | parseLExp (INT i) = SEXP (INT i)
  | parseLExp (CONS{car=SYM f,cdr}) =
    (case (Symbol.toString f) of 
	 "lambda" => parseLambdaExp cdr
       | "letrec" => parseLetRecExp cdr
       | _ => 
	 (* Just an APP. *)
	 let
	     val args = parseLExpList cdr
	 in
	     APP {f = REF f, args = args}
	 end
	 )
  | parseLExp (CONS{car,cdr}) =
    (* An APP. *)
    let
	val f = parseLExp car
	val args = parseLExpList cdr
    in
	APP {f = f, args = args}
    end


and parseLExpList NIL = []
  | parseLExpList (CONS{car,cdr}) =
    (parseLExp car)::(parseLExpList cdr)


and parseLetRecExp (CONS{car=varlams,
			 cdr=CONS{car=body,cdr=NIL}}) =
    let
	val varlamlist = parseVarLamList varlams
	val body_exp = parseLExp body
    in
	LETREC {varlams = varlamlist, body=body_exp}
    end

and parseLambdaExp (CONS{car=vars,
			 cdr=CONS{car=body,cdr=NIL}}) =
    let
	val varlist = parseVarList vars
	val body_exp = parseLExp body
    in
	LAM {vars = varlist, body = body_exp}
    end

and parseVarLamList NIL = []
  | parseVarLamList (CONS{car=varlam,
			  cdr}) =
    (parseVarLam varlam)::(parseVarLamList cdr)
    
and parseVarLam (CONS{car=SYM v,
		      cdr=CONS{car=CONS{car=_,
					cdr=lam},
			       cdr=_}}) =
    (v,parseLambdaExp lam)

and parseVarList NIL = []
  | parseVarList (CONS{car=SYM v,
		       cdr}) =
    v::(parseVarList cdr)


fun fromSExp s = parseLExp s

end







structure CPS =
struct

(*

 * Note we do not use the full power of the ML type system to encode
   the partitioning constraints.  In many cases, we expect that the
   type of the label (ConL or FunL) correctly dictates the type of the
   term.

 *)

type VAR = Symbol.super

type VARS = VAR list

datatype LIT = Int of int
	     | False


datatype LAB = FunL of Label.super
	     | ConL of Label.super


datatype PRIM = PLUS
	      | MINUS
	      | TIMES
	      | IF
	      | TEST_ZERO


datatype CALL = Call   of LAB * EXP * EXPS * EXPS
	      | LetRec of LAB * VAREXPS * VAREXPS * CALL

and EXP = 
    VarE  of VAR 
  | LamE  of LAM
  | LitE  of LIT
  | PrimE of PRIM

withtype LAM = LAB * VARS * VARS * CALL

and VAREXP = VAR * EXP

and VAREXPS = VAREXP list

and EXPS = EXP list


and PR = LAM





(* "Pretty" Printing for an AST *)


fun stringOfPR (pr) = stringOfLAM pr

and stringOfLAM (l, fvs, cvs, call) 
  = "(lambda[" ^
    (stringOfLAB l) ^
    "] (" ^ 
    (stringOfVARS fvs) ^ 
    ";" ^
    (stringOfVARS cvs) ^
    ") " ^
    (stringOfCALL call) ^
    ")"

and stringOfVAR (v) = Symbol.toString v

and stringOfVARS (v1::v2::vs) = v1 ^ " " ^ (stringOfVARS (v2::vs))
  | stringOfVARS (v::[]) = v
  | stringOfVARS [] = ""

and stringOfCALL (Call(l, expf, expfs, expcs)) 
  = (stringOfLAB l) ^ 
    ":" ^
    "(" ^
    (stringOfEXP expf) ^
    (case expfs of [] => ""
		 | _  => " ") ^
    (stringOfEXPS expfs) ^
    (case expcs of [] => ";"  
		 | _  => "; ") ^
    (stringOfEXPS expcs) ^
    ")"

  | stringOfCALL (LetRec(l, varexpfs, varexpcs, call)) 
    = "(letrec (" ^
      (stringOfVAREXPS varexpfs) ^
      (case varexpcs of [] => ";"
		      | _  => "; ") ^
      (stringOfVAREXPS varexpcs) ^
      ") " ^
      (stringOfCALL call) ^ 
      ")"
      
and stringOfVAREXP (v,e) = "(" ^
			   (stringOfVAR v) ^
			   " " ^
			   (stringOfEXP e) ^
			   ")"

and stringOfVAREXPS (ve1::ve2::ves) 
  = (stringOfVAREXP ve1) ^ " " ^ (stringOfVAREXPS (ve2::ves))
  | stringOfVAREXPS (ve::[]) = stringOfVAREXP ve
  | stringOfVAREXPS [] = ""


and stringOfEXP (VarE v) = stringOfVAR v
  | stringOfEXP (LamE lam) = stringOfLAM lam
  | stringOfEXP (LitE l) = stringOfLIT l
  | stringOfEXP (PrimE prim) = stringOfPRIM prim

and stringOfEXPS (e1::e2::es) 
  = (stringOfEXP e1) ^ " " ^ (stringOfEXPS (e2::es))
  | stringOfEXPS (e::[]) = stringOfEXP e
  | stringOfEXPS [] = ""
				      
and stringOfLIT (Int i) = Int.toString(i)
  | stringOfLIT (False) = "false" 

and stringOfPRIM PLUS = "+"
  | stringOfPRIM IF = "if"
  | stringOfPRIM TEST_ZERO = "test-zero"
  | stringOfPRIM MINUS = "-"
  | stringOfPRIM TIMES = "*"

and stringOfLAB (FunL l) = Label.toString l
  | stringOfLAB (ConL l) = Label.toString l




fun schemeOfPR (pr) = schemeOfLAM pr

and schemeOfLAM (l, fvs, cvs, call) 
  = "(lambda" ^
    " (" ^ 
    (schemeOfVARS fvs) ^ 
    (case cvs of [] => ""
	       | _ => " ") ^
    (schemeOfVARS cvs) ^
    ") " ^
    (schemeOfCALL call) ^
    ")"


and schemeOfVAR (v) = Symbol.toString v

and schemeOfVARS (v1::v2::vs) = v1 ^ " " ^ (schemeOfVARS (v2::vs))
  | schemeOfVARS (v::[]) = v
  | schemeOfVARS [] = ""

and schemeOfCALL (Call(l, expf, expfs, expcs)) 
  = "(" ^
    (schemeOfEXP expf) ^
    (case expfs of [] => ""
		 | _  => " ") ^
    (schemeOfEXPS expfs) ^
    (case expcs of [] => ""  
		 | _  => " ") ^
    (schemeOfEXPS expcs) ^
    ")"

  | schemeOfCALL (LetRec(l, varexpfs, varexpcs, call)) 
    = "(letrec (" ^
      (schemeOfVAREXPS varexpfs) ^
      (case varexpcs of [] => ""
		      | _  => " ") ^
      (schemeOfVAREXPS varexpcs) ^
      ") " ^
      (schemeOfCALL call) ^ 
      ")"
      
and schemeOfVAREXP (v,e) = "(" ^
			   (schemeOfVAR v) ^
			   " " ^
			   (schemeOfEXP e) ^
			   ")"

and schemeOfVAREXPS (ve1::ve2::ves) 
  = (schemeOfVAREXP ve1) ^ " " ^ (schemeOfVAREXPS (ve2::ves))
  | schemeOfVAREXPS (ve::[]) = schemeOfVAREXP ve
  | schemeOfVAREXPS [] = ""


and schemeOfEXP (VarE v) = schemeOfVAR v
  | schemeOfEXP (LamE lam) = schemeOfLAM lam
  | schemeOfEXP (LitE l) = schemeOfLIT l
  | schemeOfEXP (PrimE prim) = schemeOfPRIM prim

and schemeOfEXPS (e1::e2::es) 
  = (schemeOfEXP e1) ^ " " ^ (schemeOfEXPS (e2::es))
  | schemeOfEXPS (e::[]) = schemeOfEXP e
  | schemeOfEXPS [] = ""
				      
and schemeOfLIT (Int i) = Int.toString(i)
  | schemeOfLIT (False) = "false" 

and schemeOfPRIM PLUS = "+:cps"
  | schemeOfPRIM IF = "if:cps"
  | schemeOfPRIM TEST_ZERO = "test-zero:cps"
  | schemeOfPRIM MINUS = "-:cps"
  | schemeOfPRIM TIMES = "*:cps"







(* 
 
 * LAB keeps track of LAM labels. 
 
 *)

structure LAB = 
struct

exception Top
exception Bottom

type super = LAB

val mems = ref ([] : super list)


fun eq(FunL l1, FunL l2) = l1 = l2
  | eq(ConL l1, ConL l2) = l1 = l2
  | eq _  = false


val wt = eq
val st = eq

fun bot () = raise Bottom
fun top () = raise Top

fun join (s1,s2) = if (eq(s1,s2)) then s1 else (top ())
fun meet (s1,s2) = if (eq(s1,s2)) then s1 else (bot ())

fun members() = LazyList.fromList (!mems)

fun is_member (x : super) = List.exists (fn x' => eq(x,x')) (!mems)

fun toString (FunL l) = l
  | toString (ConL l) = l


fun funFromString s = 
    let
	val l = FunL (Label.fromString s)

	val _ = if List.exists (fn l' =>
				   eq(l',l)) (!mems)
		then ()
		else mems := l::(!mems)
    in
	l
    end


fun freshFunL s = 
    let
	val l = FunL (Label.fresh s)
	val _ = mems := l::(!mems)
    in
	l
    end

fun freshConL s = 
    let 
	val l = ConL (Label.fresh s)
	val _ = mems := l::(!mems)
    in
	l
    end


end





(* 
 
 * CLAB keeps track of CALL labels.
 
 *)

structure CLAB = 
struct

exception Top
exception Bottom

type super = LAB

val mems = ref ([] : super list)



fun eq(FunL l1, FunL l2) = l1 = l2
  | eq(ConL l1, ConL l2) = l1 = l2
  | eq _  = false

val wt = eq
val st = eq

fun bot () = raise Bottom
fun top () = raise Top

fun join (s1,s2) = if (eq(s1,s2)) then s1 else (top ())
fun meet (s1,s2) = if (eq(s1,s2)) then s1 else (bot ())

fun members() = LazyList.fromList (!mems)

fun is_member (x : super) = List.exists (fn x' => eq(x,x')) (!mems)

fun toString (FunL l) = l
  | toString (ConL l) = l


fun funFromString s = 
    let
	val l = FunL (Label.fromString s)
	val _ = mems := l::(!mems)
    in
	l
    end


fun freshFunL s = 
    let
	val l = FunL (Label.fresh s)
	val _ = mems := l::(!mems)
    in
	l
    end

fun freshConL s = 
    let 
	val l = ConL (Label.fresh s)
	val _ = mems := l::(!mems)
    in
	l
    end


end




(* VAR creates and tracks all variables. *)
structure VAR = Symbol 



(*
 
 * A simple direct style to CPS converter.

 *)
local
    open SExp
    open LambdaCalcExp
in


fun kont vars body = (LamE(LAB.freshConL "knt",
			   map Symbol.fromString vars,
			   [],
			   body))

fun lambda uvars kvars body = (LamE(LAB.freshFunL "lmb",
				    map Symbol.fromString uvars,
				    map Symbol.fromString kvars,
				    body))

fun appfun f uexps kexps = (Call(CLAB.freshFunL "cf",
				f,
				uexps,
				kexps))

fun appcont q uexps = (Call(CLAB.freshConL "cc",
			    q,
			    uexps,
			    []))

fun letrec varulams 
	   varclams 
	   body 
  = (LetRec(LAB.freshConL "lr",
	    map (fn (v,l) => (Symbol.fromString v,l)) varulams,
	    map (fn (v,l) => (Symbol.fromString v,l)) varclams,
	    body))


fun $ v = (VarE v)

fun % n = (LitE(Int n))

fun tz i z nz = (appfun (PrimE TEST_ZERO) [i] [z,nz])

fun minus a b k = (appfun (PrimE MINUS)	[a,b] [k])

fun plus a b k = (appfun (PrimE PLUS) [a,b] [k])



val pr_transduce = 
    (lambda [] ["halt"]
     (letrec 
      [("lp2",
	(lambda [] ["u2a","d2a"]
	 (appcont ($"u2a") 
		  [lambda ["x"] ["u2b"]
			  (appcont ($"d2a") [%2,(lambda [] ["d2b"]
							(appfun ($"lp2") [] [$"u2b",$"d2b"]))])]))),
       ("lp1",(lambda [] ["d1a"] (appcont ($"d1a") [%5,$"lp1"] )))]
      [("d",(kont ["x","upstream"] (appfun ($"upstream") [] [$"d"])))]
      (appfun ($"lp2") [$"lp1"] [$"d"])))
      
						   
val pr_transduce = case pr_transduce of LamE lam => lam

(*			   

val pr_doubly 
    
  = case
    (lambda [] ["halt"]
     (letrec 
	  []
	  [("loop1",
	    (kont ["i","x"]
	     (tz ($"i")
		 (kont [] (appcont ($"halt") [$"x"]))
		 (kont [] (letrec 
			       []
			       [("loop2",
				 (kont ["j","f","y"]
				  (tz ($"j")
				      (kont [] 
					    (minus ($"i") (%1) 
						   (kont ["i-1"]
							 (appcont 
							      ($"loop1")
							      [$"i-1",
							       $"y"]))))
				      (kont []
				       (minus ($"j") (%1)
					(kont ["j-1"]
					 (appfun 
					      ($"f")
					      [$"y"]
					      [kont ["fy"]
					       (appcont ($"loop2")
						[$"j-1",
						 $"f",
						 $"fy"])])))))))]
			       (appcont ($"loop2")
					[%10,
					 (lambda ["n"] ["k"]
						 (plus ($"i") 
						       ($"n") 
						       ($"k"))),
					 $"x"]))))))]
	  (appcont ($"loop1") [%10,%0])))
    of (LamE lam) => lam
     | _ => (print "WTF, Mate?\n";
	     raise Impossible)

				 

			     
*)



fun T (REF v) return = 
    Call(CLAB.freshConL "ckv",
	 return,
	 [VarE v],
	 [])

  | T (SEXP (INT n)) return = 
    Call(CLAB.freshConL "ckv",
	 return,
	 [LitE(Int n)],
	 [])
    
  | T (LAM({vars,body})) return =
    let
    in
	Call(CLAB.freshConL "ckl",
	     return,
	     [M(LAM{vars=vars,body=body})],
	     [])
    end

  | T (LETREC{varlams,body}) return =
    let
	val varexps = map (fn (var,lam) =>
			      (var,M lam))
		      varlams
    in
	LetRec(CLAB.freshConL "lr",
	       varexps,
	       [],
	       T body return)
    end
    
  | T (APP{f=REF v,args}) return =
    (case (Symbol.toString v) of
	 "+" => Targs (PrimE PLUS) [] return args
       | "*" => Targs (PrimE TIMES) [] return args
       | "-" => Targs (PrimE MINUS) [] return args
       | "test-zero" => 
	 let
	     val [cond,ifzero,ifnot] = args
				       
	     val tzc = VAR.fresh "tzc"
	 in
	     case cond of 
		 (* Special case: atomic argument. *)
		 (REF v) => Call(CLAB.freshFunL "tz",
				 PrimE TEST_ZERO,
				 [VarE v],
				 [(kont [] (T ifzero return)),
				  (kont [] (T ifnot  return))])
	       | _ =>
		 T cond (LamE(LAB.freshConL "tza",
			      [tzc],[],
			      Call(CLAB.freshFunL "tz",
				   PrimE TEST_ZERO,
				   [VarE tzc],
				   [(kont [] (T ifzero return)),
				    (kont [] (T ifnot  return))])))
	 end
       | _ => Targs (VarE v) [] return args)

  | T (APP{f=LAM lam,args}) return =
    let
    in
	Targs (M (LAM lam)) [] return args
    end
	      
	
  | T (APP{f,args}) return =
    let
	val vf = VAR.fresh "uf"
    in
	T f (LamE(LAB.freshConL "kf",
		  [vf],[],
		  Targs (VarE vf) [] return args))
    end

and M (LAM{vars,body}) =
    let
	val k' = VAR.fresh "k"
    in
	LamE(LAB.freshFunL "ulam",
	     vars,
	     [k'],
	     T body (VarE k'))
    end

	
and Targs fexp argvs return [] = 
    let
    in
	Call(CLAB.freshFunL "cf",
	     fexp,
	     argvs,
	     [return])
    end

  | Targs fexp argvs return ((SEXP(INT n))::args) =
    let
    in
	Targs fexp (argvs @ [LitE(Int n)]) return args
    end

  | Targs fexp argvs return ((LAM lam)::args) =
    let
    in
	Targs fexp (argvs @ [M (LAM lam)]) return args
    end

  | Targs fexp argvs return ((REF v)::args) =
    let
    in
	Targs fexp (argvs @ [VarE v]) return args
    end

  | Targs fexp argvs return (a::args) =
    let
	val va = VAR.fresh "u"
    in
	T a (LamE(LAB.freshConL "ku",
		  [va],[],
		  (Targs fexp (argvs @ [VarE va]) return args)))
    end
	

fun fromLExp lexp = 
    (LAB.freshFunL "top",
     [] : VAR list,
     [VAR.fromString "halt"],
     T lexp (VarE (VAR.fromString "halt")))
    
end



(* Syntactic lattices. *)


structure VARSs = SequenceLattice(structure Member = VAR)


structure PowVAR = Power(structure Member = VAR)



structure LIT = FlatLattice(structure Set = 
			    InfiniteSet(type super = LIT
					val eq = (op =) 
						 : super * super -> bool
					val toString = stringOfLIT))



structure PRIM = FlatLattice(structure Set = 
			     FiniteSet(type super = PRIM
				       val memlist = [IF,
						      TEST_ZERO,
						      PLUS,
						      MINUS,
						      TIMES]
				       val toString = stringOfPRIM
				       val eq = (op =) 
						: super * super -> bool ))


structure EXP = FlatLattice(structure Set = 
			    InfiniteSet(type super = EXP
					val toString = stringOfEXP
					val eq = (op =) 
						 : super * super -> bool))
		
		
structure LAM = FlatLattice(structure Set = 
			    InfiniteSet(type super = LAM
					val toString = stringOfLAM
					val eq = (op =) 
						 : super * super -> bool))

structure CALL = 
FlatLattice(structure Set = 
	    InfiniteSet(type super = CALL
			val toString = stringOfCALL
			val eq = (op =) : super * super -> bool))


structure PR = LAM





(*

fun procLabelsForLAM (l, varfs, varcs, call) = l::(procLabelsForCALL call)

and procLabelsForCALL (Call(l, exp, expfs, expcs)) = (procLabelsForEXP exp) @
						     (procLabelsForEXPs expfs) @
						     (procLabelsForEXPs expcs)

  | procLabelsForCALL (LetRec(l, vefs, vecs, call)) = [l] @ (procLabelsForVAREXPs vefs) @
						      (procLabelsForVAREXPs vecs) @
						      (procLabelsForCALL call)


and procLabelsForEXP (VarE v) = []
  | procLabelsForEXP (LamE lam) = procLabelsForLAM lam
  | procLabelsForEXP (LitE lit) = []
  | procLabelsForEXP (PrimE prim) = []

and procLabelsForEXPs (exp::exps) = (procLabelsForEXP exp) @ (procLabelsForEXPs exps)
  | procLabelsForEXPs [] = []

and procLabelsForVAREXPs varexps = procLabelsForEXPs (map (fn (v,e) => e) varexps)







fun callLabelsForLAM (l, varfs, varcs, call) = (callLabelsForCALL call)

and callLabelsForCALL (Call(l, exp, expfs, expcs)) = [l] @ 
						     (callLabelsForEXP exp) @
						     (callLabelsForEXPs expfs) @
						     (callLabelsForEXPs expcs)

  | callLabelsForCALL (LetRec(l, vefs, vecs, call)) = [l] @ 
						      (callLabelsForVAREXPs vefs) @
						      (callLabelsForVAREXPs vecs) @
						      (callLabelsForCALL call)


and callLabelsForEXP (VarE v) = []
  | callLabelsForEXP (LamE lam) = callLabelsForLAM lam
  | callLabelsForEXP (LitE lit) = []
  | callLabelsForEXP (PrimE prim) = []

and callLabelsForEXPs (exp::exps) = (callLabelsForEXP exp) @ (callLabelsForEXPs exps)
  | callLabelsForEXPs [] = []

and callLabelsForVAREXPs varexps = callLabelsForEXPs (map (fn (v,e) => e) varexps)




fun varsInLAM (l, varfs, varcs, call) = varfs @ varcs @ (varsInCALL call)
					
and varsInCALL (Call(l, exp, expfs, expcs)) = (varsInEXP exp) @
					      (varsInEXPs expfs) @
					      (varsInEXPs expcs)
					      
  | varsInCALL (LetRec(l, vefs, vecs, call)) = (varsInVAREXPs vefs) @
					       (varsInVAREXPs vecs) @
					       (varsInCALL call)
					       

and varsInEXP (VarE v) = []
  | varsInEXP (LamE lam) = varsInLAM lam
  | varsInEXP (LitE lit) = []
  | varsInEXP (PrimE prim) = []

and varsInEXPs (exp::exps) = (varsInEXP exp) @ (varsInEXPs exps)
  | varsInEXPs [] = []

and varsInVAREXPs varexps = 
    (map (fn (v,e) => v) varexps) @
    (varsInEXPs (map (fn (v,e) => e) varexps))
			    

*)

(* free *)



fun freeVarsInLAM (l, varfs, varcs, call) 
  = PowVAR.minus( freeVarsInCALL call,
		  varfs @ varcs )
    
and freeVarsInCALL (Call(l, exp, expfs, expcs)) 
  = (freeVarsInEXP exp) @
    (freeVarsInEXPs expfs) @
    (freeVarsInEXPs expcs)
						  
  | freeVarsInCALL (LetRec(l, vefs, vecs, call)) 
    = (PowVAR.minus( (freeVarsInEXPs 
			  (map (fn (v,e) => e) (vefs @ vecs) ))
		     @
		     (freeVarsInCALL call),
		     (map (fn (v,e) => v) (vefs @ vecs))))
								  
					       

and freeVarsInEXP (VarE v) = [v]
  | freeVarsInEXP (LamE lam) = freeVarsInLAM lam
  | freeVarsInEXP (LitE lit) = []
  | freeVarsInEXP (PrimE prim) = []

and freeVarsInEXPs (exp::exps) = (freeVarsInEXP exp) @ 
				 (freeVarsInEXPs exps)
  | freeVarsInEXPs [] = []


end





structure aTime0CFA =
struct

datatype super = SomeTime
	       | BottomTime

fun is_member (x : super) = true

fun eq(x : super, y : super) = x = y

fun wt(x : super, y : super) = eq(x,y)
fun st(x : super, y : super) = eq(x,y)

exception Top
exception Bottom

fun top() = raise Top
fun bot() =  raise Bottom

fun join(x, y) = if (eq(x,y))
		 then x
		 else top()

fun meet(x, y) = if (eq(x,y))
		 then x
		 else bot()

fun members () = LazyList.fromList [SomeTime]

fun toString SomeTime = "|t|"
  | toString BottomTime = "_|_"

val at0 = SomeTime

fun atick (at, _, _, _) = at

end







structure aTime1CFA =
struct

(* type super = CPS.LAB *)

datatype super = Lab of CPS.LAB
	       | Top
	       | Bot

fun is_member (x : super) = true

fun eq(x : super, y : super) = x = y

fun wt(x : super, y : super) = eq(x,y)
fun st(x : super, y : super) = eq(x,y)



fun top() = Top

fun bot() = Bot

fun join(x, y) = if (eq(x,y))
		 then x
		 else top()

fun meet(x, y) = if (eq(x,y))
		 then x
		 else bot()

fun mems' () = LazyList.cons(CPS.LAB.funFromString "at0",
			     CPS.LAB.members()) 

fun members () = LazyList.map (fn l => Lab l)
			      (mems'())
		
	    

fun toString (Lab (CPS.FunL l)) = "|" ^ (l) ^  "|"
  | toString (Lab (CPS.ConL l)) = "|" ^ (l) ^  "|"
  | toString (Bot) = "_"
  | toString (Top) = "T"

val at0 = (Lab(CPS.LAB.funFromString "at0"))

fun atick(at, l,_, _) = (Lab l)


exception Top
exception Bottom


end








signature POWDELTA_LATTICE =
sig
    type super

    val is_member : super -> bool

    val eq : super * super -> bool
    val wt : super * super -> bool
    val st : super * super -> bool

    val join : super * super -> super
    val meet : super * super -> super

    val bot : unit -> super
    val top : unit -> super

    exception Top
    exception Bottom

    val members : unit -> super LazyList.llist

    val toString : super -> string

    val cat : super * super -> super
			
    val inverse : super -> super

    val empty : super

    val down : super
end





functor aTimeDCFA(val pr : CPS.PR
		  structure PowDelta : POWDELTA_LATTICE) =
struct

(* structure aTime1CFA = aTime1CFA *)

structure aTime1CFAxPowDelta = Partial2ProductLattice(structure L1 = aTime1CFA
						      structure L2 = PowDelta)


open aTime1CFAxPowDelta

val at0 = (aTime1CFA.at0, PowDelta.empty)

fun atick(at, l, adelta_at0, adelta_at0_l) = 
    (aTime1CFA.atick(at, l, adelta_at0, adelta_at0_l), 
     adelta_at0_l)

(*
val _ = Might.debug "All Times: " 

val _ = LazyList.app (fn t => Might.debug (toString t)) (members())
*)


end






structure aLoc =
struct

datatype super = SomeLoc

fun is_member (x : super) = true

fun eq(SomeLoc,SomeLoc) = true

fun wt(x : super, y : super) = eq(x,y)
fun st(x : super, y : super) = eq(x,y)

exception Top
exception Bottom

fun top() = raise Top
fun bot() = (Might.debug "aLoc Bottom!"; raise Bottom)

fun join(x, y) = if (eq(x,y))
		 then x
		 else top()

fun meet(x, y) = if (eq(x,y))
		 then x
		 else bot()

fun members () = LazyList.fromList [SomeLoc]

fun toString x = "|l|"

end





structure DeltaSet =
struct

datatype super = E | D | U | DDP | UUP | UPDP

fun prime_hash E    = 2
  | prime_hash D    = 3
  | prime_hash DDP  = 5
  | prime_hash U    = 7
  | prime_hash UUP  = 11
  | prime_hash UPDP = 13

fun prime_unhash  2 = E
  | prime_unhash  3 = D
  | prime_unhash  5 = DDP
  | prime_unhash  7 = U
  | prime_unhash 11 = UUP
  | prime_unhash 13 = UPDP

fun is_member (x : super) = true

val eq = (op =) : super * super -> bool


fun members() = LazyList.fromList([E,D,U,DDP,UUP,UPDP])

fun toString E = "e"
  | toString D = "d"
  | toString U = "u"
  | toString DDP = "dd+"
  | toString UUP = "uu+"
  | toString UPDP = "u+d+"

fun inverse E = E
  | inverse D = U
  | inverse U = D
  | inverse UUP = DDP
  | inverse DDP = UUP
  | inverse UPDP = UPDP



(*  TODO: Add, cat-fresh *)

val empty = E

val down = D

end







structure Delta = FlatLattice(structure Set = DeltaSet)





structure PowDelta = 

struct 

structure PowDelta = Power(structure Member = Delta)


open PowDelta

fun perfect_hash (e::tl) = (DeltaSet.prime_hash e)*(perfect_hash tl)
  | perfect_hash [] = 1


fun eq(s1,s2) = perfect_hash(s1) = perfect_hash(s2)

fun gcd (m:int, 0):int = m
  | gcd (0, n:int):int = n
  | gcd (m:int, n:int):int =
    if m>n then gcd (m mod n, n) else gcd (m, n mod m) ;

fun wt(s1,s2) = 
    let
	val hash1 = perfect_hash(s1)
	val hash2 = perfect_hash(s2)
    in
	gcd(hash1,hash2) = hash1
    end

fun st(s1,s2) =
    let
	val hash1 = perfect_hash(s1)
	val hash2 = perfect_hash(s2)
    in
	gcd(hash1,hash2) = hash2
    end


    

(*
val E    = DeltaSet.E
val D    = DeltaSet.D
val DDP  = DeltaSet.DDP
val U    = DeltaSet.U
val UUP  = DeltaSet.UUP
val UPDP = DeltaSet.UPDP
*)



(*
fun members() = (LazyList.fromList[
				   [E,D,DDP,U,UUP,UPDP],
				   [E,D,DDP,U,UUP     ],
				   [E,D,DDP,U,    UPDP],
				   [E,D,DDP,U         ],
				   [E,D,DDP,  UUP,UPDP],
				   [E,D,DDP,  UUP     ],
				   [E                 ],
				   [                  ]
				   ])
*)

fun top() = (LazyList.toList(Delta.members()))
fun bot() = []

fun cat (Delta1, Delta2) = 
    let
	(*
 	val _ = case Delta1 of [] => Might.debug "EMPTY DELTA 1!!!!"
			     | _ => ()

 	val _ = case Delta2 of [] => Might.debug "EMPTY DELTA 1!!!!"
			     | _ => ()
*)

				    

	val llDelta1 = LazyList.fromList Delta1
	val llDelta2 = LazyList.fromList Delta2
		       
	val llcp = LazyList.crossLazyList (llDelta1, llDelta2)

	val llcatted = LazyList.map (DeltaSet.cat) llcp

	val catted = LazyList.toList(llcatted)
    in
	Union catted
    end

fun inverse (deltaSet : super) = (map (DeltaSet.inverse) deltaSet) : super

val empty = [DeltaSet.empty]
val down = [DeltaSet.down]

end







functor aF (structure LAB : SET) =
struct

structure aFrameStrings = FunctionLattice(structure Domain = LAB
					  structure Range = PowDelta)

open aFrameStrings


fun cat (ap1, ap2) = memoise(fn lab => PowDelta.cat(ap1 lab, ap2 lab))

fun inv (ap : super) = memoise(fn (lab : LAB.super) => (PowDelta.inverse (ap lab)))

val empty = fn lab => PowDelta.empty


fun down lab = fn lab' => if Domain.eq(lab,lab')
				then PowDelta.down
				else PowDelta.empty

fun memoise f = Might.memoise Domain.eq f
			 
end







structure DCFADomains =
struct

open CPS




structure CallLAB = CPS.CLAB




structure aTime = aTime1CFA



structure aF    = aF   (structure LAB = LAB)


type aTime = aTime.super

type aLoc = aLoc.super

type aF = aF.super



datatype aConf = Eval  of aEval
	       | Apply of aApply

and      aVal = aBasV of aBas
              | aLocV of aLoc
	      | aProcV of aProc

and      aProc = HaltP
               | aCloP of aClo
	       | PrimP of PRIM

and      aBas = aNumber
              | aFalse


withtype aBEnv   = VAR -> aTime

and      aD      = aVal list

and      aVEnv   = VAR * aTime -> aD

and      aLog    = aTime -> aF

and      aStore  = aLoc -> aD

and      vaD     = aD list

and      aClo    = LAM * aBEnv * aTime

and      aEval   = CALL * aBEnv * aVEnv * aLog * aStore * aTime

and      aApply  = aProc * vaD * vaD * aVEnv * aLog * aStore * aTime


type aVisited    = aConf list


structure aBas = FlatLattice(structure Set = 
			     FiniteSet(type super = aBas
				       val memlist = [aNumber,aFalse]
				       fun eq (aNumber,aNumber) = true
					 | eq (aFalse,aFalse) = true
					 | eq _ = false
				       fun toString aNumber = "|n|" 
					 | toString aFalse = "false"))
		 
structure aBEnv = FunctionLattice(structure Domain = VAR
				  structure Range = aTime)


structure aClo = Partial3ProductLattice(structure L1 = LAM
					structure L2 = aBEnv
					structure L3 = aTime)




structure aProcSet : SET = 
struct

type super = aProc

fun eq(HaltP,HaltP) = true
  | eq(aCloP clo1, aCloP clo2) = aClo.eq(clo1,clo2)
  | eq(PrimP prim1, PrimP prim2) = PRIM.eq(prim1,prim2)
  | eq _ = false

fun is_member (x : super) = true

exception Unenumerable

fun members() = raise Unenumerable

fun toString HaltP = "halt"
  | toString (aCloP (lam,aB,at_b)) = "<" ^ 
				     (CPS.LAM.toString lam) ^ 
				     "," ^ 
				     (aBEnv.toString aB) ^ 
				     "," ^
				     (aTime.toString at_b) ^ 
				     ">"
  | toString (PrimP p) = PRIM.toString p

end

structure aProc = FlatLattice(structure Set = aProcSet)






structure aValSet : SET = 
struct

type super = aVal

fun eq(aBasV b1, aBasV b2) = aBas.eq(b1,b2)
  | eq(aProcV p1, aProcV p2) = aProc.eq(p1,p2)
  | eq(aLocV l1, aLocV l2) = aLoc.eq(l1,l2)
  | eq _ = false

fun is_member (x : super) = true

exception Unenumerable

fun members() = raise Unenumerable

fun toString (aProcV HaltP) = "halt"
  | toString _ = "val"

end





structure aVal = FlatLattice(structure Set = aValSet)


structure aD = Power(structure Member = aVal)


structure vaD = SequenceLattice(structure Member = aD)



structure aVEnv = FunctionLattice(structure Domain = Partial2ProductLattice(structure L1 = VAR
									    structure L2 = aTime)
				  structure Range = aD)
						     



structure aLog = 
struct

structure aLogFn =  FunctionLattice(structure Domain = aTime
				    structure Range = aF)

open aLogFn

fun cat (adelta1, adelta2) = fn at => aF.cat(adelta1 at, adelta2 at)

end



structure aStore = FunctionLattice(structure Domain = aLoc
				   structure Range = aD)


structure aEval = Partial6ProductLattice(structure L1 = CALL
					 structure L2 = aBEnv
					 structure L3 = aVEnv
					 structure L4 = aLog
					 structure L5 = aStore
					 structure L6 = aTime)



structure aApply = Partial7ProductLattice(structure L1 = aProc
					  structure L2 = vaD
					  structure L3 = vaD
					  structure L4 = aVEnv
					  structure L5 = aLog
					  structure L6 = aStore
					  structure L7 = aTime)


structure aConf = 
struct

type super = aConf

fun is_member (x : super) = true

exception Unenumerable

fun members () =  raise Unenumerable

exception Top
exception Bottom

fun top () = raise Top
fun bot () = (Might.debug "aConf Bottom!"; raise Bottom)

fun eq(Eval e1, Eval e2) = aEval.eq(e1, e2)
  | eq(Apply a1, Apply a2) = aApply.eq(a1, a2)
  | eq _ = false

fun wt(Eval e1, Eval e2) = aEval.wt(e1, e2)
  | wt(Apply a1, Apply a2) = aApply.wt(a1, a2)
  | wt _ = false

fun st(Eval e1, Eval e2) = aEval.st(e1, e2)
  | st(Apply a1, Apply a2) = aApply.st(a1, a2)
  | st _ = false

fun join(Eval e1, Eval e2) = Eval(aEval.join(e1, e2))
  | join(Apply a1, Apply a2) = Apply(aApply.join(a1, a2))
  | join _ = top()

fun meet(Eval e1, Eval e2) = Eval(aEval.meet(e1, e2))
  | meet(Apply a1, Apply a2) = Apply(aApply.meet(a1, a2))
  | meet _ = bot()


fun toString (Eval e) = aEval.toString e
  | toString (Apply a) = aApply.toString a

fun toStringLn (Eval e) = aEval.toStringLn e
  | toStringLn (Apply a) = aApply.toStringLn a

end

structure aVisited = Power(structure Member = aConf)



end





structure DCFA =
struct

structure Doms = DCFADomains

open Doms
  

(* Auxiliary functions. *)

fun A ave aB at (VarE v)       = 
    let
	
    in
	ave(v, aB(v))
    end
  | A ave aB at (LamE lam)     = [aProcV (aCloP (lam,aB,at))]
  | A ave aB at (LitE (Int i)) = [aBasV aNumber]
  | A ave aB at (LitE False)   = [aBasV aFalse]
  | A ave aB at (PrimE prim)   = [aProcV (PrimP prim)]


fun amin_aProc adelta (HaltP) = adelta(aTime.at0)
  | amin_aProc adelta (aCloP ((FunL _,_,_,_),aB,at)) = aF.bot()
  | amin_aProc adelta (aCloP ((ConL _,_,_,_),aB,at)) = adelta(at)
  | amin_aProc adelta (PrimP prim) = aF.bot()

fun amin_aVal adelta (aBasV _)      = aF.bot()
  | amin_aVal adelta (aProcV aProc) = amin_aProc adelta aProc
  | amin_aVal adelta (aLocV _)      = aF.bot()

fun amin_aD adelta (v::rest) = aF.join(amin_aVal adelta v, amin_aD adelta rest)
  | amin_aD adelta []        = aF.bot()

fun amin_vaD adelta (d::rest)   = aF.join(amin_aD adelta d, amin_vaD adelta rest)
  | amin_vaD (adelta : aLog) [] = aF.bot()


fun ==> (a,b) = (a,b)

infix ==> 




(* Contour collection. *)

structure PowaTime = Power(structure Member = aTime)

structure PowaVal = Power(structure Member = aVal)


fun live_contours_LAM (lam, aB, ave, asigma, visited) =
    let
	val freeVars = freeVarsInLAM lam

	val local_contours = map aB freeVars

	val vad = map (fn v => (A ave aB aTime.at0 (VarE v))) freeVars
    in
	local_contours @ (live_contours_aDs (vad,ave,asigma,visited))
    end

and live_contours_aDs (ad::rest, ave, asigma, visited) = 
    let
	val local_contours_1 
	  = live_contours_aD  (ad, ave, asigma, visited)
	val local_contours_2 
	  = live_contours_aDs (rest, ave, asigma, visited)
    in
	local_contours_1 @ local_contours_2
    end
  | live_contours_aDs ([], ave, asigma, visited) = []


and live_contours_aD (v::rest, ave, asigma,visited) 
  = (live_contours_aVal (v,ave,asigma,visited)) @ 
    (live_contours_aD (rest,ave,asigma,visited))

  | live_contours_aD ([],ave,asigma,visited) = []


and live_contours_aVal (v as aLocV loc, ave, asigma,visited) = 
    if PowaVal.isin(v,!visited)
    then []
    else (visited := v::(!visited);
	  live_contours_aD (asigma loc, ave, asigma, visited))

  | live_contours_aVal (v as aProcV (aCloP (lam,aB_b,at_b)), 
			ave, asigma, visited) = 
    if PowaVal.isin(v,!visited)
    then []
    else (visited := v::(!visited);
	  at_b::(live_contours_LAM (lam, aB_b, ave, asigma,visited)))

  | live_contours_aVal (v as aProcV (HaltP), 
			ave, asigma, visited) = [aTime.at0]
  | live_contours_aVal _ = []





(* State gradient. *)

structure aNabla = FunctionLattice(structure Domain = aConf
				   structure Range = aF)

structure aNablaLAB = FunctionLattice(structure Domain = CallLAB
				      structure Range = aF)

structure aNablaVAR = 
struct

structure aNV = FunctionLattice(structure Domain = VAR
			      structure Range = aF)

open aNV

fun cat(anv1, af) = memoise(fn v => aF.cat(anv1 v, af))

end




(* A class 1 state gradient. *)

fun gen_nabla_LAM (varDelta : aNablaVAR.super) 
		  (FunL lab, varfs, [vcont], call) =
    let
	(* val _ = Might.debug ("NABLA: Considering LAM: " ^ lab) *)

	val varDelta' = 
	    (aNablaVAR.cat((aNablaVAR.update 
				varDelta
				(aNablaVAR.fromPairs [vcont ==> aF.empty])),
			   aF.down (FunL lab)))
    in
	aNablaLAB.memoise (gen_nabla_CALL varDelta' call)
    end

  | gen_nabla_LAM varDelta 
		  (ConL lab, varfs, [], call) =
    let
	(* val _ = Might.debug ("NABLA: Considering LAM: " ^ lab) *)

	val varDelta' = 
	    (aNablaVAR.cat(varDelta,
			   aF.down (ConL lab)))
    in
	aNablaLAB.memoise(gen_nabla_CALL varDelta' call)
    end

  | gen_nabla_LAM varDelta 
		  (FunL lab, varfs, varcs, call) =
    let
	(* val _ = Might.debug ("NABLA: Considering LAM: " ^ lab) *)
		
	(* TODO: Set all varcs to top. 
	 * Not necessary in an alphatised program.
	 *)
	val varDelta' = 
	    (aNablaVAR.cat(varDelta,
			   aF.down (FunL lab)))
    in
	aNablaLAB.memoise(gen_nabla_CALL varDelta' call)
    end
  | gen_nabla_LAM varDelta
		  (ConL lab, varfs, hd::tl, call) 
    = 
    let
	val _ = print ("We have a kont lambda with a kont var: " ^
		hd
		^ "\n")
    in
	raise Impossible
    end
	

 


and gen_nabla_CALL varDelta 
		   (Call (FunL kappa, expf, expfs, [VarE vcont])) =
    let
	(* val _ = Might.debug ("NABLA: Considering: " ^ kappa) *)

	val Delta_p = aF.inv(varDelta vcont)

	(* val _ = Might.debug ("Delta_p = " ^ (aF.toString Delta_p)) *)

	val local_constraint = aNablaLAB.update 
				   (aNablaLAB.top()) 
				   (aNablaLAB.fromPairs 
					[FunL kappa ==> Delta_p])

	(*
	 val _ = Might.debug ("local = " ^ (aNablaLAB.toString local_constraint))
	 *)
				   
	val constraints = foldl (fn ((LamE lam),c) => aNablaLAB.meet(c,gen_nabla_LAM (aNablaVAR.top()) lam)
				  | (_,c) => c) 
				(aNablaLAB.top()) 
				expfs
				
	val varDelta' = aNablaVAR.cat(varDelta,Delta_p)

	(* TODO: varDelta' = varDelta (+) Delta_p ? *)
	val expf_constraint = case expf of (LamE lam) => 
					   gen_nabla_LAM varDelta' lam
					 | _ => aNablaLAB.top()
    in
	aNablaLAB.memoise(aNablaLAB.meet(aNablaLAB.meet(constraints,
							expf_constraint),
					 local_constraint))
    end


  | gen_nabla_CALL varDelta (Call (FunL kappa, expf, expfs, expcs)) = 
    let
	val Delta_p = if List.exists (fn (LamE lam) => true
				       | _ => false) expcs
		      then aF.empty
		      else aF.top()

	(*
	 val _ = Might.debug ("NABLA: Considering (1): " ^ kappa)
	 *)
		

	val varDelta' = aNablaVAR.cat(varDelta,Delta_p)
				      

	val local_constraint = aNablaLAB.update 
				   (aNablaLAB.top())
				   (aNablaLAB.fromPairs 
					[FunL kappa ==> Delta_p])
				   
	val expcs_constraints =
	    foldl (fn ((LamE lam),c) => 
		      aNablaLAB.meet(c,gen_nabla_LAM varDelta' lam)
		    | (_,c) => c) 
		  (aNablaLAB.top()) 
		  expcs
	
	val expfs_constraints = 
	    foldl (fn ((LamE lam),c) => 
		      aNablaLAB.meet(c,gen_nabla_LAM(aNablaVAR.top()) lam)
		    | (_,c) => c) 
		  (aNablaLAB.top()) 
		  expfs
				
	val expf_constraint = case expf of (LamE lam) => 
					   gen_nabla_LAM varDelta' lam
					 | _ => aNablaLAB.top()
    in
	aNablaLAB.memoise(
			aNablaLAB.meet(aNablaLAB.meet(expcs_constraints,
						      expfs_constraints),
				       aNablaLAB.meet(local_constraint,
						      expf_constraint)))
    end


  | gen_nabla_CALL varDelta 
		   (Call (ConL kappa, expc, expfs, [])) = 
    let
	(* val _ = Might.debug ("NABLA: Considering: " ^ kappa) *)

	val Delta_p = case expc of (LamE lam) => aF.empty
				 | (VarE v) => aF.inv(varDelta v)
				 | _ => aF.top()

	(* val _ = Might.debug ("Delta_p = " ^ (aF.toString Delta_p)) *)

	val varDelta' = aNablaVAR.cat(varDelta,Delta_p)

	val local_constraint = aNablaLAB.update
			       (aNabla.top())
			       (aNablaLAB.fromPairs 
				    [ConL kappa ==> Delta_p])

	val expc_constraint = case expc of (LamE lam) => 
					   gen_nabla_LAM varDelta' lam
					 | _ => aNablaLAB.top()
			       
	val expfs_constraints = 
	    foldl (fn ((LamE lam),c) => 
		      aNablaLAB.meet(c,gen_nabla_LAM(aNablaVAR.top()) lam)
		    | (_,c) => c) 
		  (aNablaLAB.top()) 
		  expfs			       
    in
	aNablaLAB.memoise(
             aNablaLAB.meet(expc_constraint,
			    aNablaLAB.meet(local_constraint, 
					   expfs_constraints)))
    end
  | gen_nabla_CALL varDelta 
		   (LetRec (ConL kappa, varlamfs, varlamcs, call)) =
    let
	(* val _ = Might.debug ("NABLA: Considering: " ^ kappa) *)

	val Delta_p = aF.down (ConL kappa)

	val local_constraint = aNablaLAB.update
				   (aNabla.top())
				   (aNablaLAB.fromPairs 
					[ConL kappa ==> Delta_p])

	val varDelta' = 
	    aNablaVAR.update
		(aNablaVAR.cat(varDelta,Delta_p))
		(aNablaVAR.fromPairs(map (fn (vi,lamci) => 
					     (vi ==> aF.empty)) 
					 varlamcs))
		
	val lamc_constraints = 
	    foldl (fn ((v,(LamE lam)),c) => 
		      aNablaLAB.meet(c,gen_nabla_LAM varDelta' lam)
		    | (_,c) => c) 
		  (aNablaLAB.top()) 
		  varlamcs
	    

	val lamf_constraints = 	    
	    foldl (fn ((v,(LamE lam)),c) => 
		      aNablaLAB.meet(c,gen_nabla_LAM(aNablaVAR.top()) lam)
		    | (_,c) => c) 
		  (aNablaLAB.top()) 
		  varlamfs
		  
	val call_constraints = 
	    gen_nabla_CALL varDelta' call
    in
	aNablaLAB.memoise(
	  aNablaLAB.meet(aNablaLAB.meet(local_constraint,
					call_constraints),
			 aNablaLAB.meet(lamc_constraints,
					lamf_constraints)))
    end



fun nablifier (nablab : aNablaLAB.super) 
	      (Eval(Call(kappa,_,_,_), _, _, _, _, _)) = nablab(kappa)
  | nablifier (nablab) 
	      (Eval(LetRec(kappa,_,_,_), _, _, _, _, _)) = nablab(kappa)
  | nablifier (nablab) _ = aF.top()




fun make_state_gradient_2 pr =
    nablifier (aNablaLAB.memoise((gen_nabla_LAM (aNablaVAR.top()) pr)))


(* A class 0 state gradient. *)

fun state_gradient_1 (Eval(Call(kappa, (LamE (ConL _,_,_,_)), _, _), 
		     _, _, _, _, _)) =
    let
	(* Let-Cont? *)
    in
	aF.empty
    end
  | state_gradient_1 (Eval(Call(kappa, expf, expfs, expcs), 
		     aB, ave, adelta, asigma, at)) =
    let
	(* Will there always be a newborn continuation here? *)
	val has_lam_conts = List.exists (fn (LamE _) => true
					  | _ => false) 
					expcs
    in
	if (has_lam_conts)
	then aF.empty
	else aF.top()
    end
  | state_gradient_1 _ = aF.top()



fun make_state_gradient12 pr =
    aNabla.meet(state_gradient_1,make_state_gradient_2 pr)



(* input(pr) injects pr into a state. *)
fun input pr = Apply( aCloP(pr,aBEnv.bot(),aTime.at0),
		      [],
		      [[aProcV HaltP]],
		      aVEnv.bot(),
		      aLog.fromPairs [aTime.at0 ==> aF.empty] ,
		      aStore.bot(),
		      aTime.at0
		      )
		      
		      
fun make_next nabla =
let

fun next (c as Eval(Call(l, expf, expfs, expcs), 
		    aB, ave, adelta, asigma, at)) =
    let
	val _ = Might.debug ("\n*** In Call:")

	val _ = Might.debug (aConf.toString c)

	(* Remove non-procedures as transition candidates.  *)
	val all_procs = (A ave aB at expf)

	val aprocs = List.filter (fn aProcV p => true | _ => false) 
				 all_procs

	val _ = case aprocs of [] => Might.debug "Warning: no valid transitions."
			     | _ => ()

	val vadf = map (A ave aB at) expfs
	val vadc = map (A ave aB at) expcs
		   
	val min_aF = amin_vaD adelta ([all_procs] @ vadc)

	val _ = Might.debug "inv(min_aF) = "

	val _ = Might.debug (aF.toString (aF.inv min_aF))

	val nabla_c = nabla c

	val _ = Might.debug "nabla_c = "

	val _ = Might.debug (aF.toString nabla_c) 

	val DeltaP = aF.meet (aF.inv(min_aF), nabla_c) 
		      

	val _ = Might.debug "DeltaP = "

	val _ = Might.debug (aF.toString DeltaP)

	(* val DeltaP = aF.inv(min_aF) *)

	val adelta' = aLog.memoise (aLog.cat(adelta, fn at => DeltaP))

	val _ = Might.debug "adelta' = "

	val _ = Might.debug (aLog.toString adelta')
    in
	map (fn (aProcV aproc) => (Apply(aproc, vadf, vadc, ave, adelta', asigma, at))) aprocs
    end

  | next (c as Apply( aCloP(lam as (l, varfs, varcs, call),aB_b,at_b),
		      vadf,
		      vadc,
		      ave,
		      adelta,
		      asigma,
		      at ))  =
    let
	val _ = Might.debug "\n*** In Apply:"

	val _ = Might.debug (aConf.toString c)

	val visited = ref []

	val live_contours 
	  = PowaTime.Union [live_contours_LAM (lam, aB_b, ave, asigma, 
					       visited),
			    live_contours_aDs (vadf,ave,asigma,visited),
			    live_contours_aDs (vadc,ave,asigma,visited)]


	val _ = Might.debug ("Live contours: " ^ 
			     (PowaTime.toString live_contours))


	val DeltaP = aF.down l

	val at' = aTime.atick(at,l,adelta(aTime.at0), adelta (aTime.at0) l)

	val _ = 
	    if (not (PowaTime.isin(at',live_contours)))
	    then Might.debug ("gc'd contour: " ^ (aTime.toString at'))
	    else Might.debug ("un-gc'd contour: " ^ (aTime.toString at'))

	val merge_aVEnv = if (not (PowaTime.isin(at',live_contours)))
			  then fn (ve1,ve2) => aVEnv.update ve1 ve2
			  else aVEnv.join
			       
	val merge_aLog = if (not (PowaTime.isin(at',live_contours)))
			 then fn (adelta1,adelta2) => 
				 aLog.update adelta1 adelta2
			 else aLog.join
	
	val aB = aBEnv.update aB_b
			      (aBEnv.fromPairs ((map (fn v => 
							 (v ==> at')) 
						     (varfs @ varcs))))

	val varfXat = map (fn v => (v,at')) varfs

	val varcXat = map (fn v => (v,at')) varcs

	val ave_up_df = ListPair.zip (varfXat, vadf)

	val ave_up_dc = ListPair.zip (varcXat, vadc)

	val ave' 
	  = aVEnv.memoise(merge_aVEnv(ave,
				      (aVEnv.fromPairs(ave_up_df @ 
						       ave_up_dc))))

	val adelta' = merge_aLog (aLog.cat(adelta, fn at => DeltaP), 
				  aLog.fromPairs [at' ==> aF.empty])


	val _ = Might.debug "In Apply, adelta' = "

	val _ = Might.debug (aLog.toString adelta')
    in
	[Eval(call, aB, ave', adelta', asigma, at')]
    end
  | next (Apply(HaltP, vadf, vadc, ave, adelta, asgima, at)) = []
    

  | next (Eval(LetRec(l, varexpfs, varexpcs, call),
	       aB,
	       ave,
	       adelta,
	       asigma,
	       at)) =
    let
	val at' = aTime.atick(at,l,adelta(aTime.at0), adelta (aTime.at0) l)
		  
	val aB' = aBEnv.update aB
		  (aBEnv.fromPairs ((map (fn (v,e) => (v ==> at')) (varexpfs @ varexpcs))))
		  
	(* Bug? at' or at for evaluating the lambdas. *)
		  
	val ave_up_f = map (fn (v,e) => ((v,at') ==> (A ave aB' at' e))) (varexpfs @ varexpcs)

	val ave' = aVEnv.join (ave,  (aVEnv.fromPairs ave_up_f))

	val DeltaP = aF.down l

	val adelta_cat_DeltaP = aLog.cat (adelta, fn _ => DeltaP)

	val adelta' = aLog.join (adelta_cat_DeltaP, aLog.fromPairs [at' ==> aF.empty])

	val _ = Might.debug("DeltaP = ") ;

	val _ = Might.debug (aF.toString DeltaP)

	val _ = Might.debug ("adelta_cat_DeltaP = ")

	val _ = Might.debug (aLog.toString adelta_cat_DeltaP)

	val _ = Might.debug ("adelta' = ")

	val _ = Might.debug (aLog.toString adelta')
    in
	[Eval(call, aB', ave', adelta', asigma, at')]
    end

  | next (Apply(PrimP MINUS,
		[ad_lhs, ad_rhs],
		[ad_cont],
		ave,
		adelta,
		asigma,
		at)) =
    let
	val aprocs = List.filter (fn aProcV p => true | _ => false)  ad_cont
    in
	map (fn (aProcV aproc) => 
		(Apply(aproc, [[aBasV aNumber]], [], ave, adelta, asigma, at))) aprocs
    end

  | next (Apply(PrimP PLUS,
		[ad_lhs, ad_rhs],
		[ad_cont],
		ave,
		adelta,
		asigma,
		at)) =
    let
	val aprocs = List.filter (fn aProcV p => true | _ => false)  ad_cont
    in
	map (fn (aProcV aproc) => 
		(Apply(aproc, [[aBasV aNumber]], [], ave, adelta, asigma, at))) aprocs
    end

  | next (Apply(PrimP TIMES,
		[ad_lhs, ad_rhs],
		[ad_cont],
		ave,
		adelta,
		asigma,
		at)) =
    let
	val aprocs = List.filter (fn aProcV p => true | _ => false)  ad_cont
    in
	map (fn (aProcV aproc) => 
		(Apply(aproc, [[aBasV aNumber]], [], ave, adelta, asigma, at))) aprocs
    end

  | next (Apply(PrimP TEST_ZERO,
		[ad_zerop],
		[ad_true, ad_false],
		ave,
		adelta,
		asigma,
		at)) =
    let
	val aprocs = List.filter (fn aProcV p => true | _ => false)  (ad_true @ ad_false)
    in
	map (fn (aProcV aproc) => 
		(Apply(aproc, [], [], ave, adelta, asigma, at))) aprocs
    end

in
    next
end


fun closure next (c::rest) (visited : aVisited) =
    let
	(* val _ = Might.debug ("Checking for weakness: " ^
			     (Int.toString (length visited)) ^
			     " states"
			     ) *)

	(* val _ = Might.debug ( "State: " ^ (aConf.toString c) ^ "\n" ) *)

	val seen = aVisited.wt([c],visited)
		   
	(* val _ = Might.debug "Checked for weakness." *)
    in
	if seen
	then closure next rest visited
	else closure next ((next c) @ rest) (c::visited)
    end
  | closure next [] visited = visited


end






structure Main =
struct

fun println s = (print s; print "\n\n")




open CPS

val LFunL = LAB.freshFunL
val LConL = LAB.freshConL


val CFunL = CLAB.freshFunL
val CConL = CLAB.freshConL


(*
val pr1 = (LFunL("top"),[] : VAR list,["halt"],
	   Call(CConL("in"), VarE "halt", [LitE (Int(3))], [] )) : LAM




val pr2 = (LFunL("top"),[] : VAR list,["halt"],
	   LetRec(LConL("lr"), [], [("fact",
				    (LamE(LConL("fact"), ["a","n"], [], 
					  Call(CFunL("tz"), PrimE TEST_ZERO, 
					       [VarE "n"], 
					       [LamE (LConL "true", 
						      [],
						      [],
						      Call(CConL("hz"), VarE "halt", [VarE "a"], [] )),
						LamE (LConL "false",
						      [],
						      [],
						      Call(CFunL("*"), 
							   PrimE TIMES, 
							   [VarE "a", VarE "n"], 
							   [LamE (LConL "a*n", 
								  ["a*n"],
								  [],
								  (Call(CFunL "-",
									PrimE MINUS,
									[VarE "n", 
									 LitE (Int 1)],
									[LamE (LConL "n-1",
									       ["n-1"],
									       [],
									       Call(CConL("rec"),
										    VarE "fact",
										    [VarE "a*n",
										     VarE "n-1"],
										    [])
									       )]
									))) ] ))
						] )
					  ))
				    )],
		  Call(CConL("call-fact"), VarE "fact", [LitE (Int 1), LitE (Int 3)], [])))




val pr3 = (LFunL("top"),[] : VAR list,["halt"],
	   LetRec(LConL("lr"),
		  [("id",
		    LamE(LFunL("id"), ["x"], ["c"], 
			 Call(CConL("id:int"), VarE "c", [VarE "x"], [] ) ))],
		  [("f",
		    LamE(LFunL("f"), ["n"], ["k"],
			 Call(CFunL("tz:"), PrimE TEST_ZERO, 
			      [VarE "n"],
			      [LamE(LConL "if-0",
				    [],[],
				    Call(CConL "hz", VarE "k", [LitE(Int 1)], [])),
			       LamE(LConL "if-!0",
				    [],[],
				    Call(CFunL "call-id", VarE "id", 
					 [LamE(LFunL "lamk'",[],["k'"],
					       Call(CFunL "-:", PrimE MINUS, 
						    [VarE "n",LitE(Int 1)],
						    [LamE(LConL "lamn-1",
							  ["n-1"],[],
							  Call(CFunL "f:rec",
							       VarE "f",
							       [VarE "n-1"],
							       [VarE "k'"]))]))],
					 [LamE(LConL "lamg", ["g"], [],
					       Call(CFunL "g:", VarE "g", [], [VarE "k"]))]))
			       ])))],
		  Call(CFunL("f:call"), VarE "f", [LitE(Int(3))], [VarE "halt"])))
	  

*)



open DCFA

(* 
val odd_progn = "((lambda (o) ((lambda (f) ((f f) 3)) (lambda (x) (lambda (y) o)))) 10)"


val odd_progn = "((lambda (f) ((f f) 3)) (lambda (x) x))"

val odd_sexp = SExp.fromString odd_progn

val odd_lcexp = LambdaCalcExp.fromSExp odd_sexp

val odd_pr = CPS.fromLExp odd_lcexp

val _ = print (CPS.stringOfPR odd_pr)

val _ = print "\n"

val _ = print (CPS.schemeOfPR odd_pr)

val _ = print "\n"

*)

(*
val progn =
"(letrec ((fact (lambda (n) " ^
                 "(test-zero n " ^
                            "1 " ^
                            "(* n (fact (- n 1)))))))" ^
  "(fact 3))"

val sexp = SExp.fromString progn

val lcexp = LambdaCalcExp.fromSExp sexp

val cps = CPS.fromLExp lcexp

val _ = print (CPS.stringOfPR cps)

val _ = print "\n"

val _ = print (CPS.schemeOfPR cps)

val _ = print "\n"

*) *)


(*val progn =
    "((lambda (f) ((f f) 0)) (lambda (x) x))" 

val sexp = SExp.fromString progn

val lcexp = LambdaCalcExp.fromSExp sexp
*)

val pr = pr_transduce

val _ = print (CPS.stringOfPR pr)

val _ = print "\n"

val _ = print (CPS.schemeOfPR pr)

val _ = print "\n"



val PROG = pr


val c = input PROG

val nabla12 = make_state_gradient12 PROG

val aV = closure (make_next nabla12) [c] []  

val aV_strings = (map aConf.toStringLn aV) 

val _ = print "\n----\n"

val _ = print ("\n\nStates visited: " ^ (Int.toString(length aV)) ^ " \n")

val _ = app println aV_strings

val _ = print ("\n\nStates visited: " ^ (Int.toString(length aV)) ^ " \n")



val _ = Might.debug (PR.toString PROG)

val _ = Might.debug (CPS.schemeOfPR PROG)


end
