

structure Might =
struct

fun debugExp s x = (print s; print "\n"; x)

fun debug s = (print s; print "\n")



exception NotCached

fun memoise (eq : 'a * 'a -> bool) (f : 'a -> 'b) = 
    let
	val cache = ref ((fn (x' : 'a) => raise NotCached) : 'a -> 'b)
    in
	fn (x : 'a) =>
	   let
	       val fcache = !cache

	       val is_cached = ((fcache x; true) 
				handle NotCached => false)
	   in
	       if (is_cached)
	       then (fcache x)
	       else 
		   let
		       val y = f x
			    
		       val _ = cache := (fn x' => if eq(x,x') then y else fcache x')
		   in
		       y
		   end
	   end
    end

fun flatten ((hd::tl)::rest) = hd::(flatten ((tl)::rest))
  | flatten (([])::rest) = flatten rest
  | flatten [] = []

end







structure LazyList =
struct


datatype 'a llist = MORE of 'a * (unit -> 'a llist)
		  | END 

fun fromList (hd::tl) = MORE(hd,fn () => (fromList tl))
  | fromList [] = END

fun toList (MORE(hd,tl)) = hd::(toList(tl()))
  | toList (END) = []

fun map f (MORE(hd,tl)) = MORE(f hd,fn () => map f (tl()))
  | map f END = END

fun app p (MORE(hd,tl)) = (p hd; app p (tl()))
  | app p END = ()


fun fold f init (MORE(hd,tl)) = fold f (f(hd,init)) (tl())
  | fold f init END = init

fun forAll (MORE(hd,tl)) p = if (p hd)
			     then forAll (tl()) p
			     else false
  | forAll (END) p = true

fun filter (MORE(hd,tl)) p = if (p hd)
			     then MORE(hd,fn () => filter (tl()) p)
			     else filter (tl()) p
  | filter (END) p = END


fun multiply (x, MORE(hd,tl)) = MORE((x,hd), fn () => multiply(x,(tl())))
  | multiply (x, END) = END
    

fun append (MORE(hd1,tl1), ll2) = MORE(hd1, fn () => (append (tl1(), ll2)))
  | append (END, ll2) = ll2
     

fun crossLazyList (MORE(hd1, tl1), ll2) = 
    let 
	val hd1_by_ll2 = multiply(hd1,ll2)
	val rest = crossLazyList(tl1(), ll2)
    in
	append(hd1_by_ll2, rest)
    end
  | crossLazyList (END, _) = END

fun toStringRest sep toStr (MORE(hd1,tl1)) = sep ^ (toStr hd1) ^ (toStringRest sep toStr (tl1()))
  | toStringRest sep toStr (END) = "" 

fun toString sep toStr (MORE(hd1, tl1)) = (toStr hd1) ^ (toStringRest sep toStr (tl1()))
  | toString sep toStr (END) = ""

fun cons(hd,tl) = MORE(hd,fn () => tl)

end




signature SET =
sig
    type super

    val is_member : super -> bool

    val eq : super * super -> bool

    val members : unit -> super LazyList.llist

    val toString : super -> string
end



signature PARTIAL_ORDER =
sig
    type super

    val is_member : super -> bool

    val eq : super * super -> bool
    val wt : super * super -> bool

    val bot : unit -> super
    val top : unit -> super

    exception Top
    exception Bottom

    val members : unit -> super LazyList.llist

    val toString : super -> string
end




signature LATTICE =
sig
    type super

    val is_member : super -> bool

    val eq : super * super -> bool
    val wt : super * super -> bool
    val st : super * super -> bool

    val join : super * super -> super
    val meet : super * super -> super

    val bot : unit -> super
    val top : unit -> super

    exception Top
    exception Bottom

    val members : unit -> super LazyList.llist

    val toString : super -> string
end




signature POWER_SET =
sig
    structure Member : SET

    type super

    val is_member : super -> bool

    val eq : super * super -> bool
    val subset_eq : super * super -> bool
					 
    val isin : Member.super * super -> bool

    val union : super * super -> super
    val Union : super list -> super
    val intersect : super * super -> super

    val toList : super -> Member.super list
    val fromList : Member.super list -> super

    val forAll : super -> (Member.super -> bool) -> bool
    val thereExists : super -> (Member.super -> bool) -> bool

    val members : unit -> super LazyList.llist
					  
    val toString : super -> string
end





signature POWER_LATTICE =
sig
    structure Member : LATTICE

    type super 

    val is_member : super -> bool

    val eq : super * super -> bool
    val subset_eq : super * super -> bool
    val wt : super * super -> bool
    val st : super * super -> bool

    val join : super * super -> super
    val meet : super * super -> super

    val bot : unit -> super
    val top : unit -> super

    exception Top
    exception Bottom
					 
    val isin : Member.super * super -> bool

    val union : super * super -> super
    val Union : super list -> super
    val intersect : super * super -> super

    val toList : super -> Member.super list
    val fromList : Member.super list -> super

    val isEmpty : super -> bool

    val choose : super -> (Member.super * super)

    val foldl : ((Member.super * 'b) -> 'b) -> 'b -> super -> 'b
    val foldr : ((Member.super * 'b) -> 'b) -> 'b -> super -> 'b
    val map : (Member.super -> 'b) -> super -> 'b list

    val forAll : super -> (Member.super -> bool) -> bool
    val thereExists : super -> (Member.super -> bool) -> bool

    val empty : super

    val members : unit -> super LazyList.llist

    val toString : super -> string
end



(* Standard Domains *)

structure StringSet =
struct

type super = string

val eq = (op =) : string * string -> bool

fun is_member (x : string) = true

exception Unenumerable

fun members () = raise Unenumerable

fun toString x = x

end



(* Domain Constructors *)



functor FunctionLattice(structure Domain : SET
			structure Range  : LATTICE) =
struct

exception Top
exception Bottom

structure Domain = Domain
structure Range  = Range

type super = Domain.super -> Range.super

fun is_member (f : super) = true



fun is_range_bot thunk = 
    let
	(* Safely evaluate the thunk to see if it's the Range's bottom. *)
	val hasBottom = ((Range.bot(); true)
			 handle Range.Bottom => false)
    in
	if (hasBottom)
	then Range.eq(thunk(), Range.bot())
	else ((thunk(); false)
	      handle Range.Bottom => true
		   | Range.Top => false)
    end
	
	
fun is_range_top thunk = 
    let
	(* Safely evaluate the thunk to see if it's the Range's top. *)
	val hasTop = ((Range.top(); true)
		      handle Range.Top => false)
    in
	if (hasTop)
	then Range.eq(thunk(), Range.top())
	else ((thunk(); false)
	      handle Range.Top => true
		   | Range.Bottom => false)
    end

fun is_top f x =
    let
	val thunk = (fn () => f x)
    in
	is_range_top thunk
    end



fun in_domain f x = if (is_range_bot(fn () => f(x)))
		    then false
		    else true
			    


fun eq(f1,f2) = LazyList.forAll (Domain.members())
				(fn x =>
				    if (in_domain f1 x) andalso (in_domain f2 x)
				    then Range.eq(f1 x,f2 x)
				    else if (not (in_domain f1 x)) andalso (not (in_domain f2 x))
				    then true
				    else false)
				
fun wt(f1,f2) = LazyList.forAll (Domain.members())
				(fn x => 
				    if (in_domain f1 x) andalso (in_domain f2 x)
				    then Range.wt(f1 x, f2 x)
				    else if (in_domain f1 x) andalso (not (in_domain f2 x))
				    then false
				    else if (not (in_domain f1 x)) andalso (in_domain f2 x)
				    then true
				    else true)

fun st(f1,f2) = LazyList.forAll (Domain.members())
				(fn x => 
				    if (in_domain f1 x) andalso (in_domain f2 x)
				    then Range.st(f1 x, f2 x)
				    else if (in_domain f1 x) andalso (not (in_domain f2 x))
				    then true
				    else if (not (in_domain f1 x)) andalso (in_domain f2 x)
				    then false
				    else true)



fun memoise f = Might.memoise Domain.eq f


val my_top = fn x => Range.top()
val my_bot = fn x => Range.bot()


fun top () = my_top
fun bot () = my_bot



fun join (f1,f2) = fn x => if (in_domain f1 x) andalso (in_domain f2 x)
			   then Range.join (f1 x, f2 x)
			   else if (not (in_domain f1 x)) andalso (in_domain f2 x)
			   then f2(x)
			   else if (in_domain f1 x) andalso (not (in_domain f2 x))
			   then f1(x)
			   else Range.bot()


fun meet (f1,f2) = fn x => if (is_top f1 x) andalso (is_top f2 x)
			   then Range.top()
			   else if (is_top f1 x)
			   then f2(x)
			   else if (is_top f2 x)
			   then f1(x)
			   else Range.meet(f1 x, f2 x)


				
exception Unenumerable

fun update f g = fn x => if (in_domain g x) then g(x) else f(x)


fun fromPairs ((x,y)::rest) = (fn x' => if (Domain.eq(x,x'))
					then y
					else (fromPairs rest) x') 
  | fromPairs [] = bot()


fun partial_domain f = LazyList.filter (Domain.members())
		       (fn x => 
			   (if (in_domain f x)
			    then true
			    else false))

fun members() = raise Unenumerable


fun toString (f : super) = "[" ^ (LazyList.toString ", " 
						    (fn x => (Domain.toString x) ^ 
							     "==>" ^ 
							     (if (is_top f x) 
							      then "T"
							      else (Range.toString (f x))))
						    (partial_domain f) ) ^ "]"



end






functor FlatLattice(structure Set : SET) : LATTICE =
struct

exception Top
exception Bottom

structure Set = Set

type super = Set.super

val is_member = Set.is_member


val eq = Set.eq
fun wt(a : super, b : super) = if (eq(a,b)) 
			       then true
			       else false
val st = wt


fun top() = raise Top
fun bot() = (Might.debug "Set Bottom!"; raise Bottom)

fun join (a,b) = if eq(a,b) then a else top()
fun meet (a,b) = if eq(a,b) then a else bot()

val members = Set.members

val toString = Set.toString

end





functor InfiniteSet(type super 
		    val eq : super * super -> bool
		    val toString : super -> string) : SET =
struct

type super = super

fun is_member (x : super) = true

val eq = eq

exception Unenumerable

fun members() = raise Unenumerable

val toString = toString

end			  





functor FiniteSet(type super 
		  val eq : super * super -> bool
		  val memlist : super list
		  val toString : super -> string) : SET =
struct

type super = super

fun is_member x = List.exists (fn x' => eq(x,x')) memlist

val eq = eq

fun members() = LazyList.fromList(memlist)

val toString = toString

end




functor FiniteSubset(structure SuperSet : SET
		     val memlist : SuperSet.super list) =
struct

structure SuperSet = SuperSet

type super = SuperSet.super

val eq = SuperSet.eq 

fun is_member x = List.exists (fn x' => eq(x,x')) memlist

fun members() = LazyList.fromList(memlist)

val toString = SuperSet.toString

end




functor SequenceLattice(structure Member : LATTICE) : LATTICE =
struct

structure Member = Member

type super = Member.super list

fun is_member (x : super) =  true

fun eq(x1::l1,x2::l2) = Member.eq(x1,x2) andalso eq(l1,l2)
  | eq([],x2::l2) = false
  | eq(x1::l1,[]) = false
  | eq([],[]) = true


fun wt(x1::l1,x2::l2) = if (Member.wt(x1,x2))
			then wt(l1,l2)
			else false
  | wt([],x2::l2) = true
  | wt(x1::l1,[]) = false
  | wt([],[]) = true

fun st(x1::l1,x2::l2) = if (Member.st(x1,x2))
			then st(l1,l2)
			else false
  | st([],x2::l2) = false
  | st(x1::l1,[]) = true
  | st([],[]) = true

exception Unenumerable

exception Top
exception Bottom

fun top () = raise Top
fun bot () = (Might.debug "Seq Bottom!"; raise Bottom)


fun join (x1::l1, x2::l2) = (Member.join(x1,x2))::join(l1,l2)
  | join (x1::l1, []) = x1::l1
  | join ([], x2::l2) = x2::l2
  | join ([],[]) = []


fun meet (x1::l1, x2::l2) = (Member.meet(x1,x2))::join(l1,l2)
  | meet (x1::l1, []) = x1::l1
  | meet ([], x2::l2) = x2::l2
  | meet ([],[]) = []


fun members () = raise Unenumerable

fun toStringHelp (x::l) = "," ^ (Member.toString x) ^ (toStringHelp l)
  | toStringHelp [] = ""

fun toString (x::l) = "[" ^ (Member.toString x) ^ (toStringHelp l) ^ "]"
  | toString [] = "[]" 

end








functor Power(structure Member : LATTICE) =
struct

exception Top
exception Bottom
	  
structure Member = Member

type super = Member.super list


fun isin (e, D) = List.exists (fn e' => Member.eq(e,e')) D

fun subset_eq (D, D') = List.all (fn e => isin(e,D')) D

fun eq (D,D') = subset_eq(D,D') andalso subset_eq(D',D)


fun wt (D,D') = List.all (fn e => 
			     List.exists (fn e' => 
					     Member.wt(e,e') ) D') D

fun st (D,D') = List.all (fn e => 
			     List.exists (fn e' => 
					     Member.st(e,e') ) D') D

fun join (D,D') =
    let
	val ub = D @ D'

	fun build_lub (e::C) = if isin(e,C)
			       then (build_lub C)
			       else e::(build_lub C)
	  | build_lub [] = []
    in
	build_lub ub
    end

fun meet (D,D') =
    let
	val ub = D @ D'

	fun build_glb (e::C,A) = if (isin(e,D) andalso 
				     isin(e,D') andalso
				     (not (isin(e,A))))
			       then (build_glb (C,e::A))
			       else (build_glb (C,A))
	  | build_glb ([],A) = A
    in
	build_glb (ub,[])
    end

val union = join
val intersect = meet

fun Join (UD : super list) = foldl (fn (D,C) => join(D,C)) [] UD

val Union = Join
			       
			       

fun bot () = [] : super

fun top () = raise Top

val empty = bot()

exception Empty

fun choose (e::D) = (e,D)
  | choose [] = raise Empty

fun isEmpty [] = true
  | isEmpty _ = false

val map = map : (Member.super -> 'b) -> super -> 'b list
val foldl = foldl : ((Member.super * 'b) -> 'b) -> 'b -> super -> 'b
val foldr = foldr : ((Member.super * 'b) -> 'b) -> 'b -> super -> 'b

fun forAll D p = List.all p D
fun thereExists D p = List.exists p D

fun is_member (S : super) = List.all (Member.is_member) S

fun toList (e : super) = (e : Member.super list)

(* WARNING: This could be buggy if there are duplicate or weaker entries. *)
fun fromList (e : Member.super list) = (e : super)



local 
    open LazyList

    fun pow(MORE(e,tl)) = 
	let
	    val pow_l = pow(tl())
				       
	    val pow_l_union_e = 
		LazyList.map (fn S => e::S) pow_l
	in
	    LazyList.append(pow_l,pow_l_union_e)
	end
      | pow(END) = (MORE ([], fn () => END))
	
in

fun members () = pow(Member.members())

end


fun minus (S1,S2) = 
    List.filter (fn s1 => not (isin (s1, S2))) S1



fun toString (e::D) = 
    let
	fun toString_help (e::D) = "," ^ 
				   (Member.toString e) ^ 
				   (toString_help D)
	  | toString_help [] = ""

    in
	"{" ^ 
	(Member.toString e) ^
	(toString_help D) ^
	"}"
    end
  | toString [] = "{}"

end





functor Partial2ProductLattice(structure L1 : LATTICE
			       structure L2 : LATTICE) =
struct

exception Top
exception Bottom

structure L1 = L1
structure L2 = L2

type super = L1.super * L2.super

fun is_member (d1,d2) = (L1.is_member d1) andalso (L2.is_member d2)

fun eq((d1,d2), (d1',d2')) =
    L1.eq(d1,d1') andalso 
    L2.eq(d2,d2')

fun wt((d1,d2), (d1',d2')) = 
    let
    in
	L1.wt(d1,d1') andalso 
	L2.wt(d2,d2')
    end

(* TODO: Fix all st's so that they short-circuits, too. *)
fun st((d1,d2), (d1',d2')) = 
    let
	val b1 = L1.st(d1,d1')
	val b2 = L2.st(d2,d2')
    in
	b1 andalso b2
    end

fun join ((d1,d2), (d1',d2')) = (L1.join(d1,d1'),
				 L2.join(d2,d2'))

fun meet ((d1,d2), (d1',d2')) = (L1.meet(d1,d1'),
				 L2.meet(d2,d2'))

fun bot () = (L1.bot(),L2.bot())
fun top () = (L1.top(),L2.top())

fun members() = LazyList.crossLazyList (L1.members(), L2.members())

fun toString (d1,d2) = "(" ^ 
		       (L1.toString d1) ^ 
		       "," ^ 
		       (L2.toString d2) ^ 
		       ")"

end





functor Partial3ProductLattice(structure L1 : LATTICE
			       structure L2 : LATTICE
			       structure L3 : LATTICE) =
struct

exception Top
exception Bottom

structure L1 = L1
structure L2 = L2
structure L3 = L3

type super = L1.super * L2.super * L3.super

fun is_member (d1,d2,d3) = L1.is_member d1 andalso
			   L2.is_member d2 andalso
			   L3.is_member d3

fun eq((d1,d2,d3), (d1',d2',d3')) = L1.eq(d1,d1') andalso 
				    L2.eq(d2,d2') andalso
				    L3.eq(d3,d3')

fun wt((d1,d2,d3), (d1',d2',d3')) = 
    let
    in
	L1.wt(d1,d1') andalso 
	L2.wt(d2,d2') andalso 
	L3.wt(d3,d3')	
    end


fun st((d1,d2,d3), (d1',d2',d3')) = 
    let
	val b1 = L1.st(d1,d1')
	val b2 = L2.st(d2,d2')
	val b3 = L3.st(d3,d3')
    in
	b1 andalso b2 andalso b3
    end


fun join ((d1,d2,d3), (d1',d2',d3')) = (L1.join(d1,d1'),
					L2.join(d2,d2'),
					L3.join(d3,d3'))

fun meet ((d1,d2,d3), (d1',d2',d3')) = (L1.meet(d1,d1'),
					L2.meet(d2,d2'),
					L3.meet(d3,d3'))

fun bot () = (L1.bot(),L2.bot(),L3.bot())
fun top () = (L1.top(),L2.top(),L3.top())

exception Unenumerable

fun members () = raise Unenumerable


fun toString (d1,d2,d3) = "(" ^ 
			  (L1.toString d1) ^ 
			  "," ^ 
			  (L2.toString d2) ^ 
			  "," ^ 
			  (L3.toString d3) ^ 
			  ")"

end





functor Partial4ProductLattice(structure L1 : LATTICE
			       structure L2 : LATTICE
			       structure L3 : LATTICE
			       structure L4 : LATTICE) =
struct

exception Top
exception Bottom

structure L1 = L1
structure L2 = L2
structure L3 = L3
structure L4 = L4

type super = L1.super * L2.super * L3.super * L4.super


fun is_member (d1,d2,d3,d4) = L1.is_member d1 andalso
			      L2.is_member d2 andalso
			      L3.is_member d3 andalso
			      L4.is_member d4

fun eq((d1,d2,d3,d4), (d1',d2',d3',d4')) = L1.eq(d1,d1') andalso 
					   L2.eq(d2,d2') andalso
					   L3.eq(d3,d3') andalso
					   L4.eq(d4,d4') 

fun wt((d1,d2,d3,d4), (d1',d2',d3',d4')) = 
    let
    in
	L1.wt(d1,d1') andalso 
	L2.wt(d2,d2') andalso 
	L3.wt(d3,d3') andalso 
	L4.wt(d4,d4')
    end

fun st((d1,d2,d3,d4), (d1',d2',d3',d4')) = 
    let
	val b1 = L1.st(d1,d1')
	val b2 = L2.st(d2,d2')
	val b3 = L3.st(d3,d3')
	val b4 = L4.st(d4,d4')
    in
	b1 andalso b2 andalso b3 andalso b4
    end

fun join ((d1,d2,d3,d4), (d1',d2',d3',d4')) = (L1.join(d1,d1'),
					       L2.join(d2,d2'),
					       L3.join(d3,d3'),
					       L4.join(d4,d4'))

fun meet ((d1,d2,d3,d4), (d1',d2',d3',d4')) = (L1.meet(d1,d1'),
					       L2.meet(d2,d2'),
					       L3.meet(d3,d3'),
					       L4.meet(d4,d4'))

fun bot () = (L1.bot(),L2.bot(),L3.bot(),L4.bot())
fun top () = (L1.top(),L2.top(),L3.top(),L4.top())

exception Unenumerable

fun members () = raise Unenumerable

fun toString (d1,d2,d3,d4) = "(" ^ 
			  (L1.toString d1) ^ 
			  "," ^ 
			  (L2.toString d2) ^ 
			  "," ^ 
			  (L3.toString d3) ^ 
			  "," ^ 
			  (L4.toString d4) ^ 
			  ")"

end






functor Partial5ProductLattice(structure L1 : LATTICE
			       structure L2 : LATTICE
			       structure L3 : LATTICE
			       structure L4 : LATTICE
			       structure L5 : LATTICE) =
struct

exception Top
exception Bottom

structure L1 = L1
structure L2 = L2
structure L3 = L3
structure L4 = L4
structure L5 = L5

type super = L1.super * L2.super * L3.super * L4.super * L5.super

fun is_member (d1,d2,d3,d4,d5) = L1.is_member d1 andalso
				 L2.is_member d2 andalso
				 L3.is_member d3 andalso
				 L4.is_member d4 andalso
				 L5.is_member d5 

exception Unenumerable

fun members () = raise Unenumerable


fun eq((d1,d2,d3,d4,d5), (d1',d2',d3',d4',d5')) = L1.eq(d1,d1') andalso 
						  L2.eq(d2,d2') andalso
						  L3.eq(d3,d3') andalso
						  L4.eq(d4,d4') andalso
						  L5.eq(d5,d5')

fun wt((d1,d2,d3,d4,d5), (d1',d2',d3',d4',d5')) = 
    let
    in
	L1.wt(d1,d1') andalso 
	L2.wt(d2,d2') andalso 
	L3.wt(d3,d3') andalso 
	L4.wt(d4,d4') andalso 
	L5.wt(d5,d5')	
    end


fun st((d1,d2,d3,d4,d5), (d1',d2',d3',d4',d5')) = 
    let
	val b1 = L1.st(d1,d1')
	val b2 = L2.st(d2,d2')
	val b3 = L3.st(d3,d3')
	val b4 = L4.st(d4,d4')
	val b5 = L5.st(d5,d5')
    in
	b1 andalso b2 andalso b3 andalso b4 andalso b5
    end

fun join ((d1,d2,d3,d4,d5), (d1',d2',d3',d4',d5')) = (L1.join(d1,d1'),
						      L2.join(d2,d2'),
						      L3.join(d3,d3'),
						      L4.join(d4,d4'),
						      L5.join(d5,d5'))

fun meet ((d1,d2,d3,d4,d5), (d1',d2',d3',d4',d5')) = (L1.meet(d1,d1'),
						      L2.meet(d2,d2'),
						      L3.meet(d3,d3'),
						      L4.meet(d4,d4'),
						      L5.meet(d5,d5'))

fun bot () = (L1.bot(),L2.bot(),L3.bot(),L4.bot(),L5.bot())
fun top () = (L1.top(),L2.top(),L3.top(),L4.top(),L5.top())

fun toString (d1,d2,d3,d4,d5) = "(" ^ 
				(L1.toString d1) ^ 
				"," ^ 
				(L2.toString d2) ^ 
				"," ^ 
				(L3.toString d3) ^ 
				"," ^ 
				(L4.toString d4) ^ 
				"," ^ 
				(L5.toString d5) ^ 
				")"

end






functor Partial6ProductLattice(structure L1 : LATTICE
			       structure L2 : LATTICE
			       structure L3 : LATTICE
			       structure L4 : LATTICE
			       structure L5 : LATTICE
			       structure L6 : LATTICE) =
struct

exception Top
exception Bottom

structure L1 = L1
structure L2 = L2
structure L3 = L3
structure L4 = L4
structure L5 = L5
structure L6 = L6

type super = L1.super * L2.super * L3.super * L4.super * L5.super * L6.super

fun is_member (d1,d2,d3,d4,d5,d6) = L1.is_member d1 andalso
				    L2.is_member d2 andalso
				    L3.is_member d3 andalso
				    L4.is_member d4 andalso
				    L5.is_member d5 andalso
				    L6.is_member d6

exception Unenumerable

fun members () = raise Unenumerable


fun eq((d1,d2,d3,d4,d5,d6), (d1',d2',d3',d4',d5',d6')) = L1.eq(d1,d1') andalso 
							 L2.eq(d2,d2') andalso
							 L3.eq(d3,d3') andalso
							 L4.eq(d4,d4') andalso
							 L5.eq(d5,d5') andalso
							 L6.eq(d6,d6')

fun wt((d1,d2,d3,d4,d5,d6), (d1',d2',d3',d4',d5',d6')) = 
    let
    in
	L1.wt(d1,d1') andalso 
	L2.wt(d2,d2') andalso 
	L3.wt(d3,d3') andalso 
	L4.wt(d4,d4') andalso 
	L5.wt(d5,d5') andalso 
	L6.wt(d6,d6')	
    end

fun st((d1,d2,d3,d4,d5,d6), (d1',d2',d3',d4',d5',d6')) = 
    let
	val b1 = L1.st(d1,d1')
	val b2 = L2.st(d2,d2')
	val b3 = L3.st(d3,d3')
	val b4 = L4.st(d4,d4')
	val b5 = L5.st(d5,d5')
	val b6 = L6.st(d6,d6')
    in
	b1 andalso b2 andalso b3 andalso b4 andalso b5 andalso b6
    end

fun join ((d1,d2,d3,d4,d5,d6), (d1',d2',d3',d4',d5',d6')) = (L1.join(d1,d1'),
							     L2.join(d2,d2'),
							     L3.join(d3,d3'),
							     L4.join(d4,d4'),
							     L5.join(d5,d5'),
							     L6.join(d6,d6'))
							    
fun meet ((d1,d2,d3,d4,d5,d6), (d1',d2',d3',d4',d5',d6')) = (L1.meet(d1,d1'),
							     L2.meet(d2,d2'),
							     L3.meet(d3,d3'),
							     L4.meet(d4,d4'),
							     L5.meet(d5,d5'),
							     L6.meet(d6,d6'))


fun bot () = (L1.bot(),L2.bot(),L3.bot(),L4.bot(),L5.bot(),L6.bot())
fun top () = (L1.top(),L2.top(),L3.top(),L4.top(),L5.top(),L6.top())


fun toString (d1,d2,d3,d4,d5,d6) = "(" ^ 
				     (L1.toString d1) ^ 
				     "," ^ 
				     (L2.toString d2) ^ 
				     "," ^ 
				     (L3.toString d3) ^ 
				     "," ^ 
				     (L4.toString d4) ^ 
				     "," ^ 
				     (L5.toString d5) ^ 
				     "," ^ 
				     (L6.toString d6) ^ 
				     ")"


fun toStringLn (d1,d2,d3,d4,d5,d6) = "(" ^ 
				     (L1.toString d1) ^ 
				     ",\n" ^ 
				     (L2.toString d2) ^ 
				     ",\n" ^ 
				     (L3.toString d3) ^ 
				     ",\n" ^ 
				     (L4.toString d4) ^ 
				     ",\n" ^ 
				     (L5.toString d5) ^ 
				     ",\n" ^ 
				     (L6.toString d6) ^ 
				     ")"
end



functor Partial7ProductLattice(structure L1 : LATTICE
			       structure L2 : LATTICE
			       structure L3 : LATTICE
			       structure L4 : LATTICE
			       structure L5 : LATTICE
			       structure L6 : LATTICE
			       structure L7 : LATTICE) =
struct

exception Top
exception Bottom

structure L1 = L1
structure L2 = L2
structure L3 = L3
structure L4 = L4
structure L5 = L5
structure L6 = L6
structure L7 = L7


type super = L1.super * L2.super * L3.super * L4.super * L5.super * L6.super * L7.super

fun is_member (d1,d2,d3,d4,d5,d6,d7) = L1.is_member d1 andalso
				    L2.is_member d2 andalso
				    L3.is_member d3 andalso
				    L4.is_member d4 andalso
				    L5.is_member d5 andalso
				    L6.is_member d6 andalso
				    L7.is_member d7 

exception Unenumerable

fun members () = raise Unenumerable


fun eq((d1,d2,d3,d4,d5,d6,d7), (d1',d2',d3',d4',d5',d6',d7')) = L1.eq(d1,d1') andalso 
								L2.eq(d2,d2') andalso
								L3.eq(d3,d3') andalso
								L4.eq(d4,d4') andalso
								L5.eq(d5,d5') andalso
								L6.eq(d6,d6') andalso
								L7.eq(d7,d7')
								
fun wt((d1,d2,d3,d4,d5,d6,d7), (d1',d2',d3',d4',d5',d6',d7')) = 
    let
    in
	L1.wt(d1,d1') andalso 
	L2.wt(d2,d2') andalso 
	L3.wt(d3,d3') andalso 
	L4.wt(d4,d4') andalso 
	L5.wt(d5,d5') andalso 
	L6.wt(d6,d6') andalso 
	L7.wt(d7,d7')	
    end

fun st((d1,d2,d3,d4,d5,d6,d7), (d1',d2',d3',d4',d5',d6',d7')) = 
    let
	val b1 = L1.st(d1,d1')
	val b2 = L2.st(d2,d2')
	val b3 = L3.st(d3,d3')
	val b4 = L4.st(d4,d4')
	val b5 = L5.st(d5,d5')
	val b6 = L6.st(d6,d6')
	val b7 = L7.st(d7,d7')
    in
	b1 andalso b2 andalso b3 andalso b4 andalso b5 andalso b6 andalso b7
    end

fun join ((d1,d2,d3,d4,d5,d6,d7), (d1',d2',d3',d4',d5',d6',d7')) = (L1.join(d1,d1'),
								    L2.join(d2,d2'),
								    L3.join(d3,d3'),
								    L4.join(d4,d4'),
								    L5.join(d5,d5'),
								    L6.join(d6,d6'),
								    L7.join(d7,d7'))
								   
							    
fun meet ((d1,d2,d3,d4,d5,d6,d7), 
	  (d1',d2',d3',d4',d5',d6',d7')) = (L1.meet(d1,d1'),
					    L2.meet(d2,d2'),
					    L3.meet(d3,d3'),
					    L4.meet(d4,d4'),
					    L5.meet(d5,d5'),
					    L6.meet(d6,d6'),
					    L7.meet(d7,d7'))


fun bot () = (L1.bot(),L2.bot(),L3.bot(),L4.bot(),L5.bot(),L6.bot(),L7.bot())
fun top () = (L1.top(),L2.top(),L3.top(),L4.top(),L5.top(),L6.top(),L7.top())


fun toString (d1,d2,d3,d4,d5,d6,d7) = "(" ^ 
				   (L1.toString d1) ^ 
				   "," ^ 
				   (L2.toString d2) ^ 
				   "," ^ 
				   (L3.toString d3) ^ 
				   "," ^ 
				   (L4.toString d4) ^ 
				   "," ^ 
				   (L5.toString d5) ^ 
				   "," ^ 
				   (L6.toString d6) ^ 
				   "," ^ 
				   (L7.toString d7) ^ 
				   ")"

fun toStringLn (d1,d2,d3,d4,d5,d6,d7) = "(" ^ 
					(L1.toString d1) ^ 
					",\n" ^ 
					(L2.toString d2) ^ 
					",\n" ^ 
					(L3.toString d3) ^ 
					",\n" ^ 
					(L4.toString d4) ^ 
					",\n" ^ 
					(L5.toString d5) ^ 
					",\n" ^ 
					(L6.toString d6) ^ 
					",\n" ^ 
					(L7.toString d7) ^ 
					")"

end






