module Lattice where

import Data.Set as Set
import InfiniteMap as IMap


 -- A lattice library.
 
class Eq a => Lattice a where
 top  :: a
 bot  :: a
 (<:) :: a -> a -> Bool
 join :: a -> a -> a
 meet :: a -> a -> a

 joins :: [a] -> a

 lfp  :: (a -> a) -> a
 gfp  :: (a -> a) -> a


 -- By default, a lattice is flat.
 top = error "top!"
 bot = error "bot!"
 x <: y = x == y   || 
          x == bot || 
          y == top
 x `join` y | x <: y    = y
            | y <: x    = x
            | otherwise = top
 x `meet` y | x <: y    = x
            | y <: x    = y
            | otherwise = bot
 
 joins []      = bot
 joins [a]     = a
 joins (hd:tl) = hd `join` (joins tl)

 lfp f = lfp' bot
  where lfp' a = if f a == a
                 then a
                 else lfp' (f a)

 gfp f = gfp' top
  where gfp' a = if f a == a
                 then a
                 else gfp' (f a)

 -- K -> A : A map lattice.

instance (Ord k, Lattice a) => Lattice (InfiniteMap k a) where
 bot = constant bot
 top = constant top
 f <: g     = isSubmapOfBy (<:) f g
 f `join` g = lift join f g
 f `meet` g = lift meet f g



 -- A x B : A cartesian product lattice.

instance (Lattice a, Lattice b) => Lattice (a,b) where
 bot = (bot,bot)
 top = (top,top)
 (a1,b1) <: (a2,b2) = (a1 <: a2) ||
                      (a1 == a2 && b1 <: b2) 
 (a1,b1) `join` (a2,b2) = (a1 `join` a2, b1 `join` b2)
 (a1,b1) `meet` (a2,b2) = (a1 `meet` a2, b1 `meet` b2) 




 -- A* : A sequence lattice.

instance (Lattice a) => Lattice [a] where
 top = error "top sequence"
 bot = []
 [] <: [] = True
 (hd1:tl1) <: (hd2:tl2) = hd1 <: hd2 && tl1 <: tl2
 [] <: (hd:tl) = False
 (hd:tl) <: [] = False
 (hd1:tl1) `join` (hd2:tl2) = (hd1 `join` hd2):(tl1 `join` tl2)
 (hd1:tl1) `meet` (hd2:tl2) = (hd1 `meet` hd2):(tl1 `meet` tl2)

                          

 -- P(L) : A power lattice.

allIn :: Set.Set a -> (a -> Bool) -> Bool
allIn set p = Set.fold (\ x soFar -> soFar && p x) True set

existsIn :: Set.Set a -> (a -> Bool) -> Bool
existsIn set p = Set.fold (\ x soFar -> soFar || p x) False set

instance (Ord a, Lattice a) => Lattice (Set a) where
 bot = Set.empty
 top = error "top set"
 setA <: setB = allIn setA (\a -> 
                 existsIn setB (\b ->
                   a <: b))
 setA `join` setB = Set.union setA setB
 setA `meet` setB = Set.fromList [ x | x <- (Set.toList setA) ++ 
                                            (Set.toList setB),
                                       (singleton x) <: setA,
                                       (singleton x) <: setB ] 

 -- A multitude of cross-product lattices.


instance (Lattice a, Lattice b, Lattice c, Lattice d, Lattice e) 
  => Lattice (a,b,c,d,e) where
 (a1,b1,c1,d1,e1) <: (a2,b2,c2,d2,e2) = 
  a1 <: a2 && b1 <: b2 && c1 <: c2 && d1 <: d2 && e1 <: e2 

 (a1,b1,c1,d1,e1) `join` (a2,b2,c2,d2,e2) =
  (a1 `join` a2, b1 `join` b2, c1 `join` c2, d1 `join` d2, e1 `join` e2)

 (a1,b1,c1,d1,e1) `meet` (a2,b2,c2,d2,e2) =
  (a1 `meet` a2, b1 `meet` b2, c1 `meet` c2, d1 `meet` d2, e1 `meet` e2)


instance (Lattice a, Lattice b, Lattice c, Lattice d, Lattice e, Lattice f) 
  => Lattice (a,b,c,d,e,f) where
 (a1,b1,c1,d1,e1,f1) <: (a2,b2,c2,d2,e2,f2) = 
  a1 <: a2 && b1 <: b2 && c1 <: c2 && d1 <: d2 && e1 <: e2 && f1 <: f2 

 (a1,b1,c1,d1,e1,f1) `join` (a2,b2,c2,d2,e2,f2) =
  (a1 `join` a2,
   b1 `join` b2,
   c1 `join` c2,
   d1 `join` d2,
   e1 `join` e2,
   f1 `join` f2)
 (a1,b1,c1,d1,e1,f1) `meet` (a2,b2,c2,d2,e2,f2) =
  (a1 `meet` a2,
   b1 `meet` b2,
   c1 `meet` c2,
   d1 `meet` d2,
   e1 `meet` e2,
   f1 `meet` f2)


instance (Lattice a, Lattice b, Lattice c,  Lattice d, Lattice e, 
          Lattice f, Lattice g) => Lattice (a,b,c,d,e,f,g) where
 (a1,b1,c1,d1,e1,f1,g1) <: (a2,b2,c2,d2,e2,f2,g2) = 
  a1 <: a2 &&
  b1 <: b2 &&
  c1 <: c2 &&
  d1 <: d2 &&
  e1 <: e2 &&
  f1 <: f2 &&
  g1 <: g2 

 (a1,b1,c1,d1,e1,f1,g1) `join` (a2,b2,c2,d2,e2,f2,g2) =
  (a1 `join` a2,
   b1 `join` b2,
   c1 `join` c2,
   d1 `join` d2,
   e1 `join` e2,
   f1 `join` f2,
   g1 `join` g2)

 (a1,b1,c1,d1,e1,f1,g1) `meet` (a2,b2,c2,d2,e2,f2,g2) =
  (a1 `meet` a2,
   b1 `meet` b2,
   c1 `meet` c2,
   d1 `meet` d2,
   e1 `meet` e2,
   f1 `meet` f2,
   g1 `meet` g2)

