module SymbolTable where

import Prelude hiding (lex)

import Data.Set as Set
import InfiniteMap as IMap

import Control.Monad.State

import Might


 -- Symbol table.

type Symbol = Integer

type SymbolMap = IMap.InfiniteMap String Symbol
type StringMap = IMap.InfiniteMap Symbol String

data SymbolTable = ST(Integer,SymbolMap,StringMap)
 deriving (Show)


class SymbolTableMonad m where
 getSymbolTable :: m SymbolTable
 putSymbolTable :: SymbolTable -> m ()


invertSymbolMap im = 
 IMap.foldWithKey (\ _ -> IMap.empty) 
                  (\ s v tally -> tally `but` [(v,s)]) im

emptySymbolTable :: SymbolTable
emptySymbolTable = ST(0,IMap.empty,IMap.empty)

-- reverseLookup (ST(max,symMap)) i = 
--  case IMap.lookup i (invertSymbolMap symMap) of
--   Just str -> str
--   Nothing -> error "no entry for " ++ (show i)

type SymState a = State SymbolTable a

class SymbolTableContainer a where
 extractSymbolTable :: a -> SymbolTable
 injectSymbolTable  :: a -> SymbolTable -> a

instance SymbolTableContainer SymbolTable where
 extractSymbolTable = id
 injectSymbolTable old new = new

instance (SymbolTableContainer a) => SymbolTableMonad (State a) where
 getSymbolTable = do state <- get
                     return $ extractSymbolTable state
 putSymbolTable st = do state <- get
                        put $ injectSymbolTable state st
 

getSymbol :: SymbolTableContainer a => String -> (State a Symbol)
getSymbol str = 
  do (ST(max,tbl,itbl)) <- getSymbolTable
     case IMap.lookup str tbl of 
      Just sym -> do return sym
      Nothing  -> do max'  <- return $ max+1 
                     tbl'  <- return (tbl `but` [str ==> max'])
                     itbl' <- return (itbl `but` [max' ==> str])
                     putSymbolTable $ (ST(max',tbl',itbl'))
                     return max'

symbolExists :: SymbolTableContainer a => Symbol -> (State a Bool)
symbolExists sym =
 do (ST(max,tbl,itbl)) <- getSymbolTable
    return $ sym <= max

symbolForStringExists :: SymbolTableContainer a => String -> (State a Bool)
symbolForStringExists str = 
 do (ST(max,tbl,itbl)) <- getSymbolTable
    case IMap.lookup str tbl of
     Just sym -> do return True
     Nothing  -> do return False

stringFromSymbol :: SymbolTableContainer a => Symbol -> (State a String)
stringFromSymbol sym =
 do (ST(max,tbl,itbl)) <- getSymbolTable
    case IMap.lookup sym itbl of
     Just str -> return str
     Nothing -> return $ "?" ++ (show sym)

freshSymbolFrom :: SymbolTableContainer a => String -> (State a Symbol)
freshSymbolFrom prefix =
  do (ST(max,tbl,itbl)) <- getSymbolTable
     max'  <- return(max+1)
     str   <- return $ prefix ++ "$" ++ (show max')
     tbl'  <- return(tbl `but` [str ==> max'])
     itbl' <- return(itbl `but` [max' ==> str])
     putSymbolTable (ST(max',tbl',itbl'))
     return max'


freshUserSymbol :: SymbolTableContainer a => (State a Symbol)
freshUserSymbol = freshSymbolFrom "u"

freshContSymbol :: SymbolTableContainer a => (State a Symbol)
freshContSymbol = freshSymbolFrom "k"

freshSymbol :: SymbolTableContainer a => (State a Symbol)
freshSymbol = freshSymbolFrom ""

refreshSymbol :: SymbolTableContainer a => Symbol -> (State a Symbol)
refreshSymbol sym =
 do exists <- symbolExists sym
    str' <- stringFromSymbol sym
    True <- return $ exists || 
                      error ("Trying to refresh a fresh symbol: " ++ (show sym) ++ "; " ++ str')
    str <- stringFromSymbol sym
    refreshString str

atoi cur [] = cur
atoi cur ('0':tl) = atoi (cur*10 + 0) tl
atoi cur ('1':tl) = atoi (cur*10 + 1) tl
atoi cur ('2':tl) = atoi (cur*10 + 2) tl
atoi cur ('3':tl) = atoi (cur*10 + 3) tl
atoi cur ('4':tl) = atoi (cur*10 + 4) tl
atoi cur ('5':tl) = atoi (cur*10 + 5) tl
atoi cur ('6':tl) = atoi (cur*10 + 6) tl
atoi cur ('7':tl) = atoi (cur*10 + 7) tl
atoi cur ('8':tl) = atoi (cur*10 + 8) tl
atoi cur ('9':tl) = atoi (cur*10 + 9) tl


incString str = show((atoi 0 str) + 1)

incFresh :: String -> String
incFresh ('~':tl) = '~':(incString tl)
incFresh [] = "~"
incFresh (hd:tl) = hd:(incFresh tl)


refreshString :: SymbolTableContainer a => String -> (State a Symbol)
refreshString str = 
 do exists <- symbolForStringExists str
    (if exists
     then do refreshString (incFresh str)
     else do getSymbol str)


 -- Display is like Show, but it passes around a SymbolTable State monad.

class Display a where
 display         :: SymbolTableContainer b => a -> (State b String)
 displayList     :: SymbolTableContainer b => [a] -> (State b String)
 displayWith     :: SymbolTableContainer b => b -> a -> String

 displayHTML     :: SymbolTableContainer b => a -> (State b String)
 displayHTMLList :: SymbolTableContainer b => [a] -> (State b String)
 
 displayList []  = return "" 
 displayList [a] = display a
 displayList (hd:tl) =
  do hd' <- display hd
     tl' <- displayList tl
     return $ hd' ++ " " ++ tl'

 displayWith st a = evalState (display a) st

 displayHTML x = display x
 displayHTMLList lst = displayList lst

instance Display a => Display ([] a) where
 display lst = displayList lst  
 displayHTML lst = displayHTMLList lst

instance Display a => Display (Set.Set a) where
 display set = do lst <- return $ Set.toList set
                  set' <- displaySetList lst
                  return $ "{" ++ set' ++ "}" 
  where displaySetList [] = return ""
        displaySetList [a] = display a
        displaySetList (hd:tl) =
         do hd' <- display hd
            tl' <- displaySetList tl
            return $ hd' ++ "," ++ tl'

instance (Ord k, Eq a, Display k, Display a) 
  => Display (IMap.InfiniteMap k a) where
 display imap = do lst <- return $ IMap.toList imap
                   map' <- displayIMapList lst
                   return $ "{" ++ map' ++ "}"
  where displayIMapList [] = return ""
        displayIMapList [(Just k,a)] = 
         do k' <- display k
            a' <- display a
            return $ k' ++ ":=" ++ a'
        displayIMapList [(Nothing,a)] =
         do k' <- return $ "_"
            a' <- display a
            return $ k' ++ ":=" ++ a'
        displayIMapList (hd:tl) =
         do hd' <- displayIMapList [hd]
            tl' <- displayIMapList tl 
            return $ hd' ++ "," ++ tl'

instance (Display a,Display b) => Display (a,b) where
 display (a,b) = 
  do a' <- display a
     b' <- display b
     return $ "(" ++ a' ++ "," ++ b' ++ ")"

instance (Display a,Display b,Display c) => Display (a,b,c) where
 display (a,b,c) =
  do a' <- display a
     b' <- display b
     c' <- display c
     return $ "(" ++ a' ++ "," ++ b' ++ "," ++ c' ++ ")"

instance Display Char where
 display c = return $ show c
 displayList str = return $ show str
 
instance Display Integer where 
 display i = return $ show i

instance Display Int where
 display i = return $ show i

instance Display SymbolTable where
 display (ST(max,sm,ism)) = display sm
