module CPS where

import InfiniteMap as IMap
import Data.Set as Set hiding (filter)
import Control.Monad.State hiding (join)
import qualified Data.Map as Map

import Lattice
import DS
import Might
import SymbolTable

data Syn = CallSyn   !CALL
         | LamSyn    !LAM
         | ExpSyn    !EXP
         | InPrimSyn !PRIM !LAB
         | TopSyn 
         | BotSyn
 deriving (Eq,Ord)

instance Show Syn where
 show (CallSyn _) = "CallSyn"
 show (LamSyn _) = "LamSyn"
 show (InPrimSyn _ _) = "InPrimSyn"
 show TopSyn = "TopSyn"
 show BotSyn = "BotSyn"

instance Display Syn where
 display (CallSyn call) = display call
 display (LamSyn lam) = display lam
 display (InPrimSyn prim lab) = 
  do lab'  <- display lab  
     prim' <- display prim
     return $ lab' ++ "@" ++ prim'
 display TopSyn = display "TopSyn"
 display BotSyn = display "BotSyn"


nthVar (LamSyn (Lam _ vs _ )) n = vs !! n
nthVar (CallSyn (LetRec _ vargs _)) n = fst(vargs !! n)
nthVar (TopSyn) n = error "Cannot take nth var of TopSyn."
nthVar (BotSyn) n = error "Cannot take nth var of BotSyn."

type ScopeMap = Map.Map LAB [VAR]

data InlinableExp = InLam    LAM
                  | InLamLab LAB
                  | InLit    LIT
                  | InPrim   PRIM
                  | InRef    VAR
                  | InHalt
                  | TopInExp
                  | BotInExp
 deriving (Eq,Ord)

instance Display InlinableExp where
 display (InLam lam)    = display lam
 display (InLamLab lab) = display lab
 display (InLit lit)    = display lit
 display (InPrim p)     = display p
 display (InRef  v)     = display v
 display (InHalt)       = return "Halt"
 display (TopInExp)     = return "top"
 display (BotInExp)     = return "bot"

instance Lattice InlinableExp where
 top = TopInExp
 bot = BotInExp
 

data SyntaxInfo = 
 SI { bind   :: InfiniteMap VAR LAB ,
      refs   :: InfiniteMap VAR (Set.Set EXP) ,
      labSyn :: InfiniteMap LAB Syn ,
      frees  :: InfiniteMap LAB (Set.Set VAR) ,
      scope  :: ScopeMap }

mergeSyntaxInfo (SI b1 r1 ls1 f1 s1) (SI b2 r2 ls2 f2 s2) =
 SI (b1  `join`  b2)
    (r1  `join`  r2)
    (ls1 `join`  ls2)
    (f1  `join`  f2)
    (s1  `Map.union` s2)
      

botSI = SI bot bot bot bot (Map.empty)

data OptimizationInfo = 
 OI { inliningMap     :: InfiniteMap LAB InlinableExp ,
      copyMap         :: InfiniteMap LAB (Set.Set VAR) ,
      synMap          :: InfiniteMap LAB Syn ,
      binder          :: InfiniteMap VAR LAB ,
      removableVars   :: Set.Set VAR ,
      removableArgs   :: Set.Set LAB ,
      scopeMap        :: ScopeMap ,
      alreadyInlined  :: Set.Set LAB ,
      copyPropAllowed :: Bool ,
      alive           :: Set.Set LAB }

withInlined :: OptimizationInfo -> LAB -> OptimizationInfo
withInlined oi l = oi { alreadyInlined = alreadyInlined oi `union` 
                                         singleton l }

 -- Alphatisation-related utilities.
data AlphaState = AS { seenSymbols :: Set.Set Symbol ,
                       dupSymbols  :: Set.Set Symbol }

updateAlphaState sym (AS seen dup) = AS seen' dup'
 where dup' = if sym `member` seen
              then Set.insert sym dup 
              else dup
       seen' = Set.insert sym seen
 

minAlphaSet syn = dupSymbols $ execState (doMinAlphaSet syn)
                                         (AS Set.empty Set.empty) 

isAlphaUnique syn = Set.null(minAlphaSet syn)

 -- Syntax.

class Syntax a where
 alphatise :: SymbolTableContainer b => 
                (VAR -> VAR) -> a -> (State b a)

 substitute :: SymbolTableContainer b => 
                a -> VAR -> EXP -> (State b a)

 doMinAlphaSet :: a -> State AlphaState ()
 doSyntaxInfo :: a -> [VAR] -> State SyntaxInfo ()

  -- Structural information.
 isAtomic :: a -> Bool

  -- Syntactic information.
 free   :: a -> Set.Set VAR
 varsIn :: a -> Set.Set VAR
 labsIn :: a -> Set.Set LAB
 refsTo :: a -> VAR -> Set.Set LAB

 genBinder    :: a -> InfiniteMap VAR LAB
 genSynMap    :: a -> InfiniteMap LAB Syn
 genScopeMap  :: [VAR] -> a -> ScopeMap
 genArgMap    :: a -> InfiniteMap (LAB,Int) LAB
 genParentMap :: a -> InfiniteMap LAB LAB

 killdead     :: SymbolTableContainer b => OptimizationInfo -> a ->
                 (State b a)
 copyprop     :: SymbolTableContainer b => OptimizationInfo -> a -> 
                 (State b a)
 inline       :: SymbolTableContainer b => OptimizationInfo -> a ->
                 (State b a)
 reduce       :: SymbolTableContainer b => OptimizationInfo -> a -> 
                 (State b a)
 euv          :: SymbolTableContainer b => OptimizationInfo -> a ->
                 (State b a)

  
 isAtomic _ = True



 -- A CPS grammar.

 -- Call sites.
data CALL = App { callLab :: !LAB , callFun :: !EXP , callArgs :: ![EXP] }
          | LetRec { callLab :: !LAB , callVAgs :: ![(VAR,EXP)], body :: !CALL }
          | TopCALL
          | BotCALL



 -- Expressions.
data EXP = LamExp   { expLab :: !LAB, expLam  :: !LAM  }
         | PrimExp  { expLab :: !LAB, expPrim :: !PRIM }
         | RefExp   { expLab :: !LAB, expVar  :: !VAR  }
         | LitExp   { expLab :: !LAB, expLit  :: !LIT  }
         | HaltExp  { expLab :: !LAB }
         | OmegaExp { expLab :: !LAB }
         | TopEXP
         | BotEXP
 deriving (Eq,Ord)

 -- Lambdas.
data LAM = Lam { lamLab :: !LAB , lamVars :: ![VAR] , lamCall :: !CALL }
         | TopLAM
         | BotLAM




 -- Programs.
type PR = CALL



 -- SExp->CPS monadized parser.

parseUserExp (SymSExp s) =
 do lab <- freshUserSymbol
    primADD <- getSymbol "+"
    primSUB <- getSymbol "-"
    primMUL <- getSymbol "*"
    primIFZERO <- getSymbol "if-zero"
    primDISPLAY <- getSymbol "display"
    (     if s == primADD
     then do return $ PrimExp (Fn lab) (Arith ADD)
     else if s == primSUB
     then do return $ PrimExp (Fn lab) (Arith SUB)
     else if s == primMUL
     then do return $ PrimExp (Fn lab) (Arith MUL)
     else if s == primIFZERO
     then do return $ PrimExp (Fn lab) (If ISZERO)
     else if s == primDISPLAY
     then do return $ PrimExp (Fn lab) (IOOp DISPLAY)
     else do return $ RefExp  (Fn lab) (UVar s))
parseUserExp (IntSExp i) =  
 do lab <- freshUserSymbol
    return $ LitExp (Fn lab) (IntLit i)
parseUserExp term = do lam <- parseLam term
                       lab <- freshUserSymbol
                       return $ LamExp (Fn lab) lam

parseKontExp (SymSExp s) = 
 do l <- freshContSymbol 
    return $ RefExp (Cn l) (KVar s)
parseKontExp term = do lam <- parseKont term
                       lab <- freshContSymbol
                       return $ LamExp (Cn lab) lam

parseLam (ConsSExp _ (ConsSExp uargs (ConsSExp kargs 
           (ConsSExp call NilSExp)))) =
 do uargs' <- parseUVarList uargs 
    kargs' <- parseKVarList kargs
    call'  <- parseCall call 
    l      <- freshUserSymbol
    return $ Lam (Fn l) (uargs' ++ kargs') call'

parseKont (ConsSExp _ (ConsSExp uargs (ConsSExp call NilSExp))) =
 do uargs' <- parseUVarList uargs 
    call'  <- parseCall call 
    l      <- freshContSymbol
    return $ Lam (Cn l) uargs' call'

parseUserExpList :: SymbolTableContainer b => SEXP -> (State b [EXP])
parseUserExpList = mapTrueList parseUserExp

parseKontExpList :: SymbolTableContainer b => SEXP -> (State b [EXP])
parseKontExpList = mapTrueList parseKontExp

parseUserCall :: SymbolTableContainer b => SEXP -> (State b CALL)
parseUserCall sexp = 
 do f      <- parseUserExp     $ car(cdr(sexp))
    es     <- parseUserExpList $ car(cdr(cdr(sexp)))
    qs     <- parseKontExpList $ car(cdr(cdr(cdr(sexp)))) 
    l      <- freshUserSymbol
    return $ App (Fn l) f (es ++ qs)

parseKontCall :: SymbolTableContainer b => SEXP -> (State b CALL)
parseKontCall  sexp =
 do q      <- parseKontExp     $ car(cdr(sexp))
    es     <- parseUserExpList $ car(cdr(cdr(sexp)))
    l      <- freshContSymbol
    return $ App (Cn l) q es

 -- TODO: Update these to handle a broader range of expressions.
parseULam :: SymbolTableContainer b => SEXP -> (State b (VAR,EXP))
parseULam sexp =
 do u    <- parseUVar    $ car(sexp)
    lam  <- parseUserExp $ car(cdr(sexp))
    return $ (u,lam)

parseKLam :: SymbolTableContainer b => SEXP -> (State b (VAR,EXP))
parseKLam sexp =
 do k    <- parseKVar    $ car(sexp)
    lam  <- parseKontExp $ car(cdr(sexp))
    return $ (k,lam)

parseLetRecCall :: SymbolTableContainer b => SEXP -> (State b CALL)
parseLetRecCall (sexp@(ConsSExp _ _)) = 
 do ulams <- mapTrueList parseULam $ car(cdr(sexp)) 
    klams <- mapTrueList parseKLam $ car(cdr(cdr(sexp)))
    call  <- parseCall             $ car(cdr(cdr(cdr(sexp))))
    l     <- freshContSymbol
    return $ LetRec (Cn l) (ulams++klams) call

parseCall :: SymbolTableContainer b => SEXP -> (State b CALL)
parseCall (term@(ConsSExp (SymSExp f) args)) =
 do appu <- getSymbol "appu"
    appk <- getSymbol "appk"
    letrec <- getSymbol "letrec"
    (if f == appu
     then do parseUserCall term
     else if (f == appk)
     then do parseKontCall term
     else if (f == letrec)
     then do parseLetRecCall term
     else do f' <- stringFromSymbol f
             error $ "unknown call: " ++ f')

 -- Common utilities.


 -- CBV DS->CPS monadized transformation.

-- TODO: Inline linear variables.
-- TODO: Turn lets into let-conts.

t2args :: [TERM] -> [EXP] -> ([EXP] -> CALL) -> 
 SymState CALL
t2args [] targs' bodyCont = 
 do return (bodyCont (reverse targs'))
t2args (hd:tl) targs' bodyCnt | isAtomicTerm(hd) = 
                                 do e' <- m2 hd
                                    t2args tl (e':targs') bodyCnt
                              | otherwise = 
                                 do usym <- freshUserSymbol
                                    u'   <- return $ UVar usym
                                    l    <- freshUserSymbol
                                    l'   <- return (Fn l)
                                    l2   <- freshContSymbol 
                                    l2'  <- freshContSymbol
                                    e'   <- return $ RefExp l' u'
                                    call <- t2args tl (e':targs') bodyCnt
                                    t2 hd (kont l2 l2' [u'] call)

t2vterm :: (VAR,TERM) -> SymState (VAR,EXP)
t2vterm (v,lam@(LamTerm _ _)) = 
 do exp' <- m2 lam
    return (v,exp')

t2vterms :: [(VAR,TERM)] -> SymState [(VAR,EXP)]
t2vterms [] = do return []
t2vterms (hd:tl) = 
 do hd' <- t2vterm hd
    tl' <- t2vterms tl
    return (hd':tl')

t2 :: TERM -> EXP -> SymState CALL
t2 (AppTerm (PrimTerm (If cond)) [test,ifTrue,ifFalse]) q =
 do lTrue     <- freshContSymbol
    lTrue'    <- freshContSymbol
    lFalse    <- freshContSymbol
    lFalse'   <- freshContSymbol
    lIf       <- freshUserSymbol
    lCallIf   <- freshContSymbol
    lCallIf'  <- freshContSymbol
    usym      <- freshUserSymbol
    u'        <- return $ UVar usym
    lu        <- freshUserSymbol
    e         <- return $ RefExp (Fn lu) u'
    callTrue  <- t2 ifTrue q
    callFalse <- t2 ifFalse q
    lab       <- freshUserSymbol
    lab'      <- return (Fn lab)
    (if (isAtomicTerm test) 
     then do test' <- m2 test
             return $ (appf lIf (PrimExp lab' (If (cond))) [test',
                        kont lTrue lTrue' [] callTrue,
                        kont lFalse lFalse' [] callFalse])
     else t2 test (kont lCallIf lCallIf' [u']
           (appf lIf (PrimExp lab' (If cond)) [e,
              kont lTrue lTrue' [] callTrue,
              kont lFalse lFalse' [] callFalse])))

t2 (AppTerm (PrimTerm (IOOp DISPLAY)) [v]) q =
 do kl1 <- freshContSymbol
    kl2 <- freshContSymbol
    ul1 <- freshUserSymbol
    ul2 <- freshUserLabel
    v'  <- freshUserVar
    rv' <- freshRefExp v'
    t2 v (kont kl1 kl2 [v'] 
	  (appf ul1 (PrimExp ul2 (IOOp DISPLAY)) [rv',q]))

t2 (AppTerm f es) k = 
  if isAtomicTerm f
  then do f' <- m2 f
          l <- freshUserSymbol
          t2args es [] (\ targs' -> appf l f' (targs' ++ [k]))
  else do usym <- freshUserSymbol
          u'   <- return (UVar usym)
          l    <- freshUserSymbol
          l2   <- freshUserSymbol
          f'   <- return (RefExp (Fn l) u')
          lk   <- freshContSymbol 
          lk'  <- freshContSymbol
          callBody <- t2args es [] (\ targs' -> appf l2 f' (targs' ++ [k]))
          t2 f (kont lk lk' [u'] callBody)

t2 (BeginTerm term1 term2) q =
 do kl1  <- freshContSymbol
    kl2  <- freshContSymbol
    body <- t2 term2 q
    t2 term1 (kont kl1 kl2 []
                body)

t2 (LetRecTerm vterms body) q = 
 do ulams  <- t2vterms vterms 
    lrbody <- t2 body q
    l      <- freshContSymbol
    return (LetRec (Cn l) ulams lrbody)

t2 term q = 
 do e' <- m2 term
    lk <- freshContSymbol
    return (appk lk q [e'])


m2 :: TERM -> SymState EXP
m2 (RefTerm v) = 
 do l <- freshUserSymbol
    return (RefExp (Fn l) v)
m2 (LitTerm lit) = 
 do sym <- freshUserSymbol
    return (LitExp (Fn sym) lit)
m2 (LamTerm vs body) = 
 do i     <- freshContSymbol
    k'    <- return (KVar i)
    lk    <- freshContSymbol
    q'    <- return (RefExp (Cn lk) k')
    body' <- t2 body q'
    l     <- freshUserSymbol
    sym   <- freshUserSymbol
    return (LamExp (Fn sym) (Lam (Fn l) (vs ++ [k']) body'))
m2 (PrimTerm CallCC) =
 do f   <- freshUserVar
    k   <- freshContVar
    k0  <- freshContVar
    v   <- freshUserVar
    rv  <- freshRefExp v
    rf  <- freshRefExp f
    rk  <- freshRefExp k
    rk' <- freshRefExp k
    [ul1,ul2,ul3,ul4,ul5] <- generate 5 (\_ -> freshUserLabel)
    kl5 <- freshContLabel
    return $ LamExp ul1 (Lam ul2 [f,k] 
              (App ul3 rf [LamExp ul4 (Lam ul5 [v,k0] 
                              (App kl5 rk [rv])), rk']))
m2 (PrimTerm p) = 
 do l <- freshUserSymbol
    return (PrimExp (Fn l) p)


generate 0 maker = return []
generate n maker = do hd <- maker n
                      tl <- generate (n - 1) maker
                      return (hd:tl)

freshRefExp (UVar sym) = do us <- freshUserSymbol
                            return $ RefExp (Fn us) (UVar sym)
freshRefExp (KVar sym) = do ks <- freshContSymbol
                            return $ RefExp (Cn ks) (KVar sym)

t2pr :: TERM -> SymState LAM
t2pr pr =
 do halt  <- getSymbol "halt"
    halt' <- return $ KVar halt
    lk    <- freshContSymbol -- Label on halt ref.
    lk2   <- freshContSymbol -- Label on halt kont.
    lk3   <- freshContSymbol -- Label on halt app.
    x     <- freshUserSymbol
    x'    <- return $ UVar x
    l     <- freshUserSymbol
    lu    <- freshUserSymbol
    q' <- return $ RefExp (Cn lk) halt'
    call  <- t2 pr q'
    return (Lam (Fn l) [halt'] call)

tpr :: TERM -> SymbolTable -> (LAM,SymbolTable)
tpr pr = runState (t2pr pr) 

pr2call :: TERM -> SymbolTable -> (CALL,SymbolTable)
pr2call pr = runState (do ksym <- freshContSymbol 
                          t2 pr (HaltExp (Cn ksym)))

 -- TODO: Speed up Eq,Ord instances.

instance Eq LAM where
 Lam l _ _ == Lam l' _ _ = l == l'
-- lam       == lam'       = lam == lam'
 TopLAM == TopLAM = True
 BotLAM == BotLAM = True
 _ == _ = False

instance Ord LAM where
 compare BotLAM BotLAM = EQ
 compare TopLAM TopLAM = EQ
 compare BotLAM _      = LT
 compare TopLAM _      = LT
-- compare BotLAM TopLAM = LT
-- compare TopLAM BotLAM = GT
 compare _      BotLAM = GT
 compare _      TopLAM = GT
 compare (Lam l _ _) (Lam l' _ _) = compare l l'

instance Eq CALL where
 App l _ _    == App l' _ _    = l == l'
 LetRec l _ _ == LetRec l' _ _ = l == l'
 TopCALL == TopCALL = True
 BotCALL == BotCALL = True
 _ == _ = False 

-- call         == call'         = call == call'

instance Ord CALL where
 compare BotCALL BotCALL  = EQ
 compare TopCALL TopCALL  = EQ
 compare BotCALL _        = LT
 compare TopCALL _        = LT
-- compare BotCALL TopCALL  = LT
-- compare TopCALL BotCALL  = GT
 compare _      BotCALL   = GT
 compare _      TopCALL   = GT
 compare (App l _ _)    (App l' _ _)    = compare l l'
 compare (LetRec l _ _) (LetRec l' _ _) = compare l l'
 compare (App _ _ _)    (LetRec _ _ _)  = LT
 compare (LetRec _ _ _) (App _ _ _)     = GT

 -- Display instances.
instance Display EXP where
 displayHTML (LamExp lab lam) = displayHTML lam
 displayHTML (RefExp lab var) = display var
 displayHTML e = display e

 displayHTMLList [] = return ""
 displayHTMLList (hd:tl) = 
  do hd <- displayHTML hd
     tl <- displayHTMLList tl
     return $ hd ++ " " ++ tl

 display (LamExp lab lam) = display lam
 display (RefExp lab var) = do var' <- display var
                               lab' <- display lab
                               return $ var' ++ "{"++ lab' ++"}"
 display (LitExp lab lit) = return  $ show lit
 display (PrimExp lab p) = return $ show p
 display (HaltExp lab) = return "Halt"
 display (OmegaExp lab) = return "Omega"
 display (TopEXP) = return "TopEXP"
 display (BotEXP) = return "BotEXP"

 displayList []     = return ""
 displayList [e]    = display e
 displayList (e:es) = do e' <- display e
                         es' <- display es
                         return $ e' ++ " " ++ es'

instance Display LAM where
 display (Lam (Fn l) vs call) = 
  do vs'   <- display vs
     call' <- display call
     l'    <- display (Fn l)
     (if all (\ v -> case v of { UVar _ -> True  ; _ -> False }) vs
      then return $ "(kont{"++l'++  "} (" ++ vs' ++ ") " ++ call' ++ ")"
      else return $ "(lambda{"++l'++"} (" ++ vs' ++ ") " ++ call' ++ ")")

 display (Lam (Cn l) vs call) = 
  do vs'   <- display vs
     call' <- display call
     l'    <- display (Cn l)
     return $ "(kont{"++l'++"} (" ++ vs' ++ ") " ++ call' ++ ")"
    
 displayHTML lam@(Lam _ vs call) | isAtomic lam = 
  do vs <- displayHTML vs
     call <- displayHTML call
     return $ "(lambda (" ++ vs ++ ") " ++ call ++ ")"
                                 | otherwise =
  do vs <- displayHTML vs
     call <- displayHTML call
     return $ "<table clear=\"0 pixels\" border=0 cellpadding=0 cellspacing=0><tr><td valign=\"top\">" ++ 
        "(lambda (" ++ vs ++ ")</td></tr><tr><td>&nbsp;&nbsp;&nbsp;" ++ call ++ ")</td></tr></table>"

instance Display LAB where
 display (Fn l) = stringFromSymbol l
 display (Cn l) = stringFromSymbol l
 display (TopLAB) = return "TopLAB"
 display (BotLAB) = return "BotLAB"

instance Display CALL where
 display (App (Fn l) f args) =
  do f'    <- display f
     as'   <- display args
     l'    <- display (Fn l)
     return $ "({"++l'++"}" ++ f' ++ " " ++ as' ++ ")"

 display (App (Cn l) f args) =
  do f'    <- display f
     as'   <- display args
     l'    <- display (Cn l)
     return $ "({"++l'++"}" ++ f' ++ " " ++ as' ++ ")"

 display BotCALL = return "(Unreachable)"

 display (LetRec l vargs call) =
  do vargs' <- displayVLList vargs
     call'  <- display call
     return $ "(letrec (" ++ vargs' ++ ") " ++ call' ++ ")"
  where displayVLList [] = return ""
        displayVLList [(v,lam)] =
         do v'   <- display v
            lam' <- display lam
            return $ "(" ++ v' ++ " " ++ lam' ++ ")"
        displayVLList (hd:tl) =
         do hd' <- displayVLList [hd]
            tl' <- displayVLList tl
            return $ hd' ++ " " ++ tl' 

 displayHTML app@(App _ f args) | isAtomic app =
  do f    <- displayHTML f
     args <- displayHTML args
     return $ "(" ++ f ++ " " ++ args ++ ")"
                                | otherwise = 
  do f    <- displayHTML f
     args <- displayHTML args
     return $ "<table clear=\"0 pixels\" border=0 cellpadding=0 cellspacing=0><tr><td>(" ++ f ++ " " ++ args ++ ")</td></tr></table>"

 displayHTML (LetRec _ vargs body) = 
  do body <- displayHTML body
     vargs <- displayHTMLVArgs vargs
     return $ "<table clear=\"0 pixels\" border=0 cellpadding=0 cellspacing=0><tr><td valign=\"top\">" ++ "(letrec </td><td>(<table clear=\"0 pixels\" border=0 cellpadding=0 cellspacing=0>" ++ vargs ++ "</table>)</td></tr><tr><td colspan=2>&nbsp;&nbsp;" ++ body ++ ")</td></tr></table>"
  where displayHTMLVArgs [] = return ""
        displayHTMLVArgs (hd:tl) = 
         do hd <- displayHTMLVArg hd
            tl <- displayHTMLVArgs tl
            return $ hd ++ "\n" ++ tl


        displayHTMLVArg (v,arg) = 
         do v <- displayHTML v
            arg <- displayHTML arg
            return $ "<tr><td valign=\"top\">(" ++ v ++ "</td><td valign=\"top\">" ++ arg ++ ")</td></tr>"

 -- Show instances.

instance Show LAB where
 show (Fn i) = "fn" ++ (show i)
 show (Cn i) = "cn" ++ (show i)
 show TopLAB = "TopLAB"
 show BotLAB = "BotLAB" 


 -- Syntax instances.

 {- Common -}

renameVARList nv ([],alpha) = return $ ([],alpha)
renameVARList nv (var:tl,alpha) = 
         do var' <- case var of 
                     UVar sym -> do sym' <- nv sym
                                    return $ UVar sym'
                     KVar sym -> do sym' <- nv sym
                                    return $ KVar sym'
            (tl',alpha') <- renameVARList nv (tl,alpha)
            return $ (var':tl',Might.except alpha' (var,var'))


alphalab lab = 
  do lab' <- case lab of
              Fn s -> do l' <- refreshSymbol s
                         return $ Fn l'
              Cn s -> do l' <- refreshSymbol s
                         return $ Cn l'
     return lab'

instance Syntax CALL where
 free (App l f args) = Set.unions [free f, free args]
 free (LetRec l vargs call) = freeVARs - boundVARs
  where freeVARs    = Set.unions[ freeInArgs, free call ]
        freeInArgs  = Set.unions (fmap (free . snd) vargs)
        boundVARs   = Set.fromList (fmap fst vargs)
        (-) = Set.difference
 free BotCALL = Set.empty

 alphatise alpha (App lab f args) =
  do lab'  <- alphalab lab
     f'    <- alphatise alpha f
     args' <- alphatise alpha args
     return $ App lab' f' args'

 alphatise alpha (LetRec lab vas call) =
  do lab'         <- alphalab lab
     (vs,args)    <- return $ unzip vas
     (vs',alpha') <- renameVARList refreshSymbol (vs,alpha)
     args' <- alphatise alpha' args
     call' <- alphatise alpha' call
     return $ LetRec lab' (zip vs' args') call'
 alphatise alpha BotCALL = return BotCALL


 varsIn (App _ f args) = Set.unions[varsIn f,varsIn args]
 varsIn (LetRec _ vargs call) = 
  Set.unions[varsIn vargs, varsIn call]
 varsIn BotCALL = Set.empty

 labsIn (App lab f args) = Set.unions[singleton lab,labsIn f,labsIn args]
 labsIn (LetRec lab vargs call) =
  Set.unions[singleton lab, labsIn vargs, labsIn call]
 labsIn BotCALL = Set.empty

 refsTo (App lab f args) v        = (refsTo f v) `union` (refsTo args v)
 refsTo (LetRec lab vargs call) v = 
  if any (\ (v',e') -> v' == v) vargs
  then Set.empty
  else (refsTo vargs v) `union` (refsTo call v)
 refsTo BotCALL _ = Set.empty


 doSyntaxInfo BotCALL scp = return ()

 doSyntaxInfo call@(App lab f args) scp = 
   do doSyntaxInfo f scp
      doSyntaxInfo args scp
      si <- get
      free1 <- return $ (frees si) ! (expLab f)
      free2 <- return $ Set.unions [ (frees si) ! (expLab arg) | arg <- args ]
      free  <- return $ free1 `Set.union` free2
      put $ si { labSyn = (labSyn si) `but` [lab ==> CallSyn call] ,
                 frees = (frees si) `but` [ lab ==> free ] }

 doSyntaxInfo call@(LetRec lab vargs body) scp =
   do doSyntaxInfo vargs scp'
      doSyntaxInfo body scp'
       -- This is hackery.  Fix it.
      si <- get
      free1 <- return $ (frees si) ! (callLab body)
      free2 <- return $ Set.unions [ (frees si) ! (expLab arg) | 
                                     (v,arg) <- vargs  ] 
      free <- return $  (free1 `Set.union` free2) 
                         `Set.difference` (Set.fromList vs)
      put $ si { bind    = (bind si)   `but` [ v ==> lab | v <- vs ] ,
                 labSyn  = (labSyn si) `but` [ lab ==> CallSyn call ] ,
                 frees   = (frees si)  `but` [ lab ==> free ] }
   where vs = fmap fst vargs
         scp' = appendReverse vs scp
           

 doMinAlphaSet BotCALL = return ()
 doMinAlphaSet (App lab f args) = 
  do doMinAlphaSet f
     doMinAlphaSet args
     perform $ updateAlphaState (labSym lab)


 doMinAlphaSet (LetRec lab vargs call) =
  do doMinAlphaSet call
     perform $ updateAlphaState (labSym lab)
     doFoldAction vargs doMinAlphaSet
     
                                
     

 substitute BotCALL v e = return BotCALL
 substitute (App lab f args) v e = 
  do f'    <- substitute f v e
     args' <-  substitute args v e
     return $ App lab f' args' 

  -- TODO: Add safety check.
 substitute lr@(LetRec lab vargs call) v e = 
  do (if any (\ (v',e') -> v' == v) vargs
      then return lr
      else if any (\ (v',e') -> v' `member` free e) vargs
           then do lr <- alphatise id lr 
                   substitute lr v e
           else do vargs' <- substitute vargs v e
                   call'  <- substitute call  v e 
                   return $ LetRec lab vargs' call')
                

 genScopeMap scope (App l f args) = 
  Map.unions[genScopeMap scope f,
             genScopeMap scope args]

 genScopeMap scope (LetRec l vargs call) = 
  map' `Map.union` map''
  where vs     = fmap fst vargs
        scope' = appendReverse vs scope
        map'   = genScopeMap scope' vargs
        map''  = genScopeMap scope' call

 genScopeMap scope BotCALL = Map.empty

 genBinder (App _ f args) =
  (genBinder f) `join` (genBinder args)

 genBinder (LetRec lab vargs call) =
  (genBinder call `join` 
   (bot `but` [ (v,lab) | (v,_) <- vargs ]))
   `join`
  (genBinder vargs)

 genBinder BotCALL = bot

 genSynMap app@(App lab f args) = 
  (genSynMap f) `join` (genSynMap args) `join` 
  (bot `but` [lab ==> CallSyn app])

 genSynMap term@(LetRec lab vargs call) = 
  (genSynMap call `join` (bot `but` [lab ==> CallSyn term]))
   `join`
  (genSynMap vargs)
 
 genSynMap BotCALL = bot

 genParentMap (App par (PrimExp child _) args)
  = (bot `but` [child ==> par]) `join` genParentMap args

 genParentMap (App _ f args) = genParentMap f `join` genParentMap args

 genParentMap (LetRec l vargs body) =
  genParentMap vargs `join` genParentMap body

 genArgMap (App l f args) = map' `join` map'' 
  where genMaps exps = 
         fmap (\ (loc,exp) ->
           case exp of  
            LamExp lab lam -> genArgMap lam `but` [(l,loc) ==> lab]
            _              -> bot `but` [(l,loc) ==> expLab exp])
          (zip [0..] exps)

        map'  = joins (genMaps args)
        map'' = case f of 
                  LamExp lab lam -> genArgMap lam
                  _ -> bot
  
 genArgMap (LetRec l vas call) = genArgMap call `join` map'
  where genMaps exps = 
         fmap (\ (loc,exp) ->
           case exp of  
            LamExp lab lam -> genArgMap lam `but` [(l,loc) ==> lab]
            _              -> bot `but` [(l,loc) ==> expLab exp])
          (zip [0..] exps)
        map' = joins (genMaps (fmap snd vas)) 

 genArgMap BotCALL = bot

 killdead oi BotCALL = return BotCALL
 
 killdead oi app@(App l f args) | not(l `member` alive oi) = return BotCALL
                                | otherwise =
  do f    <- killdead oi f
     args <- killdead oi args
     return $ App l f args

 killdead oi (LetRec l vargs body) | not(l `member` alive oi) = return BotCALL
                                   | otherwise =
  do vargs <- killdead oi vargs
     body  <- killdead oi body
     return $ LetRec l vargs body

 copyprop oi BotCALL = return BotCALL
 
 copyprop oi (App l f args) =
  do f    <- copyprop oi f
     args <- copyprop oi args
     return $ App l f args

 copyprop oi (LetRec l vargs body) =
  do vargs <- copyprop oi vargs
     body  <- copyprop oi body
     return $ LetRec l vargs body

 inline oi BotCALL = return BotCALL

 inline oi (App l f args) =
  do f    <- inline oi f
     args <- inline oi args
     return $ App l f args

 inline oi (LetRec l vargs body) = 
  do vargs <- inline oi vargs
     body  <- inline oi body
     return $ LetRec l vargs body
     

 isAtomic (App _ f args) = isAtomic f && all isAtomic args 
 isAtomic (LetRec _ [] body) = isAtomic body
 isAtomic (LetRec _ vargs body) = False

  -- reduce must monotonically decrease the term's size.
 reduce oi BotCALL = return BotCALL
 
 reduce oi (App _ (PrimExp _ (Arith arithOp)) [LitExp _ (IntLit a),
					       LitExp _ (IntLit b),
                                               cont]) =
  do ul <- freshUserLabel
     kl <- freshContLabel
     reduce oi $ App kl cont [LitExp ul (IntLit (compute arithOp a b))]
  where compute ADD = (+)
        compute SUB = (-)
        compute MUL = (*)
        compute DIV = div

 reduce oi (App _ (PrimExp _ (If _)) [test,LamExp _ (Lam _ [] BotCALL),
                                           LamExp _ (Lam _ [] onfalse)])= 
  reduce oi onfalse

 reduce oi (App _ (PrimExp _ (If _)) [test,LamExp _ (Lam _ [] ontrue),
                                           LamExp _ (Lam _ [] BotCALL)])= 
  reduce oi ontrue

 reduce oi (App _ (PrimExp _ (If ISZERO)) [LitExp _ (IntLit 0),ctrue,cfalse]) =
  do kl <- freshContLabel
     reduce oi $ App kl ctrue []

 reduce oi (App _ (PrimExp _ (If ISZERO)) [LitExp _ (IntLit _),ctrue,cfalse]) =
  do kl <- freshContLabel
     reduce oi $ App kl cfalse []

 reduce oi (App _ (LamExp _ (Lam _ [] call)) []) =
  reduce oi call

 reduce oi (LetRec _ [] body) = reduce oi body

 reduce oi app@(App ul1 (LamExp ul2 (Lam ul3 vs body)) args) =
  if (length vs /= length args)
  then return BotCALL
  else 
   do vargs        <- do dispApp <- display app
                         return $ zipFail ("oh dear: " ++ dispApp) vs args
      vargs        <- reduce oi vargs
      body         <- reduce oi body
      (subs,vargs) <- return $ inlineAffine body vargs
      (vs,args)    <- return $ unzip vargs
      call         <- performSubs subs body 
      case vargs of 
       [] -> return call
       _  -> return $ App ul1 (LamExp ul2 (Lam ul3 vs call)) args
  where inlineAffine call [] = ([],[])
        inlineAffine call ((v,arg):tl) =
         let (subs,vargs) = inlineAffine call tl
            -- TODO: Inline while ref hunting.
          in if (case arg of
                  RefExp _ _ -> True
                  LitExp _ _ -> True
                  HaltExp _ -> True
                  _ -> False) ||
                (size(refsTo call v) <= 1) ||
                (Set.null(free arg)) 
             then ((v ==> arg):subs,vargs)
             else (subs,(v,arg):vargs)

        -- BUG/TODO: What if this lambda binds a free variable of
        -- something we inline?

        performSubs [] term = return term
        performSubs ((v,e):tl) trm  = 
         do -- e <- alphatise id e
            trm <- substitute trm v e 
            performSubs tl trm



 reduce oi (App lab f args) =
  do f'    <- reduce oi f
     args' <- reduce oi args
     return $ App lab f' args'

 reduce oi lr0@(LetRec lab vargs body) =
  do vargs <- inlineNonSelfRecursive vargs vargs
     (subs,vargs)  <- return $ inlineAffine [] vargs
     subs  <- performSubsOn subs subs -- Optimize what we're inlining.
     vargs <- performSubsOn subs vargs
     body  <- performSubs subs body
     vargs <- reduce oi vargs
     body  <- reduce oi body
     lr <- return $ LetRec lab vargs body
     (if free lr `isSubsetOf` free lr0
      then return $ lr
      else do dispSubs  <- display subs
              dispVargs <- display vargs
              dispFree  <- display (free lr)
              dispLR    <- display lr
              error $ "Free variables created by reducing LR,\n subs: " ++ 
                       dispSubs ++ "\n vargs: " ++ dispVargs ++ 
                       "\n free: " ++ dispFree ++ "\n lr:\n" ++ dispLR)
  where  -- TODO: Enhance this for variables in vargs.
        inlineAffine :: [(VAR,EXP)] -> [(VAR,EXP)] -> ([(VAR,EXP)],[(VAR,EXP)])
        inlineAffine subs [] = (subs,[])
        inlineAffine subs ((v,arg):tl) =
         let (subs',vargs') = inlineAffine subs tl 
             suband = case arg of
                       RefExp lab var -> if var == v
                                         then OmegaExp lab
                                         else arg
                       _ -> arg
          in if (case arg of
                  RefExp _ _ -> True
                  LitExp _ _ -> True
                  HaltExp _ -> True
                  _ -> False) 
                || (size(refsTo body v) <= 1 && 
                    size(refsTo vargs v) == 0) 
                || Set.null(free arg)
             then ((v ==> suband):subs',vargs')
             else (subs',(v ==> arg):vargs')

        inlineNonSelfRecursive subs [] = return []
        inlineNonSelfRecursive subs ((v,e):tl) =
         do subs' <- return $ Prelude.filter (\ (v',e') -> v'/=v &&
                                               size(refsTo e' v) == 1) 
                                subs
            (v,e) <- performSubs subs' (v,e) 
            tl    <- inlineNonSelfRecursive subs tl
            return $ ((v,e):tl)

        performSubs [] term = return term
        performSubs ((v,e):tl) trm  = 
         do trm <- substitute trm v e 
            performSubs tl trm

        performSubsOn subs [] = return [] 
        performSubsOn subs (hd:tl) = 
         do hd <- performSubs subs hd 
            tl <- performSubsOn subs tl
            return $ hd:tl
 

 euv oi BotCALL = return BotCALL

 euv oi (App lab f args) =
  do 
     (if isArithPrim f && any (\ arg -> case arg of 
                        RefExp l v -> v `member` removableVars oi
                                   || l `member` removableArgs oi
                        e -> expLab e `member` removableArgs oi) args
      then do klab <- freshContLabel
              cont <- euv oi (last args)
              return $ App klab cont []
      else do f' <- euv oi f
              args'  <- 
                return $ filter (not . (`member` removableArgs oi) . 
                                     expLab) args
              args' <- euv oi args'
              return $ App lab f' args')
  where isArithPrim (PrimExp _ (Arith _)) = True
        isArithPrim _ = False

 euv oi (LetRec lab vargs call) = 
  do call' <- euv oi call
     vargs' <-
      return $ filter (not . (`member` removableArgs oi) . expLab . snd) 
                vargs
     vargs' <- euv oi vargs'
     return $ LetRec lab vargs' call'





instance Syntax VAR where
 free _ = Set.empty
 alphatise alpha var = return $ alpha var
 varsIn v = singleton v
 labsIn v = Set.empty
 refsTo v v' = Set.empty
 doSyntaxInfo v scp = return ()
 doMinAlphaSet v = perform $ updateAlphaState (varSym v)
 substitute var v' e' = return var
 genBinder v = bot
 genParentMap v = bot
 genSynMap v = bot
 genScopeMap scope v = Map.empty
 genArgMap v = bot
 killdead oi v = return v
 copyprop oi v = return v
 reduce oi v = return v
 inline oi v = return v
 euv oi v = return v




instance Syntax LAM where
 free (Lam l vs call) = free call - Set.fromList vs
  where (-) = Set.difference

 isAtomic (Lam l vs BotCALL) = True
 isAtomic (Lam l vs (App _ f args)) = not(isLamExp f) && 
                                      all (not . isLamExp) args 
  where isLamExp (LamExp _ _) = True
        isLamExp _ = False
 isAtomic _ = False

 doSyntaxInfo lam@(Lam lab vs call) scp =
  do scp <- return $ appendReverse vs scp
     doSyntaxInfo call scp
     si <- get
     free <- return $ (frees si ! callLab call) `Set.difference` 
                      (Set.fromList vs)
     put $ si { bind = (bind si) `but` [ v ==> lab | v <- vs ] ,
                labSyn = (labSyn si) `but` [ lab ==> LamSyn lam ] ,
                frees = (frees si) `but` [ lab ==> free  ] }

 doMinAlphaSet (Lam lab vs call) = 
  do perform $ updateAlphaState (labSym lab)
     doMinAlphaSet call
     doFoldAction vs doMinAlphaSet
     

 alphatise alpha (Lam lab vs call) =
  do lab'         <- alphalab lab
     (vs',alpha') <- renameVARList refreshSymbol (vs,alpha)
     call'        <- alphatise alpha' call
     return $ Lam lab' vs' call'

 varsIn (Lam _ vs call) = Set.unions[Set.fromList vs,varsIn call]

 labsIn (Lam lab vs call) = Set.singleton lab `union` labsIn call
 
 refsTo (Lam lab vs call) v = if any (v==) vs
                              then Set.empty
                              else refsTo call v

 substitute lam@(Lam lab vs call) v' e' = 
  if any (v'==) vs
  then return $ Lam lab vs call -- Stop.
  else if any (\v -> v `member` free e')  vs -- Alphatise?
       then do lam <- alphatise id lam
               substitute lam v' e'
       else do call <- substitute call v' e'
               return $ Lam lab vs call
 
 genBinder (Lam lab vs call) =
  genBinder call `join` (bot `but` [ (v,lab) | v <- vs ])

 genSynMap (term@(Lam lab vs call)) =
  genSynMap call `join` (bot `but` [ lab ==> LamSyn term ])

 genScopeMap scope (Lam lab vs call) = genScopeMap scope' call
  where scope' = appendReverse vs scope

 genParentMap (Lam lab vs call) = genParentMap call

 genArgMap (Lam lab vs call) = genArgMap call
 
 reduce oi (Lam lab vs call) = 
  do call' <- reduce oi call
     return $ Lam lab vs call'

 killdead oi (Lam lab vs call) = 
  do call <- killdead oi call
     return $ Lam lab vs call

 copyprop oi (Lam lab vs call) = 
  do call <- copyprop oi call 
     return $ Lam lab vs call

 inline oi (Lam lab vs call) =
  do call <- inline oi call
     return $ Lam lab vs call
 
 euv oi (Lam lab vs call) =
  do vs'   <- return $ filter (not . (`member` removableVars oi)) vs
     call' <- euv oi call
     return $ Lam lab vs' call'

     

instance Syntax EXP where
 free (LamExp _ lam) = free lam
 free (RefExp _ var) = Set.singleton var
 free (LitExp _ _)   = Set.empty
 free (PrimExp _ _)  = Set.empty
 free (HaltExp _)    = Set.empty
 free (OmegaExp _)   = Set.empty

 isAtomic (LamExp _ lam) = isAtomic lam
 isAtomic _ = True

 alphatise alpha (OmegaExp lab) =
  do lab <- alphalab lab
     return $ OmegaExp lab

 alphatise alpha (LamExp lab lam) = 
  do lam' <- alphatise alpha lam
     lab' <- alphalab lab
     return $ LamExp lab' lam'

 alphatise alpha (RefExp lab var) = 
  do lab' <- alphalab lab
     return $ RefExp lab' (alpha var)

 alphatise alpha (PrimExp lab p) =
  do lab' <- alphalab lab
     return $ PrimExp lab' p
 
 alphatise alpha (LitExp lab lit) = 
  do lab' <- alphalab lab
     return $ LitExp lab' lit
 alphatise alpha (HaltExp lab) = 
  do lab' <- alphalab lab
     return $ HaltExp lab'

 doSyntaxInfo (LamExp lab lam) scp = 
  do doSyntaxInfo lam scp
     si <- get
     free <- return $ frees si ! lamLab lam
     put $ si { frees = (frees si) `but` [ lab ==> free ] ,
                scope = Map.insert lab scp (scope si)  }

 doSyntaxInfo exp scp = 
  do si <- get
     put $ si { frees = (frees si) `but` [ expLab exp ==> Set.empty ] ,
                scope = Map.insert (expLab exp) scp (scope si)  }


 doMinAlphaSet (LamExp lab lam) = 
  do perform $ updateAlphaState (labSym lab) 
     doMinAlphaSet lam
 doMinAlphaSet exp =
  do perform $ updateAlphaState (labSym (expLab exp))


 reduce oi (LamExp lab lam) =
  do lam' <- reduce oi lam
     return $ eta(LamExp lab lam')
  where eta e@(LamExp l1 (Lam l2 vs (App l3 f args))) =
         if alphaMatch vs args then f else e
        eta e = e

        alphaMatch []       [] = True
        alphaMatch (v1:vtl) ((RefExp _ v2):rtl) | v1 == v2 = alphaMatch vtl rtl
                                                | otherwise = False
        alphaMatch _        _  = False
 reduce oi e = return e

 killdead oi (LamExp lab lam) = 
  do lam <- killdead oi lam
     return $ LamExp lab lam

 killdead oi e = return e

 copyprop oi (LamExp lab lam) = 
  do lam <- copyprop oi lam
     return $ LamExp lab lam
 copyprop oi (RefExp lab var) =
  do scopeMap <- return $ scopeMap oi 
     copyMap  <- return $ copyMap oi 
     firstCopy (scopeMap Map.! lab) (copyMap ! lab)
  where firstCopy [] copies = return $ RefExp lab var
        firstCopy (hd:tl) copies = 
         if hd `member` copies 
         then return $ RefExp lab hd
         else firstCopy tl copies 
 copyprop oi e = return e

 inline oi (LamExp lab lam) = 
  do lam <- inline oi lam
     return $ LamExp lab lam
 inline oi (RefExp lab var) = 
  do inMap <- return $ inliningMap oi
     scopeMap <- return $ scopeMap oi
     case inMap ! lab of 
      InHalt          -> return $ HaltExp lab
      InLit lit       -> return $ LitExp lab lit
      InLamLab lamlab -> 
       case synMap oi ! lamlab of
         LamSyn lam -> 
          if not(lab `member` refsTo lam var)
          then if allIn (free lam) (\ v ->
                   any (==v) (scopeMap Map.! lab))
               then do lam <- alphatise id lam
                       return $ LamExp lab lam
               else do dispLam  <- display lam
                       dispFree <- display (free lam)
                       dispScope <- display (scopeMap Map.! lab)
                       error $ "Super-beta inlining introduced a free variable:\n " ++ dispLam ++ "\n free: " ++ dispFree ++ "\n scope: " ++ dispScope
          else return $ RefExp lab var
         TopSyn     -> return $ RefExp lab var
         oth        -> do dispOth <- display oth
                          error $ "Invalid inlining candidate: " ++ dispOth
      _ -> return $ RefExp lab var
 inline oi e = return e

 euv oi (LamExp lab lam) = 
  do lam <- euv oi lam 
     return $ LamExp lab lam

 euv oi (RefExp lab var) | var `member` removableVars oi =
                            return $ LitExp lab FalseLit
                         | otherwise = return $ RefExp lab var
 euv oi e = return e

 varsIn (OmegaExp _) = Set.empty
 varsIn (HaltExp _) = Set.empty
 varsIn (RefExp _ var) = singleton var
 varsIn (LamExp _ lam) = varsIn lam
 varsIn (LitExp _ lit) = Set.empty
 varsIn (PrimExp _ p) = Set.empty

 labsIn (LamExp lab lam) = singleton lab `union` labsIn lam
 labsIn e                = singleton (expLab e)

 refsTo (LamExp lab lam) v = refsTo lam v
 refsTo (RefExp lab var) v = if v == var then singleton lab else Set.empty
 refsTo _                v = Set.empty

 substitute (LamExp lab lam) v' e' = 
  do lam <- substitute lam v' e'
     return $ LamExp lab lam
 substitute (RefExp lab var) v' e' | var == v' = do e' <- alphatise id e'
                                                    return $ e'
                                   | otherwise = return $ RefExp lab var
 substitute e                v' e' = return e

 genBinder (LamExp _ lam) = genBinder lam
 genBinder _ = bot

 genSynMap (PrimExp lab prim) = bot `but` [lab ==> InPrimSyn prim lab]
 genSynMap (LamExp _ lam) = genSynMap lam
 genSynMap _ = bot
 
 genParentMap (LamExp _ lam) = genParentMap lam
 genParentMap _ = bot

 genScopeMap scope (LamExp _ lam) = genScopeMap scope lam  
 genScopeMap scope (RefExp lab var) = 
  Map.insert lab (reverse scope) Map.empty
 genScopeMap scope _ = Map.empty

 genArgMap (LamExp _ lam) = genArgMap lam
 genArgMap _              = bot

instance (Syntax a, Syntax b) => Syntax (a,b) where
 free (a,b) = free a `union` free b
 
 alphatise alpha (a,b) = 
  do a' <- alphatise alpha a
     b' <- alphatise alpha b
     return $ (a',b')

 varsIn (a,b) = varsIn a `union` varsIn b
 labsIn (a,b) = labsIn a `union` labsIn b
 refsTo (a,b) v = (refsTo a v) `union` (refsTo b v)

 doSyntaxInfo (a,b) =
  do doSyntaxInfo a
     doSyntaxInfo b

 doMinAlphaSet (a,b)  = 
  do doMinAlphaSet a
     doMinAlphaSet b 

 substitute (a,b) v e = 
  do a <- substitute a v e
     b <- substitute b v e
     return (a,b)

 genBinder (a,b) = genBinder a `join` genBinder b
 genSynMap (a,b) = genSynMap a `join` genSynMap b

 genScopeMap scope (a,b) = genScopeMap scope a `Map.union`  
                           genScopeMap scope b

 genParentMap (a,b) = genParentMap a `join` genParentMap b

 genArgMap (a,b) = genArgMap a `join` genArgMap b
 
 killdead oi (a,b) = 
  do a <- killdead oi a
     b <- killdead oi b
     return $ (a,b) 

 copyprop oi (a,b) = 
  do a <- copyprop oi a
     b <- copyprop oi b
     return $ (a,b)

 inline oi (a,b) =
  do a <- inline oi a
     b <- inline oi b
     return $ (a,b)

 reduce oi (a,b) =
  do a' <- reduce oi a
     b' <- reduce oi b
     return (a',b')

 euv oi (a,b) =  
  do a <- euv oi a
     b <- euv oi b
     return (a,b)


instance Syntax a => Syntax [a] where
 free syns = Set.unions (fmap free syns)

 alphatise alpha []      = return []
 alphatise alpha (hd:tl) = 
  do hd <- alphatise alpha hd
     tl <- alphatise alpha tl
     return $ hd:tl
 
 killdead oi [] = return [] 
 killdead oi (hd:tl) =
  do hd <- killdead oi hd
     tl <- killdead oi tl
     return $ hd:tl

 copyprop oi [] = return []
 copyprop oi (hd:tl) =
  do hd <- copyprop oi hd
     tl <- copyprop oi tl 
     return $ hd:tl

 inline oi = listLift (inline oi) (inline oi)
 reduce oi = listLift (reduce oi) (reduce oi)
 euv oi = listLift (euv oi) (euv oi)
 
 varsIn [] = Set.empty
 varsIn (hd:tl) = varsIn hd `union` varsIn tl


 labsIn lst = Set.unions (fmap labsIn lst)

 refsTo lst v = Set.unions (fmap (\ el -> refsTo el v) lst)

 doSyntaxInfo [] scp  = return ()
 doSyntaxInfo (hd:tl) scp = 
  do doSyntaxInfo hd scp
     doSyntaxInfo tl scp
 
 doMinAlphaSet lst = doFoldAction lst doMinAlphaSet

 substitute [] v' e' = return []
 substitute (hd:tl) v' e' = 
  do hd <- substitute hd v' e'
     tl <- substitute tl v' e'
     return $ hd:tl

 genBinder lst = joins (fmap genBinder lst)

 genSynMap [] = bot
 genSynMap (hd:tl) = (genSynMap hd) `join` (genSynMap tl)

 genScopeMap scope [] = Map.empty
 genScopeMap scope (hd:tl) = 
  genScopeMap scope hd `Map.union` genScopeMap scope tl

 genArgMap lst = joins (fmap genArgMap lst)

 genParentMap lst = joins (fmap genParentMap lst)

 -- Sugar.

listLift fhd ftl [] = return []
listLift fhd ftl (hd:tl) = 
 do hd <- fhd hd
    tl <- ftl tl
    return $ hd:tl

coupleLift fa fb (a,b) = 
 do a <- fa a 
    b <- fb b
    return $ (a,b)

 -- Lattice instances.

instance Lattice Syn where
 top = TopSyn
 bot = BotSyn

instance Lattice EXP where
 top = TopEXP
 bot = BotEXP

instance Lattice CALL where
 top = TopCALL
 bot = BotCALL
 (App l1 _ _)    <: (App l2 _ _)    = l1 == l2
 (LetRec l1 _ _) <: (LetRec l2 _ _) = l1 == l2
 call1 <: call2 | call1 == bot = True
                | call2 == top = True
                | otherwise = False

instance Lattice LAM where
 top = TopLAM
 bot = BotLAM
 (Lam l1 _ _) <: (Lam l2 _ _) = l1 == l2
 lam1 <: lam2 | lam1 == bot = True
              | lam2 == top = True
              | otherwise = False

instance Lattice LAB where
 top = TopLAB
 bot = BotLAB
 l1 <: l2 | l1 == l2  = True
          | l1 == bot = True
          | l2 == top = True
          | otherwise = False


 -- Sugar.

lambda l1 l2 vs call = LamExp (Fn l1) (Lam (Fn l2) vs call)
kont l1 l2 vs call = LamExp (Cn l1) (Lam (Cn l2) vs call)

appf l f as = App (Fn l) f as
appk l q as = App (Cn l) q as

n l i = LitExp (Fn l) (IntLit i)


