module Main where

 -- Personal library.
import Might 

 -- Data structures.
import Data.Set as Set 
import qualified Data.Map as Map
import InfiniteMap as IMap
import Lattice

 -- State, IO.
import Control.Monad.State as MState hiding (join,fix)
import IO
import System

 -- DCFA libraries.
import CPS
import DS
import SymbolTable
import AbstractFrameStrings
import DCFADomains


 -- Common utilities.

class Partitionable a where
 isCont    :: a -> Bool
 mayBeHalt :: a -> Bool

instance Partitionable a => Partitionable (Set.Set a) where
 isCont set = existsIn set (\ a -> isCont a)
 mayBeHalt set = existsIn set (\ a -> mayBeHalt a)

instance Partitionable Val where
 isCont (Clo (Lam (Fn _) vs _) _ _) = 
  all (\ v -> case v of 
                KVar _ -> False
                UVar _ -> True) vs
 isCont (Clo (Lam (Cn _) _ _) _ _) = True
 isCont (Halt _) = True
 isCont _ = False

 mayBeHalt (Halt _) = True
 mayBeHalt _        = False


 -- Auxiliary semantic functions.

eval be ve t (LamExp _ lam)          = singleton (Clo lam be t)
eval be ve t (RefExp _ v)            = ve ! (v, be ! v)
eval be ve t (LitExp _ (IntLit i))   = singleton (IntVal i)
eval be ve t (LitExp _ (FalseLit))   = singleton (FalseVal)
eval be ve t (HaltExp _)             = singleton (Halt t0)
eval be ve t (OmegaExp l)            = singleton (Omega l t0)
eval be ve t (PrimExp l p)           = singleton (Prim l p)
eval be ve t TopEXP                  = error "cannot evaluate top"
eval be ve t BotEXP                  = error "cannot evaluate bot"

 -- Initial state.

t0 = Fn (-1)
b0 = Fn (-1)

init2 :: CALL -> Conf
init2 call = Eval call bot bot
	     bot (b0,t0)

 -- Finding reachable (live) contours.

data CN = Time CNt
        | Binding VAR CNb
 deriving (Eq,Ord)

data ReachableState = RS { explored :: !(Set.Set CN) , varEnv :: !VEnv }

type ExploreState a = MState.State ReachableState a

class Explorable a where
 explore :: a -> ExploreState ()

instance Explorable a => Explorable [a] where
 explore [] = return ()
 explore (hd:tl) = do explore hd
                      explore tl
instance Explorable a => Explorable (Set.Set a) where
 explore set = explore (Set.toList set)


instance Explorable CN where
 explore (Binding v b) = 
  do ve       <- retrieve varEnv
     explored <- retrieve explored
     (if Binding v b `member` explored
      then return ()
      else do st <- MState.get 
              MState.put $ st {explored = Set.insert 
                                          (Binding v b)
                                          explored }
              explore (ve ! (v,b)))

 explore (Time t) = return ()

instance Explorable Val where
 explore (Clo lam be t) =
  do st   <- MState.get
     explored <- retrieve explored
     MState.put $ st { explored = Set.insert (Time t) explored }
     explore [ Binding  v (be!v) | v <- Set.toList(free lam) ] 
     return ()

 explore _ = return ()

instance Explorable Conf where
 explore (Eval call be ve cc (b,t)) =
  do st <- MState.get 
     explored <- retrieve explored
     MState.put $ st { explored = Set.insert (Time t) explored }
     explore [ Binding v (be!v) | v <- Set.toList(free call) ]
     return () 

 explore (Apply proc ds ve  cc (b,t)) =
  do explore proc
     explore ds

reaches :: Conf -> Set.Set CN
reaches conf = explored (execState ex init)
 where init = RS { explored = Set.empty, varEnv = venv conf }
       ex = explore conf

restrictVARxCNb im rch = filterWithKey bot (\ (v,b) _ -> 
                           Binding v b `member` rch) im

restrictCNt im rch = filterWithKey bot (\ t _ ->
                      Time t `member` rch) im


 -- Transition relation.

data StackModel = LUDDY | STEELE
 deriving (Show)

data StateComparsion = WeakerThan | LessThan
 deriving (Show)

 -- Inlinable: maps a reference label to a lambda term 
 --            which may be inlined there.
type Inlinable = InfiniteMap LAB InlinableExp
type Useful = Set.Set VAR

type Copyable = InfiniteMap LAB (Set.Set VAR)

type CallLambdaChain = InfiniteMap LAB (Set.Set LAB)
type LambdaCallChain = InfiniteMap LAB (Set.Set LAB)



 -- Analysis options.

data Options = Options { fileName        :: String,
                         stackModel      :: !StackModel ,
                         stateCompare    :: !StateComparsion ,
                         useFrameStrings :: !Bool ,
		         useGC           :: !Bool ,
		         useCC           :: !Bool ,
		         contours        :: !ContourSet ,
			 optAnalyze      :: !Bool
		       }
 deriving (Show)

initOptions = 
 Options { fileName        = error "No source file selected.",
           stackModel      = STEELE,
           stateCompare    = WeakerThan ,
           useFrameStrings = False ,
	   useGC           = True ,
	   useCC           = True ,
	   contours        = CN_1CFA  ,
	   optAnalyze       = False 
	 }

options opts [] = opts

options opts(("--stack-model"):("STEELE"):tl) =
 options opts { stackModel = STEELE } tl
options opts (("--stack-model"):("LUDDY"):tl) =
 options opts { stackModel = LUDDY } tl
options opts (("--stack-model"):sm:tl) = 
 error $ "Unknown stack model: " ++ sm



options opts(("-O"):tl) =
 options (opts { optAnalyze = True }) tl


options opts(("--no-gc"):tl) =
 options (opts { useGC = False }) tl


options opts(("--no-cc"):tl) =
 options (opts { useCC = False }) tl


options opts (("--contour-set"):("0CFA"):tl) =
 options (opts { contours = CN_0CFA }) tl
options opts (("--contour-set"):("1CFA"):tl) =
 options (opts { contours = CN_1CFA }) tl

options opts (("--state-comparison"):("WT"):tl) =
 options opts { stateCompare = WeakerThan } tl
options opts (("--state-comparison"):("LT"):tl) =
 options opts { stateCompare = LessThan } tl
options opts (("--state-comparison"):sc:tl) =
 error $ "Unknown state comparison operation: " ++ sc

options opts (fn':tl) = 
 options opts { fileName = fn' } tl


 -- Count inlinable.

countInlinable inl =
 IMap.fold (\ _ -> 0 ) (\ ie count -> 
  case ie of 
    TopInExp -> count
    BotInExp -> count 
    _ -> 1 + count )
  inl
  
  
   
  


 -- DCFA State.

data ContourSet = CN_0CFA | CN_1CFA 
 deriving (Eq,Show)


data Stats = Stats { numStates      :: Integer, 
                     numCollections :: Integer  }
 deriving (Show)

data DCFASt = DCFASt { pr          :: PR, 
                       symbolTable :: SymbolTable, 
                       vars        :: Set.Set VAR,
                       inlinable   :: Inlinable,
                       notDead     :: Set.Set LAB,
                       copy        :: Copyable,
                       useful      :: (Useful -> Useful),
                       removable   :: ((Set.Set VAR,Set.Set LAB) ->
                                       (Set.Set VAR,Set.Set LAB)),
                       genRA0      :: Set.Set VAR -> Set.Set LAB,
                       opts        :: Options ,
                       history     :: ![Conf] ,
                       chains      :: (CallLambdaChain,LambdaCallChain) ,
                       stats       :: Stats ,
                       seen        :: !PowConf ,
                       debugLog    :: [String] }
                       
debug s = 
 do state <- get
    put $ state { debugLog = s : (debugLog state) } 

putDebugLog :: DCFASt -> IO ()
putDebugLog state =
 do putStr "*** BEGIN DEBUG\n\n"
    debugLog <- return $ reverse(debugLog state)
    printList debugLog
    putStr "*** END DEBUG\n\n"
 where printList [] = return ()
       printList (hd:tl) = 
        do putStr hd
           putStr "\n\n"
           printList tl
    
class DebugLogContainer a where
 extractDebugLog :: a -> [String]
 injectDebugLog :: [String] -> a

instance SymbolTableContainer DCFASt where
 extractSymbolTable dcfaSt = symbolTable dcfaSt
 injectSymbolTable dcfaSt st' = dcfaSt { symbolTable = st' }

extractStackModel dcfaSt = stackModel (opts dcfaSt)

getStackModel = 
 do state <- MState.get
    return $ extractStackModel state

initDCFASt pr symtable opts = 
 DCFASt { Main.pr     = pr,
          symbolTable = symtable, 
          vars        = varsIn pr,
          inlinable   = bot, 
          notDead     = Set.empty,
          copy        = IMap.constant (varsIn pr),
          useful      = id,
          removable   = id,
          genRA0      = const Set.empty,
          opts        = opts, 
          history     = [],
          chains      = (bot,bot) ,
          stats       = Stats 0 0 ,
          seen        = Set.empty ,
          debugLog    = [] }

 -- Logging functions.
record conf = 
 do state         <- MState.get
    Stats tot col <- retrieve stats
    MState.put $ state { history = (conf:(history state)) ,
                         stats = Stats (tot+1) col }

recordCollection :: MState.State DCFASt ()
recordCollection = 
 do state         <- MState.get
    Stats tot col <- retrieve stats
    put $ state { stats = Stats tot (col+1) }


 -- analyze computes analysis results for an individual state.

 -- TODO Bring analyze back in?


analyze :: Conf -> MState.State DCFASt ()

analyze c@(Eval (call@(App l f args)) be ve cc (b,t)) =
 do opts <- retrieve opts

    dcfaSt   <- MState.get 

    inl   <- retrieve inlinable
    inl'  <- return $ inlinableExp (f,procs)
    inl'' <- return $ foldr (\ expval tally ->
                  tally `join` inlinableExp expval) inl' argsXds

    MState.put $ dcfaSt { inlinable = inl  `join` inl'  `join` inl'' }    

    return ()
 where procs   = eval be ve t f
       ds      = fmap (eval be ve t) args
       cs      = Prelude.filter isCont ds
       argsXds = zip args ds

       inlinableExp (RefExp lab var,d) = result where
          valCandidate = Set.fold (\ val soFar -> val `join` soFar) bot d
          flowCond = not(valCandidate == top)

           -- If we were to inline, what would we inline?
          inlinedExp = case valCandidate of
                        Clo (Lam lab _ _) be_b t_b -> InLamLab lab
                        Prim l p -> InPrim p
                        IntVal i -> InLit (IntLit i)
                        FalseVal -> InLit (FalseLit)
                        Halt _ -> InHalt
                        TopVal -> top
                        _ -> error "expression not inlinable"

          envCondX = case valCandidate of
           Clo lam _ _ -> allIn (free lam) (\ v -> (be ! v) /= bot) 
           _ -> True


           -- No free vars? Trivially inlinable.
          envCond0 = case valCandidate of
                      Clo lam be_b t_b -> Set.null(free lam)
                      Prim l _ -> True
                      IntVal _ -> True
                      FalseVal -> True
                      Halt _ -> True
                      _ -> False

           -- Each free var equal in current scope? Inlinable.
          envCond1 = case valCandidate of 
                      Clo lam be_b t_b -> beEq (free lam) be be_b 
--                            || (allIn (free lam) 
--                                (\ v -> (be ! v /= bot) && -- In scope?
--                                        (log ! (be ! v)) `eqv_cn` (log ! t_b)))
                      _ -> False

           -- Trivial motion since closure birth? Inlinable.
--           envCond2 = case valCandidate of
--                       Clo lam be_b t_b -> (log!t_b) `eqv_cn` emptyString
--                       _ -> False

          result =
           if envCondX && flowCond && (envCond0 || envCond1)
           then bot `but` [lab ==> inlinedExp]
           else bot `but` [lab ==> top]

       inlinableExp _ = bot 

        -- beEq vars be1 be2 checks that be1|vars = be2|vars.
       beEq vars be1 be2 = allIn vars (\ v ->
                         -- Both variables must be in scope.
                        (be1 ! v) /= bot && 
		        (be2 ! v) /= bot &&

                         -- Equal by contour counting?
                         ((be1 ! v) == (be2 ! v) &&
                          (cc ! (v,be1 ! v)) == ONE))

analyze _ = return ()

       

 -- next computes the subsequent states of a given state.
next :: Conf -> MState.State DCFASt [Conf]

next c@(Apply(Halt t0') [d] ve  cc (b,t)) = return $! []

next c@(Apply(Omega _ t0') ds ve cc (b,t)) = return $! []

next c@(Apply(Clo(Lam l vs call) be tb) ds ve cc (b,t)) = 
 do sm       <- getStackModel
    opts     <- retrieve opts
    dispConf <- display c


    cc' <- return $! if (useGC opts && ccZombies) then restrictVARxCNb cc reachable else cc
    ve' <- return $! if (useGC opts && veZombies) then restrictVARxCNb ve reachable else ve
    cc'' <- return $!  cc' + (bot `but` [vb ==> ONE | vb <- newBindings])


    ve''     <- return $! ve' `join` (bot `but` [ (v,b) ==> d 
       | (v,d) <- 
            -- zipFail ("argument mismatch in Apply state\n:" ++dispConf) vs ds u 
            zip vs ds
            ])

    (if veZombies || ccZombies
     then recordCollection
     else return ())
    
    return $! 
     if length vs == length ds
     then [Eval call be' ve'' 
	   (if (useCC opts) then cc'' else cc)
           (b,t)]
     else []
  where newBindings = [v ==> b | v <- vs ]
        veZombies   = any (\ vb -> (ve ! vb) /= bot) newBindings
        ccZombies   = any (\ vb -> (cc ! vb) /= ZERO) newBindings
        reachable   = reaches c

        be' = be `but` newBindings

 {- TODO: Verify optimality of lazy collecting.
    Only collect before a zombie might be created. -}
next c@(Eval (call@(App l f args)) be ve cc (b,t)) =
 do sm    <- getStackModel
    opts  <- retrieve opts
    dispCall <- display call
    (b',t') <- return $! if (contours opts == CN_0CFA)
                         then (b,t)
                         else (l,l)
--     succs <- return $! fmap (\proc -> 
--                         let 
--                          in case proc of
--                              IntVal _ -> error $ "calling int in: "
--                                         ++ dispCall
--                              _ -> Apply proc ds ve cc (b',t'))
--               (Set.toList (eval be ve t f))
    succs <- return $! [  Apply proc ds ve cc (b',t') |
                           proc <- (Set.toList (eval be ve t f))  ]
    return succs
 where 
       ds = [ eval be ve t e | e <- args ] 
       cs = [ c | c <- ds, isCont c ] 


next c@(Eval lr@(LetRec l vargs call) be ve  cc (b,t)) =
 do opts <- retrieve opts

    (b',t') <- return $! if (contours opts == CN_0CFA)
                         then (b,t)
                         else (l,l)

    (vs,_) <- return $! unzip vargs
    newBindings <- return $! [ v ==> b' | v <- vs ] 

    veZombies <- return $! any (\ vb -> (ve ! vb) /= bot) newBindings
    ccZombies <- return $! any (\ vb -> (cc ! vb) /= ZERO) newBindings


    be'  <- return $! be `but` newBindings


    cc'  <- return $! if (useGC opts && ccZombies) then restrictVARxCNb cc reachable else cc
    ve'  <- return $! if (useGC opts && veZombies) then restrictVARxCNb ve reachable else ve
    cc'' <- return $! cc' + (bot `but` [ vb ==> ONE | vb <- newBindings])

    ve'' <- return $! ve' `join` (bot `but`
                    [ (v,b') ==> (eval be' ve t' arg) | (v,arg) <- vargs ])


    (if veZombies || ccZombies
     then recordCollection
     else return ())
    return $! [Eval call be' ve'' 
	    (if (useCC opts) then cc'' else cc)
	    (b',t')]
 where 


--        b'    = l
--        t'    = l

       reachable = reaches c

 -- NOTE: the stack is popped in primitive operations.
 -- We have to pop it to simulate going through an Eval state.

next c@(Apply(Prim l (Arith arithOp)) [x,y,cont] ve  cc (b,t)) =
 do sm    <- getStackModel
    opts  <- retrieve opts
    cc'   <- return $ cc
    (b',t') <- return $! if (contours opts == CN_0CFA)
                         then (b,t)
                         else (l,l)
    return $! [Apply proc [singleton TopInt] ve
                cc'
                (b',t')
              | proc <- Set.toList cont ]
 where ans = absArith arithOp (Set.toList x) (Set.toList y)
--        absArith ADD [IntVal a] [IntVal b] = singleton (IntVal (a + b))
--        absArith MUL [IntVal a] [IntVal b] = singleton (IntVal (a * b))
--        absArith SUB [IntVal a] [IntVal b] = singleton (IntVal (a - b))
--        absArith DIV [IntVal a] [IntVal b] = singleton (IntVal (a `div` b))
       absArith _   a          b          = singleton TopInt
       -- TODO: If it cannot possibly be two ints, we can stop.



next c@(Apply(Prim l (IOOp DISPLAY)) [x,cont] ve  cc (b,t)) =
 do sm    <- getStackModel
    opts  <- retrieve opts
    cc'   <- return $ cc
    (b',t') <- return $! if (contours opts == CN_0CFA)
                         then (b,t)
                         else (l,l)
    return $! [Apply proc [] ve 
                cc'
                (b',t') -- TODO: tick goes here.
              | proc <- Set.toList cont ]

next c@(Apply(Prim l (If test)) [cond,ctrue,cfalse] ve  cc (b,t)) =
 do 
    opts <- retrieve opts
    sm <- getStackModel
    cc' <- return $ cc
    (b',t') <- return $! if (contours opts == CN_0CFA)
                         then (b,t)
                         else (l,l)
    return $! [Apply proc [] ve
                cc'
                (b',t')  -- TODO: tick goes here.
               | proc     <- subsequents ]
 where subsequents = absIf test (Set.toList cond)
       absIf ISZERO [IntVal 0] = Set.toList ctrue
       absIf ISZERO lst        = if any (\ val -> case val of
                                                   IntVal 0 -> True
                                                   TopVal -> True
                                                   TopInt -> True
                                                   _ -> False) lst
                                 then (Set.toList ctrue) ++ (Set.toList cfalse)
                                 else Set.toList cfalse

next c@(Apply(Prim l (If _)) ds ve  cc (b,t)) =
 error "if primitive with invalid arguments"


next c =
 do c' <- display c
--    error $ "Explosion on: " ++ c'
    return $! []


 -- visited computes the set of reachable states.
vis pr = if isAlphaUnique pr
         then do state <- get
                 put $ state { pr      = pr ,
                          seen    = Set.empty ,
                          history = [] }
                 visited [init2 pr]
         else error "Trying to analyze non-alpha-unique program."



visited :: [Conf] -> (State DCFASt PowConf)
visited [] = retrieve seen
visited (hd:tl) = 
 do 
    seen <- retrieve seen
    opts <- retrieve opts
    test <- return $ stateCompare opts
    (if (optAnalyze opts)
     then analyze hd
     else return ())
    record hd
    str <- display hd
--    analyze hd
    (if case test of
         WeakerThan -> (singleton hd) <: seen
         LessThan -> hd `member` seen
     then do visited tl
     else do new <- next hd
             state <- get
             put $ state { seen = seen `union` singleton hd }
             visited (new ++ tl))




 -- Extension parsing.
extension "" = ""
extension ('.':ext) = ext
extension (_:tl) = extension tl


dcfa pr symtab opts =
 do put $ initDCFASt pr symtab opts





 -- main
main :: IO ()
main = 
 do  -- Parse command line options.
    args <- getArgs
  --  putStr $! (show args)
    opts@(Options file sm sc useFS useGC useCC cnSet optAnalyze) 
      <- return $ options initOptions args
    putStr $ "options = "
    putStr $ (show opts)
    putStr "\n"

     -- Read the source.
    ext          <- return $ extension file
    src          <- readFile file

     -- Tokenize.
    (tokens,st)  <- return $ tokens src

     -- Create an S-Expression from the tokens.
    sexp         <- return $ car(parseSExpList tokens)

     -- Parse into an AST.
    (cpsAST,st) <- if ext == "scm" 
                   then do dsAST <- return $ MState.evalState 
                                               (parseTerm sexp) st
                           return $ pr2call dsAST st
                   else do return $ MState.runState (parseCall sexp) st

     -- Alphatise the AST.
    (cpsAST,st) <- return $ MState.runState (alphatise id cpsAST) st

     -- Check for simple errors: top-level free variables.
--     putStr "Original:\n"
--     putStr $ evalState (display sexp) st
--     putStr "\n\n"
--     putStr $ evalState (display cpsAST) st
--     putStr "\n\n"

    freeVars <- return $ evalState (display (free cpsAST)) st
    -- putStr "\n\n"
    putStr $ if not(Set.null (free cpsAST)) 
             then error $ "syntax error: top level free variables: " ++
                          freeVars
             else ""


     -- Run DCFA.
     
    state <- return $ initDCFASt cpsAST st opts

    finalSt <- return $ execState (vis cpsAST) state


--    putStr $ "analysis complete\n\n" 

    putStr $ "num states = " ++ (show (Set.size (seen finalSt)))
    putStr $ "\n"


    inliningResults <- return $! inlinable finalSt

	   --putStr $ (display inliningResults) ;

    putStr $ "num inlinable expressions = "
    putStr $ (show (countInlinable (inliningResults)))
    putStr $ "\n"

    return ()  
