module DCFADomains where

 -- Personal library.
import Might 

 -- Data structures.
import Data.Set as Set 
import qualified Data.Map as Map
import InfiniteMap as IMap
import Lattice

 -- State, IO.
import Control.Monad.State as MState hiding (join,fix)
import IO
import System

 -- DCFA libraries.
import CPS
import DS
import SymbolTable
import AbstractFrameStrings

 -- DCFA domains.
type CNb = LAB
type CNt = LAB
type CNa = LAB

type Log = InfiniteMap CNt F

type BEnv = InfiniteMap VAR CNb

type VEnv = InfiniteMap (VAR,CNb) D

data Val = Halt CNt
         | Omega LAB CNt
         | Clo LAM BEnv CNt
         | Loc CNa
         | TopInt
         | BotInt
         | IntVal Integer
         | FalseVal
         | Prim LAB PRIM
         | TopVal
         | BotVal
 deriving (Eq,Ord)

type D = Set.Set Val

type Loc = Val

type Store = InfiniteMap Loc D

data CNCount = ZERO | ONE | MORE
 deriving (Eq,Show)

type CNCounter = InfiniteMap (VAR,CNb) CNCount

type Eval = (CALL,BEnv,VEnv,Log,CNCounter,(CNb,CNt))

type Apply = (Val,[D],VEnv,Log,CNCounter,(CNb,CNt))

data Conf = Eval { call :: CALL ,
                   benv :: BEnv ,
                   venv :: VEnv ,
                   log  :: Log ,
                   cnc  :: CNCounter ,
                   cn   :: (CNb,CNt) }
          | Apply { proc  :: Val ,
                    parms :: [D] ,
                    venv  :: VEnv ,
                    log   :: Log ,
                    cnc   :: CNCounter ,
                    cn    :: (CNb,CNt) }
 deriving (Eq,Ord)

type PowConf = Set.Set Conf

instance Ord CNCount where
 compare ZERO ZERO = EQ
 compare ZERO _    = LT
 compare ONE  ZERO = GT
 compare ONE  ONE  = EQ
 compare ONE  MORE = LT 
 compare MORE MORE = EQ
 compare MORE _    = GT
 

 -- Contour counting.

-- addCounter :: Ord k => InfiniteMap k CNCount -> InfiniteMap k CNCount -> 
--                InfiniteMap k CNCount
-- addCounter = IMap.lift addCount

instance Num CNCount where
 ZERO + a    = a
 a    + ZERO = a
 ONE  + ONE  = MORE 
 MORE + _    = MORE
 _    + MORE = MORE
 fromInteger 0 = ZERO
 fromInteger 1 = 1
 fromInteger n | n >= 2 = MORE
 abs a = a
 signum a = ONE

instance (Show k, Show a, Ord k, Num a) => Num (InfiniteMap k a) where
 a + b = IMap.lift (+) a b
  

 -- Display instances.

instance Display Val where
 display (Clo lam be t) = 
  do lam' <- display lam
     be'  <- display be
     t'   <- display t
     return $ "Clo " ++ lam' ++ " " ++ be' ++ " " ++ t'
 
 display (TopVal) = return "TopVal"
 display (BotVal) = return "BotVal"
 display (TopInt) = return "TopInt"
 display (BotInt) = return "BotInt"
 display (Prim l p) = return (show p)
 display (IntVal i) = return (show i)
 display (FalseVal) = return "#f"
 display (Halt t) = return "Halt"
 display (Omega l t) = return "Omega"

instance Display CNCount where 
 display ZERO = return "0"
 display ONE  = return "1"
 display MORE = return "*"

instance Display Delta where
 display d = return $ show d

instance Display Conf where
 display (Eval call be ve log cc (b,t)) =
  do call' <- display call
     be'   <- display be
     ve'   <- display ve
     log'  <- display log
     cc'   <- display cc
     b'    <- display b
     t'    <- display t
     return $ "Eval(" ++ call' ++ ",\n "
                      ++ be' ++  ",\n "
                      ++ ve' ++  ",\n "
                      ++ cc' ++  ",\n "
                      ++ log' ++ ",\n "
                      ++ "(" ++ b' ++ "," ++ t' ++ ")" ++ ")"

 display (Apply proc ds ve log cc (b,t)) =
  do proc' <- display proc
     ds'   <- display ds
     ve'   <- display ve
     log'  <- display log
     cc'   <- display cc
     b'    <- display b
     t'    <- display t
     return $ "Apply(" ++ proc' ++ ",\n "
                       ++ ds' ++  ",\n "
                       ++ ve' ++  ",\n "
                       ++ cc' ++  ",\n "
                       ++ log' ++ ",\n "
                       ++ "(" ++ b' ++ "," ++ t' ++ ")" ++ ")"

 displayList [] = do return ""
 displayList (hd:tl) = 
  do hd' <- display hd
     tl' <- displayList tl
     return $ hd' ++ "\n\n" ++ tl'


 -- Lattice instances.

instance Lattice CNCount where
 bot = ZERO
 top = MORE
 a   <:   b = a <= b
 a `join` b = max a b
 a `meet` b = min a b

instance Lattice Val where
 bot = BotVal
 top = TopVal

 BotVal     <: _            = True
 _          <: TopVal       = True
 BotInt     <: (IntVal _)   = True
 (IntVal _) <: TopInt       = True

 a <: b | a == b    = True
        | otherwise = False
 
 (IntVal lit1) `join` (IntVal lit2) | lit1 == lit2 = (IntVal lit1)
                                    | otherwise = TopInt
 a `join` b | a <: b     = b
            | b <: a     = a
            | otherwise  = TopVal

 (IntVal lit1) `meet` (IntVal lit2) | lit1 == lit2 = (IntVal lit1)
                                    | otherwise = BotInt
 a `meet` b | a <: b     = a
            | b <: a     = b
            | otherwise  = BotVal

 -- Ordered by the speed of the test.
instance Lattice Conf where
 (Eval call be ve log cc cn) <: (Eval call' be' ve' log' cc' cn')
   = cn   <: cn'   &&
     call <: call' &&
     be   <: be'   &&
     cc   <: cc'   &&
     ve   <: ve'   &&
     log  <: log'  
    
 (Apply proc ds ve log cc cn) <: (Apply proc' ds' ve' log' cc' cn')
   = cn   <: cn'    &&
     proc <: proc'  &&
     ds   <: ds'    &&
     cc   <: cc'    &&
     ve   <: ve'    &&
     log  <: log'

 _ <: _ = False

