module DS where

import Prelude hiding (lex)

import Data.Set as Set
import InfiniteMap as IMap
import Control.Monad.State

import Lattice

import Might
import SymbolTable



 -- A lexer for S-Expressions.

data LEXSTATE = StartSt | SymSt String | Comment

data TOKEN = EOF | LPAR | RPAR | SYM Integer | INT Integer | SHARP String
 deriving (Eq,Ord,Show)

nonID = Set.fromList"() \n\t`',;"

toInt prefix [] = Just prefix
toInt prefix ('0':tl) = toInt (prefix * 10 + 0) tl
toInt prefix ('1':tl) = toInt (prefix * 10 + 1) tl
toInt prefix ('2':tl) = toInt (prefix * 10 + 2) tl
toInt prefix ('3':tl) = toInt (prefix * 10 + 3) tl
toInt prefix ('4':tl) = toInt (prefix * 10 + 4) tl
toInt prefix ('5':tl) = toInt (prefix * 10 + 5) tl
toInt prefix ('6':tl) = toInt (prefix * 10 + 6) tl
toInt prefix ('7':tl) = toInt (prefix * 10 + 7) tl
toInt prefix ('8':tl) = toInt (prefix * 10 + 8) tl
toInt prefix ('9':tl) = toInt (prefix * 10 + 9) tl
toInt prefix (c:tl) = Nothing


lex StartSt []        = do return (EOF,[])
lex Comment ('\n':tl) = do lex StartSt tl
lex Comment (_:tl)    = do lex Comment tl
lex StartSt (';':tl)  = do lex Comment tl
lex StartSt (' ':tl)  = do lex StartSt tl
lex StartSt ('\t':tl) = do lex StartSt tl
lex StartSt ('\n':tl) = do lex StartSt tl
lex StartSt ('(':tl)  = do return (LPAR,tl)
lex StartSt (')':tl)  = do return (RPAR,tl)
lex StartSt ('#':'f':tl) = do return (SHARP "f",tl)
lex StartSt (c:tl)    = do lex (SymSt [c]) tl
lex (SymSt rid) (c:tl) | c `member` nonID =
                          do id <- return $ reverse rid
                             tok <- case toInt 0 id of
                                     Just int -> do return (INT int)
                                     Nothing -> do i <- getSymbol id 
                                                   return (SYM i)
                             return (tok,c:tl) 
                       | otherwise = do lex (SymSt (c:rid)) tl


lexString [] = do return [ EOF ]
lexString input = do (tok,input') <- lex StartSt input
                     rest <- if tok == EOF
                             then return $ [] 
                             else lexString input'
                     return $ tok:rest


tokens input = runState (lexString input) emptySymbolTable



 -- An S-Expression grammar.

data SEXP = SymSExp Integer
          | IntSExp Integer
          | ConsSExp SEXP SEXP
          | NilSExp
          | SharpSExp String
 deriving (Eq,Ord,Show)

car (ConsSExp first rest) = first
car mismatch = error $ "cannot take car of: " ++ (show mismatch)

cdr (ConsSExp first rest) = rest

parseSExpList toks = fst(parse toks) 
 where parse [] = error "No EOF!"
       parse [EOF] = (NilSExp,[])
       parse (RPAR:tl) = (NilSExp,tl)
       parse (LPAR:tl) = (ConsSExp subexp rest,tl'')
        where (subexp,tl') = parse tl
              (rest,tl'') = parse tl'
       parse ((SHARP s):tl) = (ConsSExp (SharpSExp s) rest,tl')
        where (rest,tl') = parse tl
       parse ((INT n):tl) = (ConsSExp (IntSExp n) rest,tl')
        where (rest,tl') = parse tl
       parse ((SYM s):tl) = (ConsSExp (SymSExp s) rest,tl')
        where (rest,tl') = parse tl
       parse bogus = error (show bogus)


instance Display SEXP where
 display (NilSExp)     = return "()"
 display (SharpSExp s) = return $ "#" ++ s
 display (SymSExp s)   = stringFromSymbol s
 display (IntSExp i)   = return $ show i

 display (ConsSExp car NilSExp) = 
  do car' <- display car
     return $ "(" ++ car' ++ ")"
 display (ConsSExp car cdr@(ConsSExp _ _)) = 
  do car' <- display car
     cdr' <- displayConsList cdr 
     return $ "(" ++ car' ++ " " ++ cdr'  ++ ")"
  where
   displayConsList (NilSExp) = return ""
   displayConsList (ConsSExp car NilSExp) = display car
   displayConsList (ConsSExp car cdr@(ConsSExp _ _))  =
    do car' <- display car
       cdr' <- displayConsList cdr
       return $ car' ++ " " ++  cdr'
   displayConsList (ConsSExp car cdr) = 
    do car' <- display car
       cdr' <- display cdr
       return $ car' ++ " . " ++ cdr'
        


 display (ConsSExp car cdr) =
  do car' <- display car
     cdr' <- display cdr
     return $ car' ++ " . " ++ cdr'



 -- A Direct-Style grammar.

 -- Labels.
data LAB = Fn { labSym :: Integer }
         | Cn { labSym :: Integer }
         | TopLAB
         | BotLAB
 deriving (Eq,Ord)


 -- Variables.
data VAR = UVar { varSym :: Integer }
         | KVar { varSym :: Integer }
         | TopVAR
         | BotVAR
 deriving (Eq,Ord,Show) 


 -- Literals.
data LIT = IntLit Integer
         | FalseLit
 deriving (Eq,Ord)

 -- Primops.
data ARITH = ADD | SUB | MUL | DIV
 deriving (Eq,Ord)

data TEST = ISZERO
 deriving (Eq,Ord)

data CONSOP = CONS | CAR | CDR 
 deriving (Eq,Ord)

data CHANOP = SWITCH 
 deriving (Eq,Ord)

data IOOP = DISPLAY
 deriving (Eq,Ord)

data PRIM = Arith ARITH
          | If TEST
          | CallCC
          | ConsOp CONSOP
          | ChanOp CHANOP
          | IOOp IOOP
 deriving (Eq,Ord)


 -- Lattice instances.
instance Lattice VAR where
 top = TopVAR
 bot = BotVAR


 -- Show instances.
instance Show ARITH where
 show ADD = "+"
 show SUB = "-"
 show MUL = "*"
 show DIV = "/"

instance Show TEST where
 show ISZERO = "if-zero"

instance Show IOOP where
 show DISPLAY = "display"

instance Show PRIM where
 show (Arith oper) = show oper
 show (If cond) = show cond
 show (IOOp iop) = show iop

instance Show LIT where
 show (IntLit i) = show i
 show (FalseLit) = "#f"




 -- Direct-style expressions.

data TERM = AppTerm TERM [TERM]
          | LamTerm [VAR] TERM
          | LetRecTerm [(VAR,TERM)] TERM
          | RefTerm VAR
          | LitTerm LIT
          | BeginTerm TERM TERM
          | PrimTerm PRIM

isAtomicTerm (RefTerm v) = True
isAtomicTerm (PrimTerm p) = True
isAtomicTerm (LitTerm lit) = True
isAtomicTerm (LamTerm vs body) = True
isAtomicTerm _ = False

freshUserVar :: SymbolTableContainer b => (State b VAR)
freshUserVar = do us <- freshUserSymbol
                  return $ UVar us

freshContVar :: SymbolTableContainer b => (State b VAR)
freshContVar = do ks <- freshContSymbol
                  return $ KVar ks

freshUserLabel :: SymbolTableContainer b => (State b LAB)
freshUserLabel = do us <- freshUserSymbol
                    return $ Fn us

freshContLabel :: SymbolTableContainer b => (State b LAB)
freshContLabel = do ks <- freshContSymbol
                    return $ Cn ks

 -- Display instances.

instance Display LIT where
 display lit = return $ show lit

instance Display PRIM where
 display prim = return $  show prim

instance Display VAR where
 display (UVar sym) = stringFromSymbol sym
 display (KVar sym) = stringFromSymbol sym

 displayList []     = return "" 
 displayList [v]    = display v
 displayList (v:vs) = do v' <- display v
                         vs' <- display vs
                         return $ v' ++ " " ++ vs'

instance Display TERM where
 display (PrimTerm prim)   = return $ show prim
 display (LitTerm lit    ) = return $ show lit
 display (RefTerm (UVar i)) = stringFromSymbol i
 display (LamTerm vs body) = 
  do vs'   <- display vs
     body' <- display body
     return $ "(lambda (" ++ vs' ++ ") " ++ body' ++ ")"
 display (AppTerm f es) = 
  do f'  <- display f
     es' <- display es
     return $  "(" ++ f' ++ " " ++ es' ++ ")"



mapTrueList f (NilSExp) = do return $ []
mapTrueList f (ConsSExp hd tl) = 
 do hd' <- f hd
    tl' <- mapTrueList f tl
    return $ hd':tl'


parseUVar :: SymbolTableContainer b => SEXP -> (State b VAR)
parseUVar (SymSExp v) = do return $ (UVar v)

parseKVar :: SymbolTableContainer b => SEXP -> (State b VAR)
parseKVar (SymSExp v) = do return $ (KVar v)

parseUVarList :: SymbolTableContainer b => SEXP -> (State b [VAR])
parseUVarList = mapTrueList parseUVar

parseKVarList :: SymbolTableContainer b => SEXP -> (State b [VAR])
parseKVarList = mapTrueList parseKVar

parseLamTerm :: SymbolTableContainer b => SEXP -> (State b TERM)
parseLamTerm sexp =
  do args  <- return $ car(cdr(sexp))
     body  <- return $ car(cdr(cdr(sexp)))
     args' <- parseUVarList args
     body' <- parseTerm body
     return $ LamTerm args' body'

parseLetTerm :: SymbolTableContainer b => SEXP -> (State b TERM)
parseLetTerm sexp = 
 do vterms       <- return $ car(cdr(sexp))
    body         <- return $ car(cdr(cdr(sexp)))
    vterms'      <- parseVTermList vterms
    (vs',terms') <- return $ unzip vterms'
    body'        <- parseTerm body
    return $ (AppTerm (LamTerm vs' body') terms')

parseVTerm :: SymbolTableContainer b => SEXP -> (State b (VAR,TERM))
parseVTerm (ConsSExp (SymSExp sym) (ConsSExp term NilSExp)) =
 do term' <- parseTerm term
    return (UVar sym,term')

parseVTermList :: SymbolTableContainer b => SEXP -> (State b [(VAR,TERM)])
parseVTermList = mapTrueList parseVTerm


parseLetRecTerm sexp =
 do vterms  <- parseVTermList $ car(cdr(sexp))
    body    <- parseTerm      $ car(cdr(cdr(sexp)))
    return $ LetRecTerm vterms body

parseBeginTerm sexp = 
 do s1 <- parseTerm (car(cdr(sexp)))
    s2 <- parseTerm (car(cdr(cdr(sexp))))
    return $ BeginTerm s1 s2

parseArgTerms NilSExp = do return []
parseArgTerms (ConsSExp arg args) = 
  do e   <- parseTerm arg
     es  <- parseArgTerms args
     return $ e:es

parseAppTerm sexp = 
 do f     <- return $ car(sexp)
    args  <- return $ cdr(sexp)
    f'    <- parseTerm f
    args' <- parseArgTerms args
    return $ AppTerm f' args'

parseTerm (SymSExp i) = 
 do mul   <- getSymbol "*"
    sub   <- getSymbol "-"
    add   <- getSymbol "+"
    ifz   <- getSymbol "if-zero"
    ifn   <- getSymbol "if-nil"
    ccc   <- getSymbol "call/cc"
    cons  <- getSymbol "cons"
    car   <- getSymbol "car"
    cdr   <- getSymbol "cdr"
    nil   <- getSymbol "nil"
    swch  <- getSymbol "switch"
    dsp   <- getSymbol "display"
    (if   i == mul 
     then return $ PrimTerm (Arith MUL)
     else if i == sub
     then return $ PrimTerm (Arith SUB)
     else if i == add
     then return $ PrimTerm (Arith ADD)
     else if i == ifz
     then return $ PrimTerm (If ISZERO)
     else if i == ccc
     then return $ PrimTerm (CallCC)
     else if i == dsp
     then return $ PrimTerm (IOOp DISPLAY)
--      else if i == swch
--      then switchTerm
--      else if i == cons
--      then consTerm
--      else if i == car
--      then carTerm
--      else if i == cdr
--      then cdrTerm
     else return $ RefTerm (UVar i))
parseTerm (SharpSExp "f") = do return $ LitTerm (FalseLit)
parseTerm (IntSExp i) = do return $ LitTerm (IntLit i)
parseTerm (term@(ConsSExp (SymSExp f) args)) =
 do lambda <- getSymbol "lambda"
    letrec <- getSymbol "letrec" 
    let'   <- getSymbol "let"
    begin  <- getSymbol "begin"
    (if (f == lambda)
     then do parseLamTerm term
     else if (f == letrec)
     then do parseLetRecTerm term
     else if (f == let') 
     then do parseLetTerm term
     else if (f == begin) 
     then do parseBeginTerm term
     else do parseAppTerm term)
parseTerm (term@(ConsSExp f args)) = do parseAppTerm term

 -- Sugar.

consTerm :: SymbolTableContainer b => (State b TERM)
consTerm = 
 do a     <- freshUserVar
    b     <- freshUserVar
    field <- freshUserVar
    return $ LamTerm [a,b] 
              (LamTerm [field] 
               (AppTerm (PrimTerm (If ISZERO)) 
                        [RefTerm field,RefTerm a,RefTerm b]))

carTerm :: SymbolTableContainer b => (State b TERM)
carTerm =
 do cell <- freshUserVar
    return $ LamTerm [cell] (AppTerm (RefTerm cell) [LitTerm (IntLit 0)])

cdrTerm :: SymbolTableContainer b => (State b TERM)
cdrTerm =
 do cell <- freshUserVar
    return $ LamTerm [cell] (AppTerm (RefTerm cell) [LitTerm (IntLit 1)])

 -- Transducer sugar.

switchTerm :: SymbolTableContainer b => (State b TERM)
switchTerm =
 do toks <- lexString (error "fix this")
--     "(lambda (chan) (call/cc (lambda (k0) ((cdr k) (cons (car chan) k0)))))"
    sexp <- return $ car(parseSExpList toks)
    term <- parseTerm sexp
    return term
    


 -- Transformers.
