module Main where

 -- Personal library.
import Might 

 -- Data structures.
import Data.Set as Set 
import qualified Data.Map as Map
import InfiniteMap as IMap
import Lattice

 -- State, IO.
import Control.Monad.State as MState hiding (join,fix)
import IO
import System

 -- DCFA libraries.
import CPS
import DS
import SymbolTable
import AbstractFrameStrings
import DCFADomains


 -- Common utilities.

class Partitionable a where
 isCont    :: a -> Bool
 mayBeHalt :: a -> Bool

instance Partitionable a => Partitionable (Set.Set a) where
 isCont set = existsIn set (\ a -> isCont a)
 mayBeHalt set = existsIn set (\ a -> mayBeHalt a)

instance Partitionable Val where
 isCont (Clo (Lam (Fn _) vs _) _ _) = 
  all (\ v -> case v of 
                KVar _ -> False
                UVar _ -> True) vs
 isCont (Clo (Lam (Cn _) _ _) _ _) = True
 isCont (Halt _) = True
 isCont _ = False

 mayBeHalt (Halt _) = True
 mayBeHalt _        = False


 -- Log functions.

topLog = IMap.constant topF

catLog :: Log -> F -> Log
catLog log f = if f == bot 
               then bot
               else IMap.lift catF log (IMap.constant f)

emptyAt :: CNt -> Log
emptyAt t = IMap.insert t emptyString bot

class Aged a where
 age :: Log -> a -> F

instance Aged Val where
 age log (Halt t0) = log ! t0
 age log clo@(Clo _ be t) = 
  if isCont clo
  then log ! t
  else error "Non-continuation closure used in age value calculation."
 age log (Prim _ p) = error $ "Primitive ("++(show p)
                                 ++") used in age calculation."
 age log (IntVal i) = error $ "Integer ("++(show i)
                                 ++") used in age calculation."
 age log (TopInt)   = error "TopInt used in age calculation."
 age log (BotInt)   = error "BotInt used in age calculation."
 age log (TopVal)   = error "TopVal used in age calculation."
 age log (BotVal)   = error "BotVal used in age calculation."
 age log (FalseVal) = error "#f used in age calculation."

instance Aged a => Aged (Set.Set a) where
 age log set = Set.fold (\m old -> (age log m) `join` old) bot set

instance Aged a => Aged [a] where
 age log lst = foldl (\ old m -> (age log m) `join` old) bot lst


 -- Auxiliary semantic functions.

eval be ve t (LamExp _ lam)          = singleton (Clo lam be t)
eval be ve t (RefExp _ v)            = ve ! (v, be ! v)
eval be ve t (LitExp _ (IntLit i))   = singleton (IntVal i)
eval be ve t (LitExp _ (FalseLit))   = singleton (FalseVal)
eval be ve t (HaltExp _)             = singleton (Halt t0)
eval be ve t (OmegaExp l)            = singleton (Omega l t0)
eval be ve t (PrimExp l p)           = singleton (Prim l p)
eval be ve t TopEXP                  = error "cannot evaluate top"
eval be ve t BotEXP                  = error "cannot evaluate bot"

 -- Initial state.

t0 = Fn (-1)
b0 = Fn (-1)

init2 :: CALL -> Conf
init2 call = Eval call bot bot (bot `but` [t0 ==> emptyString]) 
	     bot (b0,t0)

 -- Finding reachable (live) contours.

data CN = Time CNt
        | Binding VAR CNb
 deriving (Eq,Ord)

data ReachableState = RS { explored :: Set.Set CN , varEnv :: VEnv }

type ExploreState a = MState.State ReachableState a

class Explorable a where
 explore :: a -> ExploreState ()

instance Explorable a => Explorable [a] where
 explore [] = return ()
 explore (hd:tl) = do explore hd
                      explore tl
instance Explorable a => Explorable (Set.Set a) where
 explore set = explore (Set.toList set)


instance Explorable CN where
 explore (Binding v b) = 
  do ve       <- retrieve varEnv
     explored <- retrieve explored
     (if Binding v b `member` explored
      then return ()
      else do st <- MState.get 
              MState.put $ st {explored = Set.insert 
                                          (Binding v b)
                                          explored }
              explore (ve ! (v,b)))

 explore (Time t) = return ()

instance Explorable Val where
 explore (Clo lam be t) =
  do st   <- MState.get
     explored <- retrieve explored
     MState.put $ st { explored = Set.insert (Time t) explored }
     explore [ Binding  v (be!v) | v <- Set.toList(free lam) ] 
     return ()

 explore _ = return ()

instance Explorable Conf where
 explore (Eval call be ve log cc (b,t)) =
  do st <- MState.get 
     explored <- retrieve explored
     MState.put $ st { explored = Set.insert (Time t) explored }
     explore [ Binding v (be!v) | v <- Set.toList(free call) ]
     return () 

 explore (Apply proc ds ve log cc (b,t)) =
  do explore proc
     explore ds

reaches :: Conf -> Set.Set CN
reaches conf = explored (execState ex init)
 where init = RS { explored = Set.empty, varEnv = venv conf }
       ex = explore conf

restrictVARxCNb im rch = filterWithKey bot (\ (v,b) _ -> 
                           Binding v b `member` rch) im

restrictCNt im rch = filterWithKey bot (\ t _ ->
                      Time t `member` rch) im


 -- Transition relation.

data StackModel = LUDDY | STEELE
 deriving (Show)

data StateComparsion = WeakerThan | LessThan
 deriving (Show)

 -- Inlinable: maps a reference label to a lambda term 
 --            which may be inlined there.
type Inlinable = InfiniteMap LAB InlinableExp
type Useful = Set.Set VAR

type Copyable = InfiniteMap LAB (Set.Set VAR)

type CallLambdaChain = InfiniteMap LAB (Set.Set LAB)
type LambdaCallChain = InfiniteMap LAB (Set.Set LAB)

 -- Analysis options.

data Options = Options { fileName        :: String,
                         stackModel      :: StackModel,
                         stateCompare    :: StateComparsion,
                         useFrameStrings :: Bool }
 deriving (Show)

initOptions = 
 Options { fileName        = error "No source file selected.",
           stackModel      = STEELE,
           stateCompare    = WeakerThan ,
           useFrameStrings = False }

options opts [] = opts

options opts(("--stack-model"):("STEELE"):tl) =
 options opts { stackModel = STEELE } tl
options opts (("--stack-model"):("LUDDY"):tl) =
 options opts { stackModel = LUDDY } tl
options opts (("--stack-model"):sm:tl) = 
 error $ "Unknown stack model: " ++ sm

options opts (("--state-comparison"):("WT"):tl) =
 options opts { stateCompare = WeakerThan } tl
options opts (("--state-comparison"):("LT"):tl) =
 options opts { stateCompare = LessThan } tl
options opts (("--state-comparison"):sc:tl) =
 error $ "Unknown state comparison operation: " ++ sc

options opts (fn':tl) = 
 options opts { fileName = fn' } tl


 -- DCFA State.

data Stats = Stats { numStates      :: Integer, 
                     numCollections :: Integer  }
 deriving (Show)

data DCFASt = DCFASt { pr          :: PR, 
                       symbolTable :: SymbolTable, 
                       vars        :: Set.Set VAR,
                       inlinable   :: Inlinable,
                       notDead     :: Set.Set LAB,
                       copy        :: Copyable,
                       useful      :: (Useful -> Useful),
                       removable   :: ((Set.Set VAR,Set.Set LAB) ->
                                       (Set.Set VAR,Set.Set LAB)),
                       genRA0      :: Set.Set VAR -> Set.Set LAB,
                       opts        :: Options ,
                       history     :: [Conf] ,
                       chains      :: (CallLambdaChain,LambdaCallChain) ,
                       stats       :: Stats ,
                       seen        :: PowConf ,
                       debugLog    :: [String] }
                       
debug s = 
 do state <- get
    put $ state { debugLog = s : (debugLog state) } 

putDebugLog :: DCFASt -> IO ()
putDebugLog state =
 do putStr "*** BEGIN DEBUG\n\n"
    debugLog <- return $ reverse(debugLog state)
    printList debugLog
    putStr "*** END DEBUG\n\n"
 where printList [] = return ()
       printList (hd:tl) = 
        do putStr hd
           putStr "\n\n"
           printList tl
    
class DebugLogContainer a where
 extractDebugLog :: a -> [String]
 injectDebugLog :: [String] -> a

instance SymbolTableContainer DCFASt where
 extractSymbolTable dcfaSt = symbolTable dcfaSt
 injectSymbolTable dcfaSt st' = dcfaSt { symbolTable = st' }

extractStackModel dcfaSt = stackModel (opts dcfaSt)

getStackModel = 
 do state <- MState.get
    return $ extractStackModel state

initDCFASt pr symtable opts = 
 DCFASt { Main.pr     = pr,
          symbolTable = symtable, 
          vars        = varsIn pr,
          inlinable   = bot, 
          notDead     = Set.empty,
          copy        = IMap.constant (varsIn pr),
          useful      = id,
          removable   = id,
          genRA0      = const Set.empty,
          opts        = opts, 
          history     = [],
          chains      = (bot,bot) ,
          stats       = Stats 0 0 ,
          seen        = Set.empty ,
          debugLog    = [] }

 -- Logging functions.
record conf = 
 do state         <- MState.get
    Stats tot col <- retrieve stats
    MState.put $ state { history = (conf:(history state)) ,
                         stats = Stats (tot+1) col }

recordCollection :: MState.State DCFASt ()
recordCollection = 
 do state         <- MState.get
    Stats tot col <- retrieve stats
    put $ state { stats = Stats tot (col+1) }


 -- analyze computes analysis results for an individual state.

analyzeUVE :: Conf -> State DCFASt ()

analyzeUVE (Eval (App kappa f args) be ve log cc (b,t)) =
 do state <- get
    put $ state
    return ()
 where procs = eval be ve t f
       cs      = Prelude.filter isCont ds
       ds      = fmap (eval be ve t) args

        -- The useful variables generator.

       usefulGen alreadyUseful = 
        let
          refSet (RefExp _ var) = singleton var
          refSet _              = Set.empty

           -- TODO: Enforce condition that primitives must appear in function position. 

           -- Function position variables are useful.
           -- Predicate variables are useful. 
           -- Variables as continuations to a primop are useful.
           -- Variables given to halt are useful.
          locallyUseful = 
           case f of 
            RefExp  _ var       -> singleton var
            PrimExp _ (If _)    -> Set.unions (fmap refSet args)
            PrimExp _ (Arith _) -> refSet (last args)
            HaltExp _           -> Set.unions (fmap refSet args)
            _                   -> Set.empty
             
          usefulValArguments (Clo (Lam _ vs _) be t) = args
            where args = Set.fromList (fmap fst usefulNxArgs)
                  usefulNxArgs = Prelude.filter 
                          ((`member` alreadyUseful) . snd) allNxArgs
                  allNxArgs = zip [0..] vs
          usefulValArguments (Halt _) = Set.fromList [0..length args-1]
          usefulValArguments (Omega _ _) = Set.empty
 
          usefulDArguments d = Set.unions (fmap usefulValArguments (Set.toList d))

           -- Loop over all procedures invoked here.  
          globallyUseful = Set.unions (fmap (\ proc ->
           case proc of

            {- Do we output these values?  
               If so: all arguments are useful. -}
            Prim _ (IOOp _) -> Set.unions (fmap refSet args)

            {- Does any continuation bind a useful variable?
               If so: all variable arguments are useful.  -}
            Prim _ p -> 
             let usefulArgs = Set.unions (fmap usefulDArguments cs)
              in if not(Set.null(usefulArgs)) || any mayBeHalt cs
                 then Set.unions (fmap refSet args)
                 else Set.empty

            {- Does this lambda have useful arguments?  
               If so: corresponding variable arguments are useful. -}
            clo@(Clo (Lam _ vs _) _ _) ->
             let usefulArgs = usefulValArguments clo
                 allNxArgs = zip [0..] args
                 usedNxArgs = Prelude.filter ((`member` usefulArgs) . fst)
                               allNxArgs
                 usedArgs = Set.unions (fmap (refSet . snd) usedNxArgs)
              in usedArgs 

            {- All arguments to halt are useful. -}
            Halt _ -> Set.unions(fmap refSet args)


            Omega _ _ -> Set.empty

            _ -> error "Not executable.") (Set.toList procs))
         in alreadyUseful   `union` 
            locallyUseful   `union`  
            globallyUseful


analyzeUVE _ = return ()



analyze :: Conf -> State DCFASt ()
analyze (Eval(LetRec l vargs call) be ve log cc (b,t)) =
 do dcfaSt   <- MState.get
    (cl,lc)  <- retrieve chains 
    cl'      <- return $ cl `join` (bot `but` [l ==> singleton l])
    lc'      <- return $ lc `join` (bot `but` [l ==> singleton l])
    MState.put $ dcfaSt { chains = (cl',lc') ,
                          notDead = notDead dcfaSt `union` singleton l }

analyze (Eval(App kappa f args) be ve log cc (b,t)) =
 do dcfaSt   <- MState.get 
    topProp  <- return $ IMap.constant (vars dcfaSt) 

    copy     <- retrieve copy
    copy'    <- return $ copyProp topProp (f,procs)
    copy''   <- return $ foldr (\ expval tally -> 
                  tally `meet` copyProp topProp expval) 
                  copy' argsXds
    inl   <- retrieve inlinable
    inl'  <- return $ inlinableExp (f,procs)
    inl'' <- return $ foldr (\ expval tally ->
                  tally `join` inlinableExp expval) inl' argsXds

    (cl,lc)  <- retrieve chains
    MState.put $ dcfaSt { inlinable = inl  `join` inl'  `join` inl'' ,
                          copy      = copy `meet` copy' `meet` copy'',
                          genRA0    = \ rv -> genRA0 dcfaSt rv `union`
                                           initiallyRemovable rv,
                          removable = removable dcfaSt . removeGen ,
                          useful    = useful dcfaSt . usefulGen ,
                          chains    = (cl `join` cl', lc `join` lc') ,
                          notDead = (notDead dcfaSt) `union` 
                                    singleton kappa  }
 where procs   = eval be ve t f
       ds      = fmap (eval be ve t) args
       cs      = Prelude.filter isCont ds
       argsXds = zip args ds

       lambdaLabels = Set.fold (\ p lbls ->
                       case p of 
                        Clo (Lam lab _ _) _ _ -> lab:lbls
                        _ -> lbls) [] procs 

       cl' = bot `but` [ kappa ==> Set.fromList lambdaLabels ]
       lc' = bot `but` [ lamlab ==> singleton kappa | lamlab <- lambdaLabels ]

        -- Super-beta copy propagation.


       copyProp topProp (RefExp lab var,d) = prop
        where prop = topProp `but` [lab ==> IMap.foldWithKey 
                       (\ t' -> Set.empty) 
                       (\ v' t' tally -> 
                           let d' = ve ! (v',t')
                            in if v' /= var && absEq d d'
                               then tally `union` singleton v'
                               else tally)   be]
       copyProp topProp _ = topProp
         

        -- Super-beta inlining.
       inlinableExp (RefExp lab var,d) = result where
          valCandidate = Set.fold (\ val soFar -> val `join` soFar) bot d
          flowCond = not(valCandidate == top)

           -- If we were to inline, what would we inline?
          inlinedExp = case valCandidate of
                        Clo (Lam lab _ _) be_b t_b -> InLamLab lab
                        Prim l p -> InPrim p
                        IntVal i -> InLit (IntLit i)
                        FalseVal -> InLit (FalseLit)
                        Halt _ -> InHalt
                        TopVal -> top
                        _ -> error "expression not inlinable"

          envCondX = case valCandidate of
           Clo lam _ _ -> allIn (free lam) (\ v -> (be ! v) /= bot) 
           _ -> True

           -- No free vars? Trivially inlinable.
          envCond0 = case valCandidate of
                      Clo lam be_b t_b -> Set.null(free lam)
                      Prim l _ -> True
                      IntVal _ -> True
                      FalseVal -> True
                      Halt _ -> True
                      _ -> False

           -- Each free var equal in current scope? Inlinable.
          envCond1 = case valCandidate of 
                      Clo lam be_b t_b -> beEq (free lam) be be_b 
                           || (allIn (free lam) 
                               (\ v -> (be ! v /= bot) && -- In scope?
                                       (log ! (be ! v)) `eqv_cn` (log ! t_b)))
                      _ -> False

           -- Trivial motion since closure birth? Inlinable.
          envCond2 = case valCandidate of
                      Clo lam be_b t_b -> (log!t_b) `eqv_cn` emptyString
                      _ -> False

          result =
           if envCondX && flowCond && (envCond0 || envCond1 || envCond2)
           then bot `but` [lab ==> inlinedExp]
           else bot `but` [lab ==> top]

       inlinableExp _ = bot 

        -- beEq vars be1 be2 checks that be1|vars = be2|vars.
       beEq vars be1 be2 = allIn vars (\ v ->
                         -- Both variables must be in scope.
                        (be1 ! v) /= bot && (be2 ! v) /= bot &&
                         -- Equal by strong condition? 
                        ((log ! (be1 ! v)) `eqv` (log ! (be2 ! v)) 
                         ||     
                         -- Equal by contour counting?
                         ((be1 ! v) == (be2 ! v) &&
                          (cc ! (v,be1 ! v)) == ONE)))

        -- TODO: Rework cc to work with both bindings and times.
       absEq d1 d2 = absEq' (Set.toList d1) (Set.toList d2) 
        where absEq' [Clo lam1 be1 t1] [Clo lam2 be2 t2] = 
               lam1 == lam2 && beEq (free lam1) be1 be2
              absEq' [IntVal i1] [IntVal i2] = i1 == i2
              absEq' [FalseVal]  [FalseVal]  = True
              absEq' _           _           = False

       contUsesResult rv = existsIn (last ds) (\ cont ->
        case cont of 
         Halt _ -> True
         Clo (Lam _ vs _) _ _ -> any (not . (`member` rv)) vs)

        
       initiallyRemovable rv =
        Set.fold ( \proc tally ->
         case proc of 
          Clo (Lam _ vs _) _ _ -> 
           tally `union` (Set.fromList $ fmap snd $ Prelude.filter ((`member` rv) . fst) $ zipFail "shit" vs (fmap expLab args))
          Prim _ (Arith _) -> tally `union` 
                               if contUsesResult rv
                               then Set.empty
                               else Set.fromList $ fmap expLab (drop 1 args)
          _ -> Set.empty) Set.empty procs
           

       removeGen (rv,ra) =
        let  -- Compute local restrictions.

             -- rv :: Set.Set VAR
             -- ra :: Set.Set LAB (from EXP)

            allArgsHere = Set.fromList $ fmap expLab args
       
 
             -- Which arguments are not removable because of this site?
            notRA = Set.fold ( \ proc tally ->
             case proc of 
              Clo (Lam _ vs _) _ _ -> 
               tally `union` (Set.fromList $ fmap snd $ Prelude.filter (not . (`member` rv) . fst) $ zipFail "oh shit" vs $ fmap expLab args)
              Prim _ (Arith _) -> if contUsesResult rv
                                  then Set.fromList $ fmap expLab (drop 1 args) 
                                  else Set.empty

              Prim _ (If _ ) -> allArgsHere

              Prim _ (IOOp _) -> allArgsHere
 
              Halt _ -> allArgsHere) Set.empty procs
           
             -- Which variables are not removable because of this site?
            notRV = Set.fold ( \ proc tally ->
             case proc of 
              Clo (Lam _ vs _) _ _ -> 
               tally `union` (Set.fromList $ fmap fst $ Prelude.filter (not . (`member` ra) . snd) $ zipFail "crap" vs $ fmap expLab args)
              _ -> Set.empty) Set.empty procs

         in (rv `Set.difference` notRV, ra `Set.difference` notRA)


       usefulGen alreadyUseful = 
        let
          refSet (RefExp _ var) = singleton var
          refSet _              = Set.empty

           -- TODO: Enforce condition that primitives must appear in function position. 

           -- Function position variables are useful.
           -- Predicate variables are useful. 
           -- Variables as continuations to a primop are useful.
           -- Variables given to halt are useful.
          locallyUseful = 
           case f of 
            RefExp  _ var       -> singleton var
            PrimExp _ (If _)    -> Set.unions (fmap refSet args)
            PrimExp _ (Arith _) -> refSet (last args)
            HaltExp _           -> Set.unions (fmap refSet args)
            _                   -> Set.empty
             
          usefulValArguments (Clo (Lam _ vs _) be t) = args
            where args = Set.fromList (fmap fst usefulNxArgs)
                  usefulNxArgs = Prelude.filter 
                          ((`member` alreadyUseful) . snd) allNxArgs
                  allNxArgs = zip [0..] vs
          usefulValArguments (Halt _) = Set.fromList [0..length args-1]
          usefulValArguments (Omega _ _) = Set.empty
 
          usefulDArguments d = Set.unions (fmap usefulValArguments (Set.toList d))

           -- Loop over all procedures invoked here.  
          globallyUseful = Set.unions (fmap (\ proc ->
           case proc of

            {- Do we output these values?  
               If so: all arguments are useful. -}
            Prim _ (IOOp _) -> Set.unions (fmap refSet args)

            {- Does any continuation bind a useful variable?
               If so: all variable arguments are useful.  -}
            Prim _ p -> 
             let usefulArgs = Set.unions (fmap usefulDArguments cs)
              in if not(Set.null(usefulArgs)) || any mayBeHalt cs
                 then Set.unions (fmap refSet args)
                 else Set.empty

            {- Does this lambda have useful arguments?  
               If so: corresponding variable arguments are useful. -}
            clo@(Clo (Lam _ vs _) _ _) ->
             let usefulArgs = usefulValArguments clo
                 allNxArgs = zip [0..] args
                 usedNxArgs = Prelude.filter ((`member` usefulArgs) . fst)
                               allNxArgs
                 usedArgs = Set.unions (fmap (refSet . snd) usedNxArgs)
              in usedArgs 

            {- All arguments to halt are useful. -}
            Halt _ -> Set.unions(fmap refSet args)


            Omega _ _ -> Set.empty

            _ -> error "Not executable.") (Set.toList procs))

           -- Loop over all continuations passed here.

--           useful' = Set.fold (\ proc tally -> 
--             case proc of
--              Prim _ p -> if anyUseful 
--                          then tally `union` fromIf `union` localVars 
--                          else tally `union` fromIf
--                          where anyUseful = any (\ c -> 
--                                 existsIn c (\ val -> 
--                                  case val of
--                                   Clo (Lam _ vs _) _ _ -> 
--                                     any (`member` tally) vs
--                                   Halt _ -> True
--                                   _ -> False)) cs -- Compute cs
--                                localVars = foldr (\ e vars ->
--                                 case e of  
--                                  RefExp _ var -> vars `union` singleton var
--                                  _ -> vars) Set.empty args
--                                fromIf = case p of 
--                                          If _ -> case args of 
--                                                     (RefExp lab var):_ -> 
--                                                       singleton var
--                                                     _ -> Set.empty
--                                          _ -> Set.empty

--              Halt _ -> case args of 
--                         (RefExp lab var):_ -> singleton var `union` tally
--                         _ -> tally

--              Clo (Lam _ vs _) _ _ ->
--               foldr (\ (v,e) tally' -> 
--                 case e of
--                  RefExp _ usedVar -> if v `member` tally' 
--                                      then tally' `union` singleton usedVar
--                                      else tally'
--                  _ -> tally')
--                tally ((zip vs args))

--              _ -> tally) (alreadyUseful `union` locallyUseful) procs
         in alreadyUseful   `union` 
            locallyUseful   `union`  
            globallyUseful  `union` 
--            useful'   `union`
            Set.empty

analyze c@(Apply(Clo (Lam l vs call) be tb) ds ve log cc (b,t)) =
 do state <- get
    put $ state { notDead = (notDead state) `union` singleton l }

analyze c@(Apply(Prim lab _) ds ve log cc (b,t)) =
 do (cl,lc)  <- retrieve chains
    cs       <- return [ c | c <- ds, isCont c ]
    conts    <- return $ flattenListOfSets cs
    contLabs <- return $ Set.unions (fmap labelSet conts)
    cl'      <- return $ bot `but` [lab ==> contLabs ]
    lc'      <- return $ bot `but` [contLab ==> singleton lab | 
                            contLab <- Set.toList contLabs ]
    state    <- get
    put $ state { chains = (cl `join` cl',
                            lc `join` lc') }
    return ()
 where labelSet (Clo (Lam lab _ _) _ _) = singleton lab
       labelSet _ = Set.empty
       
analyze _ = do return () 


 -- next computes the subsequent states of a given state.
next :: Conf -> MState.State DCFASt [Conf]

next c@(Apply(Halt t0') [d] ve log cc (b,t)) = return []

next c@(Apply(Omega _ t0') ds ve log cc (b,t)) = return []

next c@(Apply(Clo(Lam l vs call) be tb) ds ve log cc (b,t)) = 
 do sm       <- getStackModel
    opts     <- retrieve opts
    dispConf <- display c
    ve''     <- return $ ve' `join` (bot `but` [ (v,b) ==> d 
       | (v,d) <- zipFail ("argument mismatch in Apply state\n:" ++dispConf) vs ds ])
    (if logZombies || veZombies || ccZombies
     then recordCollection
     else return ())
    change <- return $ case (l,sm) of
                        (Cn _,LUDDY) -> inv(log ! (tb))
                        (_) -> emptyString
    log''   <- if useFrameStrings opts
               then  return $ (log' `catLog` change `catLog` (downFor l)) 
                                  `join`
                              (emptyAt t)
               else return $ topLog
    return [Eval call be' ve'' log'' cc'' (b,t)]
  where newBindings = [v ==> b | v <- vs ]
        logZombies  = log ! t /= bot
        veZombies   = any (\ vb -> (ve ! vb) /= bot) newBindings
        ccZombies   = any (\ vb -> (cc ! vb) /= ZERO) newBindings
        reachable   = reaches c
        cc' = if ccZombies then restrictVARxCNb cc reachable else cc
        ve' = if veZombies then restrictVARxCNb ve reachable else ve
        log' = if logZombies then restrictCNt log reachable else log

        cc'' = cc' + (bot `but` [vb ==> ONE | vb <- newBindings])
        be' = be `but` newBindings

 {- TODO: Verify optimality of lazy collecting.
    Only collect before a zombie might be created. -}
next c@(Eval (call@(App l f args)) be ve log cc (b,t)) =
 do sm    <- getStackModel
    opts  <- retrieve opts
    dispCall <- display call
    succs <- return $ fmap (\proc -> 
                        let log' = log `catLog` (inv change)
                            change = case (l,sm) of 
                                      (Fn _,STEELE) -> age log cs
                                      (Cn _,STEELE) -> age log proc
                                      _ -> emptyString
                            log'' = if useFrameStrings opts
                                    then log'
                                    else topLog
                         in case proc of
                             IntVal _ -> error $ "calling int in: "
                                        ++ dispCall
                             _ -> Apply proc ds ve log'' cc (b',t'))
              (Set.toList (eval be ve t f))
    return succs
 where b' = l -- TODO: tick function goes here
       t' = l -- TODO: tick function goes here 
       ds = [ eval be ve t e | e <- args ] 
       cs = [ c | c <- ds, isCont c ] 


next c@(Eval lr@(LetRec l vargs call) be ve log cc (b,t)) =
 do opts <- retrieve opts
    (if logZombies || veZombies || ccZombies
     then recordCollection
     else return ())
    log''' <- if useFrameStrings opts
              then return log''
              else return topLog      
    return [Eval call be' ve'' log''' cc'' (b',t')]
 where newBindings = [ v ==> b' | v <- vs ] 
       logZombies  = log ! t /= bot
       veZombies   = any (\ vb -> (ve ! vb) /= bot) newBindings
       ccZombies   = any (\ vb -> (cc ! vb) /= ZERO) newBindings
       cc'  = if ccZombies then restrictVARxCNb cc reachable else cc
       ve'  = if veZombies then restrictVARxCNb ve reachable else ve
       log' = if logZombies then restrictCNt log reachable else log

       cc'' = cc' + (bot `but` [ vb ==> ONE | vb <- newBindings])
       be'  = be `but` newBindings
       ve'' = ve' `join` (bot `but`
                    [ (v,b') ==> (eval be' ve t' arg) | (v,arg) <- vargs ])
       log'' = log' `join` (emptyAt t')
       b'    = l
       t'    = l
       (vs,_) = unzip vargs
       reachable = reaches c

 -- NOTE: the stack is popped in primitive operations.
 -- We have to pop it to simulate going through an Eval state.
next c@(Apply(Prim l (Arith arithOp)) [x,y,cont] ve log cc (b,t)) =
 do sm    <- getStackModel
    opts  <- retrieve opts
    log'  <- return $ log `join` (emptyAt t)
    cc'   <- return $ cc
    return $ [Apply proc [ans] ve
               (if useFrameStrings opts
                then log' `catLog` 
                      case sm of
                       LUDDY -> emptyString
                       STEELE -> inv(age log' proc)
                else topLog)
                cc'
                (l,l) -- TODO: tick goes here.
              | proc <- Set.toList cont ]
 where ans = absArith arithOp (Set.toList x) (Set.toList y)
       absArith ADD [IntVal a] [IntVal b] = singleton (IntVal (a + b))
       absArith MUL [IntVal a] [IntVal b] = singleton (IntVal (a * b))
       absArith SUB [IntVal a] [IntVal b] = singleton (IntVal (a - b))
       absArith DIV [IntVal a] [IntVal b] = singleton (IntVal (a `div` b))
       absArith _   a          b          = singleton TopInt
       -- TODO: If it cannot possibly be two ints, we can stop.


next c@(Apply(Prim l (IOOp DISPLAY)) [x,cont] ve log cc (b,t)) =
 do sm    <- getStackModel
    opts  <- retrieve opts
    log'  <- return $ log `join` (emptyAt t)
    cc'   <- return $ cc
    return $ [Apply proc [] ve 
               (if useFrameStrings opts
                then log' `catLog` 
                      case sm of
                       LUDDY -> emptyString
                       STEELE -> inv(age log' proc)
                else topLog)
                cc'
                (l,l) -- TODO: tick goes here.
              | proc <- Set.toList cont ]

next c@(Apply(Prim l (If test)) [cond,ctrue,cfalse] ve log cc (b,t)) =
 do log' <- return $ log `join` (emptyAt t)
    opts <- retrieve opts
    sm <- getStackModel
    cc' <- return $ cc
    return $ [Apply proc [] ve
               (if useFrameStrings opts
                then log' `catLog` case sm of
                               LUDDY -> emptyString
                               STEELE -> inv(age log' proc)
                else topLog)
                cc'
                (l,l)  -- TODO: tick goes here.
               | proc     <- subsequents ]
 where subsequents = absIf test (Set.toList cond)
       absIf ISZERO [IntVal 0] = Set.toList ctrue
       absIf ISZERO lst        = if any (\ val -> case val of
                                                   IntVal 0 -> True
                                                   TopVal -> True
                                                   TopInt -> True
                                                   _ -> False) lst
                                 then (Set.toList ctrue) ++ (Set.toList cfalse)
                                 else Set.toList cfalse

next c@(Apply(Prim l (If _)) ds ve log cc (b,t)) =
 error "if primitive with invalid arguments"


next c =
 do c' <- display c
    error $ "Explosion on: " ++ c'


 -- visited computes the set of reachable states.
vis pr = if isAlphaUnique pr
         then do state <- get
                 put $ state { pr      = pr ,
                          seen    = Set.empty ,
                          history = [] }
                 visited [init2 pr]
         else error "Trying to analyze non-alpha-unique program."

visited :: [Conf] -> (State DCFASt PowConf)
visited [] = retrieve seen
visited (hd:tl) = 
 do seen <- retrieve seen
    opts <- retrieve opts
    test <- return $ stateCompare opts
    record hd
    str <- display hd
--    analyze hd
    (if case test of
         WeakerThan -> (singleton hd) <: seen
         LessThan -> hd `member` seen
     then do visited tl
     else do new <- next hd
             state <- get
             put $ state { seen = seen `union` singleton hd }
             visited (new ++ tl))

 -- Extension parsing.
extension "" = ""
extension ('.':ext) = ext
extension (_:tl) = extension tl


dcfa pr symtab opts =
 do put $ initDCFASt pr symtab opts


 -- main
main :: IO ()
main = 
 do  -- Parse command line options.
    args <- getArgs
    opts@(Options file sm sc useFS) <- return $ 
      options initOptions args

     -- Read the source.
    ext          <- return $ extension file
    src          <- readFile file

     -- Tokenize.
    (tokens,st)  <- return $ tokens src

     -- Create an S-Expression from the tokens.
    sexp         <- return $ car(parseSExpList tokens)

     -- Parse into an AST.
    (cpsAST,st) <- if ext == "scm" 
                   then do dsAST <- return $ MState.evalState 
                                               (parseTerm sexp) st
                           return $ pr2call dsAST st
                   else do return $ MState.runState (parseCall sexp) st

     -- Alphatise the AST.
    (cpsAST,st) <- return $ MState.runState (alphatise id cpsAST) st

     -- Check for simple errors: top-level free variables.
    putStr "Original:\n"
    putStr $ evalState (display sexp) st
    putStr "\n\n"
    putStr $ evalState (display cpsAST) st
    putStr "\n\n"

    freeVars <- return $ evalState (display (free cpsAST)) st
    putStr "\n\n"
    putStr $ if not(Set.null (free cpsAST)) 
             then error $ "syntax error: top level free variables: " ++
                          freeVars
             else ""

     -- Run DCFA.
