 -- A small, macro-less Scheme.
module App.MicroScheme.Main where

import qualified Data.Map as Map
import qualified Data.Set as Set

import qualified System (getArgs)
import qualified Control.Monad.State as St

import qualified Compiler.Symbol as Sym
import qualified Compiler.SExpression as SX
import qualified Compiler.HOIL as DS

import qualified Compiler.ChickenBackend as Ch
import qualified Compiler.Backend as B

dsAnt = DS.nullAnn



parseExp :: SX.LexicalSExp -> DS.DSExp ()
parseExp (SX.Nil annt) = DS.Nil dsAnt
parseExp (SX.Int int annt) = DS.Int int dsAnt
parseExp (SX.Symbol sym annt) = 
 case Sym.toString sym of
  "+"  -> (DS.Prim (DS.Sum [0,1] False) dsAnt)
  "-"  -> (DS.Prim (DS.Difference [0,1] False) dsAnt)
  "*"  -> (DS.Prim (DS.Product [0,1] False) dsAnt)
  "="  -> (DS.Prim (DS.EqNum [0,1] False) dsAnt)
  "<" -> (DS.Prim (DS.LessThanEq [0,1] False) dsAnt)
  "<=" -> (DS.Prim (DS.LessThanEq [0,1] False) dsAnt)
  _ -> DS.Var sym dsAnt
parseExp (SX.Bool bool annt) = DS.Bool bool dsAnt
 
parseExp (SX.Cons f@(SX.Symbol{SX.symbol=Sym.Sym{Sym.toString=s}}) tl an)=
 case s of 

  "define" -> error "internal defines not allowed"


  -- TODO: Add support for variable arity.
  -- (lambda (<var>*) <body>)
  "lambda" ->
    let 
     sxvars = SX.car tl
     sxbody = (SX.car . SX.cdr) tl
     dsvars = SX.mapToList parseExp sxvars
     dsbody = parseExp sxbody
    in DS.Lam [] dsvars dsbody dsAnt

  -- (let* ((<var> <exp>)*) <body>)
  "let*" ->
    let 
     varexpsxs = SX.mapToList SX.toList (SX.car tl)
     sxbody = (SX.car . SX.cdr) tl

     toNestedLets [] = parseExp sxbody
     toNestedLets (([v,e]):tl) = 
      DS.Let (parseExp v,parseExp e,dsAnt) (toNestedLets tl) dsAnt

    in toNestedLets varexpsxs

  -- (letrec ((<var> <exp>)*) <body>)
  "letrec" ->
    let 
     varexpsxs = SX.mapToList SX.toList (SX.car tl)
     sxbody = (SX.car . SX.cdr) tl

     parseVarExp [v,e] = (parseExp v, parseExp e,dsAnt)
    in DS.LetRec (map parseVarExp varexpsxs) (parseExp sxbody) dsAnt

  -- (if <cond> <consequent> <alternate>)
  "if" -> 
    let
     sxcond = SX.car tl
     sxconsequent = (SX.car . SX.cdr) tl
     sxalternate  = (SX.car . SX.cdr . SX.cdr) tl
     cond = parseExp sxcond
     consequent = parseExp sxconsequent
     alternate = parseExp sxalternate
    in DS.If cond consequent alternate DS.Direct dsAnt

  -- (try <exp> <handler>)
  "try" ->
   let
    sxexp = SX.car tl
    sxhandler = (SX.car . SX.cdr) tl
    
    exp = parseExp sxexp
    handler = parseExp sxhandler
   in DS.Try exp handler dsAnt


  -- (raise <exn>)
  "raise" ->
   let

    sxexn = SX.car tl
    
    exn = parseExp sxexn
   in DS.Raise exn dsAnt
  
  _ -> DS.App (parseExp f) [] (SX.mapToList parseExp tl) dsAnt
   
     
parseExp (SX.Cons f args annt) =
 DS.App (parseExp f) [] (SX.mapToList parseExp args) dsAnt

  

test :: IO ()
test =
 do
 args <- System.getArgs

 arg0 <- return $! args !! 0

 contents <- readFile arg0

 sx <- return $! SX.fromStringWithName arg0 contents
 
 (sx,st) <- return $! St.runState (SX.canonicalize sx) Sym.empty

 putStrLn "canonicalized: "
 putStrLn (show sx)
 putStrLn ""


 putStrLn "parsed: "
 pr <- return $! parseExp sx
 putStrLn (Ch.chickenized pr)
 putStrLn ""

 pr <- return $! DS.annotateFreeVars pr  

 pr <- return $! if (DS.hasFreeVarsAnnotation pr) 
                  then pr
                  else error "failed to annotate everything"



 putStrLn "alphatised & annotated:"
 (pr,st) <- return $! St.runState 
             (do pr <- DS.alphatise pr
                 pr <- return $! DS.annotateFreeVars pr
                 return $! pr) st
 putStrLn (Ch.chickenized pr)
 putStrLn ""



 putStrLn "a-normalized:"
 (pr,st) <- return $! St.runState (DS.anormalize pr) st
 putStrLn (Ch.chickenized pr)
 putStrLn ""


  -- Annotate mutable variables.
 (pr,mutSet) <- return $! DS.annotateMutability pr
 
--  putStrLn "mutability marked: " 
--  putStrLn (Ch.chickenized pr)
--  putStrLn ""

 putStrLn "variables immutified: " 
 pr <- return $! DS.immutifyVariables Set.empty pr
 putStrLn (Ch.chickenized pr)
 putStrLn ""
 

 putStrLn "exception-passing transformed: " 
 (pr,st) <- return $! St.runState (DS.unexceptionalize pr) st
 putStrLn (Ch.chickenized pr)
 putStrLn ""

 pr <- return $! DS.annotateFreeVars pr  

 pr <- return $! if (DS.hasFreeVarsAnnotation pr) 
                  then pr
                  else error "failed to annotate everything"


 putStrLn "CPS converted: "
 (pr,st) <- return $! St.runState (DS.continuateProgram pr) st
 putStrLn (Ch.chickenized pr)
 putStrLn ""


  -- Annotate free vars.
 pr <- return $! DS.annotateFreeVars pr  
 pr <- return $! if (DS.hasFreeVarsAnnotation pr) 
                  then pr
                  else error "failed to annotate everything"


 putStrLn "flat-closure converted: "
 (pr,st) <- return $! St.runState (DS.flatClosureConvert pr) st
 putStrLn (Ch.chickenized pr)
 putStrLn ""

  -- BUG? Alphatise here?

 putStrLn "letrec-hoisted:"
 pr <- return $! DS.unsafeHoistLetRecs pr
 putStrLn (Ch.chickenized pr)
 putStrLn ""


 putStrLn "lambda-hoisted:"
 (pr,st) <- return $! St.runState 
             (do pr <- DS.topHoistLambdas pr
                 return $! pr) st
 putStrLn (Ch.chickenized pr)
 putStrLn ""


 putStrLn "with args flattened:"
 (pr,st) <- return $! St.runState 
             (do DS.flattenApps pr) st
 putStrLn (Ch.chickenized pr)
 putStrLn ""
 

  -- With boundVars annotation.
 pr <- return $! DS.annotateBoundVars Map.empty pr
 
 putStrLn "as GCC code:"
 putStrLn ""
 putStrLn $! B.cpmToGCC pr
 putStrLn ""

 writeFile "tmp/prog.c" (B.cpmToGCC pr)


 return $! ()

