module Compiler.GCFA where

import qualified Control.Monad.State as St

import qualified Compiler.RxRS as CPS
import qualified Compiler.Symbol as Sym

import qualified Data.List as List
import qualified Data.Map as Map
import qualified Data.Set as Set
import qualified Data.Maybe as Maybe

-- import qualified Math.InfiniteMap as IMap

import qualified Monad (liftM)

import Lattice.Lattice 
import Lattice.StandardLattice 

import qualified Might

import System.IO.Unsafe


-- type IMap k a = IMap.InfiniteMap k a
type Map k a = Map.Map k a
type Set e = Set.Set e
type Symbol = Sym.Symbol


 -- Convenience.
a ==> b = (a,b)



 -- Abstract domains.
type Time = [CPS.Label]

type Bind = (Sym.Symbol,Time) 
type StoreLoc = (CPS.Label,Time,Int)
type ArrayLoc = (CPS.Label,Time,Int)
type StringLoc = (CPS.Label,Time,Int)

type Frame = (CPS.Label,Time)

type Log = Map.Map Time F

data HeapDom 
 = HDBind !Bind
 | HDStoreLoc !StoreLoc
 | HDArrayLoc !ArrayLoc
 | HDStringLoc !StringLoc
 | HDLogTime !Time 
 | HDFrame !Frame -- To count how many times this frame has been pushed.
 | HDTop
 | HDBot
 deriving (Eq,Ord,Show)

type GCWorklist = [HeapDom]

type BEnv = Map Symbol Time



data Val ext
   -- Lattice elements.
 = ValTop
 | ValBot
   
   -- Structures.
 | ValStoreLoc !StoreLoc
 | ValArrayLoc !ArrayLoc
 | ValStringLoc !StringLoc

   -- Procedures & primitives.
 | ValPrim !(CPS.Term ext)
 | ValClo {
    cloLAM  :: !(CPS.Term ext) ,
    cloBEnv :: !BEnv ,
    cloTime :: !Time 
   }

   -- Types: 
 | ValType { typeName :: !Sym.Symbol, typeMembers :: ![Sym.Symbol] }

   -- *Immutable* strings.
 | ValString !String
 
   -- Booleans.
 | ValSymbol !Sym.Symbol

   -- Booleans.
 | ValBool !Bool

   -- Integers.
 | ValNum

   -- Lists.
 | ValNil

   -- Characters.
 | ValChar

   -- No value at all.
 | ValVoid
 deriving (Eq,Ord,Show)


type D ext = Set (Val ext)

 -- TODO: For now, collapse everything together.  Later, unmerge.
data Index = IndLength
             | IndAny
 deriving (Eq,Ord,Show)

type Field = Sym.Symbol

type Obj ext = Map Field (D ext)

type Array ext = Map Index (D ext)

type VEnv ext = Map Bind (D ext)

type Store ext = Map StoreLoc (Obj ext)

type ArrStore ext = Map ArrayLoc (Array ext)

type StrStore ext = Map StringLoc (AString)

data AString
 = AStringTop 
 | AStringBot
 | AString !String
 deriving (Eq,Ord,Show)

 -- This is a non-standard lattice, so it's handled specially below.
data Count
 = ZERO | ONE | MORE | CountTop | CountBot
 deriving (Eq,Ord)

data Counter
 = Counter {
    muDef :: !Count,
    muMap :: !(Map HeapDom Count)
   }
 deriving (Eq,Show)


data Heap ext
 = Heap {
    heapVEnv :: !(VEnv ext),
    heapStore :: !(Store ext) ,
    heapArrays :: !(ArrStore ext) ,
    heapStrings :: !(StrStore ext) ,
    heapLog :: !Log ,
    heapCounter :: !(Counter)
   }
 | HeapTop
 deriving (Eq)

type Parm ext = (D ext)

data Parms ext
 = Parms { 
  ordered :: ![Parm ext] ,
  labels :: ![Symbol] ,
  labeled :: ![Parm ext] ,
  rest :: !(Maybe (Parm ext))
  }
 | ParmsTop
 | ParmsBot
 deriving (Eq,Show)

data State ext
 = Eval {
    stateCALL :: !(CPS.Term ext) ,
    stateBEnv :: !BEnv, 

    stateHeap    :: !(Heap ext) ,
    stateTime    :: !Time
   }
 | Apply { 
    stateProc    :: !(Val ext) , 
    stateParms   :: !(Parms ext) ,

    stateHeap    :: !(Heap ext) ,
    stateTime    :: !Time 
   } 
 | StateTop
 | StateBot
 deriving (Eq)

instance Show (State ext) where
 show (Eval call be h t) =
  "call: " ++ (show call) ++ "\n" ++
  "benv: " ++ (show be) ++ "\n" ++ 
  "heap: " ++ (show h) ++ "\n" ++
  "time: " ++ (show t) 
 show (Apply proc parms h t) =
  "proc: " ++ (show proc) ++ "\n" ++
  "args: " ++ (show parms) ++ "\n" ++
  "heap: " ++ (show h) ++ "\n" ++
  "time: " ++ (show t) 

instance Show (Heap ext) where
 show (Heap ve store arr str log mu) =
  "\n" ++
  " venv:  " ++ (show ve) ++ "\n" ++
  " store: " ++ (show store) ++ "\n" ++
  " arr:   " ++ (show arr) ++ "\n" ++
  " str:   " ++ (show str) ++ "\n" ++
  " log:   " ++ (show log) ++ "\n" ++
  " mu:    " ++ (show mu)

 -- Semantics domain helpers.

isStoreLocVal (ValStoreLoc{}) = True
isStoreLocVal _ = False

isArrayLocVal (ValArrayLoc{}) = True
isArrayLocVal _ = False

isApply (Apply{}) = True
isApply _ = False

isEval (Eval{}) = True
iSEval _ = False

isEvalJoin (Eval (CPS.App f _ _) _ _ _) = CPS.isJoinPoint f
isEvalJoin _ = False

isApplyJoin (Apply (ValClo lam be birthdate) _ _ _) = CPS.isJoinPoint lam
isApplyJoin _ = False


restrictBEnv :: BEnv -> (CPS.FreeVarMap ext) -> BEnv
restrictBEnv map1 map2 = Map.intersectionWith (\ l r -> l) map1 map2


unmarkApplyJoin state@(Apply (ValClo lam be birthdate) _ _ _) = 
 state { stateProc = (ValClo (CPS.unmarkJoinPoint lam) be birthdate) }
unmarkApplyJoin state = state

isApplicableVal (ValPrim{}) = True
isApplicableVal (ValClo{}) = True
isApplicableVal _ = False

isContD d = any isContVal (Set.toList d) 

isContVal (ValClo lam be t) = CPS.isContTerm lam
isContVal _ = False

isRestEmpty (Nothing) = True
isRestEmpty (Just dloc) =
 case Set.toList dloc of
  [ValNil] -> True
  _ -> False


initialHeap = Heap {
  heapVEnv = bot,
  heapStore = bot,
  heapArrays = bot, 
  heapStrings = bot,
  heapLog = Map.singleton t0 emptyFrameString , 
  heapCounter = muZero 
 }


 -- GCFA auxiliary functions.

a :: (Eq ext) => BEnv -> (Heap ext) -> Time -> (CPS.Term ext) -> (D ext)
a be h t term@(CPS.Lam _ call _) = Set.singleton (ValClo term be' t)
 where 
  be' = be `restrictBEnv` (CPS.free term)

a be h t term@(CPS.Lit (CPS.LitInt _) _) = Set.singleton (ValNum)
a be h t term@(CPS.Lit (CPS.LitSymbol sym) _) = Set.singleton (ValSymbol sym)
a be h t term@(CPS.Lit (CPS.LitBool val) _) = Set.singleton (ValBool val)
a be h t term@(CPS.Lit (CPS.LitChar char) _) = Set.singleton (ValChar)
a be h t term@(CPS.Lit (CPS.LitNil) _) = Set.singleton (ValNil)
a be h t term@(CPS.Lit (CPS.LitString val) _) = Set.singleton (ValString val)

a be h t term@(CPS.Void ann) = Set.singleton ValVoid

a be h t term@(CPS.Ref sym _) = 
 let
  in case Map.lookup (sym, t) (heapVEnv h) of
      Just d -> d
      Nothing -> error $ "Corrupted heap on: " ++ (show sym) 
 where 
  t = case Map.lookup sym be of 
       Just t -> t
       Nothing -> error $ "Corrupted local env: Free variable: " ++ (show sym)


a be h t term@(CPS.Prim prim _) =
 Set.singleton (ValPrim term)

a be h t term@(CPS.Var sym _) = 
 error $ "Binding Var used in place of Ref in `a' function: " ++ (show term)

a be h t term = 
 error $ "Unhandled case in `a' function: " ++ (show term)


t0 = [] 


initState call = Eval call bot initialHeap t0


 -- PRIMITIVE HANDLING.

 -- Baked constants.
dTrue = Set.singleton (ValBool True)
dFalse = Set.singleton (ValBool False)
dEither = Set.fromList [ValBool True,ValBool False]
dNil = Set.singleton (ValNil)
dVoid = Set.singleton (ValVoid)


consType = ValType Sym.cons' [Sym.car',Sym.cdr']

 -- Unspecified.

evalPrim CPS.UNSPECIFIED s [] Nothing = return $! (stateHeap s,Set.empty)

 -- This primitive only has meaning in the abstract.  It is used for debugging.
evalPrim CPS.ABSTRACE s _ _ =
 do
 True <- Might.debugM $ "abstrace:\n" ++ (show s) ++ "\n"
 return $! (stateHeap s,dTrue)

 -- Arithemtic.
evalPrim CPS.SUM s _ r = return $! (stateHeap s,Set.singleton ValNum)
evalPrim CPS.DIFF s _ r = return $! (stateHeap s,Set.singleton ValNum)
evalPrim CPS.PROD s _ r = return $! (stateHeap s,Set.singleton ValNum)
evalPrim CPS.DIV s _ r = return $! (stateHeap s,Set.singleton ValNum)
evalPrim CPS.REMAINDER s _ r = return $! (stateHeap s,Set.singleton ValNum)
evalPrim CPS.QUOTIENT s _ r = return $! (stateHeap s,Set.singleton ValNum)
evalPrim CPS.EXPT s _ r = return $! (stateHeap s,Set.singleton ValNum)

evalPrim CPS.MAX s _ r = return $! (stateHeap s,Set.singleton ValNum)
evalPrim CPS.MIN s _ r = return $! (stateHeap s,Set.singleton ValNum)



evalPrim CPS.EXACT2INEXACT s _ r  = 
 return $! (stateHeap s, Set.singleton ValNum)

 -- Arithmetic relations.
evalPrim CPS.LT s _ r = 
 do
 return $! (stateHeap s,dEither)
evalPrim CPS.LEQ s _ r = 
 do
 return $! (stateHeap s,dEither)
evalPrim CPS.GT s _ r = 
 do
 return $! (stateHeap s,dEither)
evalPrim CPS.GEQ s _ r = 
 do
 return $! (stateHeap s,dEither)
evalPrim CPS.EQNUM s _ r = 
 do
 return $! (stateHeap s,dEither)
evalPrim CPS.ZEROP s _ r = 
 do
 return $! (stateHeap s,dEither)
evalPrim CPS.NUMBERP s _ r = 
 do
 return $! (stateHeap s,dEither)





 -- Strings.
evalPrim CPS.NUMBER2STRING state _ _ =
 do
 allocTopString state

evalPrim CPS.STRING_APPEND state _ _ =
 do
 allocTopString state







 -- Equality.
evalPrim CPS.EQP s _ Nothing =
 return $! (stateHeap s,dEither)
evalPrim CPS.EQVP s _ Nothing =
 return $! (stateHeap s,dEither)
evalPrim CPS.EQUALP s _ Nothing =
 return $! (stateHeap s,dEither)



 -- I/O.
evalPrim CPS.WRITE s _ Nothing = return $! (stateHeap s,Set.singleton ValVoid)
evalPrim CPS.DISPLAY s _ Nothing = return $! (stateHeap s,Set.singleton ValVoid)
evalPrim CPS.NEWLINE s _ Nothing = return $! (stateHeap s,Set.singleton ValVoid)


 -- Logic.  
 -- TODO: Improve these.
evalPrim CPS.NOT s _ Nothing = return $! (stateHeap s,dEither)
evalPrim CPS.AND s [] Nothing = return $! (stateHeap s,dTrue)
evalPrim CPS.AND s dargs Nothing = return $! (stateHeap s,joins dargs)

evalPrim CPS.OR s [] Nothing = return $! (stateHeap s,dFalse)
evalPrim CPS.OR s dargs Nothing = return $! (stateHeap s,joins dargs)



 -- Vectors.
evalPrim CPS.VECTOR_LENGTH state [dloc] Nothing = 
 do
 return $! (stateHeap state, Set.singleton ValNum)


evalPrim CPS.VECTORREF state [dloc,dindex] Nothing = 
 do
 let arrstore = heapArrays $ stateHeap $ state
 locs <- return $! Set.toList dloc
 locs <- return $! filter isArrayLocVal locs
 locs <- return $! map (\ (ValArrayLoc loc) -> loc) locs
 dval <- return $!
  joins [ d | loc <- locs,
          arr <- return $! (arrstore `at` loc),
          d <- return $! arr `at` IndAny  ]
 return $! (stateHeap state, dval)

evalPrim CPS.VECTORSET state [dloc,dindex,dval] Nothing = 
 do
 let arrstore = heapArrays $ stateHeap $ state
 locs <- return $! Set.toList dloc
 locs <- return $! filter isArrayLocVal locs
 locs <- return $! map (\ (ValArrayLoc loc) -> loc) locs
 arr <- return $! Map.singleton IndAny dval
 arrstore' <- return $! foldl (\ arrstore' loc ->
                          arrstore' `join` Map.fromList[loc ==> arr]) arrstore locs
 return $! ((stateHeap state){ heapArrays=arrstore' },dVoid)
 
 -- TODO: Allow for narrowing.
evalPrim CPS.VECTOR state delements Nothing =
 do
 h <- return $! stateHeap state
 [loc] <- arrayAlloc state 1
 arr <- return $! mkArray (Set.singleton ValNum) delements
 dloc <- return $! Set.singleton (ValArrayLoc loc)
 arrstore <- return $! heapArrays h
 mu <- return $! heapCounter h
 (arrstore',mu') <- return $! insertArray (loc,arr) arrstore mu
 h' <- return $! h { heapArrays  = arrstore' , heapCounter = mu' }
 return $! (h',dloc)


evalPrim CPS.MAKE_VECTOR state [len] Nothing =
 evalPrim CPS.MAKE_VECTOR state [len,Set.empty] Nothing

evalPrim CPS.MAKE_VECTOR state [len,ddefault] Nothing =
 do
 h <- return $! stateHeap state
 [loc] <- arrayAlloc state 1
  -- BUG: This won't scale when we narrow.
 arr <- return $! mkArray (Set.singleton ValNum) [ddefault]
 dloc <- return $! Set.singleton (ValArrayLoc loc)
 arrstore <- return $! heapArrays h
 mu <- return $! heapCounter h
 (arrstore',mu') <- return $! insertArray (loc,arr) arrstore mu
 h' <- return $! h { heapArrays  = arrstore' , heapCounter = mu' }
 return $! (h',dloc)





 -- Lists.
 -- TODO: Make this more accurate.  (Actually search the heap.)
 -- TODO: Have it actually return (cons elem '()) instead of #t.
evalPrim CPS.MEMV state (e:l) Nothing = 
 do
 (h,d) <- evalPrim CPS.LIST state [e] Nothing
 return $! (h,d `join` dFalse)
evalPrim CPS.MEMQ state (e:l) Nothing = 
 do
 (h,d) <- evalPrim CPS.LIST state [e] Nothing
 return $! (h,d `join` dFalse)

evalPrim CPS.MEMV state _ Nothing = error $ "Add way of signalling primop error."


evalPrim CPS.CAR state [dloc] Nothing =
 do
 (Heap _ store _ _ _ _) <- return $! stateHeap state
 locs <- return $! Set.toList dloc
 locs <- return $! filter isStoreLocVal locs
 locs <- return $! map (\ (ValStoreLoc loc) -> loc) locs
 d <- return $! 
  joins [ obj `at` Sym.car' |  loc <- locs, obj <- return $! store `at` loc ]
 return $! (stateHeap state,d)


evalPrim CPS.CDR state [dloc] Nothing =
 do
 (Heap _ store _ _ _ _) <- return $! stateHeap state
 locs <- return $! Set.toList dloc
 locs <- return $! filter isStoreLocVal locs
 locs <- return $! map (\ (ValStoreLoc loc) -> loc) locs
 d <- return $! 
  joins [ obj `at` Sym.cdr' |  loc <- locs, obj <- return $! store `at` loc ]
 return $! (stateHeap state,d)
 

evalPrim CPS.NULLP state [d] Nothing = 
 do
 vals <- return $! Set.toList d
 case vals of
  [ValNil] -> return $! (stateHeap state,dTrue)
  _ -> return $! 
   if elem ValNil vals 
   then (stateHeap state,dEither)
   else (stateHeap state,dFalse)


evalPrim CPS.PAIRP state [d] Nothing = 
 do
 vals <- return $! Set.toList d
 if all isPairVal vals
  then return $! (stateHeap state,dTrue)
  else 
   if any mayBePairVal vals
    then return $! (stateHeap state,dEither)
    else return $! (stateHeap state,dFalse)
 where
  store = heapStore $ stateHeap $ state

  isPairVal (ValStoreLoc loc) = 
   let
    obj = store `at` loc
    ty = obj `at` Sym.objtype' 
   in ty == Set.singleton consType
  isPairVal _ = False

  mayBePairVal (ValStoreLoc loc) = 
   let
    obj = store `at` loc
    ty = obj `at` Sym.objtype' 
   in Set.member consType ty
  mayBePairVal _ = False
   


 -- Lists.

evalPrim CPS.LIST state [] Nothing = 
 do
 return $! (stateHeap state,Set.singleton ValNil)


evalPrim CPS.LIST state args Nothing = 
 do
 h@(Heap _ store _ _ _ mu) <- return $! stateHeap state
 storeLocs <- storeAlloc state (length args) 
 conses <- return $! mkList args (map ValStoreLoc (tail storeLocs))
 (store',mu') <- return $! insertObjs (zip storeLocs conses) store mu
 return $! (h{heapStore = store',heapCounter = mu'},
           Set.singleton $! ValStoreLoc $! head storeLocs)


 -- TODO: Optimize set-car!, set-cdr! using counting to do strong updates.
evalPrim CPS.SETCDR state [dloc,dval] Nothing =
 do
 h@(Heap _ store _ _ _ _) <- return $! stateHeap state
 locs <- return $! Set.toList dloc
 locs <- return $! filter isStoreLocVal locs
 locs <- return $! map (\ (ValStoreLoc loc) -> loc) locs
 obj <- return $! Map.fromList[Sym.cdr' ==> dval]
 store' <- return $! foldl (\ store' loc -> 
                       store' `join` Map.fromList[loc ==> obj]) store locs
 return $! (h{heapStore=store'},dVoid)


evalPrim CPS.SETCAR state [dloc,dval] Nothing =
 do
 h@(Heap _ store _ _ _ _) <- return $! stateHeap state
 locs <- return $! Set.toList dloc
 locs <- return $! filter isStoreLocVal locs
 locs <- return $! map (\ (ValStoreLoc loc) -> loc) locs
 obj <- return $! Map.fromList[Sym.car' ==> dval]
 store' <- return $! foldl (\ store' loc -> 
                       store' `join` Map.fromList[loc ==> obj]) store locs
 return $! (h{heapStore=store'},dVoid)



evalPrim CPS.CONS state [car,cdr] Nothing =
 do
 h@(Heap _ store _ _ _ mu) <- return $! stateHeap state
 obj <- return $! mkConsObj car cdr
 storeLocs <- storeAlloc state 1
 (store',mu') <- return $! insertObjs (zip storeLocs [obj]) store mu
 return $! (h{heapStore = store',heapCounter = mu'},
           Set.singleton $! ValStoreLoc $! head storeLocs)



 -- BOTH of these append operations are unsound.
evalPrim CPS.APPEND state (dloc:rest) _ =
 do
 () <- return $! unsafePerformIO 
         (do
          putStrLn "Warning: Using unsound append!"
          return $! ())
 return $! (stateHeap state,dloc)

evalPrim CPS.APPEND state _ (Just dlocArgs) =
 do
 (_,dloc) <- evalPrim CPS.CAR state [dlocArgs] Nothing
 () <- return $! unsafePerformIO 
         (do
          putStrLn "Warning: Using unsound append!"
          return $! ())
 return $! (stateHeap state,dloc)




 -- Errors.
-- evalPrim CPS.ERROR state _ _ = 
--  do
--  return $! (,bot)


evalPrim prim _ _ _ = 
 error $ "Unhandled primitive in evalPrim: " ++ (show prim)



heapCar :: (Heap ext) -> (D ext) -> (D ext)
heapCar (Heap { heapStore = store }) dloc =
 let
  locs =  Set.toList dloc
  locs' = filter isStoreLocVal locs
  locs'' = map (\ (ValStoreLoc loc) -> loc) locs'
 in joins [ obj `at` Sym.car' |  loc <- locs'', obj <- return $! store `at` loc ]


heapCdr :: (Heap ext) -> (D ext) -> (D ext)
heapCdr (Heap { heapStore = store }) dloc =
 let
  locs =  Set.toList dloc
  locs' = filter isStoreLocVal locs
  locs'' = map (\ (ValStoreLoc loc) -> loc) locs'
 in joins [ obj `at` Sym.car' |  loc <- locs'', obj <- [store `at` loc] ]


allocTopString ::
 State ext -> 
 (St.State (AnalysisState ext) (Heap ext,D ext))
allocTopString state =
 do
 [strLoc] <- stringAlloc state 1
 let h@(Heap { heapStrings = strStore } ) = stateHeap state
 let strStore' = strStore `join` (Map.singleton strLoc top)
 return $! (h{ heapStrings = strStore' }, Set.singleton (ValStringLoc strLoc))


listSmushed heap carseen cdrseen [] = carseen
listSmushed heap carseen cdrseen (val@(ValStoreLoc loc):tl)
 | loc `Set.member` cdrseen = listSmushed heap carseen cdrseen tl
 | otherwise =
   listSmushed heap (carseen `join` (heapCar heap (Set.singleton val)))
               (Set.insert loc cdrseen)
               ((Set.toList (heapCdr heap (Set.singleton val)) ++ tl))
listSmushed heap carseen cdrseen (nonloc:tl) = 
 listSmushed heap carseen cdrseen tl
 


firstn :: 
 (State ext) ->
 Int ->
 (D ext) ->
 (St.State (AnalysisState ext) ([D ext],D ext))
firstn state n dloc
 | n < 0 = error $ "firstn of a negative?"
 | n == 0 = return $! ([],dloc)
 | n > 0 = 
 do
 (_,car) <- evalPrim (CPS.CAR) state [dloc] Nothing
 (_,cdr) <- evalPrim (CPS.CDR) state [dloc] Nothing
 (tail,dnext) <- firstn state (n-1) cdr
 return $! (car:tail,dnext)

 


mkList :: 
 [D ext] -> [Val ext] -> [Obj ext]
mkList [d] [] = [mkConsObj d (Set.singleton ValNil)]
mkList (d:tl) (loc:locs) = (mkConsObj d (Set.singleton loc)):(mkList tl locs)

insertObjs ::
 [(StoreLoc,Obj ext)] -> (Store ext) -> Counter -> (Store ext,Counter)
insertObjs [] store mu = (store,mu)
insertObjs ((loc,obj):tl) store mu = 
 insertObjs tl
  (store `join` (Map.singleton loc obj))
  (incrementStoreLocs mu [loc])
  


mkConsObj :: (D ext) -> (D ext) -> (Obj ext)
mkConsObj first rest = 
 Map.fromList[(Sym.car',first),(Sym.cdr',rest),(Sym.objtype',
  Set.singleton (ValType Sym.cons' [Sym.car',Sym.cdr']))]


insertArray ::
 (ArrayLoc,Array ext) -> (ArrStore ext) -> Counter -> (ArrStore ext,Counter)
insertArray (loc,arr) arrstore mu =
 (arrstore `join` (Map.singleton loc arr), 
  incrementArrayLocs mu [loc])

mkArray :: (D ext) -> [D ext] -> (Array ext)
mkArray dlen dcontents =
 Map.fromList [IndLength ==> dlen,IndAny ==> joins dcontents]





 -- The transition relation.
next ::
 (Eq ext) =>
 (State ext) ->
 (St.State (AnalysisState ext) [State ext])



next state@(Apply 
  (ValPrim (CPS.Prim (CPS.CPS CPS.APPLY) _))
  (Parms [dproc,dloc] [lab] [cont] Nothing)
  h
  t)
 | Sym.cont' == lab = 
  do
  t' <- tick state t
  return $! 
   do
   proc <- Set.toList dproc
   return $! Apply proc (Parms [] [lab] [cont] (Just dloc)) h t



next state@(Apply 
  (ValPrim (CPS.Prim (CPS.CPS prim) _))
  parms
  h
  t)
 | prim == CPS.HALT || prim == CPS.ERROR =
   do
   return $! []




next state@(Apply 
  (ValPrim (CPS.Prim (CPS.CPS prim) _))
  (Parms ordered [lab] [cont] rest)
  h
  t)
 | Sym.cont' == lab = 
   do
   t' <- tick state t
   (h',answer) <- evalPrim prim state ordered rest 
   return $! 
    [ Apply cont (Parms [answer] [] [] Nothing) h' t' | 
      cont <- Set.toList cont ]



next state@(Apply { stateProc = ValPrim (CPS.Prim haltp _) }) 
 | haltp == CPS.HALT || haltp == CPS.ERROR =
 do 
  -- Done!
 return $! [] 

-- next state@(Apply 
--  (ValPrim (CPS.Prim CPS.APPLY _))
--  parms@(Parms [dproc,dloc] [] [] Nothing)
--  h
--  t
--  ) =
--  do
--  t' <- tick state t
--  return $! 
--   do
--   proc <- Set.toList dproc
--   return $! Apply proc (Parms [] [] [] (Just dloc)) h t



 -- (begin ...)
next state@(Eval
 (CPS.Seq
  [CPS.App (CPS.Prim prim _) args@(CPS.Args ordered [] [] Nothing _) _,
   call] _)
  be 
  h@(Heap ve store arrstore strstore log mu)
  t) 
   -- Evaluate for effect only.
 | prim == CPS.HALT || prim == CPS.ERROR = return $! []
 | otherwise = 
 do
 let args = map (a be h t) ordered
 t' <- tick state t
 (h',answer) <- evalPrim prim state args Nothing
  -- Mark nothing as having happened since this time.
 let log' = log `join` (Map.singleton t' emptyFrameString)
 return $ [Eval call be h'{heapLog = log'} t']




 -- (set! v exp ...)
next state@(Eval
 (CPS.Set (CPS.Ref sym _) exp call ann)
 be
 h@(Heap ve _ _ _ log mu)
 t) =
 do
 let d = a be h t exp
 t' <- tick state t
 let log' = log `join` (Map.singleton t' emptyFrameString)
  -- Contract: Free variables are bound in time t0.
 let bindtime = Map.findWithDefault (error "Top-level free var?") sym be
 bind <- return $! (sym,bindtime)
  -- Strong update?
 let ve' = case mu `muat` (HDBind bind) of
            ONE ->  Map.insert bind d ve
            _ -> ve `join` (Map.fromList [(bind,d)])
 return $! [Eval call be (h{heapVEnv=ve',heapLog=log'}) t']



next state@(Eval
 (CPS.Let 
   forms
   [CPS.App (CPS.Prim prim _) args _]
   call _)
  be 
  h
  t)
 | prim == CPS.HALT || prim == CPS.ERROR = return $! []



 -- let for primtive; result unused.
next state@(Eval
 (CPS.Let 
   [CPS.Forms [] [] [] (Just (CPS.Var sym _)) _]
   [CPS.App (CPS.Prim prim _) args@(CPS.Args ordered [] [] Nothing _) _]
   call _)
  be 
  h@(Heap ve store arrstore strstore log mu)
  t) 
   -- This variable will never be used; evaluate purely for effect.
  | not(sym `Map.member` (CPS.free call)) =
   do
   t' <- tick state t
   let args = map (a be h t) ordered
   let frame = (sym,t)
   let mu' = incrementFrames mu [(sym,t)]
   let delta_p = downFor frame
   let log' = (logcat mu' log delta_p) `join`
             (Map.singleton t' emptyFrameString)
   (h',answer) <- evalPrim prim state args Nothing
   return $ [Eval call be (h'{heapLog = log',heapCounter=mu'}) t']




next state@(Eval
 (CPS.Let 
   [CPS.Forms [CPS.Var sym _] [] [] Nothing _]
   [CPS.App (CPS.Prim prim _) args@(CPS.Args ordered [] [] Nothing _) _]
   call _)
  be 
  h@(Heap ve store arrstore strstore log mu)
  t) =
   do
   let args = map (a be h t) ordered

   t' <- tick state t
   (h',answer) <- evalPrim prim state args Nothing

   let binds = [(sym,t')]
   let be' = (be `Might.insertAll` binds) `restrictBEnv` (CPS.free call)

   let bindds = [((sym,t'),answer)]
   let ve' = ve `join` (Map.fromList bindds)

   let (Heap _ _ _ _ _ mu') = h'   

   let frame = (sym,t)

   let mu'' = incrementBinds mu' binds
   let mu''' = incrementFrames mu'' [frame]

   let delta_p = downFor frame
   let log' = (logcat mu''' log delta_p) `join`
             (Map.singleton t' emptyFrameString)


   return $ [Eval call be' (h'{heapVEnv = ve', heapCounter = mu''', heapLog = log' }) t']



 -- TODO: Make this smarter.
next state@(Eval
 (CPS.Cond [cond,CPS.Lit (CPS.LitBool True) _] [thn,els] ann)
 be
 h@(Heap{heapLog=log})
 t) =
 do
 t' <- tick state t
 let log' = log `join` (Map.singleton t' emptyFrameString)
 return $! 
  do
  call <- [thn,els]
  return $! Eval call be h{heapLog = log'} t' 


next state@(Eval call@(CPS.Cond{}) be h t) =
 error $ "Malformed (cond ...) during analysis: " ++ (show call)
 
  




next state@(Eval
 (CPS.Let 
   [CPS.Forms [CPS.Var sym _] [] [] Nothing _]
   [exp]
   call _)
  be 
  h@(Heap ve store arrstore strstore log mu)
  t) =
   do
   let answer = a be h t exp

   t' <- tick state t

   binds <- return $! [(sym,t')]
   be' <- return $! (be `Might.insertAll` binds) `restrictBEnv` (CPS.free call)

   bindds <- return $! [((sym,t'),answer)]
   ve' <- return $! ve `join` (Map.fromList bindds)

   let frame = (sym,t)
 
   let mu' = incrementBinds mu binds
   let mu'' = incrementFrames mu' [frame]

   let delta_p = downFor frame
   let log' = (logcat mu'' log delta_p) `join`
              (Map.singleton t' emptyFrameString)

   return $ [Eval call be' (h{heapVEnv = ve', heapCounter = mu'', heapLog = log' })  t']
 




next state@(Eval 
  letrec@(CPS.LetRec vars lams call _)
  be 
  h@(Heap ve store arrstore strstore log mu)
  t) =
 do 
 let variables = map (\ (CPS.Forms [v] [] [] Nothing _) -> v) vars

 t' <- tick state t
 let binds = [ (sym,t') | CPS.Var sym _ <- variables ] 
 let be' = be `Might.insertAll` binds 
 let denotables = map (a be' h t') lams
 let bindds = zip binds denotables
 let ve' = ve `join` (Map.fromList bindds) 

 let frame = (CPS.labelOf letrec,t)
 let mu' = incrementBinds mu binds
 let mu'' = incrementFrames mu' [frame]

 let delta_p = downFor frame
 let log' = (logcat mu'' log delta_p) `join`
             (Map.singleton t' emptyFrameString)

 let be'' = be' `restrictBEnv` (CPS.free call)
  -- log' <- error "no log'"
 return $! [Eval call be'' (Heap ve' store arrstore strstore log' mu'')  t']




next state@(Eval 
  (CPS.App f (CPS.Args ordered labels labeled Nothing _) ann)
  be 
  h@(Heap _ _ _ _ log mu)
  t) =
 do 
 t' <- tick state t

 let procs = a be h t f 
 True <-
  if Set.size procs > 2 then Might.debugM "Forking!" else return True

 let ordered' = map (a be h t) ordered
 let labeled' = map (a be h t) labeled

  -- Pull out the conts.
 let conts = filter isContD labeled'

  -- Fork to all procedures.
 return $!
  do
  proc <- Set.toList procs

   -- If we have continuations, rest the stack.
  let delta_p = case conts of 
                  [] -> inv (age log proc) -- This is a continuation.
                  [cont] -> inv (age log cont)
                  _ -> error "Multiple continuations not yet supported."
  let log' = logcat mu log delta_p  

  let h' = h{ heapLog = log' }

  return $! Apply proc (Parms ordered' labels labeled' Nothing) h' t'
   

 -- TODO: Support all types of parameters for Apply states.




 -- Continuation application.
next state@(Apply
 (ValClo
  lam@(CPS.Lam forms@(CPS.Forms ordered [] [] Nothing _) call _)
  be
  birthdate)
 parms@(Parms ordered' [] [] Nothing) 
 _
 t) 
   -- # args. mismatch?
 | not(length ordered == length ordered') = 
   error $ "Args to continuation don't match!" 
 | otherwise =
  do
  t' <- tick state t
  let frame = (CPS.labelOf lam, t)
  let delta_p = downFor frame
  Just (variables,denotables,h') <- matchFormsParms state forms parms
  (Heap ve store arrstore strstore log mu) <- return $! h'
  let binds = [ (sym,t') | sym <- variables ] 
  let bindds = zip binds denotables
  let be' = (be `Might.insertAll` binds) `restrictBEnv` (CPS.free call)
  let ve' = ve `join` (Map.fromList bindds)
  let mu' = incrementBinds mu binds
  let mu'' = incrementFrames mu' [frame]
  let log' = (logcat mu'' log delta_p) `join`
             (Map.singleton t' emptyFrameString)
  return $! [Eval call be' h'{ heapVEnv=ve', heapLog=log', heapCounter=mu'' } t']




 -- User function application.

next state@(Apply
 (ValClo
  lam@(CPS.Lam forms@(CPS.Forms ordered [lab] [k] rest _) call _)
  be
  birthdate)
 parms@(Parms ordered' [lab'] [c'] _) -- TODO: Does this allow rest?
 _
 t) 
 | not(lab == Sym.cont' && lab' == Sym.cont') = 
   error $ "Labeled arguments don't match in:\n" ++ (show state)
 | otherwise = 
  do
  t' <- tick state t
  let frame = (CPS.labelOf lam, t)
  let delta_p = downFor frame
  match <- matchFormsParms state forms parms
  case match of
   Just (variables,denotables,h') -> do
    (Heap ve store arrstore strstore log mu) <- return $! h'
    let binds = [ (sym,t') | sym <- variables ] 
    let bindds = zip binds denotables
    let be' = (be `Might.insertAll` binds) `restrictBEnv` (CPS.free call)
    let ve' = ve `join` (Map.fromList bindds)
    let mu' = incrementBinds mu binds
    let mu'' = incrementFrames mu' [frame]
    let log' = (logcat mu'' log delta_p) `join`
               (Map.singleton t' emptyFrameString)
    return $! [Eval call be' h'{ heapVEnv=ve', heapLog=log' , heapCounter=mu'' } t']
   Nothing -> do
    True <- Might.debugM $ "Warning: mismatched args state hit."
    return $! []
    -- error $ "params not matched in:\n" ++ (show state)
-- (Heap ve' store arrstore strstore log' mu'') 

 





 -- TODO: This case redundant?
next state@(Apply
 (ValClo
  lam@(CPS.Lam forms@(CPS.Forms [] [] [] (Just v) _) call _)
  be
  birthdate)
 parms@(Parms _ [] [] _) 
 h@(Heap ve store arrstore strstore log mu)
 t) 
   -- Will this variable ever be used?
 | not((CPS.varSym v) `Map.member` (CPS.free call)) =
  do
  t' <- tick state t 
  let frame = (CPS.labelOf lam, t)
  let delta_p = downFor frame
  let mu' = incrementFrames mu [frame]
  let log' = (logcat mu' log delta_p) `join`
             (Map.singleton t' emptyFrameString)
  return $! [Eval call be (h{ heapCounter = mu', heapLog = log' }) t']


 -- POSSIBLE BUG WARNING:

 -- If these states are hit, it's probably a bug in GCFA, or the
 -- original program.
next state@(Apply { stateProc = val })
 | (not . isApplicableVal) val =
 do
  -- TODO: Flag these states in AnalysisState for debugging.
 return $! []


next state = error $ "state not yet handled: " ++ (show state)





matchFormsParms ::
 (State ext) ->
 (CPS.Term ext) ->
 (Parms ext) ->
 (St.State (AnalysisState ext)
   (Maybe ([Sym.Symbol],[D ext],Heap ext)))


 -- Clean up apply parameters first.
matchFormsParms state
 (CPS.Forms ordered [] [] Nothing _)
 (Parms     [] [] [] (Just dloc)) =
  -- Make a list out of ordered.  Bind that to sym.
 do
 (ordered',dnextloc) <- firstn state (length ordered) dloc
 return $!
  Just (map CPS.varSym ordered,ordered',stateHeap state)


matchFormsParms state
 (CPS.Forms ordered  [] [] Nothing _)
 (Parms     ordered' [] [] Nothing) =
 do
 if length ordered == length ordered'
  then 
  return $! Just (map CPS.varSym ordered, ordered',stateHeap state)
  else
  return $! Nothing

matchFormsParms state
 (CPS.Forms ordered  [lab] [CPS.Var sym _] Nothing _)
 (Parms     ordered' [lab'] [d'] Nothing) 
 | lab == Sym.cont' && lab' == Sym.cont' =
 do
 if length ordered == length ordered' 
  then 
  return $! Just (sym:(map CPS.varSym ordered), d':ordered',stateHeap state)
  else
  return $! Nothing


matchFormsParms state
 (CPS.Forms ordered [lab] [CPS.Var sym _] (Just (CPS.Var rest _)) _)
 (Parms     ordered' [lab'] [c'] Nothing) 
 | length ordered' < length ordered =
   return $! Nothing -- Argument mismatch.

 | lab == Sym.cont' && lab' == Sym.cont' &&
   length ordered == length ordered' =
   do
   return $! Just (rest:sym:(map CPS.varSym ordered),dNil:c':ordered',stateHeap state)

 | lab == Sym.cont' && lab' == Sym.cont' &&
   length ordered < length ordered' =
   do
   (h',restd) <- evalPrim CPS.LIST state tail Nothing
   return $! Just (rest:sym:syms,restd:c':ds,h')
   where
    (symds,tail) = zipRemainder (map CPS.varSym ordered) ordered'

    (syms,ds) = unzip symds

    zipRemainder [] [] = ([],[])
    zipRemainder [] dl = ([],dl)
    zipRemainder (sym:stl) (d:dtl) = 
     let
      (symds,tail) = zipRemainder stl dtl
     in ((sym,d):symds,tail)
     


matchFormsParms state
 (CPS.Forms [] [] [] (Just (CPS.Var sym _)) _)
 (Parms     ordered [] [] Nothing) =
  -- Make a list out of ordered.  Bind that to sym.
 do
 (h',dloc) <- evalPrim CPS.LIST state ordered Nothing
 return $!
  Just ([sym],[dloc],h')

matchFormsParms state forms parms = error $
 "Unmatched Forms Parms: " ++ (show forms) ++ "\n v.\n" ++ (show parms)



 -- GARBAGE COLLECTION


valTouches :: (Val ext) -> [HeapDom]
valTouches (ValStoreLoc loc) = [HDStoreLoc loc]
valTouches (ValArrayLoc loc) = [HDArrayLoc loc]
valTouches (ValStringLoc loc) = [HDStringLoc loc]
valTouches (ValPrim{}) = []
valTouches (ValClo lam be birthdate) =
 let
  freeInLam = map fst (Map.toList $! CPS.free lam)
  binds = Might.mapMaybe (\ sym -> if sym `Map.member` be
                                     then Just (sym,be Map.! sym)
                                     else Nothing) freeInLam
  in (HDLogTime birthdate):(map HDBind binds)
valTouches (ValType{}) = []
valTouches (ValString{}) = []
valTouches (ValSymbol{}) = []
valTouches (ValBool{}) = []
valTouches (ValNum{}) = []
valTouches (ValNil{}) = []
valTouches (ValChar{}) = []
valTouches (ValVoid{}) = []

dTouches d = foldl (++) [] (map valTouches (Set.toList d))

stateTouches :: (State ext) -> [HeapDom]
stateTouches (Eval call be h t) = 
 let
  freeInCall = map fst (Map.toList $! CPS.free call)
  binds = Might.mapMaybe (\ sym -> if sym `Map.member` be
                                     then Just (sym,be Map.! sym)
                                     else Nothing) freeInCall
  in (HDLogTime t):(map HDBind binds)
stateTouches (Apply proc (Parms ordered _ labeled rest) h t) =
 let
  procTouches = valTouches proc
  restList = Maybe.maybeToList rest
  restTouches = foldl (++) [] (map dTouches restList)
  orderedTouches = foldl (++) [] (map dTouches ordered)
  labeledTouches = foldl (++) [] (map dTouches labeled) 
  in (procTouches ++ orderedTouches ++ labeledTouches ++ restTouches)

data GCState ext = GCState {
  gcVisited :: !(Heap ext) ,
  gcHeapWorklist :: ![HeapDom] ,
  gcValWorklist :: ![Val ext] ,
  gcHeap :: !(Heap ext) 
 }
 deriving (Show)


gc :: (State ext) -> (State ext)
gc state = 
 let
  in state { stateHeap = reaches state }



reaches :: (State ext) -> (Heap ext) 
reaches state =
 let
  work = stateTouches state
  oldheap = stateHeap state
  visited = (Heap bot bot bot bot bot muZero)
  gcstate' = runGC (GCState visited ((HDLogTime []):work) [] oldheap)
  (GCState h' _ _ _) = gcstate'
 in h'

  
 


runGC :: (GCState ext) -> (GCState ext)
runGC gcs@(GCState vis [] [] h) = gcs
runGC (GCState vis@(Heap visve _ _ _ _ mu) (hd@(HDBind b):tl) valwork h) =
 let
    -- Pull hd off the worklist.
   d = case Map.lookup b (heapVEnv h) of 
        Just d -> d
        Nothing -> error $ "Error in GC on HDBind: " ++ (show b)

   vals = Set.toList d

   visve' = Map.insert b d visve

   mu' = muInsert hd ((heapCounter h) `muat` hd) mu

   -- Yay for laziness.
  in case Map.lookup b visve of
        -- Already scanned this.
      Just _ -> runGC (GCState vis tl valwork h) 

        -- Haven't seen this.
      Nothing ->
       runGC (GCState (vis{heapVEnv=visve',heapCounter=mu'}) tl
                       (vals ++ valwork) h)

runGC (GCState vis@(Heap _ store _ _ _ mu) (hd@(HDStoreLoc l):tl) valwork h)=
 let
    -- Pull hd off the worklist.
   obj = case Map.lookup l (heapStore h) of 
        Just obj -> obj
        Nothing -> error $ "Error in GC on HDStoreLoc: " ++ (show l)

   d = joins $! map snd (Map.toList obj)
   vals = Set.toList d

   store' = Map.insert l obj store

   mu' = muInsert hd ((heapCounter h) `muat` hd) mu

   -- Yay for laziness.
  in case Map.lookup l store of
        -- Already scanned this.
      Just _ -> runGC (GCState vis tl valwork h) 

        -- Haven't seen this.
      Nothing ->
       runGC (GCState (vis{heapStore=store',heapCounter=mu'}) tl
                       (vals ++ valwork) h)

runGC (GCState vis@(Heap _ _ arrstore _ log mu) (hd@(HDArrayLoc l):tl) valwork h)=
 let
    -- Pull hd off the worklist.
   arr = case Map.lookup l (heapArrays h) of 
        Just arr -> arr
        Nothing -> error $ "Error in GC on HDArrayLoc: " ++ (show l)

   d = joins $! map snd (Map.toList arr)
   vals = Set.toList d

   astore' = Map.insert l arr arrstore

   mu' = muInsert hd ((heapCounter h) `muat` hd) mu

   -- Yay for laziness.
  in case Map.lookup l arrstore of
        -- Already scanned this.
      Just _ -> runGC (GCState vis tl valwork h) 

        -- Haven't seen this.
      Nothing ->
       runGC (GCState (vis{heapArrays=astore',heapCounter=mu'}) tl
                       (vals ++ valwork) h)

runGC (GCState vis@(Heap _ _ _ sstore _ mu) (hd@(HDStringLoc l):tl) valwork h)=
 let
    -- Pull hd off the worklist.
   str = case Map.lookup l (heapStrings h) of 
        Just str -> str
        Nothing -> error $ "Error in GC on HDStringLoc: " ++ (show l)

   sstore' = Map.insert l str sstore

   mu' = muInsert hd ((heapCounter h) `muat` hd) mu

  in runGC (GCState (vis{heapStrings=sstore',heapCounter=mu'}) tl 
                     (valwork) h)

runGC (GCState vis@(Heap _ _ _ _ newlog newmu) (hd@(HDLogTime t):tl) valwork h) 
    -- In theory, log can be bot only with logging turned off.
 | heapLog h == bot = runGC (GCState vis tl valwork h)
 | otherwise =
  let
   oldlog = heapLog h
   oldmu = heapCounter h
   birthdate = t

   -- Transfer this entry for the log over.
   delta_p =
    Map.findWithDefault
     (error $ "Illegal time: " ++ (show birthdate))
     birthdate oldlog

   newlog' = Map.insert birthdate delta_p newlog

   -- Now, find the frames in this delta, and preserve the counts.
   frames = map HDFrame (findFrames delta_p)

   frameCounts = map (\ frame -> (frame,oldmu `muat` frame)) frames

   newmu' = muInsertMany frameCounts newmu


   vis' = vis { heapLog = newlog', heapCounter = newmu' }

  in runGC (GCState vis' tl valwork h)

-- runGC (GCState vis hdwork ((val@(ValClo lam be birthdate)):tl) h) =
--  let

--   in runGC (GCState vis' ((valTouches val) ++ hdwork) tl h)
  
runGC (GCState vis hdwork (val:tl) h) =
 runGC (GCState vis ((valTouches val) ++ hdwork) tl h)

runGC gcs = error $ "unmatched in runGC: " ++ (show gcs)


--  let
--   cmp (l,d) (l',d') = compare l l'

--   (slabels,slabeled) = unzip $! List.sortBy cmp (zip labels labeled)
--   (slabels',slabeled') = unzip $! List.sortBy cmp (zip labels' labeled')

--   restds = take 

--   safe = (slabels == slabels') && 
--       (Maybe.isNothing rest || length ordered' >= length ordered)
--   in 
--    if 
--    then else if
--       (slabels == slabels') && 
--       (Maybe.isJust rest || length ordered == length ordered')
--    else
--     Nothing






 -- The analysis itsef.


data StackPolicy = StackSteele | StackFortran
 deriving (Show)

data AnalysisOptions = Opts {
  optkCFA :: !Int ,
  optPolyStore :: !Bool ,
  optGlobalHeap :: !Bool ,
  optCount :: !Bool ,
  optLog :: !Bool ,
  optGC :: !Bool ,
  optCountOff :: !Int ,
  optJoinConds :: !Bool ,
  optPruneCacheThreshold :: !Int ,
  optStackPolicy :: !StackPolicy 
 }
 deriving (Show)

defaultOptions = Opts {

   -- Precision options.
  optkCFA = 0 ,
  optPolyStore = False , -- TODO: Implement this.
  optGlobalHeap  = True , 
  optCount = False ,
  optLog = False ,
  optGC = True ,
  optJoinConds = False ,
  optPruneCacheThreshold = -1 ,

   -- Platform options.
  optStackPolicy = StackSteele ,
   
   -- Debugging options.
  optCountOff = 100
 }


data AnalysisState ext = AState {
  anTermCheckInterval :: !Int ,
  anSinceTermCheck :: !Int ,
  anOptions :: !AnalysisOptions ,
  anGlobalHeap :: !(Heap ext) ,
  anNumVisited :: !Int ,
  anJoinmap :: !(Map (StateHash ext) [State ext]) ,
  anJoinlist :: ![State ext] ,
  anWorklist :: ![State ext] ,
  anVisited :: !(Map (StateHash ext) [Heap ext]) ,

   -- Optimization info.
  anEscaping :: !(Set (CPS.Term ext)) ,
  anDead :: !(Set (CPS.Term ext))
 }
 deriving (Show)


initAnalysisState = AState {
  anTermCheckInterval = 10 ,
  anSinceTermCheck = 10 ,
  anOptions = defaultOptions ,
  anGlobalHeap = bot ,
  anNumVisited = 0 ,
  anJoinmap = Map.empty ,
  anJoinlist = [] ,
  anWorklist = [] ,
  anVisited = Map.empty ,
 
   -- Optimization info.
  anEscaping = Set.empty ,
  anDead = Set.empty
 }



 -- Analysis functions.


 -- analyzeDead removes call sites from the dead list.
analyzeDead :: (State ext) -> (St.State (AnalysisState ext) ())

analyzeDead state@(Apply{}) = return $! ()

analyzeDead state@(Eval { stateCALL = call@(CPS.Set _ app _ _) }) =
 do
 anstate <- St.get
 let dead = anDead anstate
 let dead' = Set.delete call dead
 let dead'' = Set.delete app dead'
 St.put $! anstate { anDead = dead'' }

analyzeDead state@(Eval { stateCALL = call@(CPS.Let _ [app] _ _) }) =
 do
 anstate <- St.get
 let dead = anDead anstate
 let dead' = Set.delete call dead
 let dead'' = Set.delete app dead'
 St.put $! anstate { anDead = dead'' }

analyzeDead state@(Eval { stateCALL = call@(CPS.Seq [app,rest] _) }) =
 do
 anstate <- St.get
 let dead = anDead anstate
 let dead' = Set.delete call dead
 let dead'' = Set.delete app dead'
 St.put $! anstate { anDead = dead'' }

analyzeDead state@(Eval { stateCALL = call@(CPS.App{}) }) =
 do
 anstate <- St.get
 St.put $! anstate { anDead = Set.delete call (anDead anstate) }

analyzeDead state@(Eval { stateCALL = call@(CPS.LetRec{}) }) =
 do
 anstate <- St.get
 St.put $! anstate { anDead = Set.delete call (anDead anstate) }

analyzeDead state@(Eval { stateCALL = call@(CPS.Cond{}) }) =
 do
 anstate <- St.get
 St.put $! anstate { anDead = Set.delete call (anDead anstate) }

analyzeDead unhandled = error $ "Unsure how to analyze for deadness: " ++
                           (show unhandled)


analyzeEscaping :: (State ext) -> (St.State (AnalysisState ext) ())

 -- BUG: Eventually, this has to check Parms too.
analyzeEscaping state@(Apply { stateProc = (ValClo lam be birthdate) } ) =
 do
 anstate <- St.get
 if not $ optLog $ anOptions $ anstate 
  then error $ "Cannot do escape analysis without logging."
  else do
   let log = heapLog $ stateHeap $ state
   let escapes = anEscaping anstate

   let delta_p = case Map.lookup birthdate log of
                 Just delta_p -> delta_p
                 Nothing -> error $ "Escape analysis found bug in logging."

   let thisEscapes = not(hasNoUp delta_p)

   let escapes' =
        if thisEscapes
        then Set.insert lam escapes
        else escapes

   St.put $! anstate { anEscaping = escapes' }


analyzeEscaping _ =
 do
 return $! ()




 -- Global heap management.
installGlobalHeap h =
 do
 anstate <- St.get
 St.put $! anstate { anGlobalHeap = h }


heapMustBeMonotonic :: St.State (AnalysisState ext) Bool
heapMustBeMonotonic =
 do
 anstate <- St.get 
 let opts = anOptions anstate
 return $! optGC opts || optCount opts


widenGlobalHeap :: [State ext] -> (St.State (AnalysisState ext) ())

widenGlobalHeap [] =
 do
  -- Nothing to merge?  Keep the one we have.
 return ()

widenGlobalHeap states = 
 do
 monotonicHeap <- heapMustBeMonotonic

 if monotonicHeap
  then do
    -- Throw out the old heap.  Merge all new ones together.
   let newHeap = joins (map stateHeap states)
   installGlobalHeap newHeap

  else do
   oldHeap <- Might.retrieve anGlobalHeap
   let newHeap = joins (oldHeap:(map stateHeap states))
   installGlobalHeap newHeap




pushStates :: [State ext] -> (St.State (AnalysisState ext) ())
pushStates todo =
 do
 st <- St.get 
 let shouldJoinConds = optJoinConds $ anOptions $ st
 if shouldJoinConds 
  then do
   let (joins,normals) = List.partition isApplyJoin todo
   pushJoinStates joins
   St.put st { anWorklist = normals ++ (anWorklist st) }
  else do
   St.put st { anWorklist = todo ++ (anWorklist st) }



pushJoinStates :: [State ext] -> (St.State (AnalysisState ext) ())
pushJoinStates todo =
 do
 st <- St.get 
 let shouldJoinConds = optJoinConds $ anOptions $ st
 if shouldJoinConds
  then do
   let joinmap = anJoinmap st
   let joinmap' = push joinmap todo 
   St.put $! st { anJoinmap = joinmap' }
  else do
   pushStates todo
 where
   push map [] = map
   push map (hd:tl) = 
    let
     map' =
      Map.insertWith
       ( \ new old -> new ++ old) (hashState hd) [hd] map
    in push map' tl



hasWorkLeft :: (St.State (AnalysisState ext) Bool) 
hasWorkLeft =
 do
 st <- St.get
 return $! (not (null $! anWorklist st)) || 
           (not (Map.null $! anJoinmap st)) ||
           (not (null $! anJoinlist st))



popState :: (St.State (AnalysisState ext) (State ext)) 
popState =
 do
 st <- St.get
 let Opts{ optGC = shouldGC, optJoinConds = shouldJoinConds } = anOptions st 
 let joinlist = anJoinlist st
 let worklist = anWorklist st
 let joinmap = anJoinmap st
 if (not $ null joinlist)
  then do
    if shouldJoinConds then return () else error $ "Assert Fail: No joining!" 
    St.put $! st { anJoinlist = tail joinlist }
--    if (Might.debug $! "came from joinlist") then return () else return ()
    return $! unmarkApplyJoin $! head joinlist
  else if (not $ null worklist)
  then do
    St.put $! st { anWorklist = tail worklist }
--    if (Might.debug $! "came from worklist") then return () else return ()
    return $! head worklist
  else if (not $! Map.null joinmap)
  then do
   if shouldJoinConds then return () else error $ "Assert Fail: No joining!!"
    -- Convert the joinmap into the joinlist, merging all joins at the
    -- same index.
--   if (Might.debug $! "came from joinmap") then return () else return ()
   let gc' = if shouldGC then gc else \ x -> x
   let joinlist' = map (joins . (map gc') . snd) (Map.toList joinmap)
   St.put $! st { anJoinlist = tail joinlist' }
   return $! unmarkApplyJoin $! head joinlist'
  else error $ "no more states!"



 -- The seen set stores heaps.
seenBy :: (Heap ext -> Heap ext -> Bool) -> (State ext) -> 
 (St.State (AnalysisState ext) Bool)
seenBy leq state =
 do
 if isApply state 
  then do
   return $! False 
  else do
  st <- St.get
  let heap = stateHeap state
  let seen = anVisited st
  let stateHash@(EvalHash call t be) = hashState state
  let seenHeaps = case Map.lookup stateHash seen of
                            Just lst -> lst
                            Nothing -> [] 

  -- Quit early if less than someone.  Throw away any heap that this
  -- heap is greater than.

  let (found,seenHeaps') = filterSeen leq heap seenHeaps 
  let pruneThreshold = optPruneCacheThreshold $ anOptions $ st
  if (pruneThreshold > 1) && (length seenHeaps' >= pruneThreshold)
   then do
    if (pruneThreshold <= 1) 
      then error "Assert Fail: no prune!" 
      else return ()
     -- Collapse everyone here.
    let seen' = Map.insert stateHash [] seen
    St.put $! st { anVisited = seen' }

--    True <- Might.debugM "Collapsing cache line."

     -- Enqueue the merged state.
    let mergedHeap = joins seenHeaps'
    pushStates [Eval call be mergedHeap t ]

     -- Clearly, we didn't see this guy, but we'll lie, 'cause we're
     -- going to repush a state stronger than him.
    return $! True
   else do
    let seen' = Map.insert stateHash seenHeaps' seen
    St.put $! st { anVisited = seen' }
    return $! found



filterSeen leq heap [] = (False,[heap])
filterSeen leq heap (hd:tl) = 
 if (leq heap hd)
 then (True,hd:tl)
 else if (leq hd heap)
 then filterSeen leq heap tl
 else
  let
   (found,tl') = (filterSeen leq heap tl)
   in (found,hd:tl')


modTermCheckInterval f =
 do
 anstate <- St.get 
 St.put $! anstate { anTermCheckInterval = f (anTermCheckInterval anstate) }


bumpSinceTermCheck :: 
 Int -> (St.State (AnalysisState ext) ())
bumpSinceTermCheck n = 
 do
 anstate <- St.get
 k <- return $! anSinceTermCheck anstate
 St.put $! anstate { anSinceTermCheck = k + n }
 return $! ()


shouldCheck :: 
 (St.State (AnalysisState ext) Bool)
shouldCheck = 
 do
 anstate <- St.get
 k <- return $! anSinceTermCheck anstate
 interval <- return $! anTermCheckInterval anstate
 if k >= interval
  then 
   do
   St.put $! anstate { anSinceTermCheck = 0 }
   return $! True
  else
   do
   return $! False



bumpVisited k =
 do
 st <- St.get
 St.put $! st { anNumVisited = (anNumVisited st) + k }





 -- State-hashing fun.

 -- StateHash is a perfect hash on the parts for which <: is ==.
data StateHash ext = 
  EvalHash {
   stateHashCall :: !(CPS.Term ext) ,
   stateHashTime :: !Time ,
   stateHashBEnv :: !BEnv 
   }
 | ApplyHash {
   stateHashProc :: !(Val ext) ,
   stateHashTime :: !Time
  }
 deriving (Eq,Ord,Show)



hashState :: State ext -> StateHash ext

hashState (Eval call be h t) =
 EvalHash call t be

hashState (Apply proc parms h t) = 
 ApplyHash proc t
 


interpretAbstractly :: 
 (Eq ext) =>
 (CPS.Term ext) ->
 St.State (AnalysisState ext) ()

interpretAbstractly call = 
 do 
 initAnalState <- St.get
 St.put $! initAnalState { anDead = CPS.callSites call }
 pushStates [initState call]
 explore




explore :: (Eq ext) => St.State (AnalysisState ext) ()
explore =
 do
 anstate <- St.get
 let opts = anOptions anstate
 let shouldCount = optCount opts
 let shouldLog = optLog opts
 let countOff = optCountOff opts
 let shouldGC = optGC opts
 let globalHeaped = optGlobalHeap opts
 let numVisited = anNumVisited anstate
 True <- if (numVisited `mod` countOff == 0)
        then Might.debugM $!
               "states: " ++ (show $ numVisited)
        else return $! True

 moreWork <- hasWorkLeft
 if not moreWork
  then return $! ()
  else do
   curr <- popState

    -- Invariant validation.
   if not shouldLog || (isApply curr || hasLogEntryForCurrentTime curr)
    then return ()
    else error $ "State time does not have log entry:\n" ++ (show curr)

    -- Analysis.
   analyzeDead curr

   if shouldLog then analyzeEscaping curr else return ()

    --True <- Might.debugM $ "counter: " ++ (show $ heapCounter $ stateHeap curr) ++ "\n"
    --True <- Might.debugM $ "current: " ++ (show curr) ++ "\n"
    -- True <- Might.debugM $ "touches: " ++ (show (stateTouches curr)) ++ "\n"


    -- If we're not counting, widen this guy's counter to top.
   curr <- return $!
    if shouldCount
    then curr
    else curr { stateHeap = (stateHeap curr) { heapCounter = top } }


    -- If we're not frame-string logging, narrow this guy's log to bot.
   curr <- return $!
    if shouldLog
    then curr
    else curr { stateHeap = (stateHeap curr) { heapLog = bot } }

    -- Garbage collect this state?  (Narrowing)
   curr <- if shouldGC then return $! gc curr else return $! curr

    -- Widen to a global heap for constraint-solving? (Widening)
   curr <-
    if not globalHeaped
    then return curr
    else
     do
     globalHeap <- Might.retrieve anGlobalHeap
     let newHeap = stateHeap curr `join` globalHeap
     installGlobalHeap newHeap
     curr <- return $! curr { stateHeap = newHeap }
      -- Narrow again with a GC?
     if shouldGC then return $! gc curr else return $! curr


    -- If it's not been seen, seenBy records it.
   let compareOp = wtHeap shouldCount shouldLog
   seenThis <- seenBy compareOp curr

   if seenThis 
    then do
     explore
    else do
     bumpVisited 1
     succ <- next curr

--      True <-Might.debugM $ "succs: " ++ (Might.showWithSep"\n\n" succ) ++ "\n"
--      True <-Might.debugM $ "eos\n"
     if (isEvalJoin curr) 
      then do
       pushJoinStates succ
      else do
       pushStates succ

--      if globalHeaped
--       then widenGlobalHeap succ
--       else return ()

      -- Keep exploring.
     explore


 
run :: (Eq ext) => (CPS.Term ext) -> (AnalysisState ext)
run call = St.execState (interpretAbstractly call) (initAnalysisState) 
 




 -- Hard-coded 0CFA.
tick :: (State ext) -> Time -> (St.State (AnalysisState ext) Time)

tick state t =
 do
 (Opts { optkCFA = k }) <- Might.retrieve anOptions
 tick_k k state t

tick_k :: Int -> (State ext) -> Time -> (St.State (AnalysisState ext) Time)
tick_k k (Eval { stateCALL = call }) t = 
 return $! take k ((CPS.label $ CPS.ann $ call):t)

tick_k k (Apply { stateProc = ValClo lam be birthdate }) t =
 return $! take (k+1) ((CPS.label $ CPS.ann $ lam):t)

tick_k k (Apply { stateProc = ValPrim prim }) t =
 return $! take (k+1) ((CPS.label $ CPS.ann $ prim):t)

tick_k k state t =
 return $ error $ "There is no next time for: " ++ (show state)



 -- Smash it all together.
storeAlloc :: (State ext) -> Int -> (St.State (AnalysisState ext) [StoreLoc])
-- storeAlloc (Eval { stateCALL = call, stateTime = t }) i = 
--  do 
--  return $! [ ((CPS.label . CPS.ann) call, t, n) | n <- [1..i] ]
-- storeAlloc (Apply { stateProc = ValClo lam _ , stateTime = t}) i =
--  do
--  return $! [ ((CPS.label . CPS.ann) lam, t, n) | n <- [1..i] ]
storeAlloc state i =
 do
 return $! [ (Sym.init',[],0) | n <- [1..i] ]


stringAlloc :: (State ext) -> Int -> (St.State (AnalysisState ext) [StringLoc])
stringAlloc state i =
 do
 return $! [ (Sym.init',[],0) | n <- [1..i] ]

arrayAlloc :: (State ext) -> Int -> (St.State (AnalysisState ext) [ArrayLoc])
arrayAlloc state i =
 do
 return $! [ (Sym.init',[],0) | n <- [1..i] ]



 -- Note: If things go shitty, try fleshing these out into "full" flat
 -- lattices with top & bot.
instance Lattice Sym.Symbol where
 top = Sym.SymTop
 bot = Sym.SymBot








 -- Counter library.

muZero = Counter ZERO Map.empty

muLookup x (Counter def f) = Map.lookup x f

muInsert x ZERO (Counter ZERO f) = (Counter ZERO (Map.delete x f))
muInsert x y (Counter ZERO f) = (Counter ZERO (Map.insert x y f))

muInsertMany [] f = f
muInsertMany ((x,y):tl) f = 
 muInsertMany tl (muInsert x y f)
 

muGetOnes f (Counter ZERO g) = (Counter ZERO (Map.filter f g))

(Counter def f) `muat` x = Map.findWithDefault def x f


 -- Count & Counter are a special hacks because ZERO is *not* bot.
instance Lattice Count where
 top = CountTop
 bot = CountBot

 CountBot <: _ = error $ "these should not be compared"
 _ <: CountTop = True
 
 ONE  <: ONE  = True
 ZERO <: ZERO = True
 MORE <: MORE = True
 
 _ <: _ = False
 
instance Might.Hashable Count where
 hash ZERO = 1
 hash ONE = 2
 hash MORE = 3
 hash CountTop = 4
 hash CountBot = 5


 -- TODO: Include the default.
instance Might.Hashable Counter where 
 hash (Counter def f) = Might.hash f
 hashM m (Counter def f) = (Might.hashM m f) `mod` m

instance Lattice Counter where
 bot = Counter bot Map.empty
 top = Counter top Map.empty
 
 (Counter ZERO f) <: (Counter ZERO g) =
--   let
--     -- Check entries in dom(g) but not in dom(f).
--     h = g `Map.difference` f
--    in
--     (Map.isSubmapOfBy (<:) f g) &&
--     all (==ZERO) (map snd (Map.toList h))
  Might.isSubmapOfByOption lt f g
   where
    lt (Just a) (Just b) = a <: b
    lt (Just a) Nothing = a <: ZERO
    lt Nothing (Just b) = ZERO <: b


 (Counter CountTop f) `join` (Counter CountTop g)
  | Map.null f && Map.null g = Counter CountTop Map.empty

 (Counter ZERO f) `join` (Counter ZERO g)
    -- There are three sets to consider:
    -- 1. Keys only in f.
    -- 2. Keys only in g.
    -- 3. Keys in both f and g.
  = (Counter ZERO fg)
  where
   fg = Might.combineMaps joinRight f g
   joinRight key Nothing Nothing = error "joinRight: How was this called?"
   joinRight key (Just ZERO) (Just ZERO) =
    Nothing
   joinRight key (Just l) (Just r) =
    Just $ l `join` r
   joinRight key (Just ZERO) Nothing =
    Nothing
   joinRight key (Just l) Nothing =
    Just $ l `join` ZERO
   joinRight key Nothing (Just ZERO) =
    Nothing
   joinRight key Nothing (Just r) =
    Just $ ZERO `join` r

 (Counter CountBot f) `join` mu 
  | Map.null f = mu 

 mu `join` (Counter CountBot f)
  | Map.null f = mu 

 join a b = 
  error $ "Failed match in join: " ++ (show a) ++ "; " ++ (show b)



 (Counter _ f) `meet` (Counter _ g) = 
  error $ "`meet' on counters not required, I hope."
  


incrementStringLocs = increment HDStringLoc
incrementArrayLocs = increment HDArrayLoc
incrementStoreLocs = increment HDStoreLoc
incrementBinds = increment HDBind
incrementFrames = increment HDFrame

increment increment (Counter CountTop f) entries
 | Map.null f = Counter CountTop f

increment injector (Counter def f) entries =
 (Counter def f')
 where 
  f' = Map.unionWith (\ count1 count2 -> count1 + count2) f newEntries
  ones = ONE:ones
  entries1s = zip (map injector entries) ones
  newEntries = Map.fromList entries1s





 -- OUTPUT.


-- contextSensitiveSingularValues ::
--  (Map.Map (StateHash ext) [Heap ext]) -> [(StateHash ext,Counter)]
-- contextSensitiveSingularValues visited =
--  map (\ (hash,heaps) -> (hash,trim $! joins heaps)) (Map.toList visited)
--  where
--  trim heap = muGetOnes (heapCounter heap)

 
-- singularValues ::
--  (Map.Map (StateHash ext) [Heap ext]) -> Counter
-- singularValues visited =
--  foldl (\ counter' (hash,counter)  -> counter' `join` counter)
--        bot
--        (contextSensitiveSingularValues visited)
 




instance Lattice (CPS.Term ext) where
 top = CPS.TermTop
 bot = CPS.TermBot

instance Lattice (Heap ext) where
 top = HeapTop
 bot = (Heap bot bot bot bot bot bot)

 (Heap ve s a st log mu) <: (Heap ve' s' a' st' log' mu') =
  (ve,s,a,st,log,mu) <: (ve',s',a',st',log',mu')

 (Heap ve s a st log mu) `join` (Heap ve' s' a' st' log' mu') =
  (Heap 
   (ve `join` ve')
   (s `join` s')
   (a `join` a') 
   (st `join` st')
   (log `join` log')
   (mu `join` mu'))


(Heap ve s a st log _) `wtIgnoreCounter` (Heap ve' s' a' st' log' _) =
  (ve,s,a,st,log) <: (ve',s',a',st',log')


wtHeap shouldCount shouldLog
 (Heap ve s a st log mu) (Heap ve' s' a' st' log' mu') =
 (ve,s,a,st) <: (ve',s',a',st') &&
 ((not shouldCount) || mu <: mu') &&
 ((not shouldLog || log <: log'))






instance Lattice (State ext) where
 top = StateTop
 bot = StateBot

 (Eval call be h t) <: (Eval call' be' h' t') =
  (call,be,h,t) <: (call',be',h',t')

 (Apply proc parms h t) <: (Apply proc' parms' h' t') =
  (proc,parms,h,t) <: (proc',parms',h',t')

 (Apply proc parms h t) `join` (Apply proc' parms' h' t') =
  (Apply
   (proc `join` proc')
   (parms `join` parms')
   (h `join` h')
   (t `join` t'))


instance Lattice (Parms ext) where
 top = ParmsTop
 bot = ParmsBot


 (Parms ordered labels labeled rest) `join`
  (Parms ordered' labels' labeled' rest') = 
  if not(length ordered  == length ordered') ||  
     not(labels == labels') ||  
     not(length labeled == length labeled') ||
     not(Maybe.isJust rest  ==  Maybe.isJust rest')
  then error $ "trying to merge incompatible params!"
  else 
  (Parms (ordered `join` ordered')
          labels
          (labeled `join` labeled')
          (rest `join` rest'))

 -- No labeled args: 
 (Parms ordered [] [] rest) <: 
  (Parms ordered' [] [] rest') = 
  (all (\ (o,o') -> o <: o') (zip ordered ordered')) &&
  case (rest,rest') of
   (Just o, Just o') -> o <: o'
   (Nothing,Nothing) -> True
   _ -> False

 -- Nasty.  Labeled args:
 (Parms ordered labels labeled rest) <: 
  (Parms ordered' labels' labeled' rest') = 
  (all (\ (o,o') -> o <: o') (zip ordered ordered')) &&
  (let
     cmp (l,d) (l',d') = compare l l'

     (slabels,slabeled) = unzip $! List.sortBy cmp (zip labels labeled)
     (slabels',slabeled') = unzip $! List.sortBy cmp (zip labels' labeled')
    in 
     slabels == slabels' &&
     (all (\ (o,o') -> o <: o') (zip slabeled slabeled'))) &&
  case (rest,rest') of
   (Just o, Just o') -> o <: o'
   (Nothing,Nothing) -> True
   _ -> False


instance Might.Hashable HeapDom where
 hash (HDBind x) = Might.hash x
 hash (HDStoreLoc x) = Might.hash x
 hash (HDArrayLoc x) = Might.hash x
 hash (HDStringLoc x) = Might.hash x
 hash (HDTop) = 2
 hash (HDBot) = 3

 hashM m (HDBind x) = Might.hashM m x
 hashM m (HDStoreLoc x) = Might.hashM m x
 hashM m (HDArrayLoc x) = Might.hashM m x
 hashM m (HDStringLoc x) = Might.hashM m x
 hashM m (HDTop) = 2
 hashM m (HDBot) = 3


instance Lattice HeapDom where
 top = HDTop
 bot = HDBot

instance Might.Hashable (Val ext) where
 hash _ = error $ "hash not implemeneted" 

 hashM m (ValStoreLoc l) = Might.hashM m l
 hashM m (ValArrayLoc l) = Might.hashM m l
 hashM m (ValStringLoc l) = Might.hashM m l

 hashM m (ValPrim prim) = Might.hashM m prim
 hashM m (ValClo lam be time) =
  ((Might.hashM m lam) * (Might.hashM m be) * (Might.hashM m time)) `mod` m

 hashM m (ValType name mems) =
  ((Might.hashM m name) * (Might.hashM m mems)) `mod` m

 hashM m (ValString s) = (Might.hashM m s) `mod` m

 hashM m (ValSymbol s) = (Might.hashM m s) `mod` m

 hashM m (ValNum) = 2 `mod` m
 hashM m (ValNil) = 3 `mod` m
 hashM m (ValChar) = 4 `mod` m
 hashM m (ValVoid) = 5 `mod` m
 hashM m ValTop = 6 `mod` m
 hashM m ValBot = 7 `mod` m



 hashM m v = error $ "hash not implemented for: " ++ (show v)

instance Lattice (Val ext) where
 top = ValTop
 bot = ValBot 
 
instance Lattice AString where
 top = AStringTop
 bot = AStringBot


instance Show (Count) where
 show ZERO = "0"
 show ONE = "1"
 show MORE = "oo"
 show CountTop = "Top:Count"
 show CountBot = "Bot:Count"

instance Num (Count) where
 ZERO + ZERO = ZERO
 ONE  + ZERO = ONE
 MORE + ZERO = MORE

 ZERO + ONE  = ONE
 ONE  + ONE  = MORE
 MORE + ONE  = MORE

 ZERO + MORE = MORE
 ONE  + MORE = MORE
 MORE + MORE = MORE

 CountTop + _ = CountTop
 _ + CountTop = CountTop
 
 fromInteger 0 = ZERO
 fromInteger 1 = ONE
 fromInteger _ = error "Cannot convert >1 into Count."
 
 c1 * c2 = error "multiply not defined on Count"
 abs c = error "abs not defined on Count"
 signum c = error "signum not defined on Count"


 -- DCFA

 -- Delta.
data Delta = E | D | U | DDP | UUP | UPDP
 deriving(Eq,Ord,Show)

class Uppable a where
 hasNoUp :: a -> Bool

instance Uppable Delta where
 hasNoUp E = True
 hasNoUp D = True
 hasNoUp DDP = True
 hasNoUp U = False
 hasNoUp UUP = False
 hasNoUp UPDP = False



instance Lattice Delta where
 top = error "top Delta"
 bot = error "bot Delta"
 (<:) = (==)


 -- Pow(Delta).
type PowDelta = Set.Set Delta

topPowDelta = Set.fromList[ E, D, U, DDP, UUP, UPDP ]
emptyPowDelta = Set.fromList [ E ]
downPowDelta = Set.fromList [ D ]


data F
 = AbsFrameString {
    fsDefault :: !PowDelta , 
    fsMap :: !(Map.Map Frame PowDelta)
   }
 deriving (Eq,Show)

instance Uppable a => Uppable (Set.Set a) where
 hasNoUp a = all hasNoUp (Set.toList a)

instance Uppable F where
 hasNoUp (AbsFrameString def m) 
  | def == emptyPowDelta =
    all hasNoUp (map snd (Map.toList m))
 


(AbsFrameString def f) `fsat` x = 
 case Map.lookup x f of
  Just y -> y
  Nothing -> def


abscat :: Count -> Delta -> Delta -> PowDelta
abscat MORE E   a    = Set.fromList[a]
abscat MORE a   E    = Set.fromList[a]

abscat MORE D   D    = Set.fromList[DDP]
abscat MORE D   DDP  = Set.fromList[DDP]
abscat MORE D   U    = Set.fromList[E]
abscat MORE D   UUP  = Set.fromList[U, UUP]
abscat MORE D   UPDP = Set.fromList[DDP, UPDP]
		      
abscat MORE U   D    = Set.fromList[E, UPDP]
abscat MORE U   DDP  = Set.fromList[D, DDP, UPDP]
abscat MORE U   U    = Set.fromList[UUP]
abscat MORE U   UUP  = Set.fromList[UUP]
abscat MORE U   UPDP = Set.fromList[UPDP]

abscat MORE DDP D    = Set.fromList[DDP]
abscat MORE DDP DDP  = Set.fromList[DDP]
abscat MORE DDP U    = Set.fromList[D, DDP]
abscat MORE DDP UUP  = Set.fromList[E, D, U, DDP, UUP]
abscat MORE DDP UPDP = Set.fromList[D, DDP, UPDP]		      

abscat MORE UUP D    = Set.fromList[U, UUP]
abscat MORE UUP DDP  = Set.fromList[E, D, U, DDP, UUP, UPDP]
abscat MORE UUP U    = Set.fromList[UUP]
abscat MORE UUP UUP  = Set.fromList[UUP]
abscat MORE UUP UPDP = Set.fromList[UPDP]

abscat MORE UPDP D    = Set.fromList[UPDP]
abscat MORE UPDP DDP  = Set.fromList[UPDP]
abscat MORE UPDP U    = Set.fromList[U, UUP, UPDP]
abscat MORE UPDP UUP  = Set.fromList[U, UUP, UPDP]
abscat MORE UPDP UPDP = Set.fromList[E, D, U, DDP, UUP, UPDP]

 -- abscat ONE E E = Set.fromList[E] 
abscat ONE U D = emptyPowDelta
abscat ONE D U = emptyPowDelta
abscat ONE a1 a2 = abscat MORE a1 a2
-- abscat ONE E E = emptyPowDelta
-- abscat ONE U E = Set.fromList[U] 
-- abscat ONE D E = Set.fromList[D] 
-- abscat ONE E U = Set.fromList[U] 
-- abscat ONE E D = Set.fromList[D] 
-- abscat ONE a1 a2 =
--  error $ "Can't abscat ONE anything but D and U; got: " ++ (show a1) ++ " + " ++ (show a2)
-- abscat ONE a1 a2 = abscat MORE a1 a2

abscat ZERO E  E = emptyPowDelta
abscat ZERO a1 a2 =
 error $ "Catting frame strings which have not been allocated: "
   ++ (show a1) ++ " v. " ++ (show a2)

abscat CountTop a1 a2 = abscat MORE a1 a2

deltacat ::
 Count -> (Set.Set Delta) -> (Set.Set Delta) -> (Set.Set Delta)
deltacat count l r = 
 Set.unions[ abscat count a1 a2 | a1 <- Set.toList l, a2 <- Set.toList r ]


cat :: Counter -> F -> F -> F
cat counter (AbsFrameString defL mL) (AbsFrameString defR mR) 
 | defL == Set.empty || defR == Set.empty = 
  AbsFrameString 
   Set.empty
   (Map.intersectionWithKey
     (\ k l r -> 
      let
       count = (counter `muat` (HDFrame k))
       in deltacat count l r)
     mL mR)
 | defL == emptyPowDelta && defR == emptyPowDelta = 
  AbsFrameString
   (emptyPowDelta)
   (Map.unionWithKey 
     (\ k l r -> 
      let
       count = (counter `muat` (HDFrame k))
       count' =
        if count == ZERO && ((not $ l == emptyPowDelta) || (not $ r == emptyPowDelta))
        then error $ "String not alloc'd:\n" ++ (show mL) ++ "\n v. \n" ++ (show mR) ++ "\n\ncounter: " ++ (show counter) ++ "\n\nframe: " ++ (show k)
        else count
       in deltacat count' l r)
     mL mR)

  
logcat :: Counter -> Log -> F -> Log
logcat mu log f = 
 Map.map (\ g -> cat mu g f) log


-- catF :: F -> F -> F
-- catF (AbsFrameString f) (AbsFrameString g) =
--  if (f == bot) 
--  then bot
--  else IMap.lift cat f g


class Invertible a where
 inv :: a -> a

instance Invertible Delta where
 inv E = E
 inv D = U
 inv U = D
 inv UUP = DDP
 inv DDP = UUP
 inv UPDP = UPDP

instance (Ord a, Invertible a) => Invertible (Set.Set a) where
 inv set = Set.map inv set

instance (Ord k, Invertible a) => Invertible (Map.Map k a) where
 inv im = Map.map inv im

instance Invertible F where
 inv (AbsFrameString def f) = AbsFrameString (inv def) (inv f)

instance Lattice F where
 bot = error "bottom F?"
 top = AbsFrameString topPowDelta Map.empty

 (AbsFrameString defL mL) <: (AbsFrameString defR mR) 
  | defR == topPowDelta && Map.null mR = True -- _ <: top
  | defL == emptyPowDelta && defR == emptyPowDelta =
    Might.isSubmapOfByOption lt mL mR
    where
     lt (Just a) (Just b) = a <: b
     lt (Just a) Nothing = a <: defR
     lt Nothing (Just b) = defL <: b

 (AbsFrameString defL mL) `join` (AbsFrameString defR mR)
  | defL == emptyPowDelta && defR == emptyPowDelta =
    (AbsFrameString
     emptyPowDelta 
     (Might.combineMaps comb mL mR))
    where
     comb k (Just a) (Just b) = Just (a `join` b)
     comb k (Just a) Nothing = Just (a `join` emptyPowDelta)
     comb k Nothing (Just b) = Just (emptyPowDelta `join` b)



emptyFrameString :: F
emptyFrameString = AbsFrameString emptyPowDelta Map.empty

downFor :: Frame -> F
downFor frame = 
 (AbsFrameString emptyPowDelta (Map.singleton frame downPowDelta))



findFrames :: F -> [Frame]
findFrames (AbsFrameString def m) 
 | def == emptyPowDelta =
   map fst (filter (\ (k,pd) -> not (pd == emptyPowDelta)) (Map.toList m))



class Ageable a where
 age :: Log -> a -> F


-- age :: Log -> (Val ext) -> F

instance Ageable a => Ageable (Set.Set a) where
 age log d = joins (map (age log) (Set.toList d))

instance Ageable (Val ext) where
 age log (ValClo lam be birthdate) =
  case Map.lookup birthdate log of
   Just delta_p -> delta_p
   Nothing -> error $ "Could not find " ++ (show (lam,birthdate)) ++ "in log."
 age log (ValPrim (CPS.Prim CPS.HALT _)) = 
  case Map.lookup [] log of
   Just delta_p -> delta_p

 age log valerror = error $ "Unhandled case in age: " ++ (show valerror)




 -- Validation & debugging.

hasLogEntryForCurrentTime state = 
 let
   log = heapLog $ stateHeap $ state
   time = stateTime $ state
  in
   case Map.lookup time log of 
    Just _ -> True
    Nothing -> False
     
