module Compiler.HOIL where

 -- A library for a generic Higher-Order Intermediate Language.



 -- UNRESOLVED ISSUES:



 -- LESSONS LEARNED:

 {- Allow mixed DS/ANF/CPS. -}

 -- More specifically, don't use CPS-converted primops: instead, allow
 -- let-forms inside CPS.

 {- Don't let primops take handler arguments. -}

 -- Primops should be *primitive* operations.  Code that raises
 -- exceptions for primops should be made explicit.

 -- For example:
 {- (+ x y) => (let-values ((sum overflow) (+/overflow x y)) 
                 (if overflow
                     (raise `(+-overflow ,x ,y ,sum))
                     sum))                                    -}




 -- TODO:

 -- eta-expand non-procedure-position primops, e.g.,  
 -- ( ... * ... ) => ( ... (lambda (a b) (* a b)) ... )

 -- depolymorphize: Create a fresh copy of each letrec-bound function
 -- for each type/use.

 -- * Write nestedClosureConvert.

 -- * Mark user/cont during CPS conversion.

 -- * Rewrite CPS conversion to handle multiple return values.

 -- * Redo let to handle multiple return values.

 -- * Tag call/cc with exceptional in exception-passing transform.

 -- * Sort out app*, mk*, sug*: when app? when mk? when sug?

 -- * Drop handlers from PrimOps.

 -- * Factor out primops which return multiple values, e.g.: + and +/overflow

 -- * Create special fields for cont, exn & env in Lam & App.

 -- * Synchronize pre/postcondition comments with actual code.



import qualified Control.Monad.State as St

import qualified Data.Map as Map
import qualified Data.Set as Set

import qualified Compiler.Symbol as Symbol

import qualified Might

import qualified Lattice.Lattice as Lat





 {- The direct-style lambda-calculus AST. -}

data DSExp ext = 

   -- Atomic expressions.
   Nil { ann :: (Annotation ext) }

 | Bool { bool :: Bool ,
	  ann :: (Annotation ext) }

 | Symbol { symbol :: Symbol.Symbol ,
	    ann :: (Annotation ext) }

 | Int { int :: Int, 
	 ann :: (Annotation ext) }

 | Var { symbol :: !Symbol.Symbol ,
	 ann :: (Annotation ext) }

 | Prim { primop :: !PrimOp ,
	  ann :: (Annotation ext) }

 | Ret { vals :: ![DSExp ext],
	 ann :: (Annotation ext) }

 | Lam { keywordFormals :: ![(Symbol.Symbol,
			      DSExp ext,
			      (Annotation ext))] ,
--	envReg :: !(Maybe (DSExp ext)) ,
	 formals :: ![DSExp ext] ,
--	 rest :: !(Maybe (DSExp ext)) ,
	 body :: !(DSExp ext),
	 ann :: (Annotation ext) }

 | CallCC { argTakesExn :: !Bool ,
	   ann :: (Annotation ext) }


   -- Call expressions.
 | Seq { firstExp :: !(DSExp ext) ,
	 secondExp :: !(DSExp ext) ,
	 ann :: (Annotation ext) }


 | Set { varexp :: !(DSExp ext,
		     DSExp ext,
		     (Annotation ext)) ,
	 body :: !(DSExp ext) ,
	 ann :: (Annotation ext) }

 | Let { varexp :: !(DSExp ext,
		      DSExp ext,
		      (Annotation ext)),
	 body :: !(DSExp ext),
	 ann :: (Annotation ext) }

 | LetRec { varexps :: ![(DSExp ext,
			  DSExp ext,
			  (Annotation ext))],
	    body :: !(DSExp ext),
	    ann :: (Annotation ext) }

 | If { cond :: !(DSExp ext) ,
	consequent :: !(DSExp ext) ,
	alternate :: !(DSExp ext) , 
	ifStyle :: CallingStyle ,
	ann :: (Annotation ext) }

 | App { f :: !(DSExp ext),
--	 envExp :: !(Maybe (DSExp ext)) ,
	 keywordArgs :: ![(Symbol.Symbol, DSExp ext, (Annotation ext))] ,
	 args :: ![DSExp ext] ,
	 ann :: (Annotation ext) }

 | Try { try :: !(DSExp ext) ,
	 handler :: !(DSExp ext) ,
	 ann :: (Annotation ext) }

 | Raise { exception :: !(DSExp ext) ,
	   ann :: (Annotation ext) }





 {- Calling styles. -}

data CallingStyle = Direct | ContinuationPassing
 deriving (Eq,Ord)






 {- Primitve operations. -}

 -- checkType lists the arguments (0-indexed) which require a run-time
 -- type check.


 -- TODO: Get rid of the handler-taking.
data PrimOp = 
  -- Comparison.
   EqNum { 
	   checkType :: ![Int], 
	   takesHandler :: !Bool }
 | LessThan { 
	      checkType :: ![Int],
	    takesHandler :: !Bool}
 | LessThanEq { 
		checkType :: ![Int],
	      takesHandler :: !Bool }

  -- Halt
 | Halt { takesHandler :: !Bool }

  -- Tuples
 | Tuple { 
	   takesHandler :: !Bool }
 | Select { selectIndex :: !Int ,
	    checkType :: ![Int] ,
	    takesHandler :: !Bool }

  -- Arithmetic.
 | Sum { 
	 checkType :: ![Int],
       takesHandler :: !Bool }
 | Difference { 
		checkType :: ![Int],
	      takesHandler :: !Bool }
 | Product { 
	     checkType :: ![Int],
	     takesHandler :: !Bool }

  -- Lists.
 | Cons { 
	takesHandler :: !Bool }
 | Car { 
	 checkType :: ![Int], 
       takesHandler :: !Bool }
 | Cdr { 
	 checkType :: ![Int],
       takesHandler :: !Bool }

  -- Side effects.
 | CellNew { 
	     takesHandler :: !Bool }
 | CellSet { 
	     checkType :: ![Int],
	   takesHandler :: !Bool }
 | CellContents { 
		  checkType :: ![Int],
		takesHandler :: !Bool }

 deriving (Eq,Ord)





 {- AST Printing. -}

instance Show (CallingStyle) where
 show ContinuationPassing = "'cps"
 show Direct = "'ds"

showExn False = "'no-take-hnd"
showExn True = "'take-hnd"


instance Show (PrimOp) where
 show (EqNum checks hnd) = 
  "(prim-= " ++ 
   "'ds" ++
   " '(" ++ (Might.showWithSep " " checks) ++ ") " ++
   (showExn hnd) ++ 
  ")"

 show (LessThan checks hnd) =
  "(prim-< " ++ 
   "'ds" ++
   " '(" ++ (Might.showWithSep " " checks) ++ ") " ++
   (showExn hnd) ++ 
  ")"

 show (LessThanEq checks hnd) = 
  "(prim-<= " ++ 
   "'ds" ++
   " '(" ++ (Might.showWithSep " " checks) ++ ") " ++
   (showExn hnd) ++ 
  ")"

 show (Sum checks hnd) =
  "(prim-+ " ++ 
   "'ds" ++
   " '(" ++ (Might.showWithSep " " checks) ++ ") " ++
   (showExn hnd) ++ 
  ")"

 show (Difference checks hnd) =
  "(prim-- " ++ 
   "'ds" ++
   " '(" ++ (Might.showWithSep " " checks) ++ ") " ++
   (showExn hnd) ++ 
  ")"  

 show (Product checks hnd) = 
  "(prim-* " ++ 
   "'ds" ++
   " '(" ++ (Might.showWithSep " " checks) ++ ") " ++
   (showExn hnd) ++ 
  ")"  


 show (Tuple hnd) = 
  "(prim-tuple " ++ 
   "'ds" ++
   " " ++ 
   (showExn hnd) ++ 
  ")"  


 show (Select index checks hnd) = 
  "(prim-select " ++ 
   (show index) ++
   " " ++
   "'ds" ++
   " '(" ++ (Might.showWithSep " " checks) ++ ") " ++
   (showExn hnd) ++ 
  ")"  

 show (Halt{}) = "halt"


  -- TODO: Properly print these primops.
 show (Cons{}) = error "cons"
 show (Car{}) = error "car"
 show (Cdr{}) = error "cdr" 
 show (CellNew{}) = "cell-new"
 show (CellSet{}) = "cell-set!"
 show (CellContents{}) = "cell-contents"


showKeywords [] = ""
showKeywords [(s,e,a)] = (show s) ++ ": " ++ (show e)
showKeywords (hd:tl) =
 (showKeywords [hd]) ++ " " ++ (showKeywords tl)

instance Show (DSExp ann) where
 show (Nil{}) = "'()"
 show (Bool{bool=True}) = "#t"
 show (Bool{bool=False}) = "#f"
 show (Int{int=i}) = show i
 show (Var{symbol=s}) = show s
 show (Symbol{symbol=s}) = "'" ++ (show s)
 show (Set{varexp=(v,e,a)}) =
  "(set! " ++ (show v) ++ " " ++ (show e) ++ ")\n"
 show (Seq{firstExp=e1,secondExp=e2}) =
  "(begin " ++ (show e1) ++ "\n " ++ (show e2) ++ ")"
 show (Lam [] formals body annt) = 
  "(lambda (" ++ (Might.showWithSep " " formals) ++ ")\n "
    ++ (show body) ++ ")"
 show (Lam keywords formals body annt) = 
  "(lambda (" ++ (Might.showWithSep " " formals) ++ " " ++
                 (showKeywords keywords) ++ ")\n "
    ++ (show body) ++ ")"

--  show (Let (Var sym (ExpAnn{mutated=Just b}),e,a) body annt) =
--   "(let ((" ++ (show sym) ++ "{" ++ (show b) ++ "} " ++ (show e) ++ "))\n " ++ (show body) ++ ")"

 show (Let (v,e,a) body ann) = 
  "(let ((" ++ (show v) ++ " " ++ (show e) ++ "))\n " ++ (show body) ++ ")"
 show (LetRec varexps body ann) =
  "(letrec (" ++ (showVarExps varexps) ++ ")\n " ++ (show body) ++ ")"
 show (If cond cons alt Direct ann) =
  "(if " ++ (show cond) ++ " " ++ (show cons) ++ " " ++ (show alt) ++ ")"
 show (If cond cons alt ContinuationPassing ann) =
  "(if-cps " ++ (show cond) ++ " " ++ (show cons) ++ " " ++ (show alt) ++ ")"
 show (App f [] [] ann) = 
  "(" ++ (show f) ++ ")"
 show (App f [] args ann) = 
  "(" ++ (show f) ++ " " ++ (Might.showWithSep " " args) ++ ")"
 show (App f keywords args ann) = 
  "(" ++ (show f) ++ " " ++ (Might.showWithSep " " args) 
      ++ " " ++ (showKeywords keywords)
      ++ ")"
--  show (App f (Just env) args annt) =
--   "(" ++ (show f) ++ " " ++ (Might.showWithSep " " (env:args)) ++ ")"
 show (Prim{primop=p}) = (show p)
 show (Try exp handler ann) =
  "(try " ++ (show exp) ++ " " ++ (show handler) ++ ")"
 show (Raise exn ann) =
  "(raise " ++ (show exn) ++ ")"
 show (CallCC False ann) =
  "call/cc" 
 show (CallCC True ann) = 
  "call/cc+exn"


showVarExps [] = ""
showVarExps [(v,e,a)] = "(" ++ (show v) ++ " " ++ (show e) ++ ")"
showVarExps ((v,e,a):tl) = (showVarExps [(v,e,a)]) ++ "\n" ++ (showVarExps tl)





 {- Syntactic predicates. -}


 -- A trivial expression is one which:

 -- 1. Always returns to the current continuation. 
 -- 2. Always terminates.
 -- 3. Does not capture or invoke a continuation upon evaluation.

isTrivExp (Nil{}) = True
isTrivExp (Int{}) = True
isTrivExp (Symbol{}) = True
isTrivExp (Var{}) = True
isTrivExp (Lam{}) = True
isTrivExp (Prim{}) = True

 -- TODO: Why is this false?
isTrivExp (CallCC{}) = False

isTrivExp (App (Prim{}) envReg args annt) = 
 all isTrivExp args

isTrivExp (App f envReg args annt) = 
 False

isTrivExp (Let{}) = False
isTrivExp (LetRec{}) = False
isTrivExp (Try{}) = False
isTrivExp (Raise{}) = False

isTrivExp (Seq{}) = False

isTrivExp unhandled = 
 error $ "unhandled in isTrivExp: " ++ (show unhandled)





 {- Abbreviated HOIL AST constructors. -}

 -- TODO: Have these self-annotate as much as possible.

(==>) a b = (a,b,nullAnn)

genVar str = 
 do
 sym <- Symbol.genSym str
 return $! mkVar sym

mkVar sym =
 Var sym (nullAnn{freeVars=Just $! Map.singleton sym 1})

mkLam [] formals body = 
 case (freeVars (ann body)) of
  Just _ -> shallowAnnotateFreeVars $! 
   Lam [] formals body nullAnn
  Nothing -> 
   Lam [] formals body nullAnn

mkApp f keywords args = App f keywords args nullAnn

mkClo lam env = 
 case ((freeVars (ann lam)),(freeVars (ann env))) of
  (Just free1,Just free2) ->
   sugClo lam env (nullAnn{freeVars=Just $! mergeFreeMaps free1 free2})
  _ -> sugClo lam env nullAnn

mkPrim op = shallowAnnotateFreeVars $! Prim op nullAnn

mkLet [] body = body

mkLet ((var,exp):tl) body = Let (var,exp,nullAnn) (mkLet tl body) nullAnn

mkIf cond cons alt = If cond cons alt Direct nullAnn

mkTuple exps = 
 appTuple exps nullAnn

mkSelect exp i =
 App (mkPrim (Select{selectIndex=i,
		     checkType=[0],
		     takesHandler=False}))
     []
     [exp]
     (nullAnn{freeVars=freeVars (ann exp)})



appCellContents exp annt =
 App (mkPrim (CellContents{
			   checkType=[0],
			   takesHandler=False}))
     []
     [exp] 
     annt

appCellSet var exp annt =
 App (mkPrim (CellSet{
		      checkType=[0],
		      takesHandler=False}))
     []
     [var,exp]
     annt

appCellNew exp annt =
 App (mkPrim (CellNew{
		      takesHandler=False}))
     []
     [exp]
     annt



primHalt :: DSExp ann
primHalt = Prim (Halt{
		      takesHandler=False}) nullAnn


appCallCC argTakesExn exp annt =
 App (CallCC argTakesExn nullAnn) [] [exp] annt
 

appTuple exps annt =
 App (mkPrim (Tuple{
		  takesHandler=False})) 
     []
     exps  
     (annt{freeVars=
      foldl (\ free exp ->
       case (free,freeVars (ann exp)) of
        (Just free1,Just free2) -> Just (mergeFreeMaps free1 free2)
        _ -> Nothing)
       (Just Map.empty)
       exps
      })

appSelect exp i annt =
 App (mkPrim (Select{selectIndex=i,
		   takesHandler=False,
		   checkType=[0]}))
      []
      [exp]
      (annt{freeVars=freeVars (ann exp)})


 -- TODO: Annotate the following appropriately.
sugClo lam env annt =
 appTuple [lam,env] annt

sugAppClo f keywords args annt =
 mkApp
  (mkSelect f 0) -- The function.
  ((Symbol.env',(mkSelect f 1),nullAnn):keywords) -- The environment.
  args



mkIsExceptional exp = 
 mkApp (mkPrim (EqNum{checkType=[],takesHandler=False})) 
 []
 [exnCode,mkSelect exp 0]



varsBoundByKeywords [] = []
varsBoundByKeywords ((_,v,_):tl) = v:(varsBoundByKeywords tl)

-- varsBoundByOptions [] = []
-- varsBoundByOptions ((Just v):tl) = v:(varsBoundByOptions tl)
-- varsBoundByOptions (Nothing:tl) = varsBoundByOptions tl


-- argsSuppliedByOptions [] = []
-- argsSuppliedByOptions ((Just exp):tl) = exp:(argsSuppliedByOptions tl)
-- argsSuppliedByOptions (Nothing:tl) = argsSuppliedByOptions tl


argsSuppliedByKeywords [] = []
argsSuppliedByKeywords ((_,exp,_):tl) = exp:(argsSuppliedByKeywords tl)



 -- BUG: O(n^2)
boundBy (Lam keywords formals body annt) = 
 Set.fromList $! map symbol ((varsBoundByKeywords keywords) ++ formals)
boundBy (LetRec varexps body annt) =
 Set.fromList $! map (\ (v,e,a) -> (symbol v)) varexps
boundBy (Let (v,e,a) body annt) = 
 Set.singleton (symbol v)

varsBoundBy (Lam keywords formals body annt) = 
 (varsBoundByKeywords keywords)  ++ formals
varsBoundBy (LetRec varexps body annt) = 
 map (\ (v,e,a) -> v) varexps
varsBoundBy (Let (v,e,a) body annt) = 
 [v]

argsSuppliedBy (App f keywords args annt) =
 (argsSuppliedByKeywords keywords) ++ args


 -- These codes are placed at the front of some transform-produced
 -- tuples to distinguish the type of the tuple.
exnCode = Int 0 nullAnn
normCode = Int 1 nullAnn

 -- For undefined values:
undefinedSymbol = Symbol Symbol.undefined' nullAnn







 {- Annotations. -}

 -- Each AST node has an annotation.

data Annotation ext =
 ExpAnn { 
   -- A globally unique label:
  lab :: !Label, 

   -- The free variables for this exp:
  freeVars :: !FreeVars , 

   -- Variables bound above this exp:
  boundVars :: !BoundVars , 

   -- (Var only) Is this var mutated via set!?
  mutated :: !(Maybe Bool) , 

   -- (Lam only) Which variables are local to this lambda?
   -- This includes Let-bound variables within it.
  localVars :: !(LocalVars) ,

   -- Arbitrary external annotation:
  ext :: Maybe ext }

 




 -- A FreeMap counts free occurrences.
type FreeMap = Map.Map Symbol.Symbol Int

type FreeVars = Maybe FreeMap

type LocalVars = Maybe (Set.Set Symbol.Symbol)

type BoundMap = Map.Map Symbol.Symbol VarInfo

type BoundVars = Maybe BoundMap

data BindingType = LetRecBound
		 | LetBound 
		 | LambdaBound

data VarInfo = VarInfo { scopeDepth :: !Int, 
			 bindingType :: !BindingType }
	     

bumpBindingDepth boundMap =
 Map.map (\ (VarInfo d bt) -> VarInfo (d+1) bt) boundMap


mergeFreeVars (Just fvmap1) (Just fvmap2) 
 = Just $! Map.unionWith (+) fvmap1 fvmap2
mergeFreeVars Nothing Nothing = error $ "No maps supplied to mergeFreeVars."

mergeFreeMaps fvmap1 fvmap2
 = Map.unionWith (+) fvmap1 fvmap2

extractFreeVars ann = freeVars ann
injectFreeVars ann fvs = ann { freeVars = fvs }
invalidatedFreeVars ann = ann { freeVars = Nothing }



 -- Unique labels.
type Label = Maybe (Symbol.Symbol)


 -- Default annotations.
nullAnn :: Annotation ext
nullAnn =
 ExpAnn
 Nothing
 Nothing
 Nothing
 Nothing
 Nothing
 Nothing







 {- Annotation verifiers. -}

hasFreeVars (ExpAnn{freeVars=Just _}) = True
hasFreeVars (ExpAnn{freeVars=Nothing}) = False

hasFreeVarsAnnotation (Nil ann) = 
 hasFreeVars ann

hasFreeVarsAnnotation (Bool bool ann) =
 hasFreeVars ann

hasFreeVarsAnnotation (Int int ann) =
 hasFreeVars ann

hasFreeVarsAnnotation (Symbol sym ann) =
 hasFreeVars ann

hasFreeVarsAnnotation (Prim {ann=annt}) =
 hasFreeVars annt

hasFreeVarsAnnotation (Var sym ann) =
 hasFreeVars ann

hasFreeVarsAnnotation (Lam keywords formals body ann) =
 hasFreeVars ann &&
 hasFreeVarsAnnotation body 

hasFreeVarsAnnotation (Let (v,e,a) body ann) =
 hasFreeVars ann &&
 hasFreeVarsAnnotation e &&
 hasFreeVarsAnnotation body

hasFreeVarsAnnotation (LetRec varexps body ann) =
 hasFreeVars ann &&
 hasFreeVarsAnnotation body &&
 all (\ (v,e,a) -> hasFreeVarsAnnotation e) varexps

hasFreeVarsAnnotation exp@(App f envExp args ann) =
 hasFreeVars ann &&
 hasFreeVarsAnnotation f &&
-- hasFreeVarsAnnotation envExp && 
 all hasFreeVarsAnnotation (argsSuppliedBy exp)

hasFreeVarsAnnotation (If cond cons alt style ann) = 
 hasFreeVars ann && 
 hasFreeVarsAnnotation cond &&
 hasFreeVarsAnnotation cons &&
 hasFreeVarsAnnotation alt

hasFreeVarsAnnotation (Try exp handler ann) =
 hasFreeVars ann && 
 hasFreeVarsAnnotation exp &&
 hasFreeVarsAnnotation handler

hasFreeVarsAnnotation (Raise exn ann) =
 hasFreeVars ann &&
 hasFreeVarsAnnotation exn

hasFreeVarsAnnotation (CallCC takesExn ann) = 
 hasFreeVars ann 

hasFreeVarsAnnotation (Set (v,e,a) body annt) = 
 hasFreeVars annt &&
 hasFreeVarsAnnotation body &&
 hasFreeVarsAnnotation e

hasFreeVarsAnnotation (Seq first second annt) =
 hasFreeVars annt &&
 hasFreeVarsAnnotation first &&
 hasFreeVarsAnnotation second


-- hasFreeVarsAnnotation unhandled = error $ "unhandled in hasFreeVarsAnnotation: " ++ (show unhandled)




 {- Annotate bound variables. -}

annotateBoundVars boundMap exp@(Nil{}) =
 exp{ann=(ann exp){boundVars=Just boundMap}}

annotateBoundVars boundMap exp@(Int{}) =
 exp{ann=(ann exp){boundVars=Just boundMap}}

annotateBoundVars boundMap exp@(Bool{}) =
 exp{ann=(ann exp){boundVars=Just boundMap}}

annotateBoundVars boundMap exp@(Var{}) =
 exp{ann=(ann exp){boundVars=Just boundMap}}

annotateBoundVars boundMap exp@(Prim{}) =
 exp{ann=(ann exp){boundVars=Just boundMap}}

annotateBoundVars boundMap exp@(Symbol{}) =
 exp{ann=(ann exp){boundVars=Just boundMap}}

annotateBoundVars boundMap exp@(Lam keywords formals body annt) = 
 let
  boundHere = boundBy exp 

  boundMap' = foldl
                (\ bmap sym -> 
                      Map.insert sym (VarInfo 0 LambdaBound) bmap)
                (bumpBindingDepth boundMap)
                (Set.toList boundHere)

  body' = annotateBoundVars boundMap' body
 in Lam keywords formals body' (annt{boundVars=Just boundMap})

annotateBoundVars boundMap (Let (v,e,a) body annt) =
 let
  e' = annotateBoundVars boundMap e 
  boundMap' = 
   Map.insert (symbol v) (VarInfo 0 LetBound) (bumpBindingDepth boundMap) 
  body' = annotateBoundVars boundMap' body
 in Let (v,e',a) body' (annt{boundVars=Just boundMap})

annotateBoundVars boundMap (LetRec varexps body annt) =
 let
  boundMap' = foldl (\ bmap (v,e,a) ->
    Map.insert (symbol v) (VarInfo 0 LetRecBound) bmap)
    (bumpBindingDepth boundMap) varexps

  varexps' = map (\ (v,e,a) -> (v,annotateBoundVars boundMap' e,a)) varexps

  body' = annotateBoundVars boundMap' body

 in LetRec varexps' body' (annt{boundVars=Just boundMap})

annotateBoundVars boundMap (App f keywords args annt) = 
 let
  f' = annotateBoundVars boundMap f
  args' = map (annotateBoundVars boundMap) args
  keywords' = map (\ (s,e,a) -> (s,annotateBoundVars boundMap e,a)) keywords
--   envExp' = 
--    case envExp of
--     Just env -> Just $! annotateBoundVars boundMap env
--     Nothing -> envExp
 in App f' keywords' args' (annt{boundVars=Just boundMap})

annotateBoundVars boundMap (If cond cons alt style annt) =
 let
  cond' = annotateBoundVars boundMap cond
  cons' = annotateBoundVars boundMap cons
  alt' = annotateBoundVars boundMap alt
 in If cond' cons' alt' style (annt{boundVars=Just boundMap})

annotateBoundVars boundMap (Seq first second annt) =
 let
  first' = annotateBoundVars boundMap first
  second' = annotateBoundVars boundMap second   
 in Seq first' second' annt


annotateBoundVars boundMap unhandled =
 error $ "unhandled in annotateBoundVars: " ++ (show unhandled) 




 {- Annotate free variables. -}

annotateFreeVars :: (DSExp ann) -> (DSExp ann)

annotateFreeVars (Nil ann) =
 Nil $! injectFreeVars ann (Just Map.empty)

annotateFreeVars (Int int ann) =
 Int int (injectFreeVars ann (Just Map.empty))

annotateFreeVars (Bool bool ann) =
 Bool bool (injectFreeVars ann (Just Map.empty))

annotateFreeVars (Var sym ann) =
 Var sym (injectFreeVars ann (Just (Map.singleton sym 1)))

annotateFreeVars (Symbol sym annt) =
 Symbol sym (injectFreeVars annt (Just Map.empty))

annotateFreeVars exp@(CallCC takesExn annt) =
 shallowAnnotateFreeVars $! exp

annotateFreeVars exp@(Prim {ann=annt}) =
 shallowAnnotateFreeVars $! exp

annotateFreeVars (If cond cons alt style annt) =
 let
  cond' = annotateFreeVars cond
  cons' = annotateFreeVars cons
  alt' = annotateFreeVars alt
  Just freeInCond' = freeVars (ann cond')
  Just freeInCons' = freeVars (ann cons')
  Just freeInAlt' = freeVars (ann alt')
  annt' = annt { freeVars = 
   Just $!
   mergeFreeMaps
    freeInCond'
    (mergeFreeMaps freeInCons'
                   freeInAlt')}
 in If cond' cons' alt' style annt'

annotateFreeVars (Lam keywords formals body annt) =
 let
  body' = annotateFreeVars body
 in shallowAnnotateFreeVars $! Lam keywords formals body' annt
                        
annotateFreeVars (Let (v,e,a) body annt) =
 let
  body' = annotateFreeVars body
  e' = annotateFreeVars e

--   boundSym = symbol v

--   Just freeInBody = freeVars (ann body')
--   Just freeInExp = freeVars (ann e')

--   freeInEither = mergeFreeMaps (Map.delete boundSym freeInBody) freeInExp

--   annt' = injectFreeVars annt (Just freeInEither)
 in shallowAnnotateFreeVars $! Let (v,e',a) body' annt


annotateFreeVars (LetRec varexps body annt) =
 let
  vars = map (\ (v,e,a) -> v) varexps
  varexps' = map (\ (v,e,a) -> (v,annotateFreeVars e,a)) varexps
  body' = annotateFreeVars body
  Just freeInBody = extractFreeVars (ann body')
  (bound,freeInExps) = foldl 
    (\ (boundMap,freeMap) (v,e,a) -> 
     let Just freeInExp = extractFreeVars $! ann e
     in (Map.insert (symbol v) 1 boundMap,
         mergeFreeMaps freeMap freeInExp))
    (Map.empty,Map.empty)
    varexps'

  freeInAnything = mergeFreeMaps freeInExps freeInBody

  annt' = injectFreeVars annt (Just $! freeInAnything `Map.difference` bound)
 in LetRec varexps' body' annt'


annotateFreeVars (App f keywords args annt) = 
 let
  f' = annotateFreeVars f
  args' = map annotateFreeVars args
  keywords' = map (\ (s,e,a) -> (s,annotateFreeVars e,a)) keywords
 in shallowAnnotateFreeVars $! App f' keywords' args' annt

annotateFreeVars (Try exp handler annt) =
 let
  exp' = annotateFreeVars exp
  handler' = annotateFreeVars handler
  annt' = annt {freeVars=mergeFreeVars (freeVars (ann exp')) 
                                       (freeVars (ann handler')) }
 in Try exp' handler' annt'

annotateFreeVars (Raise exn annt) =
 let
  exn' = annotateFreeVars exn
  annt' = annt { freeVars = freeVars (ann exn') }
 in Raise exn' annt'

annotateFreeVars (Set (v,e,a) body annt) =
 let
  e' = annotateFreeVars e
  body' = annotateFreeVars body
 in shallowAnnotateFreeVars $! Set (v,e',a) body' annt

annotateFreeVars (Seq first second annt) =
 let
  first' = annotateFreeVars first
  second' = annotateFreeVars second
 in shallowAnnotateFreeVars $! Seq first' second' annt

annotateFreeVars unhandled = error $ "unhandled in annotateFreeVars: " ++ (show unhandled)









 -- Shallow annotate free variables.
 -- Use for for recomputing free variables during modifications.
shallowAnnotateFreeVars :: (DSExp ann) -> (DSExp ann)

shallowAnnotateFreeVars (Nil ann) =
 Nil $! injectFreeVars ann (Just Map.empty)

shallowAnnotateFreeVars (Int int ann) =
 Int int (injectFreeVars ann (Just Map.empty))

shallowAnnotateFreeVars (Bool bool ann) =
 Bool bool (injectFreeVars ann (Just Map.empty))

shallowAnnotateFreeVars (Var sym ann) =
 Var sym (injectFreeVars ann (Just (Map.singleton sym 1)))

shallowAnnotateFreeVars (If cond cons alt style annt) =
 let
  Just freeInCond' = freeVars (ann cond)
  Just freeInCons' = freeVars (ann cons)
  Just freeInAlt' = freeVars (ann alt)
  annt' = annt { freeVars = 
   Just $!
   mergeFreeMaps
    freeInCond'
    (mergeFreeMaps freeInCons'
                   freeInAlt')}
 in If cond cons alt style annt'

shallowAnnotateFreeVars exp@(Lam keywords formals body annt) =
 let
  boundHere = boundBy exp

  boundSyms' = Map.fromList $! [ (sym,1) | sym <- Set.toList boundHere ]

  freeInBody = case extractFreeVars (ann body) of
                Just free -> free
                Nothing -> error $! "body not annotated: " ++ (show body)
  annt' = injectFreeVars annt (Just (freeInBody `Map.difference` boundSyms'))
 in Lam keywords formals body annt'
                        
shallowAnnotateFreeVars (Let (v,e,a) body annt) =
 let
  boundSym = symbol v

  Just freeInBody = freeVars (ann body)
  freeInExp = case freeVars (ann e) of
               Just free -> free
               Nothing -> error $! "let-exp not annotated: " ++ (show e) ++ "\nin: " ++ (show $ Let (v,e,a) body annt)

  freeInEither = mergeFreeMaps (Map.delete boundSym freeInBody) freeInExp

  annt' = injectFreeVars annt (Just freeInEither)
 in Let (v,e,a) body annt'


shallowAnnotateFreeVars (LetRec varexps body annt) =
 let
  vars = map (\ (v,e,a) -> v) varexps
  Just freeInBody = extractFreeVars (ann body)
  (bound,freeInExps) = foldl 
    (\ (boundMap,freeMap) (v,e,a) -> 
     let Just freeInExp = extractFreeVars $! ann e
     in (Map.insert (symbol v) 1 boundMap,
         mergeFreeMaps freeMap freeInExp))
    (Map.empty,Map.empty)
    varexps

  freeInAnything = mergeFreeMaps freeInExps freeInBody

  annt' = injectFreeVars annt (Just $! freeInAnything `Map.difference` bound)
 in LetRec varexps body annt'


shallowAnnotateFreeVars exp@(App f envExp args annt) = 
 let
   -- Find free variables in the procedure.
  freeInFun = case extractFreeVars (ann f) of
               Just free -> free
               Nothing -> error $! "not annotated: " ++ (show f)

   -- Find free variables in the arguments.
  freeInAny = foldl 
   (\ free arg -> 
    case (extractFreeVars (ann arg)) of
     Just freeInArg -> mergeFreeMaps free freeInArg
     Nothing -> error $! "arg not annotated: " ++ (show arg) ++ "\n" ++ "in: " ++ (show $ exp))
   (freeInFun)
   (argsSuppliedBy exp)
  annt' = injectFreeVars annt (Just $! freeInAny)
 in App f envExp args annt'

shallowAnnotateFreeVars (Try exp handler annt) =
 let
  annt' = annt { freeVars = 
   mergeFreeVars (freeVars (ann exp)) (freeVars (ann handler)) }
 in Try exp handler annt'

shallowAnnotateFreeVars (Raise exp annt) =
 let
 in Raise exp (annt {freeVars=freeVars (ann exp)})

shallowAnnotateFreeVars (CallCC takesExn annt) = 
 let
 in CallCC takesExn (annt {freeVars=Just Map.empty})

shallowAnnotateFreeVars p@(Prim {ann=annt}) =
 p { ann = annt {freeVars = Just Map.empty }}

shallowAnnotateFreeVars (Seq first second annt) =
 let
  annt' =
   annt{freeVars=mergeFreeVars (freeVars (ann first)) (freeVars (ann second))}
 in Seq first second annt'

shallowAnnotateFreeVars (Set (v,e,a) body annt) = 
 let
  Just freeInExp = freeVars (ann e)
  annt' = annt{freeVars=Just (mergeFreeMaps freeInExp (Map.singleton (symbol v) 1))}
 in Set (v,e,a) body annt'



 {- Find keywords invocations. -}



findKeywords :: 
 (DSExp ann) -> Set.Set (Symbol.Symbol)

findKeywords exp@(Nil{}) = Set.empty
findKeywords exp@(Int{}) = Set.empty
findKeywords exp@(Bool{}) = Set.empty
findKeywords exp@(Symbol{}) = Set.empty
findKeywords exp@(Prim{}) = Set.empty
findKeywords exp@(Var{}) = Set.empty

findKeywords exp@(Lam keywords formals body annt) = 
 let
  keysInBody = findKeywords body
  keysHere = keySetOf keywords 
 in keysInBody `Set.union` keysHere

findKeywords exp@(App f keywords args annt) =
 let 
  keysInFun = findKeywords f
  keysHere = keySetOf keywords
  keysInKeywords = Set.unions (map (\ (s,e,a) -> findKeywords e) keywords)
  keysInArgs = Set.unions (map findKeywords args)
 in Set.unions [keysInFun,keysHere,keysInArgs,keysInKeywords]

findKeywords exp@(If cond cons alt style annt) = 
 let
  keysInCond = findKeywords cond
  keysInCons = findKeywords cons
  keysInAlt = findKeywords alt
 in Set.unions [keysInCond,keysInCons,keysInAlt]

findKeywords exp@(LetRec varexps body annt) =
 let
  keysInVarExps = Set.unions (map (\ (v,e,a) -> findKeywords e) varexps)
  keysInBody = findKeywords body
 in keysInVarExps `Set.union` keysInBody

findKeywords exp@(Let (v,e,a) body annt) =
 let
  keysInExp = findKeywords e
  keysInBody = findKeywords body
 in keysInExp `Set.union` keysInBody

findKeywords exp@(Set (v,e,a) body annt) =
 let
  keysInExp = findKeywords e
  keysInBody = findKeywords body
 in keysInExp `Set.union` keysInBody

findKeywords unhandled = 
 error $ "unhandled in findKeywords: " ++ (show unhandled)


data KeySetInfo =
   KeysPassed !(Set.Set Symbol.Symbol)
 | KeysReceived !(Set.Set Symbol.Symbol)


keySetOf :: [(Symbol.Symbol,DSExp ext,(Annotation ext))] -> Set.Set (Symbol.Symbol)
keySetOf keywords = Set.fromList $! map (\ (s,e,a) -> s) keywords




 {- Counting the number of bindings to each variable. -}

countBindings :: (DSExp ann) -> (Map.Map Symbol.Symbol Int)

countBindings exp@(Nil{}) = Map.empty
countBindings exp@(Int{}) = Map.empty
countBindings exp@(Bool{}) = Map.empty
countBindings exp@(Symbol{}) = Map.empty
countBindings exp@(Prim{}) = Map.empty
countBindings exp@(Var{}) = Map.empty

countBindings exp@(Lam keywords formals body annt) = 
 let
  boundHere = boundBy exp 

  countLocal =
   foldl 
    (\ cnt sym -> Map.insert sym 1 cnt)
    Map.empty
    (Set.toList boundHere)

  countBody = countBindings body

 in Map.unionWith (+) countLocal countBody

countBindings exp@(Let (v,e,a) body annt) = 
 let
  countForExp = countBindings e
  countForBody = countBindings body 
  countForAll = Map.unionWith (+) countForExp countForBody
  countPlusOne = Map.unionWith (+) (Map.singleton (symbol v) 1) countForAll
 in countPlusOne

countBindings exp@(LetRec varexps body annt) = 
 let
  countBody = countBindings body

  countHere =
   foldl 
    (\ cnt (v,e,a) -> 
     Map.unionWith (+)  
      (Map.singleton (symbol v) 1)
      (Map.unionWith (+)
       (countBindings e)
       cnt))
    countBody
    varexps

 in countHere

countBindings exp@(App f envExp args annt) =
 let     
  countHere = foldl 
   (\ cnt arg -> Map.unionWith (+) (countBindings arg) cnt)
   (countBindings f)
   (argsSuppliedBy exp)
 in countHere

countBindings exp@(If cond cons alt style annt) =
 let
  in Map.unionsWith (+) (map countBindings [cond,cons,alt])

countBindings (Seq first second annt) =
 Map.union (countBindings first) (countBindings second)

countBindings unhandled =
 error $ "unhandled in countBindings: " ++ (show unhandled)







 {- Annotate mutable variables at their binding site. -}

annotateMutability :: (DSExp ann) -> (DSExp ann,Set.Set Symbol.Symbol)
annotateMutability exp@(Nil{}) =
 (exp,Set.empty)

annotateMutability exp@(Int{}) =
 (exp,Set.empty)

annotateMutability exp@(Bool{}) =
 (exp,Set.empty)

annotateMutability exp@(Symbol{}) =
 (exp,Set.empty)

annotateMutability exp@(Var{}) =
 (exp,Set.empty)

annotateMutability exp@(Prim{}) =
 (exp,Set.empty)


annotateMutability exp@(Set (v,e,a) body annt) =
 let
  (e',expMutSet) = annotateMutability e
  (body',bodyMutSet) = annotateMutability body
  mutSet' = Set.unions[expMutSet, bodyMutSet, Set.singleton (symbol v)]
 in (Set (v,e',a) body' annt,mutSet')

annotateMutability exp@(Let (v,e,a) body annt) =
 let
  (e',expMutSet) = annotateMutability e
  (body',bodyMutSet) = annotateMutability body
  mutSet' = Set.unions[expMutSet, Set.delete (symbol v) bodyMutSet] 
 in (Let (withMutability ((symbol v) `Set.member` bodyMutSet) v,e',a) body' annt,mutSet')

annotateMutability exp@(Lam keywords formals body annt) =
 let
  boundHere = boundBy exp

  (body',mutSet) = annotateMutability body

  keywords' = 
   map (\ (s,v,a) ->
    (s,withMutability (Set.member (symbol v) mutSet) v,a)) keywords

  formals' =  
   map (\ v ->
    withMutability (Set.member (symbol v) mutSet) v) formals


  mutSet' = mutSet `Set.difference` boundHere

 in (Lam keywords' formals' body' annt,mutSet')

annotateMutability exp@(LetRec varexps body annt) = 
 let
  boundHere = boundBy exp
  
  (body',bodyMutSet) = annotateMutability body 

  (reversedVarexps',mutSet) = 
   foldl 
    (\ (ves,mutSet) (v,e,a) ->
     let
      (e',mutSet') = annotateMutability e
     in ((v,e',a):ves,mutSet `Set.union` mutSet'))
    ([],bodyMutSet)
    varexps

  varexps' = reverse reversedVarexps'

  varexps'' =
   map (\ (v,e,a) -> (withMutability (Set.member (symbol v) mutSet) v,e,a))
       varexps'
 
  mutSet' = (bodyMutSet `Set.union` mutSet) `Set.difference` boundHere

 in (LetRec varexps'' body' annt,mutSet')

annotateMutability exp@(App f keywords args annt) =
 let
  (f',mutSetFun) = annotateMutability f

  (keywords',mutSetKeys) =
   foldr
    (\ (s,arg,a) (keys',mutSetKeys) ->
     let
      (arg',argMutSet) = annotateMutability arg
      mutSetKeys' = mutSetKeys `Set.union` argMutSet
     in ((s,arg',a):keys',mutSetKeys'))
    ([],Set.empty)
    keywords


  (args',mutSetArgs) =
   foldr
    (\ arg (args',mutSetArgs) ->
     let
      (arg',argMutSet) = annotateMutability arg
      mutSetArgs' = mutSetArgs `Set.union` argMutSet
     in (arg':args',mutSetArgs'))
    ([],Set.empty)
    args
  
  mutSet = Set.unions [mutSetArgs,mutSetFun,mutSetKeys]
 in (App f' keywords' args' annt,mutSet)

annotateMutability exp@(If cond cons alt style annt) =
 let
  (cond',mutSet) = annotateMutability cond
  (cons',mutSet') = annotateMutability cons
  (alt',mutSet'') = annotateMutability alt
  mutSet''' = Set.unions[mutSet,mutSet',mutSet'']
 in (If cond' cons' alt' style annt,mutSet''')

annotateMutability (Try exp handler annt) =
 let
  (exp',mutSet) = annotateMutability exp
  (handler',mutSet') = annotateMutability handler 
  mutSet'' = Set.union mutSet mutSet'
 in (Try exp' handler' annt,mutSet'')

annotateMutability (Raise exp annt) =
 let
  (exp',mutSet) = annotateMutability exp
 in (Raise exp' annt,mutSet)


annotateMutability unhandled = 
 error $ "unhandled in annotateMutability: " ++ (show unhandled)




withMutability :: Bool -> (DSExp ann) -> (DSExp ann)
withMutability b exp = 
 exp{ann=(ann exp){mutated=Just b}}




 {- Immutify variables. -}


immutifyVariables mutables exp@(Nil{}) =
 exp

immutifyVariables mutables exp@(Int{}) =
 exp

immutifyVariables mutables exp@(Bool{}) =
 exp

immutifyVariables mutables exp@(Symbol{}) =
 exp

immutifyVariables mutables exp@(Prim{}) =
 exp

immutifyVariables mutables exp@(Var sym annt) =
 if Set.member sym mutables
 then appCellContents (Var sym nullAnn) annt
 else exp

immutifyVariables mutables (Set (v,e,a) body annt) =
 let
  e' = immutifyVariables mutables e
  body' = immutifyVariables mutables body
 in Seq (appCellSet v e' nullAnn) body' annt

immutifyVariables mutables (Let (v,e,a) body annt) =
 let
  e' = immutifyVariables mutables e
 in 
   -- Is v mutable?
  case v of 
   Var _ (ExpAnn{mutated=Just True}) ->
    Let (v,appCellNew e' nullAnn,a) 
         (immutifyVariables (Set.insert (symbol v) mutables) body)
          annt
   Var _ (ExpAnn{mutated=Just False}) ->  
    Let (v,e',a) 
         (immutifyVariables mutables body)
          annt

immutifyVariables mutables exp@(Lam keywords formals body annt) = 
 let
   -- Grab the set of all variables bound here.
  boundHere = boundBy exp

   -- Create a list of the args bound here which are mutable.
  mutableArgsList = 
   filter (\ v -> case mutated (ann v) of
                    Just b -> b)
          (varsBoundBy exp)

  mutableArgs = Set.fromList (map symbol mutableArgsList)

   -- genLets rebinds all mutable args.
  genLets [] body = body
  genLets (v:tl) body = 
   mkLet [(v,appCellNew v nullAnn)]
    (genLets tl body)

   -- Compute variables which are mutable in the body.
  mutableInBody = (mutables `Set.difference` boundHere) 
                   `Set.union` 
                  (mutableArgs)

  body' = immutifyVariables mutableInBody body

 in Lam keywords formals
      (genLets mutableArgsList 
        body')
      annt

immutifyVariables mutables exp@(LetRec varexps body annt) =
 let
  boundHere = boundBy exp

  isMutableVarExp (Var sym (ExpAnn{mutated=Just b}),e,a) = b

  mutableVarExps = 
   filter isMutableVarExp varexps

  localMutables =
   (Set.fromList $! map (\ (v,e,a) -> (symbol v)) mutableVarExps)

  mutables' = 
   (mutables `Set.difference` boundHere)
    `Set.union`
   localMutables

  immutableVarExps = 
   filter (not . isMutableVarExp) varexps
    
  mutableVarExps' =
   map (\ (v,e,a) -> (v,immutifyVariables mutables' e,a))
       mutableVarExps

  immutableVarExps' = 
   map (\ (v,e,a) -> (v,immutifyVariables mutables' e,a))
       immutableVarExps

  genLets [] body = body
  genLets ((v,e,a):tl) body =
   Let (v,appCellNew undefinedSymbol nullAnn,a)
    (genLets tl body)
    nullAnn

  genSets [] body = body
  genSets ((v,e,a):tl) body =
   Seq (appCellSet v e nullAnn)
    (genSets tl body)
    nullAnn

  body' = immutifyVariables mutables' body
  
 in genLets mutableVarExps'
     (LetRec immutableVarExps' 
       (genSets mutableVarExps' body')
       nullAnn)


immutifyVariables mutables (App f keywords args annt) =
 let
  f' = immutifyVariables mutables f
--   envExp' =
--    case envExp of
--     Just env -> Just $! immutifyVariables mutables env
--     Nothing -> Nothing
  keywords' = map (\ (s,e,a) -> (s,immutifyVariables mutables e,a)) keywords
  args' = map (immutifyVariables mutables) args
 in App f' keywords' args' annt

immutifyVariables mutables (If cond cons alt style annt) =
 let
  cond' = immutifyVariables mutables cond
  cons' = immutifyVariables mutables cons
  alt' = immutifyVariables mutables alt
 in If cond' cons' alt' style annt


immutifyVariables mutables (Try exp handler annt) =
 let 
  exp' = immutifyVariables mutables exp
  handler' = immutifyVariables mutables handler
 in Try exp' handler' annt

immutifyVariables mutables (Raise exp annt) =
 let
  exp' = immutifyVariables mutables exp
 in Raise exp' annt

immutifyVariables mutables unhandled =
 error $ "unhandled in immutifyVariables: " ++ (show unhandled)



 {- Conversion to exception-passing style. -}
 
 -- Precondition: Alphatisation.

 -- Postcondition: Exceptions (try,raise) are converted to call/cc.
 --                Functions take a "current handler" argument.


 -- TODO: When unexceptionalizing Prim, add Raise's for possible problems.

unexceptionalize' exn k ret hnd (Nil ann) = Nil ann

unexceptionalize' exn k ret hnd (Int int ann) = Int int ann

unexceptionalize' exn k ret hnd (Bool bool ann) = Bool bool ann

unexceptionalize' exn k ret hnd exp@(Symbol sym ann) = exp

unexceptionalize' exn k ret hnd (Var sym ann) = Var sym ann

unexceptionalize' exn k ret hnd p@(Prim{}) = p

-- unexceptionalize' exn k ret hnd (Tuple exps ann) =
--  let
--   exps' = map (unexceptionalize' exn k ret hnd) exps
--  in Tuple exps' ann

-- unexceptionalize' exn k ret hnd (Select tpl index ann) = 
--  let
--   tpl' = unexceptionalize' exn k ret hnd tpl
--  in Select tpl' index ann

unexceptionalize' exn k ret hnd (Seq first second ann) = 
 let
  first' = unexceptionalize' exn k ret hnd first
  second' = unexceptionalize' exn k ret hnd second
 in Seq first' second' ann

 -- Constraint: The environment register must be empty; that is, we
 -- must not have closure converted.
unexceptionalize' exn k ret hnd (Lam keywords formals body annt) = 
 let
  -- TODO: Smash the handler into a keyword.
  formals' = (Var hnd nullAnn):formals  
  body' = unexceptionalize' exn k ret hnd body
 in Lam keywords formals' body' annt

unexceptionalize' exn k ret hnd (Let (v,e,a) body ann) = 
 let
  e' = unexceptionalize' exn k ret hnd e
  body' = unexceptionalize' exn k ret hnd body
 in Let (v,e',a) body' ann

unexceptionalize' exn k ret hnd (LetRec varexps body ann) = 
 let
  varexps' = map (\ (v,e,a) -> (v,unexceptionalize' exn k ret hnd e, a)) varexps
  body' = unexceptionalize' exn k ret hnd body
 in LetRec varexps' body' ann

unexceptionalize' exn k ret hnd (If cond cons alt style ann) = 
 let 
  cond' = unexceptionalize' exn k ret hnd cond
  cons' = unexceptionalize' exn k ret hnd cons
  alt' = unexceptionalize' exn k ret hnd alt  
 in If cond' cons' alt' style ann

unexceptionalize' exn k ret hnd (App p@(Prim{}) keywords args ann) =
 -- Primitives don't take environments.
 let
  args' = map (unexceptionalize' exn k ret hnd) args
 in App p keywords args' ann

unexceptionalize' exn k ret hnd (App f keywords args ann) = 
 -- Unexceptionaliztion must happen before closure conversion.
 let
  -- TODO: Add handler to keyword arg.  Annotate as "registerable."
  f' = unexceptionalize' exn k ret hnd f
  args' = map (unexceptionalize' exn k ret hnd) args
  currentHandler = Var hnd nullAnn
 in App f' keywords (currentHandler:args') ann

 -- TODO: Add support for checking.
unexceptionalize' exn k ret hnd (Set (v,e,a) body annt) =
 let
  e' = unexceptionalize' exn k ret hnd e
  body' = unexceptionalize' exn k ret hnd body
  currentHandler = Var hnd nullAnn
 in Set (v,e',a) body' annt

unexceptionalize' exn k ret hnd (Try try handler ann) =
 let
  retVar = Var ret nullAnn
  hndVar = Var hnd nullAnn
  kVar = Var k nullAnn
  exnVar = Var exn nullAnn

  try' = unexceptionalize' exn k ret hnd try
  handler' = unexceptionalize' exn k ret hnd handler  

  attempt = 
   appCallCC 
    False
    (Lam []
         [kVar] 
         (Let (hndVar,Lam [] [exnVar] 
                       (mkApp kVar [] [(mkTuple [exnCode, exnVar])])
                       nullAnn,nullAnn)
              (mkApp kVar [] [mkTuple [normCode,try']])
              nullAnn)
         nullAnn)
    nullAnn

  checkException = 
   mkIf (mkIsExceptional retVar)
        (mkApp handler' [] [hndVar,mkSelect retVar 1])
        (mkSelect retVar 1)

 in
  Let (retVar,attempt,nullAnn) 
   checkException
   nullAnn

unexceptionalize' exn k ret hnd (Raise exception ann) = 
 let 
  exception' = unexceptionalize' exn k ret hnd exception
 in App (Var hnd nullAnn) [] [exception] nullAnn


unexceptionalize' exn k ret hnd (unhandled) =
 error $ "unhandled in unexceptionalize': " ++ (show unhandled)



 -- Monadic version.

unexceptionalize exp =
 do
 kSym <- Symbol.genSym "exn-cont"
 hndSym <- Symbol.genSym "hnd"
 retSym <- Symbol.genSym "retval"
 exnSym <- Symbol.genSym "exn"
 exp' <- return $!
  -- TODO: Install a better exception handler than Nil.
  Let (Var hndSym nullAnn,Nil nullAnn,nullAnn)
   (unexceptionalize' exnSym kSym retSym hndSym exp)
   nullAnn
 return $! exp'





 -- Alphatisation.
 -- Postcondition: Any two bound variables are distinct variables.
 --
 -- Behavior: FreeVars are invalidated.

type AlphaMap = Map.Map Symbol.Symbol Symbol.Symbol

alphatise :: (Symbol.HasSymbolTable a) =>
               (DSExp ann) ->
               St.State a (DSExp ann)
alphatise = alphatise' Map.empty


alphatise' :: (Symbol.HasSymbolTable a) =>
               AlphaMap ->
               (DSExp ann) ->
               St.State a (DSExp ann)
alphatise' env (Nil ann) = 
 do
 return $! (Nil (invalidatedFreeVars ann))

alphatise' env (Int int ann) =
 do
 return $! (Int int (invalidatedFreeVars ann))

alphatise' env (Bool bool ann) =
 do
 return $! (Bool bool (invalidatedFreeVars ann))
 
alphatise' env (Var sym ann) =
 do
 return $! case Map.lookup sym env of
            Just sym' -> Var sym' (invalidatedFreeVars ann)
            Nothing -> Var sym (invalidatedFreeVars ann)

alphatise' env exp@(Lam keywords formals body ann) =
 do
 env' <- extendAlphaMap env (varsBoundBy exp)

 -- env' <- extendAlphaMap env formals
 formals' <- Might.mmap (alphatise' env') formals

--  envReg' <- alphatiseMaybe env' envReg
--  rest' <- alphatiseMaybe env' rest

 keywords' <- Might.mmap (\ (s,v,a) -> 
                    do 
                    v' <- alphatise' env' v
                    return $! (s,v',a)) keywords

--  (rest',env'') <- 
--   case rest of
--    Just v -> 
--     do newEnv <- extendAlphaMap env' [v]
--        newRestVar <- alphatise' newEnv v
--        return $! (Just newRestVar,newEnv)
--    Nothing -> return $! (rest,env')
 body' <- alphatise' env' body
 return $! Lam keywords' formals' body' (invalidatedFreeVars ann) 

alphatise' env (If cond cons alt style ann) = 
 do
 cond' <- alphatise' env cond
 cons' <- alphatise' env cons
 alt' <- alphatise' env alt
 return $! If cond' cons' alt' style (invalidatedFreeVars ann)

alphatise' env (Let (v,e,a) body ann) =
 do
 env' <- extendAlphaMap env [v]
 v' <- alphatise' env' v
 e' <- alphatise' env' e
 body' <- alphatise' env' body
 return $! Let (v',e',invalidatedFreeVars a) body' (invalidatedFreeVars ann)

alphatise' env (LetRec varexps body ann) =
 do
 vars <- return $! map (\ (v,e,a) -> v) varexps
 env' <- extendAlphaMap env vars
 
 varexps' <- Might.mmap (\ (v,e,a) ->
              do v' <- alphatise' env' v
                 e' <- alphatise' env' e
                 return $! (v',e',invalidatedFreeVars a))
              varexps
 body' <- alphatise' env' body
 return $! LetRec varexps' body' (invalidatedFreeVars ann)

alphatise' env (App f keywords args ann) = 
 do
 f' <- alphatise' env f
 args' <- Might.mmap (alphatise' env) args
-- envExp' <- alphatiseMaybe env envExp
 keywords' <- Might.mmap (\ (s,e,a) -> 
               do 
               e' <- alphatise' env e
               return $! (s,e',a))  keywords
 return $! App f' keywords' args' (invalidatedFreeVars ann)

alphatise' env (CallCC argTakesExn ann) =
 do
 return $! CallCC argTakesExn (invalidatedFreeVars ann)

alphatise' env (Try exp handler annt) =
 do
 exp' <- alphatise' env exp
 handler' <- alphatise' env handler
 return $! Try exp' handler' annt

alphatise' env (Raise exp annt) =
 do
 exp' <- alphatise' env exp
 return $! Raise exp' annt


alphatise' env p@(Prim{ann=annt}) =
 return $! p {ann = invalidatedFreeVars annt}

alphatise' env unhandled = error $ "unhandled case in alphatise': " ++ (show unhandled)



extendAlphaMap :: Symbol.HasSymbolTable a => 
                    AlphaMap ->
                    [DSExp ann] ->
                    St.State a AlphaMap
extendAlphaMap env [] = return $! env
extendAlphaMap env ((Var sym ann):tl) =
 do
 sym' <- Symbol.genSym (show sym)
 env'  <- extendAlphaMap env tl
 return $! Map.insert sym sym' env'



alphatiseMaybe :: 
 Symbol.HasSymbolTable a =>
 AlphaMap -> (Maybe (DSExp annt)) -> St.State a (Maybe (DSExp annt))
alphatiseMaybe env (Just exp) = 
 do 
 exp' <- alphatise' env exp
 return $! Just exp'

alphatiseMaybe env Nothing = 
 return $! Nothing



 {- Minimal alphatisation. -}

 -- Doesn't assign a fresh symbol unless it absolutely must; this
 -- helps to preserve readability.


data MinAlphaState = MAS { masST :: Symbol.SymbolTable ,
			   masSeen :: Set.Set Symbol.Symbol }


instance Symbol.HasSymbolTable MinAlphaState where
 extractSymbolTable (MAS st seen) = st
 injectSymbolTable (MAS st seen) st' = (MAS st' seen)


haveSeen sym =
 do 
 MAS st seen <- St.get
 return $! Set.member sym seen

markSeen sym =
 do
 MAS st seen <- St.get
 seen' <- return $! Set.insert sym seen
 St.put $! MAS st seen'


minAlphatise' :: 
               AlphaMap ->
               (DSExp ann) ->
               St.State MinAlphaState (DSExp ann)

minAlphatise' = error "undef"







 -- Flat closure conversion.
 --
 -- Precondition: Free variables must be annotated properly.
 --
 -- Postcondition: Lambda terms must not have free variables,
 --                except for letrec-bound, i.e. global, variables.
 --
 --                This makes it possible to hoist all letrecs
 --                to the top level.
 -- 
 -- Behavior: FreeVars annotations are invalidated.


type ExpMap ann = Map.Map Symbol.Symbol (DSExp ann)




flatClosureConvert ::
 (Symbol.HasSymbolTable a) =>
 (DSExp ann) ->
 St.State a (DSExp ann)

flatClosureConvert (Nil ann) =
 do
 return $! annotateFreeVars $! Nil ann

flatClosureConvert (Int int ann) = 
 do 
 return $! annotateFreeVars $! Int int ann

flatClosureConvert exp@(Symbol{}) =
 do 
 return $! annotateFreeVars $! exp

flatClosureConvert (Bool bool ann) = 
 do 
 return $! annotateFreeVars $! Bool bool ann

flatClosureConvert (Var sym ann) = 
 do
 return $! annotateFreeVars $! Var sym ann

flatClosureConvert (Prim (primop) ann) =
 error "Higher-order primitive encountered in closure conversion."

flatClosureConvert (Lam keywords formals body annt) = 
 do
 envSym <- Symbol.genSym "env"
 envVar <- return $! mkVar envSym 
 envReg <- return $! Just envVar

 Just freeSyms <- return $! extractFreeVars annt

 sortedFreeSyms <- return $! Map.toAscList freeSyms
 sortedFreeVars <- return $! map (\ (sym,_) -> mkVar sym) sortedFreeSyms

-- formals' <- return $! envVar:formals
 keywords' <- return $! (Symbol.env',envVar,nullAnn):keywords -- TODO: Annotate this as registerable.
 body' <- flatClosureConvert {- snipped env here -} body
 body'' <- return $!
  let
   rebindFree [] n body = body
   rebindFree (hd:tl) n body = 
    shallowAnnotateFreeVars $!
     Let (hd,appSelect envVar n nullAnn,nullAnn) 
      (rebindFree tl (n+1) body)
      nullAnn
  in rebindFree sortedFreeVars 0 body'
 
 closureEnv <- return $! annotateFreeVars $! appTuple sortedFreeVars nullAnn

 lam' <- return $! shallowAnnotateFreeVars $! Lam keywords' formals body'' nullAnn


 return $! shallowAnnotateFreeVars $!
  sugClo lam'
         closureEnv
         annt 

flatClosureConvert (Let (v,e,a) body annt) = 
 do
 e' <- flatClosureConvert e
 body' <- flatClosureConvert {- snipped here -} body
 annt' <- return $! annt
 Just freeInExp <- return $! freeVars $ ann $ e'
 Just freeInBody <- return $! freeVars $ ann $ body'
 freeInLet <- return $! Map.delete (symbol v) 
                        (mergeFreeMaps freeInExp freeInBody)
 return $! Let (v,e',a) body' (annt' {freeVars = Just freeInLet })

flatClosureConvert (Seq first second annt) =
 do
 first' <- flatClosureConvert first
 second' <- flatClosureConvert second
 return $! Seq first' second' annt

flatClosureConvert (LetRec varexps body annt) =
 do
 envSym <- Symbol.genSym "lrenv"
 envVar <- return $! mkVar envSym

  -- A list of the variables bound here:
 boundHere <- return $! map (\ (v,e,a) -> (symbol v)) varexps
 
  -- A FreeMap of the free variables for all clauses:
 freeInLRClausesMap <- return $! 
  foldl (\ free (v,e,a) -> 
          case freeVars (ann e) of
           Just freeMap -> mergeFreeMaps free freeMap)
        Map.empty
        varexps

-- DEBUG
--  let freeList = map (\ (v,e,a) -> (v,e,freeVars (ann e),"\n")) varexps

--  True <- error $! "freeInLR: " ++  (show freeInLRClausesMap) ++ "\n"
--   ++ "freeList: " ++ (show freeList)
-- /DEBUG

  -- A FreeMap like the prior but excluding variables bound here:
 globalFreeInLRMap <- return $! 
  foldl (\ free sym -> Map.delete sym free) freeInLRClausesMap boundHere

  -- A list version of the above:
 globalFree <- return $! 
  let
   gfree :: [Symbol.Symbol]
   gfree = map (\ (k,a) -> k) (Map.toAscList globalFreeInLRMap)
  in gfree

  -- A FreeMap for the body:

 varexps' <-
  let
   rebindFree envVar [] [] _ body = body

   rebindFree envVar (local:tl) globs 0 body =
    shallowAnnotateFreeVars $!
     mkLet [(mkVar local,mkClo (mkVar local) envVar)] 
      (rebindFree envVar tl globs 0 body)

   rebindFree envVar [] (global:tl) n body =
    shallowAnnotateFreeVars $!
     mkLet [(mkVar global,appSelect envVar n nullAnn)]
      (rebindFree envVar [] tl (n+1) body)

   convertVarExp (v,Lam keywords formals body annt,a) = 
    do
    body' <- flatClosureConvert body
    body'' <- return $! rebindFree envVar boundHere globalFree 0 body'
    keywords' <- return $! (Symbol.env',envVar,nullAnn):keywords
    return $! (v,Lam keywords' formals body'' annt,a)

   convertVarExp (v,e,a) =
    do
    e' <- flatClosureConvert e
    return $! (v,e',a) 

   convertVarExps [] = return $! []
   convertVarExps (hd:tl) = 
    do hd' <- convertVarExp hd
       tl' <- convertVarExps tl
       return $! hd':tl'
  in convertVarExps varexps


  -- Globally free Var expressions, as a list:
 globalFreeVars <- return $! map (\ sym -> mkVar sym) globalFree

 body <- flatClosureConvert body 

 body' <- 
  let
   rebindLocals [] body = body
   rebindLocals (local:tl) body = 
    shallowAnnotateFreeVars $!
     mkLet [(mkVar local,mkClo (mkVar local) envVar)]
           (rebindLocals tl body)

  in return $! rebindLocals boundHere body

 body'' <- return $!
  shallowAnnotateFreeVars $! mkLet [(envVar,mkTuple globalFreeVars)] body'
  
  -- TODO: Handle free variables here.
 return $! shallowAnnotateFreeVars $! LetRec varexps' body'' annt



 
flatClosureConvert (App p@(Prim{}) [] args annt) =
 do
 args' <- Might.mmap (flatClosureConvert) args
 p' <- return $! shallowAnnotateFreeVars p
 return $! shallowAnnotateFreeVars $! App p' [] args' annt

flatClosureConvert (App f keywords args annt) =
 do
 f' <- flatClosureConvert f
 args' <- Might.mmap (flatClosureConvert) args

 let convertKeyWord (s,e,a) =
       do
       e' <- flatClosureConvert e
       return $! (s,e',a)

 keywords' <- Might.mmap convertKeyWord keywords

 return $! shallowAnnotateFreeVars $! sugAppClo f' keywords' args' annt

flatClosureConvert (If cond cons alt style ann) =
 do 
 cond' <- flatClosureConvert cond
 cons' <- flatClosureConvert cons
 alt' <- flatClosureConvert alt
 return $! shallowAnnotateFreeVars $! If cond' cons' alt' style ann

-- CallCC not supported for closure conversion.

-- flatClosureConvert (CallCC argTakesExn annt) =
--  do
--  return $! shallowAnnotateFreeVars $! CallCC argTakesExn annt

flatClosureConvert unhandled =  
 error $ "unhandled in flatClosureConvert: " ++ (show unhandled)










 {- Nested closure conversion. -}

 -- Try to minimize environment reconstruction by nesting environments
 -- where possible.

type FreeVarDepth = [Int]

type EnvMap = Map.Map Symbol.Symbol FreeVarDepth














 {- LetRec hoisting. -}
 --
 -- Precondition: Free variable annotations.
 --
 -- Postcondition: LetRecs are hoisted as far up as possible.
 --                All LetRecs are hoisted to the top level if the
 --                source is both alphatised and closure converted.



 -- Hoist letrecs assuming an alphatised source code where only
 -- letrec-bound variables appear free in lambdas.

unsafeHoistLetRecs exp@(Nil{}) = 
 exp

unsafeHoistLetRecs (Bool bool ann) =
 Bool bool ann

unsafeHoistLetRecs exp@(Symbol{}) = 
 exp

unsafeHoistLetRecs (Int int ann) = 
 Int int ann

unsafeHoistLetRecs (Var sym ann) =
 Var sym ann

unsafeHoistLetRecs (Lam keywords formals body annt) =
 let
  body' = unsafeHoistLetRecs body
 in unsafeHoist body' (\ body'' ->
     Lam keywords formals body'' annt)

unsafeHoistLetRecs (Let (v,e,a) body annt) = 
 let
   e' = unsafeHoistLetRecs e
   body' = unsafeHoistLetRecs body
  in unsafeHoist e' (\ e'' -> 
      unsafeHoist body' (\ body'' ->
       Let (v,e'',a) body'' annt))

unsafeHoistLetRecs (Seq first second annt) =
 let 
 in unsafeHoist first (\ first' ->
     unsafeHoist second (\ second' ->
      Seq first' second' annt))

unsafeHoistLetRecs (LetRec varexps body annt) =
 let
  varexps' = map (\ (v,e,a) -> (v,unsafeHoistLetRecs e,a)) varexps
  body' = unsafeHoistLetRecs body

  (varexps'',body'') = case body' of 
   LetRec ves bod an -> (varexps' ++ ves,bod)
   exp -> (varexps',body')
    
  hoistVarExps [] = [] 
  hoistVarExps ((v,e,a):tl) = 
   case e of 
    LetRec ves bod an -> (v,bod,a):(ves ++ (hoistVarExps tl))
    exp -> (v,e,a):(hoistVarExps tl)
  
  varexps''' = hoistVarExps varexps''
 in LetRec varexps''' body'' annt


unsafeHoistLetRecs (If cond cons alt style ann) = 
 let
  cond' = unsafeHoistLetRecs cond
  cons' = unsafeHoistLetRecs cons
  alt' = unsafeHoistLetRecs alt
 in unsafeHoist cond' (\ cond'' ->
     unsafeHoist cons' (\ cons'' ->
      unsafeHoist alt' (\ alt'' ->
       If cond'' cons'' alt'' style ann)))


unsafeHoistLetRecs (App f keywords args ann) =
 let
  f' = unsafeHoistLetRecs f
  args' = map unsafeHoistLetRecs args
  keywords' = map (\ (s,e,a) -> (s,unsafeHoistLetRecs e,a)) keywords
 in unsafeHoist f' (\ f'' ->
     unsafeHoistKeywords keywords' (\ keywords'' ->
      unsafeHoistAll args' (\ args'' ->
       App f'' keywords'' args'' ann)))

unsafeHoistLetRecs exp@(CallCC takesExn ann) = 
 exp

unsafeHoistLetRecs p@(Prim{}) =
 p

unsafeHoistLetRecs unhandled = 
 error $ "unhandled in unsafeHoistLetRecs: " ++ (show unhandled)





unsafeHoist (LetRec ves bod an) k =
 LetRec ves (k bod) an
unsafeHoist exp k =
 k exp


unsafeHoistAll [] k = k []
unsafeHoistAll (hd:tl) k = 
 unsafeHoist hd (\ hd' ->
  unsafeHoistAll tl (\ tl' ->
   k (hd':tl')))

unsafeHoistKeywords [] k = k []
unsafeHoistKeywords (hd:tl) k =
 unsafeHoistKeyword hd (\ hd' ->
  unsafeHoistKeywords tl (\ tl' ->
   k (hd':tl')))

unsafeHoistKeyword (s,e,a) k = 
 unsafeHoist e (\ e' ->
  k (s,e',a))

unsafeHoistMaybe (Just e) k =
 unsafeHoist e (\ e' ->
  k $! (Just e'))
unsafeHoistMaybe (Nothing) k =
  k $! Nothing






 -- Hoist all lambdas to the top-level letrec.

 -- Precondition: Closure-converted.  LetRec-hoisted.
 --               Top-level letrec exists.

topHoistLambdas' :: 
 (Symbol.HasSymbolTable a) =>
 (DSExp ann,ExpMap ann) ->
 St.State a (DSExp ann,ExpMap ann)

topHoistLambdas' (Nil ann,glob) = 
 return $! (Nil ann,glob)

topHoistLambdas' (Int int ann,glob) = 
 return $! (Int int ann,glob)

topHoistLambdas' (Bool bool ann,glob) =
 return $! (Bool bool ann,glob)

topHoistLambdas' (exp@(Symbol{}),glob) =
 return $! (exp,glob)

topHoistLambdas' (Var sym ann,glob) =
 return $! (Var sym ann,glob)

topHoistLambdas' (p@(Prim {}),glob) = 
 return $! (p,glob)

topHoistLambdas' (lam@(Lam keywords formals body ann),glob) = 
 do
 sym' <- Symbol.genSym "lam"
 (body',glob') <- topHoistLambdas' (body,glob)
 lam' <- return $! Lam keywords formals body' ann
 glob'' <- return $! Map.insert sym' lam' glob'
 return $! (Var sym' ann,glob'')

topHoistLambdas' (exp@(CallCC{}),glob) = 
 do
 return $! (exp,glob)

topHoistLambdas' (Let (v,e,a) body ann,glob) = 
 do
 (e',glob') <- topHoistLambdas' (e,glob)
 (body',glob'') <- topHoistLambdas' (body,glob')
 return $! (Let (v,e',a) body' ann,glob'')

topHoistLambdas' (Seq first second annt,glob) =
 do
 (first',glob') <- topHoistLambdas' (first,glob)
 (second',glob'') <- topHoistLambdas' (second,glob')
 return $! (Seq first' second' annt,glob'')

topHoistLambdas' (App f keywords args ann,glob) = 
 do
 (f',glob') <- topHoistLambdas' (f,glob)
 (args',glob'') <- topHoistLambdasAll' (args,glob')
 (envExp',glob''') <- topHoistLambdasKeywords' (keywords,glob'')
 return $! (App f' envExp' args' ann,glob''')

topHoistLambdas' (If cond cons alt style ann,glob) = 
 do
 (cond',glob') <- topHoistLambdas' (cond,glob)
 (cons',glob'') <- topHoistLambdas' (cons,glob')
 (alt',glob''') <- topHoistLambdas' (alt,glob'')
 return $! (If cond' cons' alt' style ann,glob''')

topHoistLambdas' (LetRec varexps body annt,glob) = 
 do
 (varexps',glob') <- 
  let
   topHoistVarExps ([],glob) = return $! ([],glob)

   topHoistVarExps ((v,(Lam keys forms bod an),a):tl,glob) = 
    do
    (bod',g') <- topHoistLambdas' (bod,glob)
    (tl',g'') <- topHoistVarExps (tl,g')
    return $! ((v,Lam keys forms bod' an,a):tl',g'')

   topHoistVarExps ((v,e,a):tl,glob) = 
    do
    (e',g') <- topHoistLambdas' (e,glob) 
    (tl',g'') <- topHoistVarExps (tl,g')
    return $! ((v,e',a):tl',g'')

  in topHoistVarExps (varexps,glob)

 (body',glob'') <- topHoistLambdas' (body,glob')

 new_varexps <- return $! map (\ (sym,exp) -> 
                  (Var sym (ann exp),exp, ann exp)) (Map.toList glob'')
 
 varexps'' <- return $! varexps' ++ new_varexps

 return $! (LetRec varexps'' body' annt,glob'') 

topHoistLambdas' (unhandled,_) = error $ show $ "unhandled in topHoistLambdas': " ++ (show unhandled)




 -- If a top-level letrec doesn't exist, add one and then continue.
topHoistLambdas exp@(LetRec{}) = 
 do
 (exp,glob) <- topHoistLambdas' (exp,Map.empty)
 return $! exp

topHoistLambdas exp =
 do
 topHoistLambdas (LetRec [] exp (ann exp))


-- TODO: Keep annotations as updated as possible under transforms.


topHoistLambdasAll' ::
 (Symbol.HasSymbolTable a) =>
 ([DSExp ann],ExpMap ann) ->
 St.State a ([DSExp ann],ExpMap ann)

topHoistLambdasAll' ([],glob) = 
 return $! ([],glob)
topHoistLambdasAll' (hd:tl,glob) = 
 do
 (hd',glob') <- topHoistLambdas' (hd,glob)
 (tl',glob'') <- topHoistLambdasAll' (tl,glob') 
 return $! (hd':tl',glob'')
    

topHoistLambdasKeywords' ([],glob) =
 return $! ([],glob)
topHoistLambdasKeywords' ((s,e,a):tl,glob) =
 do
 (e',glob') <- topHoistLambdas' (e,glob) 
 (tl',glob'') <- topHoistLambdasKeywords' (tl,glob')
 return $! ((s,e',a):tl',glob'')
 

topHoistLambdasMaybe' :: 
 (Symbol.HasSymbolTable a) =>
 (Maybe (DSExp annt),ExpMap annt) -> 
 St.State a (Maybe (DSExp annt),ExpMap annt)
topHoistLambdasMaybe' (Just exp,glob) =
 do
 (exp',glob') <- topHoistLambdas' (exp,glob)
 return $! (Just exp',glob')
topHoistLambdasMaybe' (Nothing,glob) =
 return $! (Nothing,glob)




 -- Auxiliary functions for analyzing let(rec).

 -- Return the variables bound by a letrec.
varExpsToBoundMap :: [(DSExp ann,DSExp ann,Annotation ann)] -> FreeMap
varExpsToBoundMap varexps =
 foldl (\ bound (v,e,a) -> Map.insert (symbol v) 1 bound) Map.empty varexps
 

 -- Return the variables free in the lambda terms for a LetRec.
freeInLambdasMap :: [(DSExp ann,DSExp ann,Annotation ann)] -> FreeMap
freeInLambdasMap varexps =
 foldl (\ free (v,e,a) -> 
        case (freeVars (ann e)) of
         Just freeInE -> Map.unionWith (+) free freeInE
         Nothing -> error "Free variables not annotated.")
  Map.empty
  varexps 






 {- A-Normal Form conversion. -}

 -- Precondition: Alphatisation.

 -- Postcondition: All arguments must be atomic, ie, not applications.

 -- Behavior: Annotations are transformed like types.
 -- Behavior: FreeVars annotations, if they exist, are invalidated.
anormalize ::  (Symbol.HasSymbolTable a) => 
                (DSExp ann) -> St.State a (DSExp ann)

anormalize exp@(Nil{}) = 
 do
 return $! exp

anormalize exp@(Int{}) = 
 do
 return $! exp

anormalize exp@(Bool{}) = 
 do
 return $! exp

anormalize exp@(Var{}) = 
 do
 return $! exp

anormalize exp@(Prim{}) =
 do
 return $! exp

anormalize (Lam keywords formals body ann) = 
 do
 body' <- anormalize body 
 return $! Lam keywords formals body' ann

anormalize (Let (var,exp,annt1) body annt2) =
 do
 body' <- anormalize body
 anormalize' exp (\ var' ->
  return $! Let (var,var',annt1) body' annt2)

anormalize (LetRec varexps body annt) =
 do

 boundHere <- return $! Set.fromList $! map (\ (v,e,a) -> (symbol v)) varexps

 body' <- anormalize body

  -- Step 1: A-Normalize all of the expressions.
 varexps' <- 
  let
   anormalizeVarExps [] = return $! []
   anormalizeVarExps ((v,e,a):tl) =
    do e' <- anormalize e
       tl' <- anormalizeVarExps tl
       return $! (v,e',a):tl
  in anormalizeVarExps varexps

  -- Step 2: Lift the lets into the varexps.
  -- Clearly, we assume alphatisation.  
 varexps'' <- return $! 
  let
   liftLets [] = []

   liftLets ((v,Let (v2,e2,a2) body2 annt2,a):tl) = 
    (v2,e2,a2):(liftLets ((v,body2,a):tl))

   liftLets ((v,e,a):tl) = 
    (v,e,a):(liftLets tl)
  in liftLets varexps'


  -- Step 3: Split the varexps into two categories: lambdas & other.

  -- TODO: Find "dead" letrec'd variables, indicating statements
  -- executed purely for effect.

  -- TODO: Seperate out non-recursive expressions.

  -- Let+set the rest.

 (varexpsNonLambda,varexps''') <-  
  let
   stripNonLambda [] = ([],[])
   stripNonLambda ((ve@(v,lam@(Lam{}),a)):tl) =
    let
     (nonLam,trueRec) = stripNonLambda tl
    in (nonLam,ve:trueRec) 
   stripNonLambda (ve:tl) =
    let 
     (nonLam,trueRec) = stripNonLambda tl
    in (ve:nonLam,trueRec)

  in return $! stripNonLambda varexps''


 return $! 
  let
   reLet [] bod = bod
   reLet ((v,_,a):tl) bod = Let (v,undefinedSymbol,a) (reLet tl bod) (ann bod) 

   reSet [] bod = bod
   reSet ((v,e,a):tl) bod = 
    Set (v,e,a) 
        (reSet tl bod)
        nullAnn

  in reLet varexpsNonLambda 
      (LetRec varexps''' (reSet varexpsNonLambda body') annt)

anormalize exp@(App f keywords args ann) = 
 do
 anormalize' exp (\ v ->
  return $! v)

anormalize (If cond cons alt style annt) =
 do
 cons' <- anormalize cons
 alt' <- anormalize alt
 anormalize' cond (\ cond' ->
  return $! If cond' cons' alt' style annt)

anormalize (Try exp handler annt) = 
 do
 anormalize' exp (\ exp' ->
  do
  handler' <- anormalize handler
  return $! Try exp' handler' annt)

anormalize (Raise exp annt) =
 do 
 anormalize' exp (\ exp' ->
  return $! Raise exp' annt)
  

anormalize unhandled =
 error $ "unhandled in anormalize: " ++ (show unhandled)






anormalize' :: (Symbol.HasSymbolTable a) => 
                 (DSExp ann) -> (DSExp ann -> St.State a (DSExp ann)) ->
                 (St.State a (DSExp ann))

anormalize' exp@(Int{}) k =
 do
 k exp

anormalize' exp@(Prim{}) k =
 do
 k exp

anormalize' exp@(Var{}) k =
 do
 k exp

anormalize' (Raise exp annt) k =
 anormalize' exp (\ exp' ->
  k $! Raise exp' annt)

anormalize' (Try exp handler annt) k =
 anormalize' exp (\ exp' ->
  do 
  handler' <- anormalize handler
  k $! Try exp' handler' annt)

anormalize' exp@(Lam keywords formals body annt) k =
 do
 body' <- anormalize body
 k $! Lam keywords formals body' annt

anormalize' (App f keywords args annt) k = do
 anormalize' f (\ f' ->
  anormalizeKeywords' keywords (\ keywords' ->
   anormalizeMany' args (\ args' -> 
    do
    sym <- Symbol.genSym "val"
    var <- return $! Var sym annt
    body <- k var
    return $! Let (var,App f' keywords' args' annt,annt) body (ann body))))
   

anormalize' unhandled k =
 error $ "unhandled in anormalize':" ++ (show unhandled)



anormalizeOption' :: 
 (Symbol.HasSymbolTable a) =>
 (Maybe (DSExp annt)) ->
 ((Maybe (DSExp annt)) -> St.State a (DSExp annt)) ->
 St.State a (DSExp annt) 
anormalizeOption' (Nothing) k = k Nothing
anormalizeOption' (Just exp) k = 
 anormalize' exp (\ exp' ->
  k $! Just exp')


anormalizeMany' :: (Symbol.HasSymbolTable a) =>
                    [DSExp ann] ->
                    ([DSExp ann] -> St.State a (DSExp ann)) ->
                    (St.State a (DSExp ann))
anormalizeMany' [] k =
 k []

anormalizeMany' (exp:rest) k =
 anormalize' exp (\ exp' -> 
  anormalizeMany' rest (\ rest' ->
   k (exp':rest')))


anormalizeKeywords' [] k =
 k []
anormalizeKeywords' ((s,e,a):tl) k = 
 anormalize' e (\ e' ->
  anormalizeKeywords' tl (\ tl' ->
   k ((s,e',a):tl')))
 







 {- Continuating-passing style conversion. -}

 -- Precondition: Pre-closure conversion.  Alphatisation.

 -- Continuation conversion involves three transformation functions:

 -- `continuate' takes an expression and a Haskell-level
 -- "continuation" function to call with an expression denoting the
 -- CPS-level value of the expression.

 -- `continuateWithKont' takes an expression and a CPS-level
 -- continuation to apply to the expression's return value.

 -- `trivContinuate' takes a trivial expression and returns its
 -- CPS-level, user-world equivalent.


continuate ::
 (Symbol.HasSymbolTable a) =>
 (DSExp ann) ->
 ((DSExp ann) -> St.State a (DSExp ann)) ->
 (St.State a (DSExp ann))
  
continuate exp@(Nil{}) k =
 do
 exp' <- trivContinuate exp
 k exp'

continuate exp@(Bool{}) k =
 do
 exp' <- trivContinuate exp
 k exp'

continuate exp@(Prim{}) k =
 error $ "error in continuate: Primitives must not be directly continuated."

continuate exp@(Int{}) k =
 do
 exp' <- trivContinuate exp
 k exp'

continuate exp@(Symbol{}) k =
 do
 exp' <- trivContinuate exp
 k exp'

continuate exp@(Lam{}) k =
 do
 exp' <- trivContinuate exp
 k exp'

continuate exp@(Var{}) k =
 do
 exp' <- trivContinuate exp
 k exp'

continuate exp@(CallCC takesExn annt) k = 
 do 
 exp' <- trivContinuate exp 
 k exp'



continuate (App prim@(Prim{}) [] args annt) k =
 continuateMany args (\ args' ->
  do  
  valSym <- Symbol.genSym "trival"
  valVar <- return $! mkVar valSym
  body <- k valVar
  return $! mkLet [(valVar,App prim [] args' annt)] body)



continuate (App f keywords args annt) k = 
 do
 continuate f (\ f' ->
  continuateMany (map (\ (s,e,a) -> e) keywords) (\ keyargs' ->
   continuateMany args (\ args' ->
    do
    ansSym <- Symbol.genSym "ans"
    ansVar <- return $! Var ansSym nullAnn
    kAns <- k ansVar
    cont <- return
     $! Lam [] [ansVar] kAns nullAnn -- TODO: Annotate this.
    keywords' <- 
     return $! map (\ ((s,_,a),e) -> (s,e,a)) (zip keywords keyargs')
    return $! App f' keywords' (cont:args') annt)))


 -- TODO: Fix code bloat.   
continuate (If cond cons alt Direct annt) k =
 do
 cons' <- continuate cons k -- BUG: Exponential blow-up.
 alt' <- continuate alt k   -- 
--  cons'' <- return $! Lam Nothing [] Nothing cons' nullAnn -- TODO: Annotate.
--  alt'' <- return $! Lam Nothing [] Nothing alt' nullAnn -- TODO: Annotate.
 continuate cond (\ cond' ->
  return $! If cond' cons' alt' Direct annt)


continuate (Let (v,e,a) body annt) k =
 do
 body' <- continuate body k
 if isTrivExp e 
  then do
   e' <- trivContinuate e
   return $! Let (v,e',a) body' annt
  else do
   -- TODO: We can still cont-let-bind valExp.
   cont <- return $! Lam [] [v] body' nullAnn -- TODO: Annotate this.
   continuate e (\ e' ->
    return $! App cont [] [e'] nullAnn)


continuate (Set (v,e,a) body annt) k =
 do
 if isTrivExp e 
  then do
   body' <- continuate body k
   return $! Set (v,e,a) body' annt
  else do
   body' <- continuate body k
   continuate e (\ e' ->
    return $! Set (v,e',a) body' annt)

  
continuate (LetRec varexps body annt) k =
 do

 varexps' <-
  let
   continuateVarExps [] = return $! []
   continuateVarExps ((v,e,a):tl) =
    do
    e' <- trivContinuate e
    tl' <- continuateVarExps tl
    return $! (v,e',a):tl'  
  in continuateVarExps varexps

 body' <- continuate body k
 
 return $! LetRec varexps' body' annt


continuate exp@(Seq first second annt) k = 
 if isTrivExp first
 then 
  do
  second' <- continuate second k
  return $! Seq first second' annt
 else error $! "unimplemented Seq in continuate: " ++ (show exp)




continuate unhandled k = 
 error $ show $ "unhandled in continuate: " ++ (show unhandled)







continuateWithKont ::
 (Symbol.HasSymbolTable a) =>
 (DSExp ann) ->
 (DSExp ann) ->
 (St.State a (DSExp ann))


continuateWithKont exp@(Nil{}) q = 
 do 
 exp' <- trivContinuate exp
 return $! 
  App {f = q, keywordArgs=[], args = [exp'], ann = nullAnn}

continuateWithKont exp@(Int{}) q = 
 do 
 exp' <- trivContinuate exp
 return $! 
  App {f = q, keywordArgs=[], args = [exp'], ann = nullAnn}

continuateWithKont exp@(Bool{}) q = 
 do 
 exp' <- trivContinuate exp
 return $! 
  App {f = q, keywordArgs=[], args = [exp'], ann = nullAnn}

continuateWithKont exp@(Var{}) q = 
 do 
 exp' <- trivContinuate exp
 return $! 
  App {f = q, keywordArgs=[], args = [exp'], ann = nullAnn}

continuateWithKont exp@(Lam{}) q = 
 do 
 exp' <- trivContinuate exp
 return $! 
  App {f = q, keywordArgs=[], args = [exp'], ann = nullAnn}

continuateWithKont exp@(Let (v,e,a) body annt) q = 
 do
 body' <- continuateWithKont body q
 if isTrivExp e 
  then do
   e' <- trivContinuate e
   return $! Let (v,e',a) body' annt
  else do
   -- TODO: 
   -- TODO: We can still cont-let-bind valExp.
   cont <- return $! Lam [] [v] body' nullAnn -- TODO: Annotate this.
   continuate e (\ e' ->
    return $! App cont [] [e'] nullAnn) -- TODO: keyword cont
 

continuateWithKont (App prim@(Prim{}) [] args annt) q = 
 continuateMany args (\ args' ->
  return $! App q [] [App prim [] args' annt] nullAnn)

continuateWithKont (App f keywords args annt) q =
 continuate f (\ f' ->
  continuateMany (map (\ (s,e,a) -> e) keywords) (\ keyargs' ->
   continuateMany args (\ args' ->
    let
      oldKeysXnew = zip keywords keyargs' 
      keywords' = map (\ ((s,_,a),e) -> (s,e,a)) oldKeysXnew
     in return $! App f' (keywords') (q:args') annt)))

continuateWithKont (If cond cons alt Direct annt) q = 
 do
 cons' <- continuateWithKont cons q -- BUG: Exponential blowup.
 alt' <- continuateWithKont alt q
 continuate cond (\ cond' ->
  return $! If cond' cons' alt' Direct annt)

continuateWithKont (LetRec varexps body annt) q =
 do

 varexps' <-
  let
   continuateVarExps [] = return $! []
   continuateVarExps ((v,e,a):tl) =
    do
    e' <- trivContinuate e
    tl' <- continuateVarExps tl
    return $! (v,e',a):tl'  
  in continuateVarExps varexps

 body' <- continuateWithKont body q
 
 return $! LetRec varexps' body' annt

continuateWithKont unhandled q = 
 error $ "unhandled in continuateWithKont: " ++ (show unhandled)









trivContinuate ::
 (Symbol.HasSymbolTable a) =>
 (DSExp ann) ->
 St.State a (DSExp ann)

trivContinuate exp@(Nil{}) = return $! exp

trivContinuate exp@(Int{}) = return $! exp

trivContinuate exp@(Bool{}) = return $! exp

trivContinuate exp@(Symbol{}) = return $! exp

trivContinuate exp@(Var{}) = return $! exp

trivContinuate exp@(Prim (primop) annt) =
 -- return $! Prim (primop{callingStyle = ContinuationPassing}) annt
 error $ "error in trivContinuate: Primitives must not be directly trivContinuated."

trivContinuate exp@(CallCC True annt) =
 error $ "Cannot handle call/cc with handler yet."

trivContinuate exp@(CallCC False annt) =
 -- M call/cc =
 -- (\ (c f) .
 --   (f c (\ (k rv) . c rv))) 
 do
 fVar <- genVar "f"
 rvVar <- genVar "rv"
 contVar <- genVar "cont"
 kontVar <- genVar "kont"
 return $! 
  mkLam [] [contVar,fVar]
   (mkApp fVar [] [contVar,mkLam [] [kontVar,rvVar] (mkApp contVar [] [rvVar])])

trivContinuate exp@(App (Prim{}) [] args annt) =
 if isTrivExp exp
 then return $! exp
 else error $ "Nontrivial expression in trivContinuate."

trivContinuate (Lam keywords formals body annt) =
 do
 kSym <- Symbol.genSym "k"
 kVar <- return $! Var kSym nullAnn 
--  body' <- continuate body (\ retexp ->
--            return $! App kVar Nothing [retexp] nullAnn)
 body' <- continuateWithKont body kVar
 formals' <- return $! kVar:formals
 return $! Lam keywords formals' body' annt
 
trivContinuate unhandled =
 error $ "unhandled in trivContinuate: " ++ (show unhandled)



continuateProgram exp =
 continuateWithKont exp (primHalt)

--  (\ retexp ->
--    -- CPS conversion happens before closure conversion.
--   return $! mkApp primHalt [] [retexp])




continuateMany [] k = k []
continuateMany (hd:tl) k = 
 continuate hd (\ hd' ->
  continuateMany tl (\ tl' ->
   k $! hd':tl'))
 



 {- Flatten application arguments. -}
 --
 -- Converts all arguments to variables or "word"-sized constants. 
 -- 
 -- This is kind of like a weaker version of A-Normalization.


 -- flattenArgs walks the program, flattening existing arguments.
 -- Only applications in an argument position are flattened out.
flattenApps ::
 Symbol.HasSymbolTable a =>
 (DSExp ann) ->
 (St.State a (DSExp ann))

flattenApps exp@(Nil{}) =
 return $! exp

flattenApps exp@(Int{}) =
 return $! exp

flattenApps exp@(Bool{}) =
 return $! exp

flattenApps exp@(Var{}) =
 return $! exp

flattenApps exp@(Symbol{}) =
 return $! exp

flattenApps exp@(Prim{}) =
 return $! exp

flattenApps exp@(Lam keywords formals body annt) =
 do
 body' <- flattenApps body -- (\ exp -> return $! exp)
 return $! Lam keywords formals body' annt

flattenApps exp@(If cond cons alt ContinuationPassing annt) =  
 flattenArg cond (\ cond' ->
  flattenArg cons (\ cons' ->
   flattenArg alt (\ alt' ->
    return $! If cond' cons' alt' ContinuationPassing annt)))

flattenApps exp@(If cond cons alt Direct annt) =
 do
 cons' <- flattenApps cons
 alt' <- flattenApps alt
 flattenArg cond (\ cond' ->
  return $! If cond' cons' alt' Direct annt)

flattenApps (App f keywords args annt) = 
 do
 flattenArg f (\ f' ->
  flattenArgMany (map (\ (s,e,a) -> e) keywords) (\ keyargs' ->
   flattenArgMany args (\ args' ->
    let
      keywords' = map (\ ((s,_,a),e) -> (s,e,a)) (zip keywords keyargs')
     in return $! App f' keywords' args' annt)))
      

flattenApps (Seq first second annt) = 
 do
 flattenArg first (\ first' ->
  do
  second' <- flattenApps second
  return $! Seq first' second' annt)

flattenApps (Let (v,e,a) body annt) =
 do
 body' <- flattenApps body
 flattenArg e (\ e' ->
  return $! Let (v,e',a) body' annt)
 
flattenApps (LetRec varexps body annt) = 
 do
 varexps' <-
  Might.mmap
   (\ (v,e,a) -> 
    do
    e' <- flattenApps e -- (\ exp -> return $! exp)
    return $! (v,e',a))
   varexps
 body' <- flattenApps body 
 return $! LetRec varexps' body' annt

flattenApps unhandled =
 error $! "unhandled in flattenApps: " ++ (show unhandled)





 -- flattenArg is applied to an expression in an argument position
 -- which needs to be flattened.
flattenArg exp@(Nil{}) k =
 k $! exp

flattenArg exp@(Int{}) k =
 k $! exp

flattenArg exp@(Bool{}) k =
 k $! exp

flattenArg exp@(Symbol{}) k =
 k $! exp

flattenArg exp@(Var{}) k = 
 k $! exp

flattenArg exp@(Prim{}) k =
 k $! exp 

flattenArg (Let (v,e,a) body annt) k =
 do
 flattenArg e (\ e' ->
  do 
  letBody <- flattenArg body k
  return $! Let (v,e',a) letBody nullAnn)

flattenArg (Lam keywords formals body annt) k = 
 do
 lamSym <- Symbol.genSym "fnptr"
 lamVar <- return $! Var lamSym nullAnn
 body' <- flattenApps body -- (\ exp -> return $! exp)
 letBody <- k lamVar
 return $! 
   Let (lamVar,Lam keywords formals body' annt,nullAnn) letBody nullAnn

flattenArg exp@(App f keywords args annt) k =
 flattenArg f (\ f' ->
  flattenArgMany (map (\ (s,e,a) -> e) keywords) (\ keyargs' ->
--  flattenArgOption envExp (\ envExp' ->
   flattenArgMany args (\ args' ->
    do
    valSym <- Symbol.genSym "val"
    valVar <- return $! Var valSym nullAnn
    letBody <- k valVar
    keywords' <- return $! 
     map (\ ((s,_,a),e) -> (s,e,a)) (zip keywords keyargs')
    return $! Let (valVar,App f' keywords' args' annt,nullAnn) letBody nullAnn)))

flattenArg unhandled k = 
 error $! "unhandled in flattenArg: " ++ (show unhandled)




flattenArgOption (Just exp) k =
 flattenArg exp (\ exp' ->
  k $! Just exp')
flattenArgOption (Nothing) k =
 k $! Nothing


flattenArgMany [] k = k []
flattenArgMany (hd:tl) k =
 do
 flattenArg hd (\ hd' ->
  flattenArgMany tl (\ tl' ->
   k $! hd':tl'))
    

