 -- An expression-based AST for a generic, macroless RxRS-like scheme.

module Compiler.RxRS where

import qualified Control.Monad.State as St

import qualified Compiler.Symbol as Sym

import qualified Data.Map as Map
import qualified Data.Set as Set

import qualified Data.Maybe as Maybe

import qualified Monad (liftM)
import qualified Might

import System.IO.Unsafe

 -- COMPILE-TIME OPTIONS.

data ShowLabelsFormat = SHOWLABELS_CHICKEN | SHOWLABELS_FIRST
 deriving (Eq,Show)

showLabelsFormat = SHOWLABELS_FIRST



 -- ANNOTATIONS.

data Annotation ext = 
 Ann {
  freeVarMap :: !(Maybe (FreeVarMap ext)) ,
  label :: !Sym.Symbol ,
  isCont :: !Bool ,
  isJoin :: !Bool
 }


botAnn :: Annotation a
botAnn = Ann {
 freeVarMap = Nothing ,
 label = Sym.Sym (0-1) "{undefined}" ,
 isCont = False ,
 isJoin = False 
 }


markJoinPoint term =
 term { ann = (ann term) { isJoin = True } }

unmarkJoinPoint term =
 term { ann = (ann term) { isJoin = False } }

isJoinPoint term =
 isJoin $! ann $! term



markCont term =
 term { ann = (ann term) { isCont = True } }

unmarkCont term =
 term { ann = (ann term) { isCont = False } }

isContTerm term =
 isCont $! ann $! term

isUserTerm term =
 not (isCont $! ann $! term)


labelOf term =  label $ ann term


 -- Free variable annotations.

 -- A FreeVarMap maps from a symbol to the number of free occurences
 -- within an expression.
type FreeVarMap ext = Map.Map Sym.Symbol [Term ext]





 -- ABSTRACT SYNTAX TREE.

data Term ext = 

   TermTop

 | TermBot

 | Lit { lit :: !Literal,
       ann :: !(Annotation ext) }

 | Prim { prim :: !Primitive,
	  ann :: !(Annotation ext) }

 | Void { ann :: !(Annotation ext) } 

--    -- This evalutes to bot.  What ever it gets bound to dies.
--  | Unspec { ann :: !(Annotation ext) }

   -- A binding occurrence of a variable.
 | Var { varSym :: !Sym.Symbol,
	 ann :: !(Annotation ext) }

   -- A use of a variable.
 | Ref { varSym :: !Sym.Symbol,
	 ann :: !(Annotation ext) }

   -- Bindable formals.
 | Forms { ordered :: ![Term ext] ,
	   labels :: ![Sym.Symbol] ,
	   labeled :: ![Term ext] ,
	   rest :: !(Maybe (Term ext)) ,
	   ann :: !(Annotation ext) }

   -- Passable arguments.
 | Args { ordered :: ![Term ext] ,
	  labels :: ![Sym.Symbol] ,
	  labeled :: ![Term ext] ,
	  rest :: !(Maybe (Term ext)) ,
	  ann :: !(Annotation ext) }

   -- Lambda terms.
 | Lam { forms :: !(Term ext), -- Must be Forms.
	 body :: !(Term ext),
	 ann :: !(Annotation ext) } 

   -- Function application.
 | App { fun :: !(Term ext), 
	 args :: !(Term ext), -- Must be Args.
	 ann :: !(Annotation ext) }

   -- The top level is a sequence of terms, too.
 | Seq { exps :: ![Term ext],
	 ann :: !(Annotation ext) }

 | Do { vars :: ![Term ext],
        inits :: ![Term ext],
        steps :: ![Term ext],
        test :: !(Term ext),
        body :: !(Term ext),
        cmds :: !(Term ext),
        ann :: !(Annotation ext) }

 | Cond { conds :: ![Term ext], 
	  thens :: ![Term ext],
	  -- An else branch is modeled by a (#t ...).
	  ann :: !(Annotation ext) }

 | Case { exp :: !(Term ext) ,
	  litss :: ![Maybe [Term ext]], 
	  thens :: ![Term ext],
	  ann :: !(Annotation ext) }

   -- Multi-value return is permitted.
 | Let { varss :: ![Term ext] , -- Must be Forms.
	 exps :: ![Term ext] ,
	 body :: !(Term ext) ,
	 ann :: !(Annotation ext) }

 | Set { var :: !(Term ext),
	 exp :: !(Term ext),
	 body :: !(Term ext),
	 ann :: !(Annotation ext) }

 | LetRec { varss :: ![Term ext] , -- Must be Forms.
	    exps :: ![Term ext] ,
	    body :: !(Term ext) ,
	    ann :: !(Annotation ext) }

 | Def { forms :: !(Term ext), -- Must be Forms.
	 exp :: !(Term ext) ,
	 ann :: !(Annotation ext) }



instance Eq (Term ext) where
 t1 == t2 = ((label . ann) t1) == ((label . ann) t2)

instance Ord (Term ext) where
 compare t1 t2 = compare ((label . ann) t1) ((label . ann) t2)
 t1 < t2 = ((label . ann) t1) < ((label . ann) t2)


instance Show (Term ext) where
--  show (Seq [exp] ann) = show exp
--  show (Seq [] ann) = "#:void"
 show (Seq exps ann) =
  "(begin " ++ (Might.showWithSep " " exps) ++ ")"
 
 show (Var varSym ann) = show varSym

 show (Ref varSym ann) = show varSym

 show (Forms [] [] [] Nothing ann) = ""
 show (Forms [var] [] [] Nothing ann) = show var
 show (Forms vars [] [] Nothing ann) = 
  (Might.showWithSep " " vars)
 show (Forms ordered labels labeled rest ann) 
  | showLabelsFormat == SHOWLABELS_FIRST =
   (Might.showWithSep " " (labeled ++ ordered)) ++ 
   (case rest of
     Just v -> " . " ++ (show v)
     Nothing -> "")
  | showLabelsFormat == SHOWLABELS_CHICKEN =
   (Might.showWithSep " " ordered) ++ 
   (foldl (++) ""
    (map (\ (l,t) -> " " ++ (show l) ++ ": " ++ (show t)) (zip labels labeled))) ++ 
   (case rest of
     Just v -> " . " ++ (show v)
     Nothing -> "")


 show (Args [] [] [] Nothing ann) = ""
 show (Args [var] [] [] Nothing ann) = show var
 show (Args vars [] [] Nothing ann) = 
  (Might.showWithSep " " vars)
 show (Args ordered labels labeled rest ann) 
  | showLabelsFormat == SHOWLABELS_FIRST =
   (Might.showWithSep " " (labeled ++ ordered)) ++ 
   (case rest of
     Just v -> " . " ++ (show v)
     Nothing -> "")
  | showLabelsFormat == SHOWLABELS_CHICKEN =
   (Might.showWithSep " " ordered) ++ 
   (foldl (++) ""
    (map (\ (l,t) -> " " ++ (show l) ++ ": " ++ (show t)) (zip labels labeled))) ++ 
   (case rest of
     Just v -> " . " ++ (show v)
     Nothing -> "")


 show (Lam (Forms [] [] [] (Just var) _) body ann) =
  "(lambda " ++ (show var) ++ " " ++ (show body) ++ ")"

 show (Lam forms body ann) =
  "(lambda (" ++ (show forms) ++ ") " ++ (show body) ++ ")"

 show (App fun args ann) =
  "(" ++ (show fun) ++ " " ++ (show args) ++ ")"

 show (Cond [cond,Lit (LitBool True) _] [conseq,altern] ann) =
  "(if " ++ (show cond) ++ " " ++ (show conseq) ++ " " ++ (show altern) ++ ")"

 show (Cond conds thens ann) =
  "(cond " ++ (showCondThens condthens) ++ ")"
  where 
   condthens = zip conds thens
   showCondThens [] = ""
   showCondThens [(Lit (LitBool True) botAnn,thn)] 
    = "(else " ++ (show thn) ++ ")"
   showCondThens [(cond,thn)]
    = "(" ++ (show cond) ++ " " ++ (show thn) ++ ")"
   showCondThens (hd:tl) 
    = (showCondThens [hd]) ++ " " ++ (showCondThens tl)

 show (Case exp litss thens ann) =
  "(case " ++ (show exp) ++ " " ++
   (showLitsThens (zip litss thens))
   ++ ")"
  where
   showLitsThens [] = ""
   showLitsThens [(Nothing,thn)] =
    "(else " ++ (show thn) ++ ")"
   showLitsThens [((Just lits),thn)] =
    "((" ++ (Might.showWithSep " " lits) ++ ") " ++ (show thn) ++ ")"
   showLitsThens (hd:tl) = (showLitsThens [hd]) ++ (showLitsThens tl)


  -- let*
 show term@(Let [_] [_] (Let [_] [_] _ _) _) =
  "(let* (" ++ (showBindings term) ++ ") " ++ (showBody term) ++ ")"
  where 
   showBindings (Let [v] [e] subterm@(Let [_] [_] _ _) _) =
    "(" ++ (show v) ++ " " ++ (show e) ++ ") " ++ 
    showBindings subterm
   showBindings (Let [v] [e] body _) =
    "(" ++ (show v) ++ " " ++ (show e) ++ ")"

   showBody (Let [v] [e] subterm@(Let [_] [_] _ _) _) = showBody subterm
   showBody (Let [v] [e] body _) = show body


 show term@(Let varss exps body ann) =
  let
   varsexps = zip varss exps

   showVarsExp (Forms{ordered=[var]},exp)
    = "(" ++ (show var) ++ " " ++ (show exp) ++ ")"
   showVarsExp (forms,exp)
    = "((" ++ (show forms) ++ ") " ++ (show exp) ++ ")"

   showVarsExps [] = ""
   showVarsExps [varsexp]
    = showVarsExp varsexp
   showVarsExps (hd:tl) 
    = (showVarsExps [hd]) ++ " " ++ (showVarsExps tl)

  in "(let (" ++ (showVarsExps varsexps) ++ ") " ++ (show body) ++ ")"
        

 show (Do vars inits steps test body cmds ann) = 
  "(do (" ++
     (Might.showWithSepWith showVarInitStep " " (zip3 vars inits steps)) ++ 
    ") " ++
    "(" ++ (show test) ++ " " ++ (show body) ++ ")" ++ " " ++
   (show cmds) ++ ")"
  where
  showVarInitStep :: (Term ext, Term ext, Term ext) -> String
  showVarInitStep (Var sym _,init,Ref sym' _) 
   | sym == sym' = "(" ++ (show sym) ++ " " ++ (show init) ++ ")"
  showVarInitStep (Var sym _,init,step) =
   "(" ++ (show sym) ++ " " ++ (show init) ++ " " ++ (show step) ++ ")"


 show (LetRec varss exps body ann) =
  let
   varsexps = zip varss exps

   showVarsExp (Forms{ordered=[var]},exp)
    = "(" ++ (show var) ++ " " ++ (show exp) ++ ")"

   showVarsExps [] = ""
   showVarsExps [varsexp]
    = showVarsExp varsexp
   showVarsExps (hd:tl) 
    = (showVarsExps [hd]) ++ " " ++ (showVarsExps tl)

  in "(letrec (" ++ (showVarsExps varsexps) ++ ") " ++ (show body) ++ ")"

 show (Set var exp (Void _) ann) =
  "(set! " ++ (show var) ++ " " ++ (show exp) ++ ")"

 show (Set var exp body ann) = 
  "(begin (set! " ++ (show var) ++ " " ++ (show exp) ++ ") " 
    ++ (show body) ++ ")"

 show (Def (Forms{ordered=[var]}) exp ann) =
  "(define " ++ (show var) ++ " " ++ (show exp) ++ ")"

 show (Def{}) = error $ "{malformed define}"

 show (Lit lit ann) = (show lit)

 show (Prim prim ann) = (show prim)

 show (Void ann) = "'void-value"

-- show (Unspec ann) = "(unspecified)"


instance Might.Hashable (Term ext) where
 hash term = Might.hash $! label $! ann $! term
 hashM m term = Might.hashM m $! label $! ann $! term
 

data Literal =
   LitNil
 | LitBool !Bool
 | LitInt !Int
 | LitChar !Char
 | LitString !String
 | LitSymbol !Sym.Symbol

instance Show Literal where
 show LitNil = "'()"
 show (LitBool True) = "#t"
 show (LitBool False) = "#f"
 show (LitInt n) = show n
 show (LitChar c) = "#\\" ++ (show c)
 show (LitString str) = show str
 show (LitSymbol sym) = "'" ++ (show sym)




data Primitive =

   SUM 
 | PROD
 | DIFF
 | DIV
 | REMAINDER
 | QUOTIENT
 | EXPT

 | EXACT2INEXACT

 | MAX
 | MIN


 | NUMBER2STRING
 | STRING_APPEND

 | VECTOR_LENGTH
 | MAKE_VECTOR
 | VECTOR
 | VECTORSET
 | VECTORREF

 | APPLY 

 | CONS
 | CAR
 | CDR
 | SETCAR
 | SETCDR
 | LIST 
 | APPEND
 | MEMQ -- Candidate for removal to library.
 | MEMV -- Candidate for removal to library.
 | NULLP
 | PAIRP

 | EQNUM
 | LT
 | GT
 | LEQ
 | GEQ
 | ZEROP
 | NUMBERP

 | EQP
 | EQVP
 | EQUALP

 | NOT
 | OR
 | AND

 | WRITE
 | DISPLAY
 | NEWLINE

 | CALLCC

 | ERROR
 | HALT

 | CPS !Primitive -- Turns a primitive into a CPS version of itself.

 | UNSPECIFIED

   -- Debugging.
 | ABSTRACE

   -- Optimization.
 | DEAD -- Mark this call site as dead.

 deriving (Eq,Ord)


instance Show Primitive where
 show SUM = "+"
 show PROD = "*"
 show DIFF = "-"
 show DIV = "/"
 show REMAINDER = "remainder"
 show QUOTIENT = "quotient"
 show EXPT = "expt"

 show MAX = "max"
 show MIN = "min"

 show EXACT2INEXACT = "exact->inexact"

 show NUMBER2STRING = "number->string"
 show STRING_APPEND = "string-append"

 show APPLY = "apply"

 show CONS = "cons"
 show CAR = "car"
 show CDR = "cdr"
 show SETCAR = "set-car!"
 show SETCDR = "set-cdr!"
 show LIST = "list"
 show APPEND = "append"
 show MEMQ = "memq"
 show MEMV = "memv"
 show NULLP = "null?"
 show PAIRP = "pair?"

 show VECTOR_LENGTH = "vector-length"
 show MAKE_VECTOR = "make-vector"
 show VECTOR = "vector"
 show VECTORSET = "vector-set!"
 show VECTORREF = "vector-ref"

 show EQNUM = "="
 show Compiler.RxRS.LT = "<"
 show Compiler.RxRS.GT = ">"
 show GEQ = ">="
 show LEQ = "<="
 show NUMBERP = "number?"
 show ZEROP = "zero?"

 show EQP = "eq?"
 show EQVP = "eqv?"
 show EQUALP = "equal?"

 show NOT = "not"
  -- BUG: AND and OR are bugs 'cause they're strict.
 show OR = "or"
 show AND = "and"

 show WRITE = "write"
 show DISPLAY = "display"
 show NEWLINE = "newline"

 show ERROR = "error"

 show CALLCC = "call/cc"
 show HALT = "halt"

 show (CPS prim) = (show prim) ++ ":cps"

 show UNSPECIFIED = ".unspecified"

 show ABSTRACE = "abstrace"

 show DEAD = "DEAD-CODE"




 -- TREE WALKERS.


 -- A post-order non-Monadic walker.

mapPostOrder :: (Term ext -> Term ext) -> (Term ext) -> (Term ext)

mapPostOrder f (Seq exps ann) =
 f (Seq (map (mapPostOrder f) exps) ann)

mapPostOrder f (exp@(Var varSym ann)) =
 f exp

mapPostOrder f (exp@(Ref varSym ann)) =
 f exp 

mapPostOrder f (Forms ordered labels labeled rest ann) =
 f (Forms ordered' labels labeled' rest' ann)
  where
   ordered' = map (mapPostOrder f) ordered
   labeled' = map (mapPostOrder f) labeled
   rest' = Monad.liftM (mapPostOrder f) rest

mapPostOrder f (Args ordered labels labeled rest ann) =
 f (Args ordered' labels labeled' rest' ann)
  where
   ordered' = map (mapPostOrder f) ordered
   labeled' = map (mapPostOrder f) labeled
   rest' = Monad.liftM (mapPostOrder f) rest


mapPostOrder f (Lam vars body ann) =
 f (Lam (mapPostOrder f vars) (mapPostOrder f body) ann)

mapPostOrder f (App fun args ann) =
 f (App (mapPostOrder f fun) (mapPostOrder f args) ann)

mapPostOrder f (Cond conds thens ann) =
 f (Cond conds' thens' ann) 
  where
   conds' = map (mapPostOrder f) conds
   thens' = map (mapPostOrder f) thens
   
mapPostOrder f (Case exp litss thens ann) =
 f (Case exp' litss' thens' ann) 
  where
   exp' = mapPostOrder f exp
   litss' = map (Monad.liftM (map (mapPostOrder f))) litss
   thens' = map (mapPostOrder f) thens

mapPostOrder f (Let varss exps body ann) =
 f (Let varss' exps' body' ann) 
  where
   varss' = map (mapPostOrder f) varss
   exps' = map (mapPostOrder f) exps
   body' = mapPostOrder f body

mapPostOrder f (LetRec varss exps body ann) =
 f (LetRec varss' exps' body' ann) 
  where
   varss' = map (mapPostOrder f) varss
   exps' = map (mapPostOrder f) exps
   body' = mapPostOrder f body
   
mapPostOrder f (Do vars inits steps test body cmds ann) =
 f (Do vars' inits' steps' test' body' cmds' ann) 
  where
   mpof = mapPostOrder f
   vars' = map mpof vars
   inits' = map mpof inits
   steps' = map mpof steps
   test' = mpof test
   body' = mpof body
   cmds' = mpof cmds

mapPostOrder f (Set var exp body ann) =
 f (Set (mpof var) (mpof exp) (mpof body) ann)
  where
   mpof = mapPostOrder f

mapPostOrder f (Def vars exp ann) =
 f (Def (mpof vars) (mpof exp) ann)
  where
   mpof = mapPostOrder f

mapPostOrder f term@(Lit lit ann) = 
 f term

mapPostOrder f term@(Prim prim ann) =
 f term

mapPostOrder f term@(Void ann) =
 f term





 -- A post-order Monadic walker.

mmapPostOrder ::
 (Monad m) => 
 (Term ext -> m (Term ext)) ->
 (Term ext) ->
 (m (Term ext))

mmapPostOrder f (Seq exps ann) =
 do 
 exps' <- Might.mmap (mmapPostOrder f) exps
 f (Seq exps' ann)

mmapPostOrder f (exp@(Var varSym ann)) =
 f exp

mmapPostOrder f (exp@(Ref varSym ann)) =
 f exp 


mmapPostOrder f (Forms ordered labels labeled rest ann) =
 do
 ordered' <- Might.mmap (mmapPostOrder f) ordered
 labeled' <- Might.mmap (mmapPostOrder f) labeled
 rest' <- Might.maybeLiftM (mmapPostOrder f) rest
 f (Forms ordered' labels labeled' rest' ann)


mmapPostOrder f (Args ordered labels labeled rest ann) =
 do
 ordered' <- Might.mmap (mmapPostOrder f) ordered
 labeled' <- Might.mmap (mmapPostOrder f) labeled
 rest' <- Might.maybeLiftM (mmapPostOrder f) rest
 f (Args ordered' labels labeled' rest' ann)



mmapPostOrder f (Lam vars body ann) =
 do
 vars' <- mmapPostOrder f vars
 body' <-  mmapPostOrder f body
 f (Lam vars' body' ann)

mmapPostOrder f (App fun args ann) =
 do
 fun' <- mmapPostOrder f fun
 args' <- mmapPostOrder f args
 f (App fun' args' ann)

mmapPostOrder f (Cond conds thens ann) =
 do
 conds' <- Might.mmap (mmapPostOrder f) conds
 thens' <- Might.mmap (mmapPostOrder f) thens
 f (Cond conds' thens' ann) 
   
mmapPostOrder f (Case exp litss thens ann) =
 do
 exp' <- mmapPostOrder f exp
 litss' <- Might.mmap (Might.maybeLiftM (Might.mmap (mmapPostOrder f))) litss
 thens' <- Might.mmap (mmapPostOrder f) thens
 f (Case exp' litss' thens' ann) 

mmapPostOrder f (Let varss exps body ann) =
 do
 varss' <- Might.mmap (mmapPostOrder f) varss
 exps' <- Might.mmap (mmapPostOrder f) exps
 body' <- mmapPostOrder f body
 f (Let varss' exps' body' ann) 

mmapPostOrder f (LetRec varss exps body ann) =
 do
 varss' <- Might.mmap (mmapPostOrder f) varss
 exps' <- Might.mmap (mmapPostOrder f) exps
 body' <- mmapPostOrder f body
 f (LetRec varss' exps' body' ann) 

   
mmapPostOrder f (Set var exp body ann) =
 do
 var' <- mpof var
 exp' <- mpof exp
 body' <- mpof body
 f (Set var' exp' body' ann)
  where
   mpof = mmapPostOrder f

mmapPostOrder f (Def vars exp ann) =
 do
 vars' <- (mpof vars) 
 exp' <- (mpof exp)
 f (Def vars' exp' ann)
  where
   mpof = mmapPostOrder f

mmapPostOrder f (Do vars inits steps test body cmds ann) =
 do
 vars' <- Might.mmap (mmapPostOrder f) vars
 inits' <- Might.mmap (mmapPostOrder f) inits
 steps' <- Might.mmap (mmapPostOrder f) steps
 test' <- mmapPostOrder f test
 body' <- mmapPostOrder f body
 cmds' <- mmapPostOrder f cmds
 f (Do vars' inits' steps' test' body' cmds' ann)

mmapPostOrder f term@(Lit lit ann) = 
 f term

mmapPostOrder f term@(Prim prim ann) =
 f term

mmapPostOrder f term@(Void ann) =
 f term









mfoldPreOrder ::
 (Monad m) =>
 (a -> Term ext -> m (a,a,Term ext)) ->
 a -> 
 (Term ext) ->
 m (Term ext)

mfoldPreOrder f i t = 
 let

  mfpo init term
   | null (expChildren term) && null (formChildren term) = 
     do
     (same',deeper',term') <- f init term
     return $! term'
   | otherwise = complex init term

  complex init term@(Var{}) =
   do
   (same',deeper',term') <- f init term
   return $! term'

  complex init term@(Ref{}) =
   do
   (same',deeper',term') <- f init term
   return $! term'

  complex init term@(Forms{}) =
   do
   (same',deeper',term') <- f init term
   Forms ordered' labels' labeled' rest' formsann' <- return $! term'
   ordered'' <- Might.mmap (mfpo same') ordered'
   labeled'' <- Might.mmap (mfpo same') labeled'
   rest'' <- Might.maybeLiftM (mfpo same') rest'
   return $! Forms ordered'' labels' labeled'' rest'' formsann'

  complex init term@(Args{}) =
   do
   (same',deeper',term') <- f init term
   Args ordered' labels' labeled' rest' argsann' <- return $! term'
   ordered'' <- Might.mmap (mfpo same') ordered'
   labeled'' <- Might.mmap (mfpo same') labeled'
   rest'' <- Might.maybeLiftM (mfpo same') rest'
   return $! Args ordered'' labels' labeled'' rest'' argsann'

  complex init term@(App{}) =
   do
   (same',deeper',term') <- f init term
   App fun' args' ann' <- return term'
   fun'' <- mfpo same' fun'
   args'' <- mfpo same' args'
   return $! App fun'' args'' ann'

  complex init term@(Lam{}) = 
   do
   (same',deeper',term') <- f init term
   Lam forms' body' ann' <- return term'
   forms'' <- mfpo deeper' forms'
   body'' <- mfpo deeper' body'
   return $!
    Lam forms'' body'' ann'

  complex init term@(Seq{}) =
   do
   (same',deeper',term') <- f init term
   Seq exps' ann' <- return $! term'
   exps'' <- Might.mmap (mfpo same') exps'
   return $! Seq exps'' ann' 

  complex init term@(Let{}) =
   do
   (same',deeper',term') <- f init term
   Let varss' exps' body' ann' <- return $! term'
   varss'' <- Might.mmap (mfpo deeper') varss'
   exps'' <- Might.mmap (mfpo same') exps'
   body'' <- mfpo deeper' body'
   return $! Let varss'' exps'' body'' ann'

  complex init term@(LetRec{}) =
   do
   (same',deeper',term') <- f init term
   LetRec varss' exps' body' ann' <- return $! term'
   varss'' <- Might.mmap (mfpo deeper') varss'
   exps'' <- Might.mmap (mfpo deeper') exps'
   body'' <- mfpo deeper' body'
   return $! LetRec varss'' exps'' body'' ann'

  complex init term@(Do{}) =
   do
   (same',deeper',term') <- f init term
   Do vars' inits' steps' test' body' cmds' ann' <- return $! term'
   vars'' <- Might.mmap (mfpo deeper') vars'
   inits'' <- Might.mmap (mfpo same') inits'
   steps'' <- Might.mmap (mfpo deeper') steps'
   test'' <- mfpo deeper' test'
   body'' <- mfpo deeper' body'
   cmds'' <- mfpo deeper' cmds'
   return $! Do vars'' inits'' steps'' test'' body'' cmds'' ann'

  complex init term@(Set{}) =
   do
   (same',deeper',term') <- f init term
   Set var' exp' body' ann' <- return $! term'
   var'' <- mfpo same' var'
   exp'' <- mfpo same' exp'
   body'' <- mfpo same' body'
   return $! Set var'' exp'' body'' ann'

  complex init term@(Def{}) =
   do
   (same',deeper',term') <- f init term
   Def var' exp' ann' <- return $! term'
   var'' <- mfpo same' var'
   exp'' <- mfpo same' exp'
   return $! Def var'' exp'' ann'
   
--   complex init term@(Cond{}) =
--    do
--    (same',deeper',term') <- f init term
--    Cond conds' thens' ann' <- return $! term'
--    conds'' <- Might.mmap (mfpo same') conds'
--    thens'' <- Might.mmap (mfpo same') thens'
--    return $! Cond conds'' thens'' ann'

   -- TODO: Consider strict v. non-strict.
  complex init term@(Cond{}) =
   do
   (same',deeper',term') <- f init term
   Cond conds' thens' ann' <- return $! term'
   conds'' <- Might.mmap (mfpo same') conds'
   thens'' <- Might.mmap (mfpo same') thens'
   return $! Cond conds'' thens'' ann'

  complex init term@(Case{}) = 
   do
   (same',deeper',term') <- f init term
   Case exp' litss' thens' ann' <- return $! term' 
   let mfpolits lits = Might.mmap (mfpo same') lits
   let maymfpolits = Might.maybeLiftM mfpolits
   litss'' <- Might.mmap maymfpolits litss'
   thens'' <- Might.mmap (mfpo same') thens'
   exp'' <- mfpo same' exp'
   return $! Case exp'' litss'' thens'' ann'
  
  complex init term = error $ "unfinished complex: " ++ (show term)

 in mfpo i t



 -- Statistics.

countLambdas (Lam forms body ann) = 
 1 + (countLambdas body)

countLambdas term = 
 foldl (\ tab term -> tab + (countLambdas term)) 0 (expChildren term)


countContLambdas term@(Lam forms body ann) = 
 (if isContTerm term then 1 else 0) + (countContLambdas body)

countContLambdas term = 
 foldl (\ tab term -> tab + (countContLambdas term)) 0 (expChildren term)


countUserLambdas term@(Lam forms body ann) = 
 (if isUserTerm term then 1 else 0) + (countUserLambdas body)

countUserLambdas term = 
 foldl (\ tab term -> tab + (countUserLambdas term)) 0 (expChildren term)





gatherCallSitesJustThisTerm :: 
 (Term ext) ->
 (St.State (Set.Set (Term ext)) (Term ext))

gatherCallSitesJustThisTerm term@(App{}) =
 do
 sites <- St.get
 St.put $! Set.insert term sites
 return $! term
 

gatherCallSitesJustThisTerm term =
 do
 return $! term


gatherCallSites ::
 (Term ext) ->
 (St.State (Set.Set (Term ext)) (Term ext))
gatherCallSites = 
 mmapPostOrder gatherCallSitesJustThisTerm


callSites :: (Term ext) -> Set.Set (Term ext)
callSites term = St.execState (gatherCallSites term) Set.empty




 -- SYNTACTIC HELPERS.

 -- Syntactic constructors.

newVarRef ::
 (Sym.HasSymbolTable a) =>
 String -> St.State a (Term ext,Term ext)
newVarRef str = 
 do
 sym <- Sym.genSym str
 var <- return $! Var sym botAnn
 ref <- return $! Ref sym botAnn
 return $! (var,ref)

newContVarRef :: 
 (Sym.HasSymbolTable a) =>
 String -> St.State a (Term ext,Term ext)
newContVarRef str =
 do
 (var,ref) <- newVarRef str -- TODO: Annotate properly.
 return $! (markCont var, markCont ref)

newLet ::
 (Sym.HasSymbolTable a) =>
 String ->
 (Term ext) ->
 ((Term ext) -> St.State a (Term ext)) -> 
 (St.State a (Term ext))

newLet v term genBody =
 do
 (newvar,newref) <- newVarRef v
 body <- genBody newref
 lett <- return $! 
  Let [Forms [newvar] [] [] Nothing botAnn] [term] body botAnn
 return $! lett


newIf :: (Term ext) -> (Term ext) -> (Term ext) -> (Term ext)
newIf cond thn els =
 (Cond [cond,litTrue] [thn,els] botAnn)



formalized vars =
 Forms vars [] [] Nothing botAnn

formalizedWithKont vars k =
 Forms vars [Sym.cont'] [markCont k] Nothing botAnn

deformalized (Forms [var] [] [] Nothing ann) =
 var


newKont :: 
 (Sym.HasSymbolTable a) =>
 (Term ext -> St.State a (Term ext)) ->
 (St.State a (Term ext))
newKont k =
 do
 (u,e) <- newVarRef "uj"
 body <- k e
 return $! markCont $! Lam (Forms [u] [] [] Nothing botAnn) body botAnn


newJoinKont :: 
 (Sym.HasSymbolTable a) =>
 (Term ext -> St.State a (Term ext)) ->
 (St.State a (Term ext))
newJoinKont k =
 do
 (u,e) <- newVarRef "uj"
 body <- k e
 return $! markCont $! markJoinPoint $! Lam (Forms [u] [] [] Nothing botAnn) body botAnn



 -- TODO: Sort this into another part of the code.
maySideEffect term@(Var _ _) =
 error $ "Queried a binding variable for effectw: " ++ (show term)

maySideEffect term@(Forms _ _ _ _ _) =
 error $ "Queried formals for effect: " ++ (show term)

maySideEffect term@(Ref _ _) = True -- could be out of scope!
maySideEffect term@(Lit _ _) = False
maySideEffect term@(Void _) = False
maySideEffect term@(Lam _ _ _) = False

maySideEffect term = True
            
isUnusedIn sym exp = 
 not $ sym `Map.member` (free exp)



newApp :: (Term ext) -> [Term ext] -> (Term ext)

 -- TODO: Expand forms with an "ignore all" field.
newApp (Lam (Forms [] [] [] (Just (Var sym _)) _) body _) args
 | all (not . maySideEffect) args && sym `isUnusedIn` body = 
    -- Beta-reduce on the spot.
   body
newApp f params =
 App  f (Args params [] [] Nothing botAnn) botAnn


newAppWithCont :: (Term ext) -> [Term ext] -> (Term ext) -> (Term ext)
newAppWithCont f params q =
 App  f (Args params [Sym.cont'] [markCont q] Nothing botAnn) botAnn







 -- Common constants.

void = Void botAnn

litTrue = Lit (LitBool True) botAnn
litFalse = Lit (LitBool False) botAnn

unspec = App (Prim UNSPECIFIED botAnn) (Args [] [] [] Nothing botAnn) botAnn


 -- Syntactic predicates.

isTrue (Lit (LitBool True) _) = True
isTrue _ = False

isDef (Def{}) = True
isDef _ = False

isRef (Ref{}) = True
isRef _ = False

isLam (Lam{}) = True
isLam _ = False

isPrimApp (App (Prim{}) args ann) = True
isPrimApp _ = False

isSingularForm (Forms [var] [] [] Nothing ann) = True
isSingularForm (Forms{}) = False





   
 -- Children functions.

expChildren :: Term ext -> [Term ext]
expChildren (Seq exps ann) = exps
expChildren (Var sym ann) = []
expChildren (Ref sym ann) = []
expChildren (Forms ordered labels labeled rest ann) = []
expChildren (Args ordered labels labeled rest ann) 
 = (Maybe.maybeToList rest) ++ ordered ++ labeled
expChildren (Lam vars body ann) = [body]
expChildren (App fun args@(Args{}) ann) = fun:(expChildren args)
expChildren (Cond conds thens ann) = conds ++ thens
expChildren (Case exp litss thens ann) =
 [exp] ++ lits ++ thens
  where 
   lits = Might.flattenListOfLists $!
     foldl (\ tab pat -> (Maybe.maybeToList pat) ++ tab) [] litss
expChildren (Do vars inits steps test body cmds ann) = 
 test:body:cmds:(inits ++ steps)
expChildren (Let varss exps body ann) = body:exps
expChildren (LetRec varss exps body ann) = body:exps
expChildren (Set var exp body ann) = [var,exp,body]
expChildren (Def var exp ann) = [exp]
expChildren (Lit lit ann) = []
expChildren (Prim prim ann) = []
expChildren (Void ann) = []

expChildren other = error $ "expChildren: unmatched " ++ (show other)




 -- formChildren: binding variables directly within this expression.

formChildren :: Term ext -> [Term ext]
formChildren term@(Var var ann) = [term]
formChildren (Forms ordered labels labeled rest ann) =
 (Maybe.maybeToList rest) ++ ordered ++ labeled
formChildren (Lam forms@(Forms{}) body ann) =
 formChildren forms
formChildren (Let varss exps body ann) = 
 Might.flattenListOfLists (map formChildren varss)
formChildren (LetRec varss exps body ann) = 
 Might.flattenListOfLists (map formChildren varss)
formChildren (Do vars inits steps test body cmds ann) = 
 vars
formChildren (Def forms@(Forms{}) exp ann) =
 formChildren forms
formChildren (Seq exps ann) = []
formChildren (Ref sym ann) = []
formChildren (Args ordered labels labeled rest ann) = []
formChildren (App fun args@(Args{}) ann) = []
formChildren (Cond conds thens ann) = []
formChildren (Case exp litss thens ann) = []
formChildren (Set var exp body ann) = []
formChildren (Lit lit ann) = []
formChildren (Prim prim ann) = []
formChildren (Void ann) = []

formChildren other = error $ "formChildren: unmatched " ++ (show other)







 -- SYNTAX ANNOTATION FUNCTIONS.

 -- Label alphatization.

type Label = Sym.Symbol

relabelJustThisTerm () term =
 do
 lab' <- Sym.genSym "lab"
 return $! ((),(),term { ann = (ann term) { label = lab' } })


relabel ::
 (Sym.HasSymbolTable a) =>
 (Term ext) -> 
 (St.State a (Term ext))
relabel = mfoldPreOrder relabelJustThisTerm ()




 -- Free variable annotation.

free :: Term ext -> FreeVarMap ext
free term = 
 case freeVarMap (ann term) of
  Just free -> free
  Nothing ->
    -- TODO: Install warning for when this happens.
   free (annotateFree term)



annotateFree :: (Term ext) -> (Term ext)
annotateFree term@(Var sym ann) =
 term { ann = ann { freeVarMap  = Nothing } }

annotateFree term@(Forms ordered labels labeled rest ann) =
 term { ann = ann { freeVarMap  = Nothing } }

annotateFree term@(Lam (Forms ordered labels labeled rest _) body ann) =
 term { ann = ann { freeVarMap = Just free'' } }
  where
   syms = map varSym (formChildren term)
   exps = expChildren term
   free' = Map.unionsWith (++) (map free exps)
   free'' = Map.filterWithKey (\ k a -> not(elem k syms)) free'


annotateFree term@(Let varrs exps body ann) =
 term { ann = ann { freeVarMap = Just free'' } }
  where
   syms = map varSym (formChildren term)
   children = expChildren term
   free' = Map.unionsWith (++) (map free children)
   free'' = Map.filterWithKey (\ k a -> not(elem k syms)) free'


annotateFree term@(LetRec varss exps body _) =
 term { ann = (ann term) { freeVarMap = Just free'' } }
  where
   syms = map varSym (formChildren term)
   children = expChildren term
   free' = Map.unionsWith (++) (map free children)
   free'' = Map.filterWithKey (\ k a -> not(elem k syms)) free'
    

annotateFree term@(Do vars inits steps test body cmds _) =
  term { ann = (ann term) { freeVarMap = Just free'' } }
   where
    syms = map varSym (formChildren term)
    freeInInits = Map.unionsWith (++) (map free inits)
    children = test:body:cmds:steps
    freeInChildren = Map.unionsWith (++) (map free children)
    free' = Map.filterWithKey (\ k a -> not(elem k syms)) freeInChildren  
    free'' = Map.unionWith (++) free' freeInInits
    

annotateFree term@(Def vars exp ann) =
 term { ann = ann { freeVarMap = Just free'' } }
  where
   syms = map varSym (formChildren term)
   exps = expChildren term
   free' = Map.unionsWith (++) (map free exps)
   free'' = Map.filterWithKey (\ k a -> not(elem k syms)) free'


annotateFree term@(Ref sym ann) = 
 term { ann = ann { freeVarMap = Just $! Map.singleton sym [term] } }


annotateFree term = 
 term { ann = (ann term) { freeVarMap = Just free' } }
  where
   exps = expChildren term
   free' = Map.unionsWith (++) (map free exps)



annotateAllFree = mapPostOrder annotateFree







 -- SYNTAX VERIFICATION FUNCTIONS.

  







 -- SYNTAX TRANSFORMATION FUNCTIONS.



 -- Alphatization.


type AlphaMap = Map.Map Sym.Symbol Sym.Symbol -- For renaming.


generateAlphaMap ::
 (Sym.HasSymbolTable a) =>
 AlphaMap ->
 (Term ext) ->
 (St.State a (AlphaMap))
generateAlphaMap fvmap (Var sym ann) =
 do
 sym' <- Sym.genSym (Sym.toString sym)
 fvmap' <- return $! Map.insert sym sym' fvmap
 return $! fvmap'
generateAlphaMap fvmap forms@(Forms ordered labels labeled rest ann) =
 do
 boundHere <- return $! formChildren forms
 generateAlphaMapMany fvmap boundHere


generateAlphaMapMany fvmap [] = return $! fvmap
generateAlphaMapMany fvmap (hd:tl) =
 do
 fvmap' <- generateAlphaMap fvmap hd
 generateAlphaMapMany fvmap' tl


alphatizeJustThisTerm ::
 (Sym.HasSymbolTable a) =>
 AlphaMap ->
 (Term ext) ->
 (St.State a (AlphaMap,AlphaMap,Term ext))

alphatizeJustThisTerm fvmap (Ref sym dsAnt) =  
 do
 newSym <- return $! case Map.lookup sym fvmap of 
            Just sym' -> sym'
            Nothing -> sym
 return $! (fvmap,fvmap,Ref newSym dsAnt)

alphatizeJustThisTerm fvmap (Var sym dsAnt) =  
 do
 newSym <- return $! case Map.lookup sym fvmap of 
            Just sym' -> sym'
            Nothing -> sym
 return $! (fvmap,fvmap,Var newSym dsAnt)
 
alphatizeJustThisTerm fvmap term@(Let varss exps body ann) =
 do
 fvmap' <- generateAlphaMapMany fvmap varss
 return $! (fvmap,fvmap',term)

alphatizeJustThisTerm fvmap term@(LetRec varss exps body ann) =
 do
 fvmap' <- generateAlphaMapMany fvmap varss
 return $! (fvmap,fvmap',term)

alphatizeJustThisTerm fvmap term@(Lam vars body ann) =
 do
 fvmap' <- generateAlphaMap fvmap vars
 return $! (fvmap,fvmap',term) 

alphatizeJustThisTerm fvmap term@(Def lhs rhs ann) =
 error $ "Cannot alphatize a program with a define!"


alphatizeJustThisTerm fvmap term = return $! (fvmap,fvmap,term) 



alphatized ::
 (Sym.HasSymbolTable a) =>
 (Term ext) ->
 (St.State a (Term ext))
alphatized = mfoldPreOrder alphatizeJustThisTerm Map.empty

 








 -- [(define v (lambda ... ))]->LetRec* Conversion.

seqConvLetRecJustThis :: Term ext -> Term ext

seqConvLetRecJustThis (Seq [] ann) = (Void ann)

-- seqConvLetRec (Seq [Def vars body _] ann) = body

seqConvLetRecJustThis (Seq exps ann) 
 | not(any isDef exps) = Seq exps ann
 | otherwise = 
  let
   defs2letrec [] = (Void botAnn)
   defs2letrec [Def vars body ann] = 
    LetRec [vars] [body] (Void botAnn) ann
   defs2letrec [body] = body
   defs2letrec more =
    let
     (defs,rest) = span isDef more

     def2clause (Def vars body ann) = (vars,body)
     clauses = map def2clause defs
     (varss,bodies) = unzip clauses

     in LetRec varss bodies (nondefs2seq rest) botAnn
 
   nondefs2seq [] = (Void botAnn)
   nondefs2seq [body] = body
   nondefs2seq more =
    let
     (nondefs,rest) = span (not . isDef) more

     in case rest of 
      [] -> Seq nondefs botAnn
      _ -> Seq (nondefs ++ [defs2letrec rest]) botAnn


  in defs2letrec exps

 
seqConvLetRecJustThis term = term
 

seqConvLetRec = mapPostOrder seqConvLetRecJustThis



 -- Syntactically dead code removal.

synDeadRemoveJustThisTerm :: Term ext -> Term ext
synDeadRemoveJustThisTerm (LetRec varss exps body ann) 
 | any (not . isLam) exps = error $ "Cannot do syntactic dead code removal when expressions in letrec are non-lambdas."
 | otherwise =
 let
   varsexps = zip varss exps

   findUsed rootset newroots [] 
    | newroots `Set.isSubsetOf` rootset = rootset
    | otherwise = findUsed (rootset `Set.union` newroots) Set.empty varsexps
   findUsed rootset newroots ((Forms  [v] [] [] Nothing _,exp):tl) =
     -- Is `v' used in anthing?  If so, its free are now used.
    if varSym v `Set.member` rootset
    then 
     let
       usedInExp = Set.fromList (map fst (Map.toList (free exp)))
      in findUsed rootset (newroots `Set.union` usedInExp) tl
    else findUsed rootset newroots tl

   usedInBody = Set.fromList (map fst (Map.toList (free body)))

   used = findUsed usedInBody Set.empty varsexps

   isUsed sym = Set.member sym used

   varsexps' = filter (\ (Forms [v] [] [] Nothing _,Lam{}) -> isUsed (varSym v))
                varsexps

   (varss',exps') = unzip varsexps'
  in case varsexps' of
   [] -> body -- All died.
   _ -> LetRec varss' exps' body ann

synDeadRemoveJustThisTerm term = term


synDeadRemove :: Term ext -> Term ext
synDeadRemove = mapPostOrder synDeadRemoveJustThisTerm 



 -- Semantically dead call site removal.

semDeadRemoveJustThisTerm ::
 (Set.Set (Term ext)) -> (Term ext) -> (Term ext)

semDeadRemoveJustThisTerm dead term =
 if (term `Set.member` dead)
 then newApp (Prim DEAD botAnn) [term]
 else term


semDeadRemove ::
 (Set.Set (Term ext)) -> (Term ext) -> (Term ext)
semDeadRemove dead term = mapPostOrder (semDeadRemoveJustThisTerm dead) term




 -- Do->Letrec conversion.

doConvJustThisTerm ::
 (Sym.HasSymbolTable a) =>
 (Term ext) ->
 (St.State a (Term ext))

doConvJustThisTerm (Do vars inits steps test body cmds ann) =
 do
 (dovar,doref) <- newVarRef "do"

 let dolambody = 
       Cond [test,litTrue] [body,Seq [cmds,newApp doref steps] botAnn] botAnn

 let dolam = Lam (formalized vars) dolambody botAnn
 
 let dobody = newApp doref inits

 return $! 
  LetRec [formalized [dovar]] [dolam] dobody botAnn

doConvJustThisTerm term =
 return $! term
 

doConv ::
 (Sym.HasSymbolTable a) =>
 (Term ext) ->
 (St.State a (Term ext))
doConv = mmapPostOrder doConvJustThisTerm





 -- LetRec (and Let) conversion.

 -- Lifts out inner letrecs.

 -- WARNING: This assumes alphatization.

letrecConvJustThisTerm :: 
 (Sym.HasSymbolTable a) =>
 (Term ext) ->
 (St.State a (Term ext))

 -- Does not account for a redefine.
letrecConvJustThisTerm (LetRec vars exps body ann) =
 do
 (seq,keep,lift,set) <- return $! splitVarExps [] [] [] [] (zip vars exps)

 setsBody <- return $! genSets (reverse set) body

 seqsBody <- return $! Seq (reverse $! setsBody:seq) ann

 (vars',exps') <- return $! unzip (reverse keep)
 letrec' <- return $! LetRec vars' exps' seqsBody ann

 return $! genLets (reverse lift) letrec'
 where 
  genLets [] body = body
  genLets ((v,e):tl) body = 
   Let [v] [e] (genLets tl body) botAnn
   
  genSets [] body = body
  genSets ((v,e):tl) body = 
   Set v e (genSets tl body) botAnn
   

  splitVarExps seq keep lift set [] =
   (seq,keep,lift,set)

  splitVarExps seq keep lift set ((ve@(v,Lam{})):tl) = 
   splitVarExps seq (ve:keep) lift set tl

  splitVarExps seq keep lift set ((ve@(v,Ref sym ann)):tl)
     -- Is Ref external?  If so, lift it. 
   | let
       pullSym (Forms [Var s _] [] [] Nothing _) = s
       varvs = map pullSym (fst $! unzip tl)
       keepvs = map pullSym (fst $! unzip keep)
       liftvs = map pullSym (fst $! unzip lift)
       setvs = map pullSym (fst $! unzip set)
      in
       not (elem sym varvs) &&
       not (elem sym keepvs) &&
       not (elem sym liftvs) &&
       not (elem sym setvs)
    = splitVarExps seq keep (ve:lift) set tl

  splitVarExps seq keep lift set ((ve@(v,Prim{})):tl) =
   splitVarExps seq keep (ve:lift) set tl

  splitVarExps seq keep lift set ((ve@(v,Lit{})):tl) =
   splitVarExps seq keep (ve:lift) set tl

  splitVarExps seq keep lift set ((ve@(v,LetRec vars' exps' body' ann')):tl) =
   let
     new = (v,body')
     varexps' = (zip vars' exps') ++ [new]
    in splitVarExps seq keep lift set (varexps' ++ tl)

  splitVarExps seq keep lift set ((ve@(v',Let [v] [exp] body' ann')):tl) =
   let
     newlift = (v,unspec)
     newset = (v,exp)
    in splitVarExps seq keep (newlift:lift) (newset:set) ((v',body'):tl)

  splitVarExps seq keep lift set ((ve@(v,Seq [exp] _)):tl) =
   splitVarExps seq keep lift set ((v,exp):tl)

  splitVarExps seq keep lift set ((ve@(v,Seq exps _)):tl) =
   let
    (body:seqs) = reverse exps
    in splitVarExps (seqs ++ seq) keep lift set ((v,body):tl)

  splitVarExps seq keep lift set ve = 
   error $ "Unhandled (v,e) pair in splitVarExps: " ++ (show ve)


letrecConvJustThisTerm (Let [] [] body ann) = return $! body

letrecConvJustThisTerm term@(Let [v] [exp] body ann) = return $! term

letrecConvJustThisTerm term@(Let (var:vars) (exp:exps) body ann) =
 do
 body' <- letrecConvJustThisTerm (Let vars exps body ann) 
 return $! Let [var] [exp] body' ann


 -- Not a Let?  Not a LetRec?  Leave it alone.
letrecConvJustThisTerm term = return $! term



letrecConv ::
 (Sym.HasSymbolTable a) =>
 (Term ext) ->
 (St.State a (Term ext))
letrecConv = mmapPostOrder letrecConvJustThisTerm


 -- Free-variable let-wrapping.

letWrap program =
 let
   freeVars :: [Sym.Symbol]
   freeVars = map fst (Map.toList $ free program)

   genVarExp :: Sym.Symbol -> (Term ext, Term ext)
   genVarExp sym = (Forms [Var sym botAnn] [] [] Nothing botAnn,unspec)

   (varss,exps) = unzip $! map genVarExp freeVars

  in case varss of
   [] -> program -- No top-level free vars.
   _ -> Let varss exps program botAnn





 -- and/or->cond Conversion.

andOrConvJustThisTerm ::
 (Sym.HasSymbolTable a) =>
 (Term ext) ->
 (St.State a (Term ext))
andOrConvJustThisTerm (App (Prim AND _) (Args ordered [] [] Nothing _) _) =
 convAnd ordered
 where
  convAnd :: 
   (Sym.HasSymbolTable a) =>
   [Term ext] ->
   (St.State a (Term ext))
  convAnd [] = return $! litTrue
  convAnd [term] = return $! term
  convAnd (term:tl) =
   do 
   tl' <- convAnd tl
   return $! newIf term tl' litFalse


andOrConvJustThisTerm (App (Prim OR _) (Args ordered [] [] Nothing _) _) =
 convOr ordered
 where
  convOr :: 
   (Sym.HasSymbolTable a) =>
   [Term ext] ->
   (St.State a (Term ext))
  convOr [] = return $! litFalse
  convOr [term] = return $! term
  convOr (term:tl) =
   do 
   newLet "orop" term (\ termref ->
    do 
    tl' <- convOr tl
    return $! newIf termref termref tl')

andOrConvJustThisTerm term = return $! term


andOrConv ::
 (Sym.HasSymbolTable a) =>
 (Term ext) ->
 (St.State a (Term ext))
andOrConv = mmapPostOrder andOrConvJustThisTerm



     


 -- case->cond Conversion.


 -- TODO: Remove this eventually, and handle it in GCFA.


case2condJustThisTerm ::
 (Sym.HasSymbolTable a) =>
 (Term ext) ->
 (St.State a (Term ext))
case2condJustThisTerm (Case exp litss thens ann) =
 newLet "discrim" exp (\ exp' ->
  do
  conds' <- return $! map (lits2cond exp') litss
  return $! Cond conds' thens ann)
 where
  lits2cond :: (Term ext) -> (Maybe [Term ext]) -> (Term ext)
  lits2cond exp Nothing = litTrue
  lits2cond exp (Just lits) = 
   newApp (Prim MEMV botAnn) ((exp):[newApp (Prim LIST botAnn) lits])

case2condJustThisTerm term = return $! term



case2cond ::
 (Sym.HasSymbolTable a) =>
 (Term ext) ->
 (St.State a (Term ext))
case2cond = mmapPostOrder case2condJustThisTerm





 -- Administrative Normal Form (ANF) Conversion.

isANFAtom (Lit{}) = True
isANFAtom (Ref{}) = True
isANFAtom (Prim{}) = True
isANFAtom (Void{}) = True
isANFAtom _ = False


isANFApp (App fun (Args ordered labels labeled rest argsann) ann) =
 all isANFAtom (fun:((Maybe.maybeToList rest) ++ ordered ++ labeled))
isANFApp _ = False



 -- WARNING: A-Normalization EXPECTS alphatization.

anormalize :: (Sym.HasSymbolTable a) => 
 Term ext ->
 (Term ext -> St.State a (Term ext)) ->
 St.State a (Term ext)


anormalize term@(Ref sym ann) k =
 k term

anormalize term@(Prim prim ann) k =
 k term

anormalize term@(Lit (LitString string) ann) k =
 newLet "str" term k

anormalize term@(Lit lit ann) k =
 k term

anormalize term@(Void ann) k = 
 k term

 -- TODO: Let-bind this.
anormalize term@(Lam forms body ann) k =
 do
 body' <- anormalized body
 term' <- return $! Lam forms body' ann
 newLet "lam" term' k

 
anormalize term@(LetRec vars exps body ann) k 
 | all (\ exp ->
    case exp of
      -- Lambda terms are always valid.
     Lam{} -> True

      -- A Ref is OK, if it doesn't bind something in `vars'.
     Ref sym _ -> not (sym `elem` (map varSym (formChildren term)))

     Lit lit _ -> True

     _ -> False
     ) exps =
   do
   exps' <- Might.mmap anormalized exps
   body' <- anormalize body k
   return $! LetRec vars exps' body' ann
 | otherwise =
   error $ "Complex letrec not yet handled, exps:\n\n" ++ 
    (Might.showWithSep "\n\n" exps)

anormalize term@(Seq ((Set var exp (Void _) setann):tl) ann) k =
 anormalize exp (\ exp' ->
  do
  body' <- anormalize (Seq tl ann) k
  return $! Set var exp' body' setann)

anormalize term@(Set var exp body ann) k =
 anormalize exp (\ exp' ->
  do
  body' <- anormalize body k
  return $! Set var exp' body' ann)


anormalize 
 term@(Let [v] [App f (Args ordered labels labeled rest argsann) appann] 
        body ann) k = 
 do 
 anormalize f (\ f' ->
  anormalizeMany ordered (\ ordered' ->
   anormalizeMany labeled (\ labeled' ->
    anormalizeOption rest (\ rest' ->
     do
     body' <- anormalize body k
     return $! 
      Let [v] [App f' (Args ordered' labels labeled' rest' argsann) appann] 
       body' ann))))

anormalize term@(Let [v] [exp] body ann) k 
 | isANFApp exp || isANFAtom exp =
    do 
    -- WARNING: `k [ ]' better not have `v' free or bind `v':
    -- TODO: Have this check whether it binds v, too.
--     k_hole <- k (Void botAnn)
--     False <- return $! 
--      let
--        k_hole' = annotateFree k_hole
--        free_in_k_hole' = free k_hole'
--       in any (\ v -> (varSym v) `Map.member` free_in_k_hole') (formChildren term)
							       
    -- We're cool.  Continue.
    body' <- anormalize body k
    return $! Let [v] [exp] body' ann

 | otherwise =
    do
    anormalize exp (\ exp' ->
     do
     body' <- anormalize body k
     return $! Let [v] [exp'] body' ann)
      
   

 
anormalize term@(Cond [] [] ann) k =
 k (Void botAnn)

anormalize term@(Cond [cond] [thn] ann) k = 
 anormalize cond (\ cond' ->
  do
  thn' <- anormalized thn
  k $! Cond [cond'] [thn'] ann)

anormalize term@(Cond (cond:(Lit (LitBool True) _):_) (thn:els:_) ann) k =
 anormalize cond (\ cond' ->
  do
  thn' <- anormalized thn
  els' <- anormalized els
  k $! Cond [cond',litTrue] [thn',els'] ann)

anormalize term@(Cond (cond:conds) (thn:thns) ann) k =
 anormalize cond (\ cond' ->
  do
  thn' <- anormalized thn
  els <- anormalized (Cond conds thns ann) 
  k $! Cond [cond',litTrue] [thn',els] ann)


anormalize term@(Case exp [] [] ann) k =
 anormalize exp (\ exp' ->
  k $! (Seq [exp',Void botAnn] ann))

anormalize term@(Case exp litss thens ann) k =
 anormalize exp (\ exp' ->
  do
  thens' <- Might.mmap anormalized thens
  k $! (Case exp' litss thens' ann))

anormalize term@(Seq [] ann) k = 
 k (Void botAnn)

anormalize term@(Seq [exp] ann) k =
 anormalize exp k

anormalize term@(Seq exps ann) k =
 do
 (effexps,valexp) <- return $!
  let
    revexps = reverse exps
    valexp = head revexps
    reveffexps = tail revexps
   in (reverse reveffexps,valexp)
 effexps' <- Might.mmap anormalized effexps
 newlet <- anormalize valexp k
 return $! Seq (effexps' ++ [newlet]) botAnn


anormalize term@(App fun (Args ordered labels labeled rest argsann) ann) k =
 anormalize fun (\ fun' ->
   anormalizeMany ordered (\ ordered' ->
    anormalizeMany labeled (\ labeled' ->
     anormalizeOption rest (\ rest' ->
      do
       newapp <-
        return $! App fun' (Args ordered' labels labeled' rest' argsann) ann
       newLet "v" newapp k))))



anormalize term k = 
 error $ "anormalize case not handled: " ++ (show term)




anormalizeMany :: (Sym.HasSymbolTable a) => 
 [Term ext] ->
 ([Term ext] -> St.State a (Term ext)) ->
 St.State a (Term ext)

anormalizeMany = Might.cpsLiftMany anormalize


anormalizeOption :: (Sym.HasSymbolTable a) => 
 Maybe (Term ext) ->
 ((Maybe (Term ext)) -> St.State a (Term ext)) ->
 St.State a (Term ext)

anormalizeOption = Might.cpsLiftOption anormalize



anormalized term@(App fun (Args ordered labels labeled rest argsann) ann) =
 anormalize fun (\ fun' ->
  anormalizeMany ordered (\ ordered' ->
   anormalizeMany labeled (\ labeled' ->
    anormalizeOption rest (\ rest' ->
     return $! App fun' (Args ordered' labels labeled' rest' argsann) ann))))


anormalized term@(Lam forms body ann) =
 do
 body' <- anormalized body
 return $! Lam forms body' ann

 -- TODO: Handle things more intelligently here.  (More cases.)
anormalized term = anormalize term return







 -- Continuation-Passing Style (CPS) Conversion.


 -- continuate only works on expressions which return a single value,
 -- e.g. arguments to an App.

 -- continuateWith works on expressions which might return multiple
 -- values, e.g. the body of a lambda term.


isCPSAtom (Lit{}) = True
isCPSAtom (Void{}) = True
isCPSAtom (Ref{}) = True
isCPSAtom _ = False

 -- Some "primops" have non-primop-like behavior, e.g. apply.
isCPSPrimApp (App (Prim APPLY _) args@(Args{}) ann)  = False
isCPSPrimApp (App (Prim CALLCC _) args@(Args{}) ann)  = False

isCPSPrimApp (App (Prim{}) args@(Args{}) ann) 
 | all isCPSAtom (expChildren args) = True
isCPSPrimApp _ = False




continuate :: 
 (Sym.HasSymbolTable a) =>
 (Term ext) ->
 ((Term ext) -> (St.State a (Term ext))) ->
 (St.State a (Term ext))

continuate term@(Lit lit ann) k = 
 k term

continuate term@(Ref sym ann) k =
 k term

continuate term@(Prim CALLCC ann) k =
 do 
 term' <- continuateAtom term
 k $! term'
 

continuate term@(Prim prim ann) k =
 k $! Prim (CPS prim) ann

continuate term@(Void ann) k =
 k $! term

continuate term@(Lam{}) k = 
 do
 lam' <- continuateAtom term
 k $! lam'  
 
continuate term@(App prim@(Prim primop _) 
                       (Args ordered labels labeled rest argsann) ann) k
 | not(primop == APPLY) && not(primop == CALLCC) =
 continuateMany ordered (\ ordered' ->
  continuateMany labeled (\ labeled' ->
   continuateOption rest (\ rest' ->
    do
    (u,e) <- newVarRef "up"
    body <- k e
    app <-
     return $! App prim (Args ordered' labels labeled' rest' argsann) botAnn
    return $! Let [formalized [u]] [app] body ann)))
 
continuate term@(App fun (Args ordered labels labeled rest argsann) ann) k =
 continuateMany ordered (\ ordered' ->
  continuateMany labeled (\ labeled' ->
   continuateOption rest (\ rest' ->
    continuate fun (\ fun' ->
    do
     q <- newKont k
     return $! App fun' (Args ordered' (Sym.cont':labels) (q:labeled') rest' argsann) botAnn))))

 
continuate term@(Let [v] [exp] body ann) k 
 | isCPSPrimApp exp = 
   do
   body' <- continuate body k
   return $! Let [v] [exp] body' ann
 | isSingularForm v =
   do
   continuate exp (\ exp' ->
    do
    body' <- continuate body k
    return $! Let [v] [exp'] body' ann)
 | otherwise = 
    error $ "note sure how to continuate this"

continuate term@(LetRec varss exps body ann) k 
 -- Assume exps are all lambda terms.
 | all isLam exps =
   do
   exps' <- Might.mmap continuateAtom exps
   body' <- continuate body k
   return $! LetRec varss exps' body' ann

 | otherwise = 
   error $ "Non-lambda terms in letrec:\n" ++ (Might.showWithSep "\n\n" exps)


continuate term@(Cond [cond,Lit (LitBool True) _] [thn,els] ann) k =
 continuate cond (\ cond' ->
  do
  (k',q') <- newContVarRef "j"
  kont <- newJoinKont k
  thn' <- continuateWith thn q'
  els' <- continuateWith els q'
  k'' <- return $! formalized [k']
  return $! Let [k''] [kont] (Cond [cond',litTrue] [thn',els'] ann) ann)


continuate term k = 
 error $ "continuate unhandled case: " ++ (show term)




continuateMany :: (Sym.HasSymbolTable a) => 
 [Term ext] ->
 ([Term ext] -> St.State a (Term ext)) ->
 St.State a (Term ext)

continuateMany = Might.cpsLiftMany continuate


continuateOption :: (Sym.HasSymbolTable a) => 
 Maybe (Term ext) ->
 ((Maybe (Term ext)) -> St.State a (Term ext)) ->
 St.State a (Term ext)

continuateOption = Might.cpsLiftOption continuate






    
continuateAtom ::
 (Sym.HasSymbolTable a) =>
 (Term ext) ->
 (St.State a (Term ext))
 
continuateAtom
 term@(Lam (Forms ordered labels labeled rest formsann) body ann) =
 do
 (k',q') <- newContVarRef "k"
 forms' <- return $! Forms ordered ((Sym.cont'):labels) (k':labeled) rest formsann
 body' <- continuateWith body q'
 return $! Lam forms' body' ann
 

 -- call/cc = (lambda (f cc) (f (lambda (x k) (cc x)) cc))
continuateAtom term@(Prim CALLCC ann) =
 do
 (ccvar,ccref) <- newContVarRef "cc"
 (fvar,fref) <- newVarRef "f"
 (xvar,xref) <- newVarRef "x"
 (kvar,_) <- newContVarRef "k"
 let app_cc_x = newApp ccref [xref]
 let lam_x_k = Lam (formalizedWithKont [xvar] kvar) app_cc_x botAnn
 let app_f_lam_cc = newAppWithCont fref [lam_x_k] ccref
 let lam_f_cc = Lam (formalizedWithKont [fvar] ccvar) app_f_lam_cc botAnn
 return $! lam_f_cc


 -- continuateWith assumes that it is safe to duplicate q.
continuateWith ::
 (Sym.HasSymbolTable a) =>
 (Term ext) ->
 (Term ext) -> 
 (St.State a (Term ext))

continuateWith term q 
 | not (isContTerm q) = continuateWith term (markCont q)

continuateWith term@(Prim CALLCC _) q =
 do
 term' <- continuateAtom term
 return $! newApp q [term']

continuateWith term q 
 | isCPSAtom term =
   do
   return $! newApp q [term]
   
continuateWith term@(Lam forms body ann) q = 
 do
 lam' <- continuateAtom term
 return $! newApp q [lam']

continuateWith (Prim prim ann) q =
 return $! newApp q [Prim (CPS prim) ann]


continuateWith term@(LetRec varss exps body ann) q
 -- Assume exps are all lambda terms.
 | all isLam exps =
   do
   exps' <- Might.mmap continuateAtom exps
   body' <- continuateWith body q
   return $! LetRec varss exps' body' ann

 | otherwise = 
   error $ "Non-lambda terms in letrec:\n" ++ (Might.showWithSep "\n\n" exps)

continuateWith (Seq [] ann) q = 
 do
 return $! newApp q [void] 

continuateWith term@(Seq [exp] ann) q =
 continuateWith exp q

continuateWith (Seq (hd:tl) ann) q 
 | isCPSPrimApp hd = 
   do 
   tl' <- continuateWith (Seq tl ann) q
   return $! Seq [hd,tl'] ann
-- | (case hd of
--     Ref sym _ -> Sym.toString sym == "DEAD-VAL"
--     _ -> False) = error "converting DEAD-VAL"
 | otherwise = 
   do
   (unused,_) <- newVarRef "_" 
   callRemainder <- continuateWith (Seq tl ann) q
   cont <- return $! Lam (Forms [] [] [] (Just unused) botAnn) callRemainder botAnn
   continuateWith hd cont

continuateWith (Set var exp body ann) q = 
 continuate exp (\ exp' ->
  do
  body' <- continuateWith body q
  return $! Set var exp' body' ann)


 
 -- Pattern match on q!

continuateWith term@(App prim@(Prim primop _) 
                       (Args ordered labels labeled rest argsann) ann) 
                       (Lam (Forms [] [] [] (Just (Var restSym _)) _) body _) 
 | not(primop == APPLY) && not(primop == CALLCC) && restSym `isUnusedIn` body =
 continuateMany ordered (\ ordered' ->
  continuateMany labeled (\ labeled' ->
   continuateOption rest (\ rest' ->
    do
    app <-
     return $! App prim (Args ordered' labels labeled' rest' argsann) botAnn
    return $! Seq [app,body] ann)))


continuateWith term@(App prim@(Prim primop _) 
                       (Args ordered labels labeled rest argsann) ann) 
                       (Lam v body _) 
 | not(primop == APPLY) && not(primop == CALLCC) =
 continuateMany ordered (\ ordered' ->
  continuateMany labeled (\ labeled' ->
   continuateOption rest (\ rest' ->
    do
    app <-
     return $! App prim (Args ordered' labels labeled' rest' argsann) botAnn
    return $! Let [v] [app] body ann)))

continuateWith term@(App prim@(Prim primop _) 
                       (Args ordered labels labeled rest argsann) ann) q 
 | not(primop == APPLY) && not(primop == CALLCC) =
 continuateMany ordered (\ ordered' ->
  continuateMany labeled (\ labeled' ->
   continuateOption rest (\ rest' ->
    do
    (u,e) <- newVarRef "up"
    body <- return $! newApp q [e]
    app <-
     return $! App prim (Args ordered' labels labeled' rest' argsann) botAnn
    return $! Let [Forms [u] [] [] Nothing botAnn] [app] body ann)))
 
continuateWith term@(App fun (Args ordered labels labeled rest argsann) ann) q =
 continuateMany ordered (\ ordered' ->
  continuateMany labeled (\ labeled' ->
   continuateOption rest (\ rest' ->
    continuate fun (\ fun' ->
    do
     return $! App fun' (Args ordered' (Sym.cont':labels) (q:labeled') rest' argsann) botAnn))))


continuateWith term@(Let [v] [exp] body ann) q
 | isCPSPrimApp exp = 
   do
   body' <- continuateWith body q
   return $! Let [v] [exp] body' ann
 | isSingularForm v =
   do
   continuate exp (\ exp' ->
    do
    body' <- continuateWith body q
    return $! Let [v] [exp'] body' ann)
 | otherwise = 
    error $ "note sure how to continuateWith this: " ++ (show term)


 -- For conditionals, we mark join points.
continuateWith term@(Cond [cond] [thn] ann) q@(Ref{}) = 
 continuate cond (\ cond' ->
  do
  q <- return $! markJoinPoint q
  thn' <- continuateWith thn q
  els' <- return $! newApp q [Void botAnn] 
  return $! (Cond [cond',litTrue] [thn',els'] ann))



continuateWith term@(Cond [cond] [thn] ann) q@(Lam{}) = 
 continuate cond (\ cond' ->
  do
  q <- return $! markJoinPoint q
  (k',q') <- newContVarRef "j"
  thn' <- continuateWith thn q'
  els' <- return $! newApp q' [Void botAnn] 
  -- Save the join point in k'.
  k'' <- return $! formalized [k']
  return $! Let [k''] [q] (Cond [cond',litTrue] [thn',els'] ann) ann)

  

continuateWith term@(Cond [cond,Lit (LitBool True) _] [thn,els] ann)
  q@(Ref{}) =
 continuate cond (\ cond' ->
  do
  q <- return $! markJoinPoint q
  thn' <- continuateWith thn q
  els' <- continuateWith els q
  -- Save the join point in k'.
  return $! (Cond [cond',litTrue] [thn',els'] ann))


continuateWith term@(Cond [cond,Lit (LitBool True) _] [thn,els] ann) 
 q@(Lam{}) =
 continuate cond (\ cond' ->
  do
  q <- return $! markJoinPoint q
  (k',q') <- newContVarRef "j"
  thn' <- continuateWith thn q'
  els' <- continuateWith els q'
  -- Save the join point in k'.
  k'' <- return $! formalized [k']
  return $! Let [k''] [q] (Cond [cond',litTrue] [thn',els'] ann) ann)




continuateWith term q = 
 error $ "continuateWith unhandled case: " ++ (show term)
 

continuated term = continuateWith term (Prim HALT botAnn)