 -- Compiler.Symbol
 --
 -- A module for symbols and symbol tables.
module Compiler.Symbol where

import Char (digitToInt,isDigit)
import qualified Control.Monad.State as St
import qualified Data.Map as Map


import qualified Might

 -- Symbols.
data Symbol 
    = Sym { toCode :: !Int, toString :: !String }
    | SymTop
    | SymBot

instance Eq Symbol where
 SymBot == SymBot = True
 SymTop == SymTop = True
 (Sym i s) == (Sym i' s') = i == i'
 _ == _ = False

instance Show Symbol where
 show (Sym _ s) = s
 show SymTop = "TOP SYMBOL!!!!"
 show SymBot = "BOT SYMBOL!!!!"
 
instance Ord Symbol where
 compare SymBot SymBot = EQ
 compare SymBot _ = LT
 compare _ SymBot = GT
 compare (Sym i s) (Sym i' s') = compare i i'
 compare SymTop SymTop = EQ 
 compare _ SymTop = LT
 compare SymTop _ = GT


instance Might.Hashable Symbol where
 hash SymTop = 2 
 hash SymBot = 3 
 hash (Sym i s) = toInteger(i) + 3

 hashM m sym = (fromInteger(Might.hash sym)) `mod` m


fromString string = Sym (0-1) string


 -- Symbol tables.
data SymbolTable = SymTab { symbolMax :: !Int, 
			    symbolMap :: !(Map.Map String Symbol) }


 -- Some functions are lifted to objects containing symbol tables.
class HasSymbolTable a where
 extractSymbolTable :: a -> SymbolTable
 injectSymbolTable :: a -> SymbolTable -> a

instance HasSymbolTable SymbolTable where
 extractSymbolTable s = s
 injectSymbolTable _ s = s




 -- Low-level symbol table primitives.
lookup string symTab = Map.lookup string (symbolMap symTab)

insertLookup string (SymTab max tbl) = 
 case Map.insertLookupWithKey (\ k new old -> old) string sym' tbl of
  (Just sym,tbl) -> (sym,SymTab max tbl)
  (Nothing,tbl) -> (sym',SymTab (max+1) tbl)
 where sym' = Sym { toCode = (max+1), toString = string }


genEmpty :: St.State SymbolTable ()
genEmpty = 
 do 
    symbolize "init"

    symbolize "src"
    symbolize "lambda"
    symbolize "let"
    symbolize "letrec"
    symbolize "undefined"
    symbolize "define"
    symbolize "transformer"
    symbolize "define-syntax"
    symbolize "let-syntax"
    symbolize "syntax-rules"

    symbolize "cond"
    symbolize "case"

    symbolize "vector"
    symbolize "block-quote"
    symbolize "quote"
    symbolize "colon"
    symbolize "dot"

    symbolize "car"
    symbolize "cdr"
    symbolize "cons"
    symbolize "@"

    symbolize "."

    symbolize "env"
    symbolize "cont"
    symbolize "hnd"
    symbolize "rest"
    return $! ()


 -- An empty symbol table.
empty :: SymbolTable
empty = St.execState genEmpty (SymTab 0 Map.empty) 


 -- Predefined symbols.

init' = (symbolMap empty) Map.! "init"


src' = (symbolMap empty) Map.! "src"
lambda' = (symbolMap empty) Map.! "lambda"
let' = (symbolMap empty) Map.! "let"
letrec' = (symbolMap empty) Map.! "let"
undefined' = (symbolMap empty) Map.! "undefined"
define' = (symbolMap empty) Map.! "define"
definesyntax' = (symbolMap empty) Map.! "define-syntax"
letsyntax' = (symbolMap empty) Map.! "let-syntax"
transformer' = (symbolMap empty) Map.! "transformer"
syntaxrules' = (symbolMap empty) Map.! "syntax-rules"

cond' = (symbolMap empty) Map.! "cond"
case' = (symbolMap empty) Map.! "case"


vector' = (symbolMap empty) Map.! "vector"
block_quote' = (symbolMap empty) Map.! "block-quote"
quote' = (symbolMap empty) Map.! "quote"
colon' = (symbolMap empty) Map.! "colon"
dot' = (symbolMap empty) Map.! "dot"

car' = (symbolMap empty) Map.! "car"
cdr' = (symbolMap empty) Map.! "cdr"
cons' = (symbolMap empty) Map.! "cons"
at_symbol = (symbolMap empty) Map.! "@"

objtype' = (symbolMap empty) Map.! "."


env' = (symbolMap empty) Map.! "env"
cont' = (symbolMap empty) Map.! "cont"
hnd' = (symbolMap empty) Map.! "hnd"
rest' = (symbolMap empty) Map.! "rest"




 -- The following functions are designed to work within state monads
 -- that contain symbol table.

symbolize :: (HasSymbolTable a) => String -> St.State a Symbol
 -- Turns a string into a symbol.
symbolize string = 
 do state <- St.get
    st <- return $! extractSymbolTable state
    (sym,st) <- return $! insertLookup string st
    St.put $! injectSymbolTable state st
    return $! sym


internalize :: (HasSymbolTable a) => Symbol -> St.State a Symbol
 -- Returns the canonical symbol for this string.
internalize (Sym code string)  = 
 do state <- St.get
    st <- return $! extractSymbolTable state
    (sym,st) <- return $! insertLookup string st
    St.put $! injectSymbolTable state st
    return $! sym


 -- TODO: Rename to genSyms
genSyms [] = return $! []
 -- Returns a list of fresh symbols, one for each string.
genSyms (hd:tl) = 
 do hd' <- genSym (toString hd)
    tl' <- genSyms tl
    return $! hd':tl'


genSym :: (HasSymbolTable a) => String -> St.State a Symbol
 -- Returns a fresh symbol based on this string.
genSym string =
 do state <- St.get
    st <- return $! extractSymbolTable state
    case Compiler.Symbol.lookup string st of
     Just sym -> let (pre,num) = trailingInt (toString sym)
                  in genSym (pre ++ (show (num + (symbolMax st))))
     Nothing -> symbolize string



 -- TODO: Put this in a separate lib for strings.
genTrailingInt pre num [] = (reverse pre,reverse num)
genTrailingInt pre num (c:tl)
 | isDigit c = genTrailingInt (pre) (c:num) tl
 | True = genTrailingInt (num ++ c:pre) "" tl


trailingInt :: String -> (String,Int)
trailingInt string = 
 case genTrailingInt "" "" string of
  (pre,"") -> (pre,0)
  (pre,num) -> (pre,atoi 0 num)

charToDigit '0' = 0
charToDigit '1' = 1
charToDigit '2' = 2
charToDigit '3' = 3
charToDigit '4' = 4
charToDigit '5' = 5
charToDigit '6' = 6
charToDigit '7' = 7
charToDigit '8' = 8
charToDigit '9' = 9
charToDigit c = error $ "Compiler.Symbol.charToDigit: No such digit: " ++ (show c)

atoi num [] = num
atoi num (c:tl) = atoi ((num * 10) + (charToDigit c)) tl
