module Might where

import qualified Data.List as List
import qualified Data.Maybe as Maybe
import Data.Map as Map
import Data.Set as Set
import Control.Monad.State

import qualified Char as Char
import qualified Int as Int

import System.IO.Unsafe

 -- Syntactic sugar.

(==>) a b = (a,b)

 -- Data.Map

insertAll :: Ord k => Map k a -> [(k,a)] -> Map k a 
insertAll map lst =
 Map.unionWith (\ l r -> r) map (Map.fromList lst) 


 -- Data.Set





 -- Debugging aids.

blank = error "blank" 
todo s = error ("todo: " ++ s)
zipFail errmsg l1 l2 = if length l1 == length l2 
                       then zip l1 l2
                       else error $ "zipFail: " ++ errmsg



 -- Function manipulation.

except :: Eq a => (a -> b) -> (a,b) -> (a -> b)
except f (x,y) = \ x' -> if x == x' then y else f x'

fix init f = if f init == init
             then init 
             else f (f init)


 -- List manipulation.

foldOp f init [] =  init
foldOp f init (hd:tl) = foldOp f (f init hd) tl

appendReverse [] ans = ans
appendReverse (hd:tl) ans = appendReverse tl (hd:ans)

flattenListOfSets :: [Set.Set a] -> [a]
flattenListOfSets lst = 
 foldl (\ tally set -> (Set.toList set) ++ tally) [] lst


flattenListOfLists :: [[a]] -> [a]
flattenListOfLists [] = []
flattenListOfLists ([]:tl) =
 flattenListOfLists tl 
flattenListOfLists ((first:rest):tl) = 
 first:(flattenListOfLists (rest:tl))

 -- Foldable.

class Foldable m where 
 xfold :: (a -> b -> b) -> b -> m a -> b


instance Foldable [] where
 xfold = foldr

instance Foldable Set.Set where
 xfold = Set.fold
 

 -- Monad utilities.

doFoldAction :: (Monad m, Foldable f) => f a -> (a -> m ()) -> m ()
doFoldAction foldable action = 
 xfold (\ a mon -> mon >> action a) (return ()) foldable


 -- State monad extensions.

retrieve property = 
 do state <- get
    return $ property state

perform f =
 do state <- get
    put $ f state

showWithSep :: Show a => String -> [a] -> String
showWithSep sep [] = ""
showWithSep sep [last] = show last
showWithSep sep (hd:tl) = (show hd) ++ sep ++ (showWithSep sep tl)

showWithSepWith ::
 (a -> String) ->
 String -> [a] -> String
showWithSepWith myshow sep [] = ""
showWithSepWith myshow sep [last] = myshow last
showWithSepWith myshow sep (hd:tl) =
 (myshow hd) ++ sep ++
 (showWithSepWith myshow sep tl)


mapMaybe :: (a -> Maybe b) -> [a] -> [b]
mapMaybe f [] = []
mapMaybe f (hd:tl) = 
 case f hd of
  Just b -> b:(Might.mapMaybe f tl)
  Nothing -> Might.mapMaybe f tl


mmap :: Monad m => (a -> m b) -> [a] -> m [b]
mmap f [] = return $! []
mmap f (hd:tl) = 
 do
 hd' <- f hd 
 tl' <- mmap f tl
 return $! hd':tl'


cpsLiftMany f [] k = k []
cpsLiftMany f (hd:tl) k =
 f hd (\ hd' ->
  (cpsLiftMany f) tl (\ tl' ->
   k $! hd':tl'))

cpsLiftOption f Nothing k = k Nothing
cpsLiftOption f (Just val) k = 
 f val (\ val' ->
  k $! Just val')

maybeLiftM :: 
 (Monad m) =>
 (a -> m b) ->
 (Maybe a -> m (Maybe b))
maybeLiftM f (Just a) = 
 do
 a' <- f a
 return $! Just a'
maybeLiftM f (Nothing) = return $! Nothing
 

class Hashable a where
 hash :: a -> Integer
 hashM :: Int -> a -> Int

 hashM m a = (fromInteger (hash a)) `mod` m


instance Hashable Int where
 hash i = toInteger i

instance Hashable Char where
 hash c = toInteger (Char.ord c)

instance Hashable Bool where
 hash True = 1
 hash False = 2
 
instance (Hashable a) => Hashable [a] where
 hash lst = foldl (\ h a -> h * (hash a)) 1 lst
 hashM m lst = (foldl (\ h a -> (h * (hashM m a)) `mod` m) 1 lst) `mod` m

instance (Hashable k, Hashable a) => Hashable (Map k a) where
 hash map = 
  foldl (\ h (k,a)  -> h * ((hash k) * (hash a))) 1 (Map.toList map)

 hashM m map = 
  (foldl (\ h (k,a)  -> (h * ((hashM m k) * (hashM m a))) `mod` m) 1 (Map.toList map)) `mod` m

instance (Hashable k) => Hashable (Set k ) where
 hash map = 
  foldl (\ h k  -> h * hash k) 1 (Set.toList map)

 hashM m map = 
  (foldl (\ h k  -> (h * hashM m k) `mod` m) 1 (Set.toList map)) `mod` m
 
instance (Hashable a, Hashable b) => Hashable (a,b) where
 hash (a,b) = hash a * hash b

 hashM m (a,b) = (hashM m a * hashM m b) `mod` m

instance (Hashable a, Hashable b, Hashable c) =>  
  Hashable (a,b,c) where
 hash (a,b,c) = hash a * hash b * hash c

 hashM m (a,b,c) = (hashM m a * hashM m b * hashM m c) `mod` m

instance (Hashable a, Hashable b, Hashable c, Hashable d) =>
  Hashable (a,b,c,d) where
 hash (a,b,c,d) = hash a * hash b * hash c * hash d

 hashM m (a,b,c,d) = (hashM m a * hashM m b * hashM m c * hashM m d) `mod` m

instance (Hashable a, Hashable b, Hashable c, Hashable d, Hashable e) =>
  Hashable (a,b,c,d,e) where 
 hash (a,b,c,d,e) = hash a * hash b * hash c * hash d * hash e

 hashM m (a,b,c,d,e) = (hashM m a * hashM m b * hashM m c * hashM m d * hashM m e) `mod` m




 -- Debugging I/O.

debug :: String -> Bool
debug s = unsafePerformIO 
 (do
  putStrLn $! "debug: " ++ s
  return $! True)

debugM :: (Monad m) => String -> m Bool
debugM s = return $! unsafePerformIO 
 (do
  putStrLn $! "debug: " ++ s
  return $! True)

combineMaps ::
 (Ord k) => 
 (k -> (Maybe a) -> (Maybe b) -> Maybe c) ->
 Map.Map k a -> 
 Map.Map k b ->
 Map.Map k c
combineMaps f ma mb =
 let
   la = Map.toAscList ma
   lb = Map.toAscList mb

   keyMaybeCtoList k (Just c) = [(k,c)]
   keyMaybeCtoList k Nothing = []

   merge [] [] = []
   merge ((ka,a):tla) [] = 
    (keyMaybeCtoList ka $ f ka (Just a) Nothing) ++ (merge tla [])
   merge [] ((kb,b):tlb) = 
    (keyMaybeCtoList kb $ f kb Nothing (Just b)) ++ (merge [] tlb)
   merge la@((ka,a):tla) lb@((kb,b):tlb) =
    case compare ka kb of
     EQ -> (keyMaybeCtoList ka (f ka (Just a) (Just b)))
             ++ (merge tla tlb)
     LT -> (keyMaybeCtoList kb (f ka (Just a) Nothing))
             ++ (merge tla lb)
     GT -> (keyMaybeCtoList kb (f kb Nothing (Just b)))
             ++ (merge la tlb)

  in Map.fromAscList $ merge la lb


isSubmapOfByOption :: 
 (Ord k) =>
 (Maybe a -> Maybe b -> Bool) ->
 (Map k a) ->
 (Map k b) ->
 Bool
isSubmapOfByOption lt f g =
 let
   la = Map.toAscList f
   lb = Map.toAscList g

    -- A "merging" comparison.
   comp []              [] = True
   comp (hd:tl)         [] = False
   comp []              (hd:tl) = True
   comp la@((ka,a):tla) lb@((kb,b):tlb) =
    case compare ka kb of 
     EQ -> lt (Just a) (Just b) && comp tla tlb
     LT -> lt (Just a) Nothing && comp tla lb
     GT -> lt Nothing (Just b) && comp la tlb
  in comp la lb
 