{
module Parser.CParser where

import qualified Control.Monad.State as St
import qualified Char (isSpace,isAlpha,isDigit)

import qualified Data.Set as Set 

data ParserState = 
 PS { filename :: !String ,
      col :: !Int ,
      line :: !Int ,
      typedefs :: ![Set.Set String] ,
      input :: !String }

type ParserMonad = St.State ParserState


getPosition :: St.State ParserState Pos 
getPosition = 
 do
 pstate <- St.get
 return $! Pos (filename pstate) (line pstate) (col pstate)


isTypedefName :: String -> ParserMonad Bool
isTypedefName str = 
 do
 pstate <- St.get
 tydefs <- return $! typedefs pstate
 return $!
  any (\ set -> str `Set.member` set) tydefs

processIdent :: String -> ParserMonad Token
processIdent str = 
 do
 isTyDef <- isTypedefName str
 return $!
  if isTyDef 
  then TokenIdentifier str
  else TokenTypedefName str

processKeyword cont tok = 
 do
 advanceCol $! length (show tok)
 cont $! tok

 

}

%name cparse
%tokentype { Token }

%monad { St.State ParserState }
%lexer { lexer } { TokenEOF }


%token 
      auto            { TokenAuto }
      register        { TokenRegister }
      static          { TokenStatic }
      extern          { TokenExtern }
      typedef         { TokenTypedef }

      void            { TokenVoid }
      char            { TokenChar }
      short           { TokenShort }
      int             { TokenInt }
      long            { TokenLong }
      float           { TokenFloat }
      double          { TokenDouble }
      signed          { TokenSigned }
      unsigned        { TokenUnsigned }

      const           { TokenConst }
      volatile        { TokenVolatile } 

      struct          { TokenStruct }
      union           { TokenUnion } 

      enum            { TokenEnum } 

      identifier      { TokenIdentifier $$ }
      typedef_name    { TokenTypedefName $$ }

      '='             { TokenEq }
      '+'             { TokenPlus }
      '-'             { TokenMinus }
      '*'             { TokenStar }
      '/'             { TokenDiv }

      ';'             { TokenSemicolon }
      ':'             { TokenColon }
      ','             { TokenComma }
      '.'             { TokenDot }

      '...'          { TokenEllipses }

      '('             { TokenLPar }
      ')'             { TokenRPar }

      '{'             { TokenLBrace }
      '}'             { TokenRBrace }

      '['             { TokenLBrack }
      ']'             { TokenRBrack }



%%

 {- Adapted from p. 234-238 of K&R ANSI C, 2nd Ed. -}

 -- translation-unit
TranslationUnitR 
 : ExternalDeclaration { [$1] }
 | TranslationUnitR ExternalDeclaration { $2 : $1 }

TranslationUnit
 : TranslationUnitR { reverse $1 }

 -- external-declaration
ExternalDeclaration
 : FunctionDefinition { error "not done" }
 | Declaration { error "not done" } 

 -- function-definition
FunctionDefinition
 : DeclarationSpecifiersOpt Declarator DeclarationList CompoundStatement 
   { error "not done" }

 -- declaration
Declaration
 : DeclarationSpecifiers InitDeclaratorListOpt ';'
   { error "not done" }
 
 -- declaration-list
DeclarationList
 : DeclarationListR { reverse $1 } 

DeclarationListR
 : Declaration { [$1] }
 | DeclarationList Declaration { $2 : $1 }

 -- declaration-specifiers
DeclarationSpecifiers
 : StorageClassSpecifier DeclarationSpecifiersOpt
   { error "not done" }
 | TypeSpecifier DeclarationSpecifiersOpt
   { error "not done" }
 | TypeQualifier DeclarationSpecifiersOpt
   { error "not done" }

DeclarationSpecifiersOpt
 : DeclarationSpecifiers 
   { Just $1 } 
 | {- empty -}
   { Nothing }

 -- storage-class-specifier 
StorageClassSpecifier
 : auto  { error "not done" }
 | register { error "not done" }
 | static { error "not done" }
 | extern { error "not done" }
 | typedef { error "not done" }
 

 -- type-specifier 
TypeSpecifier
 : void { error "not done" }
 | char { error "not done" }
 | short { error "not done" }
 | int { error "not done" }
 | long { error "not done" }
 | float { error "not done" }
 | double { error "not done" }
 | signed { error "not done" }
 | unsigned { error "not done" }
 | StructOrUnionSpecifier { error "not done" }
 | EnumSpecifier { error "not done" }
 | TypedefName { error "not done" }

 -- type-qualifier 
TypeQualifier 
 : const { error "not done" }
 | volatile { error "not done" }

 -- struct-or-union-specifier
StructOrUnionSpecifier
 : StructOrUnion IdentifierOpt '{' StructDeclarationList '}'
   { error "not done" }
 | StructOrUnion Identifier
   { error "not done" }

 -- struct-or-union
StructOrUnion 
 : struct { error "not done" } 
 | union { error "not done" }

 -- struct-declaration-list
StructDeclarationList
 : StructDeclarationListR { reverse $1 }

StructDeclarationListR
 : StructDeclaration { [$1] }
 | StructDeclarationListR StructDeclaration { $2 : $1 }

 -- init-declarator-list
InitDeclaratorList
 : InitDeclaratorListR { reverse $1 }

InitDeclaratorListR
 : InitDeclarator { [$1] }
 | InitDeclaratorList ',' InitDeclarator { $3 : $1 }

 -- init-declarator
InitDeclarator
 : Declarator { error "not done" }
 | Declarator '=' Initializer { error "not done" }

 -- struct-declaration
StructDeclaration
 : SpecifierQualifierList StructDeclaratorList ';'
   { error "not done" }

 -- specifier-qualifier-list
SpecifierQualifierList 
 : TypeSpecifier SpecifierQualifierListOpt
   { error "not done" }
 | TypeQualifier SpecifierQualifierListOpt
   { error "not done" }

SpecifierQualifierListOpt
 : SpecifierQualifierList
   { $1 }
 | {- empty -}
   { [] }

 -- struct-declarator-list
StructDeclaratorList
 : StructDeclaratorListR { reverse $1 }

StructDeclaratorListR
 : StructDeclarator { [$1] }
 | StructDeclaratorList ',' StructDeclarator { $3 : $1 }

 -- struct-declarator
StructDeclarator
 : Declarator { error "not done" }
 | DeclaratorOpt { error "not done" }

 -- enum-specifier
EnumSpecifier 
 : enum IdentifierOpt '{'  EnumeratorList '}' 
   { error "not done" }
 | enum Identifier
   { error "not done" }

 -- enumerator-list 
EnumeratorList
 : EnumeratorListR
   { reverse $1 }

EnumeratorListR
 : Enumerator
   { [$1] }
 | EnumeratorListR ',' Enumerator
   { $3 : $1 }

 -- enumerator
Enumerator 
 : Identifier
   { error "not done" }
 | Identifier '=' ConstantExpression
   { error "not done" }

 -- declarator
Declarator
 : PointerOpt DirectDeclarator
   { error "not done" }

 -- direct-declarator
DirectDeclarator
 : Identifier  
   { error "not done" }
 | '(' declarator ')'
   { error "not done" }
 | DirectDeclarator '[' ConstantExpression ']'
   { error "not done" }
 | DirectDeclarator '(' ParameterTypeList ')'
   { error "not done" }
 | DirectDeclarator '{' IdentifierListOpt '}'
   { error "not done" }


 -- pointer
Pointer
 : '*' TypeQualifierListOpt 
   { error "not done" }
 | '*' TypeQualifierListOpt Pointer
   { error "not done" }

 -- type-qualifier-list
TypeQualifierList 
 : TypeQualifierListR
   { reverse $1 }

TypeQualifierListR
 : TypeQualifier 
   { [$1] }
 | TypeQualifierListR TypeQualifier
   { $2 : $1 }

 -- parameter-type-list
ParameterTypeList
 : ParameterList 
   { error "not done: parameter-type-list" }
 | ParameterList ',' '...'
   { error "not done: parameter-type-list" }

 -- parameter-list
ParameterList
 : ParameterDeclaration
 | 



 -- Terminal tokens.

Identifier
 : identifier
   { error "not done" }

IdentifierOpt
 : Identifier
   { Just $1 }
 | {- empty -} 
   { Nothing }

TypedefName
 : typedef_name 
   { error "not done" }



Exp : { error "unfinished" }

-- Exp   : let var '=' Exp in Exp { Let $2 $4 $6 }
--       | Exp1                    { Exp1 $1 }

-- Exp1  : Exp1 '+' Term           { Plus $1 $3 }
--       | Exp1 '-' Term           { Minus $1 $3 }
--       | Term                    { Term $1 }

-- Term  : Term '*' Factor         { Times $1 $3 }
--       | Term '/' Factor         { Div $1 $3 }
--       | Factor                  { Factor $1 }

-- Factor  
--       : int                     { Int $1 }
--       | var                     { Var $1 }
--       | '(' Exp ')'             { Brack $2 }


Pos : {- empty -}               {% getPosition }

{

happyError :: ParserMonad a 
happyError = error "Parse error"


data ExtDec
 = ExtDecFun of FunDef
 | ExtDecDec of Dec

data Dec 
 = UNFINISHED

data FunDef
 = NOTDONE





data Exp  
      = Let String Exp Exp
      | Exp1 Exp1

data Exp1 
      = Plus Exp1 Term 
      | Minus Exp1 Term 
      | Term Term

data Term 
      = Times Term Factor 
      | Div Term Factor 
      | Factor Factor

data Factor 
      = Int Int 
      | Var String 
      | Brack Exp


data Pos
 = Pos { posFilename :: !String, posLine :: !Int, posCol :: !Int }

data Token
      = TokenAuto
      | TokenRegister
      | TokenStatic
      | TokenExtern
      | TokenTypedef

      | TokenVoid
      | TokenChar
      | TokenShort
      | TokenInt
      | TokenLong
      | TokenFloat
      | TokenDouble
      | TokenSigned
      | TokenUnsigned

      | TokenConst
      | TokenVolatile

      | TokenStruct
      | TokenUnion

      | TokenEnum

      | TokenInteger Int

      | TokenIdentifier String
      | TokenTypedefName String

      | TokenEq
      | TokenPlus
      | TokenMinus
      | TokenStar
      | TokenDiv

      | TokenSemicolon
      | TokenColon
      | TokenComma
      | TokenDot

      | TokenEllipsis

      | TokenLPar
      | TokenRPar

      | TokenLBrace
      | TokenRBrace

      | TokenLBrack
      | TokenRBrack

      | TokenEOF

instance Show Token where
 show TokenAuto = "auto"
 show TokenRegister = "register"
 show TokenStatic = "static"
 show TokenExtern = "extern"
 show TokenTypedef = "typedef"

 show TokenVoid = "void"
 show TokenChar = "char"
 show TokenShort = "short"
 show TokenInt = "int"
 show TokenLong = "long"
 show TokenFloat = "float"
 show TokenDouble = "double"
 show TokenSigned = "signed"
 show TokenUnsigned = "unsigned"

 show TokenConst = "const"
 show TokenVolatile = "volatile"

 show TokenEnum = "enum"


advanceCol :: Int -> ParserMonad ()
advanceCol numchars =
 do
 pstate <- St.get
 St.put $! pstate { col = (col pstate) + 1, input = tail $! input pstate }


advanceLine :: ParserMonad ()
advanceLine =
 do
 pstate <- St.get
 St.put $! pstate { col = 1, line = (line pstate) + 1, 
                    input = tail $! input pstate }


lexer :: (Token -> ParserMonad a) -> ParserMonad a
lexer cont = 
 do
 pstate <- St.get
 let 
   -- End of file.
  lex [] = cont TokenEOF

   -- Ignore whitespace.
  lex (' ':cs) = do { advanceCol 1 ; lexer cont ; }
  lex ('\t':cs) = do { advanceCol 1 ; lexer cont ; }
  lex ('\n':cs) = do { advanceLine ; lexer cont ; }

   -- Keywords.
  lex ('a':'u':'t':'o':cs) =
   processKeyword cont TokenAuto
  lex ('r':'e':'g':'i':'s':'t':'e':'r':cs) =
   processKeyword cont TokenRegister
  lex ('s':'t':'a':'t':'i':'c':cs) =
   processKeyword cont TokenStatic
  lex ('e':'x':'t':'e':'r':'n':cs) =
   processKeyword cont TokenExtern
  lex ('t':'y':'p':'e':'d':'e':'f':cs) = 
   processKeyword cont TokenTypedef

  lex ('v':'o':'i':'d':cs) =
   processKeyword cont TokenVoid
  lex ('c':'h':'a':'r':cs) =
   processKeyword cont TokenChar
  lex ('s':'h':'o':'r':'t':cs) =
   processKeyword cont TokenShort
  lex ('i':'n':'t':cs) =
   processKeyword cont TokenInt
  lex ('l':'o':'n':'g':cs) =
   processKeyword cont TokenLong
  lex ('f':'l':'o':'a':'t':cs) =
   processKeyword cont TokenFloat
  lex ('d':'o':'u':'b':'l':'e':cs) =
   processKeyword cont TokenDouble
  lex ('s':'i':'g':'n':'e':'d':cs) =
   processKeyword cont TokenSigned
  lex ('u':'n':'s':'i':'g':'n':'e':'d':cs) =
   processKeyword cont TokenUnsigned

  lex ('c':'o':'n':'s':'t':cs) =
   processKeyword cont TokenVoid
  lex ('v':'o':'l':'a':'t':'i':'l':'e':cs) =
   processKeyword cont TokenChar

  lex ('s':'t':'r':'u':'c':'t':cs) =
   processKeyword cont TokenStruct
  lex ('u':'n':'i':'o':'n':cs) =
   processKeyword cont TokenUnion


   -- Parens, braces and brackets.
  lex ('(':cs) = do { advanceCol 1 ; cont TokenLPar ; } 
  lex (')':cs) = do { advanceCol 1 ; cont TokenRPar ; } 

  lex ('{':cs) = do { advanceCol 1 ; cont TokenLBrace ; } 
  lex ('}':cs) = do { advanceCol 1 ; cont TokenRBrace ; } 

  lex ('[':cs) = do { advanceCol 1 ; cont TokenLBrack ; } 
  lex (']':cs) = do { advanceCol 1 ; cont TokenRBrack ; } 


  lex (c:cs) 
      | Char.isDigit c = lexNum [c] (cs)
      | Char.isAlpha c = lexAlpha [c] (cs)


   -- Lex an alphabetic sequence.
  lexAlpha revid [] = do { tok <- processIdent (reverse revid) ; cont tok ; }
  lexAlpha revid (c:cs) 
      | Char.isAlpha c = do { advanceCol 1 ; lexAlpha (c:revid) cs ; } 
      | otherwise = do { tok <- processIdent (reverse revid) ; cont tok ; }
  
   -- Lex a number.
  lexNum revnum [] = error "unfinished"

  in lex $! input pstate



-- lexer :: String -> [Token]
-- lexer [] = []
-- lexer (c:cs) 
--       | Char.isSpace c = lexer cs
--       | Char.isAlpha c = lexVar (c:cs)
--       | Char.isDigit c = lexNum (c:cs)
-- lexer ('=':cs) = TokenEq : lexer cs
-- lexer ('+':cs) = TokenPlus : lexer cs
-- lexer ('-':cs) = TokenMinus : lexer cs
-- lexer ('*':cs) = TokenTimes : lexer cs
-- lexer ('/':cs) = TokenDiv : lexer cs
-- lexer ('(':cs) = TokenLPar : lexer cs
-- lexer (')':cs) = TokenRPar : lexer cs

-- lexNum cs = TokenInt (read num) : lexer rest
--       where (num,rest) = span Char.isDigit cs

-- lexVar cs =
--    case span Char.isAlpha cs of
--       ("let",rest) -> TokenLet : lexer rest
--       ("in",rest)  -> TokenIn : lexer rest
--       (var,rest)   -> TokenVar var : lexer rest

-- main = getContents >>= print . calc . lexer
}
