{
module Parser.SExpParser where

import qualified Compiler.SExpression as SX
import qualified Char (isSpace,isAlpha,isDigit)

import qualified Control.Monad.State as St

import System.IO.Unsafe (unsafePerformIO)

data ParserState = 
 PS { filename :: !String ,
      col :: !Int ,
      line :: !Int ,
      input :: !String }
 deriving (Show)

initParserState filename string = PS { 
 filename = filename ,
 col = 0 ,
 line = 1 ,
 input = string 
 }

type ParserMonad = St.State ParserState

type Pos = SX.Source

getPosition :: St.State ParserState Pos 
getPosition = 
 do
 pstate <- St.get
 return $! SX.SourceFile (filename pstate) (line pstate) (col pstate)

processKeyword cont tok = 
 do
 advanceCol $! length (show tok)
 cont $! tok


}

%monad { St.State ParserState }
%lexer { lexer } { TokenEOF }


%name parseSExps
%tokentype { Token }

%token 
  sym             { TokenSYM $$ }
  string          { TokenSTRING $$ }
  int             { TokenINT $$ }
  bool            { TokenBOOL $$ }
  char            { TokenCHAR $$ }
  '#('            { TokenHASHLPAR }
  '('             { TokenLPAR }
  ')'             { TokenRPAR }  
  '.'             { TokenDOT }  
  quote           { TokenQUOTE }  
  '`'             { TokenQUASIQUOTE }  
  ','             { TokenUNQUOTE }  
  ',@'            { TokenUNQUOTESPLICING }  

%%

SExpList : SExpListRev { reverse $1 }

SExpListRev : {- empty -} { [] }
              | SExpListRev SExp { $2:$1 }


SExp  : sym { SX.symbolFromString $1 }
       | int { SX.Int $1 SX.SourceUndefined }
       | string { SX.String $1 SX.SourceUndefined }
       | bool { SX.Bool $1 SX.SourceUndefined }
       | char { SX.Char $1 SX.SourceUndefined }
       | '(' SExpInnards ')' { $2 }
       | '#(' SExpInnards ')' { SX.Vector (SX.toList $2) SX.SourceUndefined }
       | quote SExp { SX.functor "quote" $2 }
       | '`' SExp { SX.functor "quasiquote" $2 }
       | ',' SExp { SX.functor "unquote" $2 }
       | ',@' SExp { SX.functor "unquotesplicing" $2 }

SExpInnards : {- empty -} { (SX.Nil SX.SourceUndefined) }
              | SExp '.' SExp { SX.Cons $1 $3 SX.SourceUndefined } 
              | SExp SExpInnards { SX.Cons $1 $2 SX.SourceUndefined }

{

data Token
 = TokenINT !Int
 | TokenSTRING !String
 | TokenSYM !String
 | TokenBOOL !Bool
 | TokenCHAR !Char
 | TokenLPAR 
 | TokenRPAR
 | TokenDOT
 | TokenQUOTE
 | TokenQUASIQUOTE
 | TokenUNQUOTE
 | TokenUNQUOTESPLICING
 | TokenHASHLPAR
 | TokenEOF
 deriving (Show)




happyError :: ParserMonad a
happyError = 
 do
 pstate <- St.get 
 error $ "Parse error, with parser state:\n" ++ (show pstate)


processAtom :: String -> ParserMonad Token
processAtom "." = return $! TokenDOT
processAtom "#t" = return $! TokenBOOL True
processAtom "#T" = return $! TokenBOOL True
processAtom "#f" = return $! TokenBOOL False
processAtom "#F" = return $! TokenBOOL False
processAtom str@('-':[])  = return $! TokenSYM str
processAtom str@('-':tl) 
 | all Char.isDigit tl = return $! TokenINT ((0-1)*(read tl))
processAtom str 
 | all Char.isDigit str = return $! TokenINT (read str)
 | otherwise = return $! TokenSYM str




advanceCol :: Int -> ParserMonad ()
advanceCol numchars =
 do
 pstate <- St.get
 St.put $! pstate { col = (col pstate) + numchars, 
		    input = drop numchars (input pstate) }


advanceLine :: ParserMonad ()
advanceLine =
 do
 pstate <- St.get
 St.put $! pstate { col = 0, line = (line pstate) + 1, 
                    input = tail $! input pstate }


isSymbolic char = elem char "!$#%&|*+-/<=>?@^_~:."


lexer :: (Token -> ParserMonad a) -> ParserMonad a
lexer cont = 
 do
 pstate <- St.get
--  () <- return $! unsafePerformIO 
--    (do
--     putStrLn $! "Lexing: " ++ (show pstate)
--     return ())
 let
   -- End of file.
  lex [] = cont TokenEOF

   -- Ignore whitespace & comments.
  lex (' ':cs) = do { advanceCol 1 ; lexer cont ; }
  lex ('\t':cs) = do { advanceCol 1 ; lexer cont ; }
  lex ('\n':cs) = do { advanceLine ; lexer cont ; }

  lex (';':cs) = do { advanceCol 1 ; lexComment cs ; }

   -- Strings.
  lex ('"':cs) = do { advanceCol 1 ; lexString [] cs ; }

   -- Characters.
  lex ('#':'\\':'s':'p':'a':'c':'e':cs) =
   do { advanceCol 7 ; cont $! TokenCHAR ' ' ; }
  lex ('#':'\\':'n':'e':'w':'l':'i':'n':'e':cs) =
   do { advanceCol 9 ; cont $! TokenCHAR '\n' ; }

  lex ('#':'\\':c:cs) 
      | c == '\n' = do { advanceLine ; cont $! TokenCHAR c ; }
      | otherwise = do { advanceCol 3 ; cont $! TokenCHAR c ; }

   -- Parens, braces and brackets.
  lex ('#':'(':cs) = do { advanceCol 2 ; cont TokenHASHLPAR ; } 
  lex ('(':cs) = do { advanceCol 1 ; cont TokenLPAR ; } 
  lex (')':cs) = do { advanceCol 1 ; cont TokenRPAR ; } 

  lex ('\'':cs) = do { advanceCol 1 ; cont TokenQUOTE ; } 
  lex ('`':cs) = do { advanceCol 1 ; cont TokenQUASIQUOTE ; } 
  lex (',':'@':cs) = do { advanceCol 2 ; cont TokenUNQUOTESPLICING ; } 
  lex (',':cs) = do { advanceCol 1 ; cont TokenUNQUOTE ; } 


--   lex ('{':cs) = do { advanceCol 1 ; cont TokenLBrace ; } 
--   lex ('}':cs) = do { advanceCol 1 ; cont TokenRBrace ; } 

--   lex ('[':cs) = do { advanceCol 1 ; cont TokenLBrack ; } 
--   lex (']':cs) = do { advanceCol 1 ; cont TokenRBrack ; } 

  lex (c:cs) 
      | Char.isDigit c = do { advanceCol 1 ; lexAtom [c] (cs) }
      | Char.isAlpha c = do { advanceCol 1 ; lexAtom [c] (cs) }
      | isSymbolic c = do { advanceCol 1 ; lexAtom [c] (cs) }

  lex other = cont $ error $ "non-exhaustive match in the lexer: " ++ (take 30 other)

   -- Lex up string.
  lexString revstr [] = error "Lexer error: EOF inside string."
  lexString revstr ('\\':c:cs) = do { advanceCol 2 ; lexString (c:revstr) cs }
  lexString revstr ('"':cs) = do { advanceCol 1 ; 
                                   cont $! TokenSTRING (reverse revstr) }
  lexString revstr ('\n':cs) = do { advanceLine ; lexString ('\n':revstr) cs }
  lexString revstr (c:cs) = do { advanceCol 1 ; lexString (c:revstr) cs }

   -- Lex up the rest of a comment.
  lexComment [] = cont TokenEOF
  lexComment ('\n':cs) = do { advanceLine ; lexer cont ; }
  lexComment (c:cs) = do { advanceCol 1 ; lexComment cs ; }

   -- Lex an atom.
  lexAtom revid [] = do { tok <- processAtom (reverse revid) ; cont tok ; }
  lexAtom revid (c:cs) 
      | Char.isAlpha c = do { advanceCol 1 ; lexAtom (c:revid) cs ; } 
      | Char.isDigit c = do { advanceCol 1 ; lexAtom (c:revid) cs ; } 
      | isSymbolic c = do { advanceCol 1 ; lexAtom (c:revid) cs ; } 
      | otherwise = do { tok <- processAtom (reverse revid) ; cont tok ; }

  in lex $! input pstate 



 


listFromStringWithName :: String -> String -> [SX.LexicalSExp]
listFromStringWithName name string =
 St.evalState parseSExps (initParserState name string)
 

}
